/**********************************************************************
 * de.langzeitarchivierung.kopal / ProcessTheadPool.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA
 ******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 **********************************************************************/

package de.langzeitarchivierung.kopal;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import de.langzeitarchivierung.kopal.processstarter.ProcessStarter;

/*******************************************************************************
 * A pool of threads which execute ProcessStarters and ProcessData objects. If
 * such an object is finished appropriate measures are taken (e.g. rescheduling
 * of failed objects, serializing them, etc).
 * 
 * @author ludwig
 * @since 20060519
 * @version 20070710
 ******************************************************************************/

public class ProcessThreadPool extends ThreadPoolExecutor
{

	// STATE (Instance variables) *****************************************

	ProcessQueue	pQueue;
	ProcessQueue	errorQueue;

	int				numberOfProcessStarters	= 0;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * @param corePoolSize
	 **************************************************************************/
	public ProcessThreadPool(int corePoolSize) {
		super(corePoolSize, corePoolSize, 10, TimeUnit.SECONDS,
				new LinkedBlockingQueue<Runnable>());
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ****

	/***************************************************************************
	 * Method to execute a thread of type runnable.
	 * 
	 * @see java.util.concurrent.ThreadPoolExecutor#afterExecute(
	 *      java.lang.Runnable, java.lang.Throwable)
	 * 
	 * @param command
	 **************************************************************************/
	@Override
	public void execute(Runnable command) {
		// To keep track how much processStarters are running and allow programm
		// exit if they all are finished.
		if (command instanceof ProcessStarter) {
			this.numberOfProcessStarters++;
		}
		super.execute(command);
	}

	/***************************************************************************
	 * Method that is invoked after executing a process.
	 * 
	 * @see java.util.concurrent.ThreadPoolExecutor#afterExecute(
	 *      java.lang.Runnable, java.lang.Throwable)
	 * 
	 * @param t
	 * @param r
	 **************************************************************************/
	@Override
	protected void afterExecute(Runnable r, Throwable t) {
		super.afterExecute(r, t);

		// To keep track how much processStarters are running and allow programm
		// exit if they all are finished.
		if (r instanceof ProcessStarter) {
			this.numberOfProcessStarters--;
			if (this.numberOfProcessStarters == 0) {
				this.pQueue.shutdownProcessQueue();
			}
			return;
		}
		else if (r instanceof ProcessData) {
			ProcessData lastProcess = (ProcessData) r;

			// If everything has finished we just throw the processData away.
			if (!lastProcess.getPolicy().hasOtherStatusThan(Status.DONE)) {
				return;
			}

			// We have not finished: We have either another reachable todo or
			// not.
			if (lastProcess.getPolicy().hasStatusWithParent(Status.TODO,
					Status.DONE)) {
				this.pQueue.addOldElement(lastProcess);
			}
			else {
				// TODO Write process to filesystem with DB.
				// TODO Don't put it in the input table again!
				// TODO Handle the errorQueue later!
				// errorQueue.addElement(lastProcess);
			}
		}
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param _pQueue The pQueue to set.
	 **************************************************************************/
	public void setProcessQueue(ProcessQueue _pQueue) {
		this.pQueue = _pQueue;
	}

	/***************************************************************************
	 * @param _errorQueue The errorQueue to set.
	 **************************************************************************/
	public void setErrorQueue(ProcessQueue _errorQueue) {
		this.errorQueue = _errorQueue;
	}

}
