/*******************************************************************************
 * de.langzeitarchivierung.kopal / WorkflowTool.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;

import de.langzeitarchivierung.kopal.formats.MetadataFormat;
import de.langzeitarchivierung.kopal.processstarter.ProcessStarter;
import de.langzeitarchivierung.kopal.ui.TestListDataListener;
import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.KopalClassLoader;
import de.langzeitarchivierung.kopal.util.KopalConsoleLogFormatter;
import de.langzeitarchivierung.kopal.util.KopalFileLogFormatter;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;
import de.langzeitarchivierung.kopal.util.KopalXMLParser;
import de.langzeitarchivierung.kopal.util.db.HibernateUtil;
import de.langzeitarchivierung.kopal.util.db.MappToHib;

/*******************************************************************************
 * WorkflowTool is the main class of the koLibRI workflow framework. It can be
 * used to e.g. commit SIPs to the Client Loader and to commit SIPs to DIAS or
 * another archive, either split up in two tools (Workflow Tool, Client Loader)
 * or as a standalone tool.
 * 
 * NOTE: Please use the database only with a single instance of the
 * WorkflowTool, because a concurrent usage of the database is not yet tested!
 * 
 * @author Ludwig
 * @version 20070710
 * @since 20050623
 ******************************************************************************/

public class WorkflowTool
{

	// CONSTANTS **********************************************************

	/** Program constants */
	private static final String	DEFAULT_CONFIG_FILE			= "config.xml";
	/** Program constants */
	public static final int		NUMBER_OF_TEST_LISTENERS	= 1;
	/** Program constants */
	public static final String	kolibriVersion				= "koLibRI v1.0 \"Engage!\"";

	// CLASS VARIABLES (Static variables) *********************************

	/** The default logger */
	private static Logger		defaultLogger				= Logger
																	.getLogger("de.langzeitarchivierung.kopal");
	/** The commandline object */
	public static CommandLine	cLine						= null;

	// STATE (Instance variables) *****************************************

	private String				logfileDir;
	private String				logfileName;
	private String				logLevel;
	private ArrayList<String>	processStarters				= new ArrayList<String>();
	private Level				lLevel;
	/** The maximum number of parallel threads */
	public int					maxNumberOfThreads			= 3;
	private boolean				useDatabase					= false;

	private ProcessThreadPool	pBoss						= null;
	private ProcessQueue		processQueue;
	// TODO The errorQueue ist not fully supported at the moment!
	private ProcessQueue		errorQueue;
	private long				programStartTime;

	/***************************************************************************
	 * Default constructor
	 * 
	 * @author Ludwig
	 * @version 20060626
	 * @since 20050623
	 **************************************************************************/
	public WorkflowTool() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) *****************

	/***************************************************************************
	 * The Main method reads the command line (and additional the config file),
	 * starts a process starter to collect list elements to process and starts
	 * processing the list.
	 * 
	 * @param args The used command line args.
	 **************************************************************************/
	public static void main(String[] args) {
		// Command line parser parses the command line.
		CommandLineParser cliParser = new PosixParser();
		Options cliOptions = cliOptions();

		try {
			cLine = cliParser.parse(cliOptions, args);
		}
		catch (ParseException e) {
			System.out.println("Commandline and its arguments are not correct!"
					+ "\nSystem message: " + e);

			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("WorkflowTool", cliOptions);

			System.exit(1);
		}

		// Print help informations.
		if (cLine.hasOption("h")) {
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("WorkflowTool", cliOptions);
			System.exit(0);
		}

		// Print system properties.
		if (cLine.hasOption("s")) {
			Properties p = System.getProperties();
			for (Enumeration e = p.propertyNames(); e.hasMoreElements();) {
				Object po = e.nextElement();
				System.out.println("Name: " + po + " Value: "
						+ p.getProperty((String) po));
			}
		}

		WorkflowTool woto = new WorkflowTool();

		// Use config file data from the command line or if not specified
		// the default config file to gather general configuration information.
		String configFileString = DEFAULT_CONFIG_FILE;
		if (cLine.hasOption("c")) {
			configFileString = cLine.getOptionValue("c");
		}
		Configurator.setConfigDocument(KopalXMLParser
				.getDocument(configFileString));
		// Configure the WorkflowTool class and the default metadata format.
		try {
			Configurator.configure(woto);
			Configurator.configure(MetadataFormat.Factory.class);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.err.println("Configuration invalid for WorkflowTool!");
			System.exit(1);
		}

		woto.run();

		System.exit(0);
	}

	/***************************************************************************
	 * The run()-method runs the WorkflowTool.
	 **************************************************************************/
	public void run() {
		this.programStartTime = System.currentTimeMillis();

		this.startLogging();
		defaultLogger.log(Level.INFO, kolibriVersion + " started at "
				+ KopalTimestamp.getTimestampWithMs(this.programStartTime)
				+ " (" + this.programStartTime + " millies)");
		if (isUseDatabase()) {
			this.initializeDb();
		}
		this.createProcessQueue();
		this.addListListeners();
		this.initializeThreadPool();
		this.startProcessStarter();

		this.process();
	}

	/***************************************************************************
	 * Defines the available command line options.
	 * 
	 * @return the options
	 **************************************************************************/
	public static Options cliOptions() {
		Options cliOptions = new Options();

		// Short, long, hasArg, desc.
		cliOptions.addOption("p", "process-starter", true, "The process "
				+ "starter module which chooses items for processing.");

		cliOptions.addOption("c", "config-file", true, "The config file to "
				+ "use.");

		cliOptions.addOption("h", "help", false, "Prints this dialog");
		cliOptions.addOption("s", "show-properties", false, "Prints the "
				+ "system properties and continues.");

		return cliOptions;
	}

	/***************************************************************************
	 * Processes the process list.
	 **************************************************************************/
	public void process() {
		defaultLogger.log(Level.INFO, "Process queue processing in progress");

		try {
			// As long as we will receive more tasks we execute them and wait
			// for new ones.
			ProcessData procData = null;
			while (this.processQueue.isAwaitingMoreProcesses()
					|| !this.processQueue.isEmpty()) {

				procData = this.processQueue.takeElement();

				if (procData != null) {
					defaultLogger.log(Level.FINE, "Scheduling process: "
							+ procData.getProcessName());

					// TODO Why has this to be synchronized?
					synchronized (procData) {
						this.pBoss.execute(procData);
					}

					defaultLogger.log(Level.FINE, "Waiting for next process");
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			defaultLogger.log(Level.SEVERE, "Error scheduling process: "
					+ e.getMessage());
		}

		// All processes are finished, so we shutdown our work.
		try {
			this.pBoss.shutdown();
			this.pBoss.awaitTermination(this.pBoss.getTaskCount() * 1000,
					TimeUnit.SECONDS);
		}
		catch (Exception e) {
			e.printStackTrace();
			defaultLogger.log(Level.SEVERE, "Error shutting down "
					+ "processThreadPool: " + e.getMessage());
		}

		// If there are no more processes to be delivered by the process
		// starter the programme exits and displays the total time elapsed
		// and writes it to the logfile.
		long totalTimeElapsed = System.currentTimeMillis()
				- this.programStartTime;

		defaultLogger.log(Level.INFO, "The process list has been processed. "
				+ "Total time elapsed: "
				+ KopalTimestamp.getDurationInHours(totalTimeElapsed));

		// If there are unfinished elements with status other than "done",
		// the programme exits with exit code 3, otherwise it exits with code 0.
		if (!this.errorQueue.isEmpty()) {
			defaultLogger.log(Level.WARNING, "Not everything has finished "
					+ "correctly! Please check the logfile", new File(
					this.logfileName));

			System.exit(3);
		}
		else {
			defaultLogger.log(Level.INFO, "Everything has been done! "
					+ "This process was logged to file '" + this.logfileName
					+ "'");
		}

		System.exit(0);
	}

	// INTERNAL (Internal - implementation details, local classes, ...) **

	/***************************************************************************
	 * Initialize the thread pool.
	 **************************************************************************/
	private void initializeThreadPool() {
		this.pBoss = new ProcessThreadPool(this.maxNumberOfThreads);
		// TODO Use configurator here?
		this.pBoss.setProcessQueue(this.processQueue);
		this.pBoss.setErrorQueue(this.errorQueue);
	}

	/***************************************************************************
	 * Start the process starters.
	 **************************************************************************/
	private void startProcessStarter() {
		ArrayList<ProcessStarter> pStarters = new ArrayList<ProcessStarter>();

		try {
			// TODO LATER: Make it work that the class can be in another
			// package, maybe this should be realized in KopalClassLoader.
			Class pStarterClass = null;

			// Use the process starter specified in command line or ...
			if (cLine.hasOption("p")) {
				pStarterClass = KopalClassLoader.getProcessStarter(cLine
						.getOptionValue("p"));
				pStarters.add((ProcessStarter) KopalClassLoader
						.getInstanceOfClass(pStarterClass));
			}
			// ... use the default process starters from the config file.
			else {
				for (Iterator<String> pIter = this.processStarters.iterator(); pIter
						.hasNext();) {
					pStarterClass = KopalClassLoader.getProcessStarter(pIter
							.next());
					pStarters.add((ProcessStarter) KopalClassLoader
							.getInstanceOfClass(pStarterClass));
				}
			}
		}
		catch (Exception e) {
			defaultLogger.log(Level.SEVERE, "Could not load and initialize "
					+ "process starters. Please check the commandline flag -p "
					+ "or the DefaultProcessStarter field in the main "
					+ "config file! System message: " + e);
			e.printStackTrace();
			System.exit(2);
		}

		// Configure and start all processStarters.
		for (Iterator<ProcessStarter> pIter = pStarters.iterator(); pIter
				.hasNext();) {
			ProcessStarter pStarter = pIter.next();
			// Configure the pStarter.
			try {
				Configurator.configure(pStarter, this);
				// Use command line for input of process starter.
				String[] remainingArgs = cLine.getArgs();
				for (int i = 0; i < remainingArgs.length; i++) {
					pStarter.setInput(remainingArgs[i]);
				}
			}
			catch (Exception e) {
				e.printStackTrace();
				System.err.println("Configuration invalid for ProcessStarter "
						+ pStarter.getClass().getName());
				System.exit(1);
			}
			this.pBoss.execute(pStarter);
			defaultLogger.log(Level.INFO, "ProcessStarter "
					+ pStarter.getClass().getName() + " started");
		}
	}

	/***************************************************************************
	 * Initialize the kopal database.
	 **************************************************************************/
	private void initializeDb() {
		try {
			Configurator.configure(HibernateUtil.class);
			HibernateUtil.initalize();
			MappToHib.initialize();
		}
		catch (Exception e) {
			e.printStackTrace();
			defaultLogger.log(Level.SEVERE, "Database initialization "
					+ "failed! System message: " + e.getMessage());
			System.exit(12);
		}

		defaultLogger.log(Level.INFO, "kopal database '"
				+ HibernateUtil.getHibernateDbName() + "' on host '"
				+ HibernateUtil.getHibernateDbHost()
				+ "' sucessfully initialized");
	}

	/***************************************************************************
	 * Defines and starts the logging infrastructure.
	 **************************************************************************/
	private void startLogging() {
		FileHandler fhFile = null;
		Handler fhConsole = null;

		// At the very beginning turn off the parent loggers console output.
		defaultLogger.setUseParentHandlers(false);

		// Create logfile name
		this.logfileName = this.logfileDir
				+ KopalTimestamp.getFilenameTimestamp() + ".log";

		// Set the file and the concole log handlers and assign them to our
		// defaultLogger.
		try {
			fhFile = new FileHandler(this.logfileName);
		}
		catch (IOException e) {
			System.out.println("Could not create logfile '" + this.logfileName
					+ "'!");

			if (!new File(this.logfileDir).isDirectory()) {
				System.out.println("Directory '" + this.logfileDir
						+ "' does not exist.");
			}
			else {
				System.out.println("System message: " + e);
			}

			System.exit(9);
		}

		// Check the log levels in the config file and set it. Use Level.INFO if
		// there is no correct level given.
		try {
			this.lLevel = Level.parse(this.logLevel);
		}
		catch (Exception e) {
			this.lLevel = Level.parse("INFO");
		}

		// Set the default logger's level.
		defaultLogger.setLevel(this.lLevel);

		// Configure the console log formatter and the file log formatter to a
		// file handler and add them to the defaultLogger.
		fhFile.setFormatter(new KopalFileLogFormatter());
		fhFile.setLevel(this.lLevel);
		defaultLogger.addHandler(fhFile);

		fhConsole = new ConsoleHandler();
		fhConsole.setFormatter(new KopalConsoleLogFormatter());
		fhConsole.setLevel(this.lLevel);
		defaultLogger.addHandler(fhConsole);

		defaultLogger.log(Level.INFO, "koLibRI logging service up "
				+ "and running. Log level set to " + this.lLevel.toString()
				+ ". Writing to file '" + this.logfileName + "'");
	}

	/***************************************************************************
	 * Creates a new process list.
	 **************************************************************************/
	private void createProcessQueue() {
		this.processQueue = new ProcessQueue();
		this.errorQueue = new ProcessQueue();

		try {
			Configurator.configure(this.processQueue);
			Configurator.configure(this.errorQueue);
		}
		catch (Exception e) {
			e.printStackTrace();
			defaultLogger.log(Level.SEVERE, "The given configuration "
					+ "is invalid for WorkflowTool!");
			System.exit(1);
		}

		defaultLogger.log(Level.FINE, "Process queue and error queue have "
				+ "been created");
	}

	/***************************************************************************
	 * Adds simple console output listeners for debugging purposes to the
	 * process list.
	 **************************************************************************/
	private void addListListeners() {
		for (int listeners = 0; listeners < NUMBER_OF_TEST_LISTENERS; listeners++) {
			this.processQueue.addListDataListeners(new TestListDataListener(
					listeners));
		}

		defaultLogger.log(Level.FINE, "List listeners have been added "
				+ "properly");
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param defaultProcessStarter The defaultProcessStarter to set.
	 **************************************************************************/
	public void setDefaultProcessStarter(String defaultProcessStarter) {
		this.processStarters.add(defaultProcessStarter);
	}

	/***************************************************************************
	 * @param _logfileDir The logfileDir to set.
	 **************************************************************************/
	public void setLogfileDir(String _logfileDir) {
		this.logfileDir = _logfileDir;
	}

	/***************************************************************************
	 * @param _logLevel The logLevel to set.
	 **************************************************************************/
	public void setLogLevel(String _logLevel) {
		this.logLevel = _logLevel;
	}

	/***************************************************************************
	 * @return Returns the processQueue.
	 **************************************************************************/
	public ProcessQueue getProcessQueue() {
		return this.processQueue;
	}

	/***************************************************************************
	 * @param _maxNumberOfThreads The maxNumberOfThreads to set.
	 **************************************************************************/
	public void setMaxNumberOfThreads(int _maxNumberOfThreads) {
		this.maxNumberOfThreads = _maxNumberOfThreads;
	}

	/***************************************************************************
	 * @return Returns the useDatabase.
	 **************************************************************************/
	public boolean isUseDatabase() {
		return this.useDatabase;
	}

	/***************************************************************************
	 * @param _useDatabase The useDatabase to set.
	 **************************************************************************/
	public void setUseDatabase(boolean _useDatabase) {
		this.useDatabase = _useDatabase;
	}

}
