/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / AdaptHtmlPage.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 **********************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.net.URI;
import java.util.HashMap;
import java.util.Iterator;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.tidy.Tidy;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.formats.Uof;
import de.langzeitarchivierung.kopal.util.FileUtils;
import de.langzeitarchivierung.kopal.util.KopalXMLParser;

/*******************************************************************************
 * This class converts all given source filenames into the given destination
 * filenames in the given HTML page.
 * 
 * NOTE: Replacation using the JTidy DOM does not really work for all links.
 * Please try the replaceFulltext switch to replace all given source filenames
 * with the given destination filenames in the given HTML page only using
 * fulltext replacement.
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20070710
 * @since 20070601
 ******************************************************************************/

public class AdaptHtmlPage implements ActionModule
{

	// STATE (Instance variables) ****************************************

	ProcessData				processData;
	Step					step;

	String					htmlPageFilename	= "";
	boolean					replaceFulltextOnly	= true;
	HashMap<String, String>	tagHashMap			= initTagMap();

	/***************************************************************************
	 * Default constructor.
	 **************************************************************************/
	public AdaptHtmlPage() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ****************

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 */
	public void go() {
		// Get the normalized base URI ot the current UOF.
		URI normBase = null;
		try {
			normBase = ((Uof) this.processData.getMetadata()).getBasePath();
		}
		catch (Exception e1) {
			e1.printStackTrace();
			this.step.setStatus(Status.ERROR, "Error getting the Uof's "
					+ "BasePath");
			return;
		}

		String message = "";
		if (this.replaceFulltextOnly) {
			message = " using fulltext replacation only";
		}
		this.step.setStatus(Status.RUNNING, "Parsing HTML page './"
				+ normBase.relativize(new File(this.htmlPageFilename).toURI())
						.toString() + "'" + message);

		// Obtain a new Tidy instance and do some configuration.
		Tidy tidy = new Tidy();
		tidy.setQuiet(true);
		tidy.setShowWarnings(false);

		// Look for the referenced links and images and replace the matches
		// found.
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(this.htmlPageFilename);
		}
		catch (FileNotFoundException e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Source HTML file not found",
					new File(this.htmlPageFilename));
			return;
		}

		// Get the hash map out of the costum data.
		HashMap filenames = (HashMap) this.processData.getCustomData().get(
				MigrateFiles.MIGRATION_FILE_TABLE);

		// Get a nice and tidy version of the HTML file as a DOM document.
		Document doc = null;
		if (!this.replaceFulltextOnly) {
			doc = tidy.parseDOM(fis, null);
		}

		// Iterate over all source filenames in the custom data hashmap.
		Iterator fileIter = filenames.keySet().iterator();
		while (fileIter.hasNext()) {
			String sourceFilename = (String) fileIter.next();
			URI sourceUri = new File(sourceFilename).toURI().normalize();
			URI destinUri = new File((String) filenames.get(sourceFilename))
					.toURI().normalize();

			// Go to error state if migrated destination file not found.
			if (!new File(destinUri).exists()) {
				this.step.setStatus(Status.ERROR, "Migrated file not found",
						new File(destinUri));
				return;
			}

			// Replace fulltext only.
			if (this.replaceFulltextOnly) {
				// Just read the document, replace all matching strings and
				// store it again.
				String sourcePage = FileUtils.readFile(new File(
						this.htmlPageFilename), "none");
				sourcePage = sourcePage.replace(normBase.relativize(sourceUri)
						.toString(), normBase.relativize(destinUri).toString());
				FileUtils.storeFile(new File(this.htmlPageFilename), "none",
						sourcePage);
			}
			// Replace links according to DOM.
			else {
				// Iterate over all the keys in the tag hashmap.
				Iterator keyIter = this.tagHashMap.keySet().iterator();
				while (keyIter.hasNext()) {
					String tagName = (String) keyIter.next();
					NodeList nl = doc.getElementsByTagName(tagName);

					// Go throught all links and replace source name with
					// destination name if source name matches.
					for (int i = 0; i < nl.getLength(); i++) {
						Element el = (Element) nl.item(i);
						String pathValue = el.getAttribute(this.tagHashMap
								.get(tagName));

						// Make an absolute path out of the linked resource.
						URI tagValUri = new File(this.processData
								.getPathToContentFiles(), pathValue).toURI()
								.normalize();

						if (tagValUri.equals(sourceUri)) {
							try {
								URI replacement = normBase
										.relativize(destinUri);
								el.setAttribute(this.tagHashMap.get(tagName),
										replacement.toString());
							}
							catch (Exception e) {
								e.printStackTrace();
								this.step.setStatus(Status.ERROR, "Error "
										+ "obtaining the UOF base path");
								return;
							}
						}
					}
				}

				// Rewrite the document using the KopalXMLParser.
				// TODO Store the file with the original charset encoding and
				// original markup! See KopalXMLParser.storeDocument()!
				KopalXMLParser.storeDocument(doc, new File(
						this.htmlPageFilename));
			}
		}

		this.step.setStatus(Status.DONE, "Adapting HTML page '"
				+ this.htmlPageFilename + "' complete");
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ********

	/***************************************************************************
	 * Initializes the tag hashmap to have all the tags to iterate over and the
	 * attributes to use for accessing the link values.
	 * 
	 * TODO Add further link tags here... Maybe take care of JavaScript hrefs,
	 * too?
	 * 
	 * @return
	 **************************************************************************/
	private HashMap<String, String> initTagMap() {
		HashMap<String, String> result = new HashMap<String, String>();

		result.put("a", "href");
		result.put("img", "src");
		result.put("body", "background");

		return result;
	}

	// GET & SET METHODS *******************************************************

	/***************************************************************************
	 * @param processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData processData) {
		this.processData = processData;
	}

	/***************************************************************************
	 * @param step The step to set.
	 **************************************************************************/
	public void setStep(Step step) {
		this.step = step;
	}

	/***************************************************************************
	 * @return Returns the htmlPageFilename.
	 **************************************************************************/
	public String getHtmlPageFilename() {
		return this.htmlPageFilename;
	}

	/***************************************************************************
	 * @param htmlPageFilename The htmlPageFilename to set.
	 **************************************************************************/
	public void setHtmlPageFilename(String htmlPageFilename) {
		this.htmlPageFilename = htmlPageFilename;
	}

	/***************************************************************************
	 * @param replaceFulltextOnly The replaceFulltextOnly to set.
	 **************************************************************************/
	public void setReplaceFulltextOnly(boolean replaceFulltextOnly) {
		this.replaceFulltextOnly = replaceFulltextOnly;
	}

}
