/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / AlreadyIngestedChecker.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.util.Iterator;
import java.util.List;

import de.ddb.lmeRobject.LmerObject;
import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.formats.Uof;
import de.langzeitarchivierung.kopal.util.db.Checksumtype;
import de.langzeitarchivierung.kopal.util.db.DBAccess;
import de.langzeitarchivierung.kopal.util.db.Dias;
import de.langzeitarchivierung.kopal.util.db.MappToHib;

/*******************************************************************************
 * Adds ingest data into the dias table of the database for this process if all
 * of the following checks succeed:
 * 
 * <li>No SIP with the current SIP's XORed checksum has already been ingested
 * (if false the module goes optionally only gives a warning)</li>
 * 
 * <li> No SIP with the same external ID is ingested right now (otherwise this
 * module waits because the module DiasIngestFeedback would not work correctly)</li>
 * 
 * <li>No SIP with the same external ID was previously ingested but the current
 * SIP version claims NOT to be an migration</li>
 * 
 * <li>No SIP with the same external ID and checksum was previously ingested</li>
 * 
 * @author jens
 * @version 20070710
 ******************************************************************************/

public class AlreadyIngestedChecker implements ActionModule
{

	// STATE (Instance variables) *****************************************

	ProcessData	processData;
	Step		step;

	long		lastCheck				= 0;
	long		waitUntilNextTry		= 15000;
	boolean		errorIfChecksumExists	= true;
	boolean		useDatabase;
	String		defaultOwner			= "";

	// CREATION (Constructors, factory methods, static/inst init) **********

	public AlreadyIngestedChecker() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	public void go() {
		this.step.setStatus(Status.RUNNING, "Checking SIP IDs in database");

		String assetId = null;
		try {
			assetId = this.processData.getMetadata().getObjectId();
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error accessing object "
					+ "metadata. System message: " + e.getMessage());
			e.printStackTrace();
			return;
		}

		if (this.useDatabase) {
			// We need to prevent that the object is ingested during this test.
			DBAccess.lockTable(Dias.class, assetId);
		}
		else {
			this.step.setStatus(Status.WARNING, "Please enable the useDatabase"
					+ " flag. No database checking without database");
			return;
		}

		// Get the SIP's checksum and the SIP's checksum type from our metadata.
		String checksumSum = null;
		String checksumSumType = null;
		try {
			LmerObject objectMetadata = ((Uof) this.processData.getMetadata())
					.getLmerObjectCopy();
			checksumSum = objectMetadata.getTransferChecksumArray(0)
					.getStringValue();
			checksumSumType = objectMetadata.getTransferChecksumArray(0)
					.getCHECKSUMTYPE();
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error accessing object "
					+ "metadata. System message: " + e.getMessage());
			DBAccess.unlockTable(Dias.class, assetId);
			e.printStackTrace();
			return;
		}

		// Check if a SIP with the same XOR checksum and default owner already
		// exists. Give a warning or go to error state.
		try {
			if (!MappToHib.alreadyIngested(checksumSum, this.defaultOwner)) {
				this.step.setStatus(Status.RUNNING, "No SIP with the current "
						+ "SIP's XORed checksum was ingested by owner '"
						+ this.defaultOwner + "' before");
			}
			// If the checksum is existing for the owner, go to error state or
			// give a warning according to configuration.
			else {
				String message = "A SIP with the current SIP's XORed checksum "
						+ "already exist in DIAS according to database for "
						+ "owner '" + this.defaultOwner
						+ "' or another ingest is in progress";
				if (this.errorIfChecksumExists) {
					this.step.setStatus(Status.ERROR, message);
					DBAccess.unlockTable(Dias.class, assetId);
					return;
				}

				this.step.setStatus(Status.WARNING, message);
			}
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error accessing database. "
					+ "System message: " + e.getMessage());
			DBAccess.unlockTable(Dias.class, assetId);
			e.printStackTrace();
			return;
		}

		// Create a list of all SIP's with external ID identically to the
		// current SIP.
		List sipsWithSameExternalId = null;
		try {
			sipsWithSameExternalId = DBAccess.getObjects(Dias.class, "urn",
					assetId);
			if (sipsWithSameExternalId.isEmpty()) {
				// Updating the database.
				MappToHib.addDias(this.processData);
				this.step.setStatus(Status.DONE, "A SIP with ID " + assetId
						+ " was not previously ingested, marking that "
						+ "ingest is in process");
				return;
			}
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error accessing database. "
					+ "System message: " + e.getMessage());
			e.printStackTrace();
			DBAccess.unlockTable(Dias.class, assetId);
			return;
		}

		// If another SIP's ingest with this external ID is in progress we have
		// to wait.
		for (Iterator sipIter = sipsWithSameExternalId.iterator(); sipIter
				.hasNext();) {
			Dias previousSip = (Dias) sipIter.next();
			if (previousSip.getDiasintid() == null) {
				this.step.setStatus(Status.TODO, "It looks like another "
						+ "SIP with ID " + assetId + " is ingested right now. "
						+ "We try again later (another possibility is an "
						+ "interruption of a previous ingest process. "
						+ "In this case you have to add the internal ID "
						+ "to the database manually)");

				// If we checked less than waitUntilNextTry seconds ago it is
				// obvious that we have not much to do. So we have to sleep().
				DBAccess.unlockTable(Dias.class, assetId);
				this.lastCheck = System.currentTimeMillis();
				if ((System.currentTimeMillis() - this.lastCheck) < this.waitUntilNextTry) {
					try {
						Thread.sleep(this.waitUntilNextTry);
					}
					catch (Exception e) {
						this.step.setStatus(Status.WARNING, "Error while "
								+ "waiting: " + e.getMessage());
						e.printStackTrace();
					}
				}
				return;
			}
		}

		// We already submitted an object with this external ID so the actual
		// SIP must be a migration. Go to error state, if no oldObjectId is
		// existing.
		String oldObjectId = (String) this.processData.getCustomIds().get(
				MappToHib.ID_DIASOLDOBJID);
		if (oldObjectId == null) {
			this.step.setStatus(Status.ERROR, "An object with the same "
					+ "external ID was previously ingested, but the "
					+ "current SIP version claims NOT to be an migration");
			DBAccess.unlockTable(Dias.class, assetId);
			return;
		}

		// If there is an oldObjectId in the custom ID list, check all internal
		// IDs of the given external ID.
		boolean foundReferencedMigration = false;
		for (Iterator sipIter = sipsWithSameExternalId.iterator(); sipIter
				.hasNext();) {
			Dias previousSip = (Dias) sipIter.next();
			if (previousSip.getDiasintid().equals(oldObjectId)) {
				foundReferencedMigration = true;
				break;
			}
		}
		if (!foundReferencedMigration) {
			this.step.setStatus(Status.ERROR, "The current SIP claims to be "
					+ "a migration, but the interal DIAS ID was not "
					+ "previously ingested according to database");
			DBAccess.unlockTable(Dias.class, assetId);
			return;
		}

		// If we already submitted an object with this external ID and checksum,
		// it probably has already been migrated.
		for (Iterator sipIter = sipsWithSameExternalId.iterator(); sipIter
				.hasNext();) {
			Dias previousSip = (Dias) sipIter.next();
			String dbChecksum = previousSip.getChecksumNew();
			Checksumtype dbChecksumType = previousSip.getChecksumtype();
			if (dbChecksum == null || dbChecksumType == null) {
				dbChecksum = previousSip.getInput().getChecksum();
				dbChecksumType = previousSip.getInput().getChecksumtype();
			}

			if (dbChecksum.equals(checksumSum)
					&& dbChecksumType.getChecksumtype().equals(checksumSumType)) {
				this.step.setStatus(Status.ERROR, "Current SIP was "
						+ "previously ingested accoding to database. "
						+ "An entry with same external ID and same XORed "
						+ "checksum is already existing");
				DBAccess.unlockTable(Dias.class, assetId);
				return;
			}
		}

		// All tests were successful --> We can update the database and add data
		// to the dias table.
		try {
			MappToHib.addDias(this.processData);
		}
		catch (Exception e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Error accessing database. "
					+ "System message: " + e.getMessage());
			DBAccess.unlockTable(Dias.class, assetId);
			return;
		}

		DBAccess.unlockTable(Dias.class, assetId);
		this.step.setStatus(Status.DONE, "The SIP with ID " + assetId
				+ " and XORed checksum " + checksumSum
				+ " was not previously ingested");
	}

	// GET & SET METHODS *******************************************************

	/***************************************************************************
	 * @param step The step to set.
	 **************************************************************************/
	public void setStep(Step step) {
		this.step = step;
	}

	/***************************************************************************
	 * @param processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData processData) {
		this.processData = processData;
	}

	/***************************************************************************
	 * @param setErrorIfchecksumExists The setErrorIfchecksumExists to set.
	 * @author Stefan Funk, SUB Göttingen 15.11.2006
	 **************************************************************************/
	public void setErrorIfChecksumExists(boolean setErrorIfChecksumExists) {
		this.errorIfChecksumExists = setErrorIfChecksumExists;
	}

	/***************************************************************************
	 * @param useDatabase The useDatabase to set.
	 * @author Stefan Funk, SUB Göttingen 13.02.2007
	 **************************************************************************/
	public void setUseDatabase(boolean useDatabase) {
		this.useDatabase = useDatabase;
	}

	/***************************************************************************
	 * @param waitUntilNextTry The waitUntilNextTry to set.
	 **************************************************************************/
	public void setWaitUntilNextTry(long waitUntilNextTry) {
		this.waitUntilNextTry = waitUntilNextTry;
	}

	/***************************************************************************
	 * @param owner The owner to set.
	 * @author Stefan Funk, SUB Göttingen 22.06.2007
	 **************************************************************************/
	public void setDefaultOwner(String defaultOwner) {
		this.defaultOwner = defaultOwner;
	}

}
