/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / DiasIngestFeedback.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 **********************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.IOException;
import java.io.StringReader;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.ingest.DiasIngest;
import de.langzeitarchivierung.kopal.ingest.DiasIngestMetadataResponse;
import de.langzeitarchivierung.kopal.ingest.DiasIngestResponse;
import de.langzeitarchivierung.kopal.retrieval.DiasAccess;
import de.langzeitarchivierung.kopal.retrieval.DiasAccessMetadataResponse;
import de.langzeitarchivierung.kopal.retrieval.DiasAccessResponse;
import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.db.DBAccess;
import de.langzeitarchivierung.kopal.util.db.Dias;
import de.langzeitarchivierung.kopal.util.db.MappToHib;

/*******************************************************************************
 * Checks if the SIP was successfully ingested and updates the database ingest
 * information.
 * 
 * @author ludwig
 * @version 20070710
 ******************************************************************************/
public class DiasIngestFeedback implements ActionModule
{

	// STATE (Instance variables) ****************************************

	ProcessData		processData;
	Step			step;

	long			lastCheck				= 0;
	long			waitUntilNextTry		= 15000;

	private boolean	useDatabase				= false;
	private boolean	removeFileOnError		= true;
	private boolean	removeTechmdOnError		= true;
	private boolean	removeDublincoreOnError	= true;
	private boolean	removeIdentifierOnError	= false;
	private boolean	removeVariousOnError	= false;
	private boolean	removeDiasOnError		= true;

	/***************************************************************************
	 * Default constructor.
	 **************************************************************************/
	public DiasIngestFeedback() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ****************

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 */
	public void go() {
		String ticketId;

		// If we checked the DIAS feedback less than waitUntilNextTry seconds
		// ago, it is obvious that we have not much to do. So we have to
		// sleep().
		if ((System.currentTimeMillis() - this.lastCheck) < this.waitUntilNextTry) {
			try {
				Thread.sleep(this.waitUntilNextTry);
			}
			catch (Exception e) {
				this.step.setStatus(Status.WARNING, "Error while waiting. "
						+ "System message: " + e.getMessage());
				e.printStackTrace();
			}
		}

		this.lastCheck = System.currentTimeMillis();

		// Get the ticket ID from the custom ID hash table and check for NULL
		// value.
		ticketId = (String) this.processData.getCustomIds().get(
				MappToHib.ID_DIASTICKETID);
		if (ticketId == null) {
			this.step.setStatus(Status.ERROR, "Ticket ID value is NULL");
			return;
		}

		// Start a DIAS ingest status request using the ticket ID.
		DiasIngest di = new DiasIngest();
		try {
			Configurator.configure(di);
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		// Request the SIP's ingest status.
		DiasIngestResponse dir = new DiasIngestResponse();
		try {
			String diasResponse = di.requestSipIngestStatus(ticketId);
			dir = DiasIngest.parseDiasResponse(diasResponse);
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		// Get the metadata response. We assume we only get one element in
		// the response.
		DiasIngestMetadataResponse dimr = new DiasIngestMetadataResponse();
		dimr = dir.getMetadata().get(0);

		// DEBUG OUTPUT!!
		//
		// this.step.setStatus(Status.RUNNING, "\nComplete DIR output:\n" //
		// + dir.toString() + "\n");
		//
		// DEBUG OUTPUT!!

		// First check if the state is "done"...
		if (dir.getMetadata().get(0).getState().equals(
				DiasIngestMetadataResponse.STATE_DONE)) {
			try {
				this.step.setStatus(Status.RUNNING, "Ingest complete for SIP "
						+ this.processData.getMetadata().getObjectId()
						+ ". DIAS says ingest state is: " + dimr.getState());

				// Update the database with the latest internal ID.
				if (updateDatabase(this.processData.getMetadata().getObjectId())) {
					this.step.setStatus(Status.DONE, "Database updated");
				}
				else {
					this.step.setStatus(Status.TODO, "Database update failed. "
							+ "We try again later");
				}
				return;
			}
			catch (Exception e) {
				e.printStackTrace();
				this.step.setStatus(Status.ERROR, "Error updating database. "
						+ "System message: " + e.getMessage());
				return;
			}
		}
		// ...then at last check if state is "error"!
		else if (dir.getMetadata().get(0).getState().equals(
				DiasIngestMetadataResponse.STATE_ERROR)) {

			this.step.setStatus(Status.RUNNING, "DIAS says ingest state is "
					+ dimr.getState() + ": " + dimr.getSubstate() + ", "
					+ dimr.getSupplementaryInformation());

			// Delete all table entries from database if wanted.
			try {
				this.step.setStatus(Status.ERROR, "Error during Ingest. "
						+ "The SIP's entries in the following tables were "
						+ "removed: " + removeTableEntries());
			}
			catch (Exception e) {
				e.printStackTrace();
				this.step.setStatus(Status.ERROR, "Error updating database. "
						+ "You will have to remove the stored entries in the "
						+ "database manually");
				return;
			}
			return;
		}
		// ...at last check if type is "busy"...
		else if (dir.getResponseType().equals(
				DiasIngestResponse.RESPONSE_TYPE_BUSY)) {
			// ...and has ingest state...
			if (dimr.getState() != DiasIngestResponse.NOT_IN_DIAS_RESPONSE) {
				this.step.setStatus(Status.TODO, "Ingest is not yet complete. "
						+ "DIAS says ingest state is: " + dimr.getState()
						+ ". We try again later");
			}
			// ...if not, just log the alert and try again later...
			else {
				this.step.setStatus(Status.TODO, "Ingest is not yet complete. "
						+ "DIAS submits ingest alert: " + dimr.getAlert()
						+ ". We try again later");
			}
		}
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ********

	/***************************************************************************
	 * Remove all the table entries, that were added to the database and that
	 * shall be removed if the ingest failed according to the configuration.
	 * 
	 * @throws Exception
	 **************************************************************************/
	private String removeTableEntries() throws Exception {
		String result = "";

		if (this.removeIdentifierOnError) {
			MappToHib.removeIdentifier(this.processData);
			result += " | identifier entries";
		}
		if (this.removeDublincoreOnError) {
			MappToHib.removeDublincore(this.processData);
			result += " | dublincore entries";
		}
		if (this.removeFileOnError) {
			MappToHib.removeFile(this.processData);
			result += " | file entries";
		}
		else if (this.removeTechmdOnError) {
			MappToHib.removeTechmdOnly(this.processData);
			result += " | techmd entries only";
		}
		if (this.removeVariousOnError) {
			MappToHib.removeVarious(this.processData);
			result += " | various entries";
		}
		if (this.removeDiasOnError) {
			MappToHib.removeDias(this.processData);
			result += " | dias entry";
		}

		return result.substring(2).trim();
	}

	/***************************************************************************
	 * Requests a Metadata response from DIAS for an ingested SIP and updates
	 * the database.
	 * 
	 * @param extId The external id to check.
	 * @return the DiasResponse to update the dias table of the database.
	 * @throws Exception
	 **************************************************************************/
	private boolean updateDatabase(String extId) throws Exception {
		DiasAccess access = new DiasAccess();
		Configurator.configure(access);
		String diasStringResponse = null;

		try {
			// Dias returns the latest asset with the specified external ID.
			diasStringResponse = access.requestShortMdByExtId(extId, "xml",
					false);
		}
		catch (IOException e) {
			e.printStackTrace();
			this.step.setStatus(Status.WARNING, "Error requesting DIAS "
					+ "information. System message: " + e.getMessage()
					+ ". We try again later");
			return false;
		}

		// Request the internal ID from the DIAS system.
		DiasAccessResponse assetData = DiasAccess
				.parseDiasResponse(new StringReader(diasStringResponse));

		if (assetData.getResponseType().equals(
				DiasAccessResponse.RESPONSE_TYPE_ERROR)) {
			this.step.setStatus(Status.WARNING, "Error accessing data from "
					+ "the DIAS. DIAS says: " + assetData.getDescription());
			return false;
		}

		String intId = null;
		DiasAccessMetadataResponse assetMetadata = assetData.getMetadata().get(
				0);

		if (assetData.getInternalAssetId().equals(
				DiasAccessResponse.NOT_IN_DIAS_RESPONSE)) {
			intId = assetMetadata.getInternalAssetId();
		}
		else {
			intId = assetData.getInternalAssetId();
		}

		// If the internal ID of the latest SIP is not in our database, it
		// hopefully was successfully ingested.
		// TODO Better get back the internal ID as a response parameter from the
		// successfull IngestStatusRequest in SubmitSipToDias!
		if (this.useDatabase
				&& DBAccess.getUniqueObject(Dias.class, "diasintid", intId) == null) {
			this.step.setStatus(Status.RUNNING, "We assume that the internal "
					+ "ID of the latest ingested SIP is equivalent to the "
					+ "internal ID of the latest used ticket");

			// Update the database.
			MappToHib.updateDias(this.processData, assetMetadata);

			return true;
		}

		return false;
	}

	// GET & SET METHODS *******************************************************

	/***************************************************************************
	 * @param processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData processData) {
		this.processData = processData;
	}

	/***************************************************************************
	 * @param step The step to set.
	 **************************************************************************/
	public void setStep(Step step) {
		this.step = step;
	}

	/***************************************************************************
	 * @param step The step to set.
	 **************************************************************************/
	public void setUseDatabase(boolean _useDatabase) {
		this.useDatabase = _useDatabase;
	}

	/***************************************************************************
	 * @param waitUntilNextTry The waitUntilNextTry to set.
	 **************************************************************************/
	public void setWaitUntilNextTry(long waitUntilNextTry) {
		this.waitUntilNextTry = waitUntilNextTry;
	}

	/***************************************************************************
	 * @param removeDiasOnError The removeDiasOnError to set.
	 **************************************************************************/
	public void setRemoveDiasOnError(boolean removeDiasOnError) {
		this.removeDiasOnError = removeDiasOnError;
	}

	/***************************************************************************
	 * @param removeDublincoreOnError The removeDublincoreOnError to set.
	 **************************************************************************/
	public void setRemoveDublincoreOnError(boolean removeDublincoreOnError) {
		this.removeDublincoreOnError = removeDublincoreOnError;
	}

	/***************************************************************************
	 * @param removeFileOnError The removeFileOnError to set.
	 **************************************************************************/
	public void setRemoveFileOnError(boolean removeFileOnError) {
		this.removeFileOnError = removeFileOnError;
	}

	/***************************************************************************
	 * @param removeIdentifierOnError The removeIdentifierOnError to set.
	 **************************************************************************/
	public void setRemoveIdentifierOnError(boolean removeIdentifierOnError) {
		this.removeIdentifierOnError = removeIdentifierOnError;
	}

	/***************************************************************************
	 * @param removeTechmdOnError The removeTechmdOnError to set.
	 **************************************************************************/
	public void setRemoveTechmdOnError(boolean removeTechmdOnError) {
		this.removeTechmdOnError = removeTechmdOnError;
	}

	/***************************************************************************
	 * @param removeVariousOnError The removeVariousOnError to set.
	 **************************************************************************/
	public void setRemoveVariousOnError(boolean removeVariousOnError) {
		this.removeVariousOnError = removeVariousOnError;
	}

}
