/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / FileCopyBase.java
 * 
 * Copyright 2005-2007 by Project kopal
 *
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 *************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Timer;
import java.util.TimerTask;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.util.FileUtils;

/*******************************************************************************
 * Actionmodule that copies all the files in this path to a new directory in the
 * temp directory named with the current timestamp in milliseconds. It then sets
 * the path to the content files to this new temp dir
 * 
 * Precondition: A set path to content files
 * 
 * Result: A new temporary directory in the temp directory, with the name of the
 * current timestamp and the path to the content files set to this new temporary
 * directory
 * 
 * @author Matthias Neubauer, German National Library
 * @author Stefan Funk, SUB Goettingen
 * @version 20070710
 * @since 20050927
 * @see java.io.File
 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule
 ******************************************************************************/

public class FileCopyBase implements ActionModule
{

	// STATE (Instance variables) ****************************************

	protected ProcessData	processData;
	protected Step			step;

	protected boolean		useDatabase					= false;
	protected String		tempDir;
	protected int			totalFileAmount				= 0;
	protected int			filesCopied					= 0;
	protected long			fileCopyNotificationTime	= 10000;
	protected String		defaultProcessStarter;

	// MANIPULATION (Manipulation - what the object does) ****************

	/***************************************************************************
	 * Starts the essential task of this class
	 **************************************************************************/
	public void go() {
		try {
			copyFiles();
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error during file copy! "
					+ "Please check if your ProcessStarter '"
					+ this.defaultProcessStarter + "' fits to this '"
					+ this.getClass().getName() + "' class. System message: "
					+ e.getMessage());
			e.printStackTrace();

			return;
		}

		this.step.setStatus(Status.DONE, "Filecopy complete");
	}

	// INTERNAL (Internal - implementation details, local classes, ...) **

	/***************************************************************************
	 * Goes through the list of files in the path to the content files, invokes
	 * the recurse method for these files and folders, and finally sets the path
	 * to content files to the newly created temp directory
	 * 
	 * @throws Exception
	 **************************************************************************/
	protected void copyFiles() throws Exception {
		// Check if path to content files is not null.
		if (this.processData.getPathToContentFiles() == null) {
			throw new FileNotFoundException("Path to content files "
					+ "is not specified");
		}

		// Check if source file does exist.
		File source = new File(this.processData.getPathToContentFiles());
		if (!source.exists()) {
			throw new FileNotFoundException("Given path to content files does"
					+ "not exist");
		}

		this.step.setStatus(Status.RUNNING, "Transfering files from '"
				+ this.processData.getPathToContentFiles() + "' to temporary "
				+ "directory");

		// Add a folder named by the current timestamp to the path.
		File destination = new File(this.tempDir
				+ File.separatorChar
				+ java.lang.Long.toString(System.currentTimeMillis())
				+ "_"
				+ FileUtils.getFilenamePersistentIdentifier(this.processData
						.getProcessName()));
		destination.mkdir();

		// Show a notification every fileCopyNotificationTime seconds, until
		// status is set to DONE.
		Timer timer = new Timer();
		timer.schedule(new ProgressNotificationTask(), 0,
				this.fileCopyNotificationTime);

		// Get the total file amount.
		this.totalFileAmount = getCompleteFileAmount(source);

		// Copy all files and folders (or the single file) to the temp
		// directory.
		if (this.totalFileAmount == 1) {
			recurseCopyFiles(source, destination);
		}
		else {
			File filelist[] = source.listFiles();
			for (int i = 0; i < filelist.length; i++) {
				recurseCopyFiles(filelist[i], destination);
			}
		}

		// Stop the timer if completed copying.
		timer.cancel();

		// Just log.
		this.step.setStatus(Status.RUNNING, "Files copied: 100%");
		this.step.setStatus(Status.RUNNING, "Re-setting path to content files "
				+ "to new temp direction");

		// Set the new path to the content files.
		this.processData.setPathToContentFiles(destination.getAbsolutePath()
				+ File.separatorChar);
	}

	/***************************************************************************
	 * Goes recursively through the diretories and invokes copy method for files
	 * 
	 * @param source - The folder containing the source data
	 * @param destination - The destination folder
	 * @throws Exception
	 **************************************************************************/
	protected void recurseCopyFiles(File source, File destination)
			throws Exception {
		if (source.isDirectory()) {
			File[] filelist = source.listFiles();
			File newdir = new File(destination, source.getName());

			// Create a new directory in the destination folder.
			newdir.mkdir();
			for (int i = 0; i < filelist.length; i++) {
				// Recurse through the directory.
				recurseCopyFiles(filelist[i], newdir);
			}
		}
		if (source.isFile()) {
			File destfile = new File(destination, source.getName());

			copyFile(source, destfile);
		}
	}

	/***************************************************************************
	 * Copies a single file from a given source to a given destination.
	 * 
	 * NOTE: Not moved to util.FileUtils, because of the filesCopied value.
	 * 
	 * @param source - The source file
	 * @param destination - The destination folder
	 * @throws Exception
	 **************************************************************************/
	protected void copyFile(File source, File destination)
			throws FileNotFoundException, IOException {
		// Open input- and output streams for the source and destination
		// file.
		FileInputStream fis = new FileInputStream(source);
		FileOutputStream fos = new FileOutputStream(destination);

		// Copy the source file to the destination file through a buffer.
		byte[] buf = new byte[1048576];
		int i = 0;
		while ((i = fis.read(buf)) != -1) {
			fos.write(buf, 0, i);
		}

		// Close the Streams.
		fis.close();
		fos.close();
		this.filesCopied++;
	}

	/***************************************************************************
	 * Computes the amount of files contained in a given directory recursively.
	 * 
	 * @param currentFile
	 * @return The file amount of course :) What else?
	 **************************************************************************/
	private int getCompleteFileAmount(File currentFile) {
		int fileAmount = 0;

		// If it is a file, directly return it's size.
		if (currentFile.isFile()) {
			fileAmount = 1;
		}
		else {
			// Otherwise get the file list and go throught all directories and
			// add the file amount.
			File fileList[] = currentFile.listFiles();

			for (int i = 0; i < fileList.length; i++) {
				if (fileList[i].isDirectory()) {
					fileAmount += getCompleteFileAmount(fileList[i]);
				}
				else {
					fileAmount++;
				}
			}
		}

		return fileAmount;
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param _tempDir The tempDir to set.
	 **************************************************************************/
	public void setTempDir(String _tempDir) {
		this.tempDir = _tempDir;
	}

	/***************************************************************************
	 * @param _processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData _processData) {
		this.processData = _processData;
	}

	/***************************************************************************
	 * @param _step The step to set.
	 **************************************************************************/
	public void setStep(Step _step) {
		this.step = _step;
	}

	/***************************************************************************
	 * @param useDb The useDatabase to set.
	 **************************************************************************/
	public void setUseDatabase(boolean useDb) {
		this.useDatabase = useDb;
	}

	/***************************************************************************
	 * @param _fileCopyNotificationTime
	 **************************************************************************/
	public void setFileCopyNotificationTime(long _fileCopyNotificationTime) {
		this.fileCopyNotificationTime = _fileCopyNotificationTime;
	}

	/***************************************************************************
	 * @param _defaultProcessStarter Default process starter.
	 **************************************************************************/
	public void setDefaultProcessStarter(String _defaultProcessStarter) {
		this.defaultProcessStarter = _defaultProcessStarter;
	}

	/***************************************************************************
	 * Just gives notification log output.
	 * 
	 * @author Stefan Funk, SUB Goettingen
	 **************************************************************************/
	class ProgressNotificationTask extends TimerTask
	{
		/**
		 * @see java.lang.Runnable#run()
		 */
		@Override
		public synchronized void run() {
			int copied = FileCopyBase.this.filesCopied;
			int fileAmount = FileCopyBase.this.totalFileAmount;

			if (fileAmount != 0 && fileAmount > copied) {
				FileCopyBase.this.step.setStatus(Status.RUNNING,
						"Files copied: " + (copied * 100 / fileAmount) + "% ("
								+ copied + " of " + fileAmount + ")");
			}
		}
	}

}
