/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / FileCopyHttpBase.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.GetMethod;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.processstarter.MonitorHttpLocationBase;
import de.langzeitarchivierung.kopal.util.HTTPUtils;

/*******************************************************************************
 * Actionmodule that copies all the files from a HTTP location in a temporary
 * directory named with the current timestamp in milliseconds. It then sets the
 * path to the content files to this new temp dir. If only files shall be copied
 * from an HTTP location, an extra directory will be created for each single
 * file and the file will be copied inside this directory.
 * 
 * Precondition: A valid URL with an index.html directory file.
 * 
 * Result: A new temporary directory in the temp directory, with the name of the
 * current timestamp and the path to the content files set to this new temporary
 * directory.
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20070710
 * @since 20061016
 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule
 ******************************************************************************/

public class FileCopyHttpBase implements ActionModule
{

	// CONSTANTS **********************************************************

	private static final int	INPUT_BUFFER_LENGTH			= 1024 * 1024;

	// STATE (Instance variables) ****************************************

	protected ProcessData		processData;
	Step						step;

	protected boolean			useDatabase;
	protected String			tempDir;
	protected HttpClient		httpClient					= new HttpClient();
	protected int				totalFileAmount				= 0;
	protected int				filesCopied					= 0;
	protected String			currentHttpLocation;
	protected long				fileCopyNotificationTime	= 10000;
	protected String			defaultProcessStarter;

	// MANIPULATION (Manipulation - what the object does) ****************

	/***************************************************************************
	 * Starts the essential task of this class.
	 **************************************************************************/
	public void go() {
		try {
			// Get the URL from the custom data HashMap.
			HashMap hm = this.processData.getCustomData();

			if (hm.containsKey(MonitorHttpLocationBase.VAR_SIPSOURCEASURI)) {
				this.currentHttpLocation = (String) hm
						.get(MonitorHttpLocationBase.VAR_SIPSOURCEASURI);
			}
			else {
				this.step.setStatus(Status.ERROR, "No SIP source in "
						+ "custom data found, must be set by the HTTP "
						+ "ProcessStarter");
				return;
			}

			downloadFiles();
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error during HTTP copy! "
					+ "Please check if your ProcessStarter '"
					+ this.defaultProcessStarter + "' fits to this '"
					+ this.getClass().getName() + "' class. System message: "
					+ e.getMessage());
			e.printStackTrace();

			return;
		}

		this.step.setStatus(Status.DONE, "Download complete");
	}

	// INTERNAL (Internal - implementation details, local classes, ...) **

	/***************************************************************************
	 * Copies files.
	 * 
	 * @throws Exception
	 **************************************************************************/
	protected void downloadFiles() throws Exception {
		File destinationPath = new File(this.tempDir);

		// Check the URL prefix.
		if (!this.currentHttpLocation.startsWith("http")) {
			throw new Exception("The path to content files is given in "
					+ "the wrong format, we need it to start with "
					+ "http:// of course");
		}

		this.step.setStatus(Status.RUNNING, "Downloading files from URL "
				+ this.currentHttpLocation + " to temporary directory");

		// Create a new temporary directory.
		destinationPath = new File(destinationPath.getAbsolutePath()
				+ File.separatorChar
				+ java.lang.Long.toString(System.currentTimeMillis()) + "_"
				+ HTTPUtils.getFilenameFromUrl(this.currentHttpLocation));
		destinationPath.mkdir();

		// Show a notification every fileCopyNotificationTime seconds, until
		// status is set to DONE.
		Timer timer = new Timer();
		timer.schedule(new ProgressNotificationTask(), 0,
				this.fileCopyNotificationTime);

		// Get the total file amount.
		this.totalFileAmount = calculateTotalFileAmount(this.currentHttpLocation);

		// Invoke copyFilesFromHttp if a directory is found, otherwise just copy
		// the file into the newly created directory.
		if (this.currentHttpLocation.endsWith("/")) {
			recurseCopyFilesFromHttp(this.currentHttpLocation, destinationPath);
		}
		else {
			copySingleFileFromHttp(this.currentHttpLocation, destinationPath);
		}

		// Stop the timer if completed copying.
		timer.cancel();

		// Just log.
		this.step.setStatus(Status.RUNNING, "Files downloaded: 100%");
		this.step.setStatus(Status.RUNNING, "Re-setting path to content files "
				+ "to new temp direction");

		// Set the new path to the content files.
		this.processData.setPathToContentFiles(destinationPath
				.getAbsolutePath()
				+ File.separatorChar);
	}

	/***************************************************************************
	 * Dopwnloads recursively all files and folders and invokes copy method for
	 * files.
	 * 
	 * @since 20061017
	 * @param currentUrl
	 * @param destinationDir
	 * @throws Exception
	 **************************************************************************/
	protected void recurseCopyFilesFromHttp(String currentUrl,
			File destinationDir) throws Exception {
		String currentHref;

		// Get a list of all contained files in one HTTP location.
		ArrayList source = HTTPUtils.getAllHrefs(this.httpClient, currentUrl);

		// Go throught all the entries.
		Iterator iter = source.iterator();
		while (iter.hasNext()) {
			currentHref = (String) iter.next();

			// If currentFile is a folder, create a new destination directory
			// and copy files from HTTP again.
			if (currentHref.endsWith("/")) {
				File newdir = new File(destinationDir, HTTPUtils
						.getFilenameFromUrl(currentHref));
				newdir.mkdir();
				recurseCopyFilesFromHttp(currentHref, newdir);
			}
			else {
				// If currentHref is a file, create the new file reference and
				// invoke copySingleFileFromHttp.
				File newfile = new File(destinationDir, HTTPUtils
						.getFilenameFromUrl(currentHref));
				copySingleFileFromHttp(currentHref, newfile);
			}
		}
	}

	/***************************************************************************
	 * Calculates the total amount of files to be copied from the HTTP location,
	 * only needed for nice logging reasons.
	 * 
	 * @param currentUrl
	 * @return int
	 * @throws Exception
	 **************************************************************************/
	protected int calculateTotalFileAmount(String currentUrl) throws Exception {
		int result = 0;
		String currentHref;

		// Get a list of all contained files in one HTTP location.
		ArrayList source = HTTPUtils.getAllHrefs(this.httpClient, currentUrl);

		// Go throught all the entries.
		Iterator iter = source.iterator();
		while (iter.hasNext()) {
			currentHref = (String) iter.next();

			// If currentFile is a folder, just use recursion.
			if (currentHref.endsWith("/")) {
				result += calculateTotalFileAmount(currentHref);
			}
			else {
				// If currentHref is a file, increase the total file amount.
				result += 1;
			}
		}

		return result;
	}

	/***************************************************************************
	 * Copies a single file from a given HTTP source to a given destination.
	 * 
	 * @param source
	 * @param newfile
	 * @throws Exception
	 * 
	 * @version 20061017
	 **************************************************************************/
	protected void copySingleFileFromHttp(String source, File newfile)
			throws Exception {
		// Get the HTTP method.
		GetMethod method = HTTPUtils.httpGetMethod(this.httpClient, source);

		// Read the response body as a stream.
		InputStream is = new BufferedInputStream(method
				.getResponseBodyAsStream());
		OutputStream os = new BufferedOutputStream(
				new FileOutputStream(newfile));

		// Read and save the file.
		try {
			byte[] buffer = new byte[INPUT_BUFFER_LENGTH];

			int read = is.read(buffer);
			while (read > 0) {
				os.write(buffer, 0, read);
				read = is.read(buffer);
			}
			os.flush();
		}
		catch (Exception e) {
			throw new Exception();
		}
		finally {
			if (is != null) {
				is.close();
			}
			if (os != null) {
				os.close();
			}
			method.releaseConnection();
		}

		// Increase the number of files copied.
		this.filesCopied += 1;
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param _tempDir The tempDir to set.
	 **************************************************************************/
	public void setTempDir(String _tempDir) {
		this.tempDir = _tempDir;
	}

	/***************************************************************************
	 * @param _processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData _processData) {
		this.processData = _processData;
	}

	/***************************************************************************
	 * @param _step The step to set.
	 **************************************************************************/
	public void setStep(Step _step) {
		this.step = _step;
	}

	/***************************************************************************
	 * @param _useDatabase The useDatabase to set.
	 **************************************************************************/
	public void setUseDatabase(boolean _useDatabase) {
		this.useDatabase = _useDatabase;
	}

	/***************************************************************************
	 * @param _fileCopyNotificationTime
	 **************************************************************************/
	public void setFileCopyNotificationTime(long _fileCopyNotificationTime) {
		this.fileCopyNotificationTime = _fileCopyNotificationTime;
	}

	/***************************************************************************
	 * @param defaultProcessStarter
	 **************************************************************************/
	public void setDefaultProcessStarter(String _defaultProcessStarter) {
		this.defaultProcessStarter = _defaultProcessStarter;
	}

	/***************************************************************************
	 * Just gives notofication log output.
	 * 
	 * @author funk
	 **************************************************************************/
	class ProgressNotificationTask extends TimerTask
	{
		/**
		 * @see java.lang.Runnable#run()
		 */
		@Override
		public synchronized void run() {
			int copied = FileCopyHttpBase.this.filesCopied;
			int fileAmount = FileCopyHttpBase.this.totalFileAmount;

			if (fileAmount != 0 && fileAmount > copied) {
				FileCopyHttpBase.this.step.setStatus(Status.RUNNING,
						"Files downloaded: " + (copied * 100 / fileAmount)
								+ "% (" + copied + " of " + fileAmount + ")");
			}
		}
	}

}
