/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / MetadataGenerator.java
 * 
 * Copyright 2005-2007 by Project kopal
 *
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 *************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Logger;

import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import de.ddb.lmeRfile.LmerFile;
import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.formats.Uof;
import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.FormatRegistry;
import de.langzeitarchivierung.kopal.util.HTMLUtils;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;
import edu.harvard.hul.ois.jhove.AESAudioMetadata;
import edu.harvard.hul.ois.jhove.Agent;
import edu.harvard.hul.ois.jhove.AgentType;
import edu.harvard.hul.ois.jhove.App;
import edu.harvard.hul.ois.jhove.ConfigHandler;
import edu.harvard.hul.ois.jhove.ErrorMessage;
import edu.harvard.hul.ois.jhove.HandlerBase;
import edu.harvard.hul.ois.jhove.InfoMessage;
import edu.harvard.hul.ois.jhove.JhoveBase;
import edu.harvard.hul.ois.jhove.Message;
import edu.harvard.hul.ois.jhove.Module;
import edu.harvard.hul.ois.jhove.NisoImageMetadata;
import edu.harvard.hul.ois.jhove.Property;
import edu.harvard.hul.ois.jhove.PropertyArity;
import edu.harvard.hul.ois.jhove.PropertyType;
import edu.harvard.hul.ois.jhove.Rational;
import edu.harvard.hul.ois.jhove.RepInfo;
import edu.harvard.hul.ois.xml.ns.jhove.JhoveDocument;

/*******************************************************************************
 * Metadata generator for a set of files for a SIP, based on
 * edu.harvard.hul.ois.JhoveBase, but not inherited from it
 * 
 * Precondition: The filelist of the ProcessData object is filled.
 * MetadataGenerator locates the files through these entries.
 * 
 * Result: The metadata objects of the UOF are filled with metadata generated by
 * JHOVE.
 * 
 * 
 * TODO: Make an option available to only ignore sigmatch errors in HTML files,
 * as they are never valid!
 * 
 * @author Matthias Neubauer, German National Library
 * @version 20070710
 * @see edu.harvard.hul.ois.jhove.JhoveBase
 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule
 * @since 20050729
 ******************************************************************************/

public class MetadataGenerator implements ActionModule
{

	// CONSTANTS **********************************************************

	protected final String				NAM							= "MetadataGenerator";
	protected final String				REL							= "1.0";
	protected final String				USE							= "ActionModule for kopal software";
	protected final String				RIGHT						= "Copyright 2005-2007 by"
																			+ " Project kopal http://kopal."
																			+ "langzeitarchivierung.de/";

	protected Logger					defaultLogger				= Logger
																			.getLogger("de.langzeitarchivierung.kopal");

	// CLASS VARIABLES (Static variables) **********************************

	/** Date of classes creation */
	public static int[]					DATE						= { 2005,
			8, 29													};

	// STATE (Instance variables) *****************************************

	/** Checksum-flag */
	protected boolean					_checksum;
	/** Ordered list of modules. */
	protected List<Module>				_moduleList;
	/** ShowRawFlag */
	protected boolean					_showRaw;
	/** Configuration file pathname. */
	protected String					jhoveConfigFile;
	/** Map of modules (for fast access by name). */
	protected Map<String, Module>		_moduleMap;
	protected boolean					_signature;
	/** ProcessData that contains the process and step informations */
	protected ProcessData				processData					= null;
	/** Step that represents the momentary action */
	protected Step						step						= null;
	/** Create flags to select whether XML data blocks are to be shown */
	protected boolean					showGifGlobalColorTable		= true;
	protected boolean					showPdfPages				= true;
	protected boolean					showPdfImages				= true;
	protected boolean					showPdfOutlines				= true;
	protected boolean					showPdfFonts				= true;
	protected boolean					showTiffMixColorMap			= true;
	protected boolean					showTiffPhotoshopProperties	= true;
	protected boolean					showWaveAESAudioMetadata	= true;
	protected boolean					showHtmlLinks				= true;
	protected boolean					showHtmlImages				= true;
	protected boolean					showIso9660FileStructure	= true;
	protected boolean					jhoveIsVerbose				= false;
	protected boolean					errorIfSigmatchDiffers		= true;
	/** Temp Dir */
	protected String					tempDir;
	/** The list of igored but unkonwn to DIAS file formats */
	protected List<String>				acceptedUnknownFileFormats	= new ArrayList<String>();
	/** The LMER linkedTo list */
	protected HashMap<File, List<File>>	linkedToList;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Default constructor
	 * 
	 * @author Matthias Neubauer, German National Library
	 * @version 20050809
	 * @since 20050802
	 **************************************************************************/
	public MetadataGenerator() {

		/* Initialize the engine. */

		this._moduleList = new ArrayList<Module>();
		this._moduleMap = new TreeMap<String, Module>();
		this._checksum = false;
		this._signature = false;
		this._showRaw = false;

		this.linkedToList = new HashMap<File, List<File>>();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * Initialization of essential JHOVE parts from the JHOVE config file
	 * 
	 * @author Matthias Neubauer, German National Library
	 * @version 20050809
	 * @since 20050802
	 **************************************************************************/
	public void iniJhove() {
		File jhoveConfig = null;
		String err = null;

		this.linkedToList = new HashMap<File, List<File>>();

		// Read the modules from the JHOVE config file
		if (this.jhoveConfigFile != null) {
			jhoveConfig = new File(this.jhoveConfigFile);
			if (!jhoveConfig.exists() || !jhoveConfig.isFile()) {
				err = "Configuration file " + jhoveConfig.getAbsolutePath()
						+ " not found or not readable;";
				jhoveConfig = null;
			}

			if (jhoveConfig != null) {
				org.xml.sax.XMLReader parser = null;
				/* Use Java 1.4 methods to create default parser. */
				try {
					javax.xml.parsers.SAXParserFactory factory = javax.xml.parsers.SAXParserFactory
							.newInstance();
					factory.setNamespaceAware(true);
					parser = factory.newSAXParser().getXMLReader();
				}
				catch (SAXException e1) {
					e1.printStackTrace();
					this.step.setStatus(Status.WARNING, "Error instantiating "
							+ "parser for the Jhove config file");
					return;
				}
				catch (ParserConfigurationException e1) {
					e1.printStackTrace();
					this.step.setStatus(Status.WARNING, "Error instantiating "
							+ "parser for the Jhove config file");
					return;
				}

				ConfigHandler configHandler = new ConfigHandler();
				parser.setContentHandler(configHandler);

				// -Commented out because of several problems---------------
				// -on a linux-system!---------------------------------------
				/*
				 * Attempt to set schema awareness to avoid validation errors.
				 */
				/*
				 * try { parser.setFeature(
				 * "http://xml.org/sax/features/validation", true);
				 * 
				 * parser.setProperty ("http://java.sun.com/xml/jaxp/" +
				 * "properties/schemaLanguage",
				 * "http://www.w3.org/2001/XMLSchema"); } catch
				 * (org.xml.sax.SAXException e){ e.printStackTrace(); }
				 */
				// ---------------------------------------------------------
				try {
					parser.parse(this.jhoveConfigFile);
				}
				catch (IOException e) {
					e.printStackTrace();
					this.step.setStatus(Status.ERROR,
							"Cannot parse Jhove Configfile");
					return;
				}
				catch (org.xml.sax.SAXException sax) {
					sax.printStackTrace();
					this.step.setStatus(Status.ERROR, "Error parsing "
							+ "configuration file", new File(
							this.jhoveConfigFile));
					return;
				}

				/*
				 * Retrieve the ordered lists of modules and output handlers
				 */
				List list = configHandler.getModule();
				List params = configHandler.getModuleParams();
				int n = list.size();
				for (int i = 0; i < n; i++) {
					String[] tuple = (String[]) list.get(i);
					List param = (List) params.get(i);
					try {
						Class cl = Class.forName(tuple[0]);
						Module module = (Module) cl.newInstance();
						module.init(tuple[1]);
						module.setDefaultParams(param);

						this._moduleList.add(module);
						this._moduleMap.put(module.getName().toLowerCase(),
								module);
					}
					catch (Exception e) {
						if (err == null) {
							err = "cannot instantiate module: " + tuple[0];
						}
					}
				}
			}

			// If we found any error, the caller needs to deal with it.
			if (err != null) {
				this.step.setStatus(Status.ERROR, err);
				return;
			}
		}

		/***********************************************************************
		 * The Bytestream module is always statically loaded.
		 **********************************************************************/

		Module module = new edu.harvard.hul.ois.jhove.module.BytestreamModule();
		module.setDefaultParams(new ArrayList());
		this._moduleList.add(module);
		this._moduleMap.put(module.getName().toLowerCase(), module);
	}

	/***************************************************************************
	 * Starts the essential task of this class
	 * 
	 * @author Matthias Neubauer, German National Library
	 * @version 20050902
	 * @since 20050802
	 **************************************************************************/
	public void go() {
		long startTime = System.currentTimeMillis();
		this.step.setStatus(Status.RUNNING, "Initializing JHOVE");
		iniJhove();

		// If an Error occured during initialization, quit this module
		if (this.step.getStatus().getType() == Status.ERROR) {
			return;
		}

		this.step.setStatus(Status.RUNNING, "Creating technical metadata");
		// Create a new KopalHandler
		KopalHandler kopalhandler = new KopalHandler();
		try {
			Configurator.configure(kopalhandler, this);
		}
		catch (Exception e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Could not configure "
					+ "KopalHandler: " + e.getMessage());
			return;
		}

		// Get the metadata object for this SIP
		Uof md = null;
		try {
			md = (Uof) this.processData.getMetadata();
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Could not initialize metadata: "
					+ e.getMessage());
			return;
		}

		// Set the base path for the METS data
		md.setBasePath(new File(this.processData.getPathToContentFiles()));

		// Go through the filelist of the ProcessData and process each of the
		// files to generate metadata for it
		List filelist = this.processData.getFileList();
		if (filelist.size() > 0) {
			for (int i = 0; i < filelist.size(); i++) {
				File file = (File) filelist.get(i);

				// Set the contentfile for the handler
				kopalhandler.setContentFile(file);

				// Process the file with JHOVE
				try {
					if (!process(file, kopalhandler)) {
						this.step.setStatus(Status.ERROR, "File '"
								+ file.getName() + "' could not be processed",
								file);
						return;
					}
				}
				catch (Exception e) {
					e.printStackTrace();
					this.step.setStatus(Status.ERROR, "File '" + file.getName()
							+ "' could not be processed. System message: "
							+ e.getMessage(), file);
					return;
				}
			}

			if (!this.linkedToList.isEmpty()) {
				// Process found object-internal file dependencies
				for (int j = 0; j < filelist.size(); j++) {
					List linkedFiles = this.linkedToList.get(filelist.get(j));

					if (linkedFiles != null) { // if links exist
						try {
							// register the links within the metadata object
							md.addLmerFileLinkedTo((File) filelist.get(j),
									linkedFiles);
						}
						catch (Exception e) {
							e.printStackTrace();
							this.step.setStatus(Status.ERROR, "Adding "
									+ "dependencies failed");
							return;
						}
					}
				}
			}
		}
		else {
			// No files to process. Error!
			this.step.setStatus(Status.ERROR, "No file to analyze!");
			return;
		}

		this.step.setStatus(Status.DONE, "Metadata successfully created in "
				+ KopalTimestamp.getDurationInHours(System.currentTimeMillis()
						- startTime));
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ******

	/***************************************************************************
	 * Processes a file and reads its information into a RepInfo object, using
	 * the KopalHandler
	 * 
	 * @author Matthias Neubauer, German National Library
	 * @param file - The content file
	 * @param handler - The KopalHandler that processes the output
	 * @exception Exception
	 * @return false - if processing should be aborted.
	 **************************************************************************/
	protected boolean process(File file, KopalHandler handler) throws Exception {
		long lastModified = -1;

		// Create a RepInfo for this file.
		RepInfo info = new RepInfo(file.getAbsolutePath());

		if (!file.exists()) {
			this.step.setStatus(Status.ERROR, "File not found", file);
			return false;
		}
		else if (!file.isFile() || !file.canRead()) {
			this.step.setStatus(Status.ERROR, "File cannot be read", file);
			return false;
		}

		// If the handler can process the file...
		else if (handler.okToProcess(file.getAbsolutePath())) {
			// ...read its size...
			info.setSize(file.length());
			// ...and read the date of the last modification.
			if (lastModified < 0) {
				lastModified = file.lastModified();
			}
			info.setLastModified(new Date(lastModified));

			/*
			 * JHOVE Comment: Invoke all modules until one returns well formed.
			 * If a module doesn't know how to validate, we don't want to throw
			 * arbitrary files at it, so we'll skip it.
			 */
			Iterator iter = this._moduleList.iterator();
			while (iter.hasNext()) {
				Module mod = (Module) iter.next();
				RepInfo infc = (RepInfo) info.clone();

				if (mod.hasFeature("edu.harvard.hul.ois.jhove.canValidate")) {
					try {
						if (!processFile(mod, false, file, infc)) {
							return false;
						}

						if (infc.getWellFormed() == RepInfo.TRUE) {
							info.copy(infc);
							break;
						}

						// If the modules does not tag the file as
						// "well-formed", check, if the file signature matched,
						// and log the possible match and maybe existing
						// messages.
						if (infc.getSigMatch().contains(mod.getName())) {
							if (!infc.getMessage().isEmpty()) {
								this.step.setStatus(Status.WARNING, "File '"
										+ file.getName()
										+ "' is NOT well-formed for module "
										+ mod.getName()
										+ ". Filetype was presumed to be "
										+ infc.getFormat()
										+ ", but the according module"
										+ " printed some warnings", file);
								logJhoveMessages(infc.getMessage());
							}
							else {
								this.step.setStatus(Status.WARNING, "File '"
										+ file.getName()
										+ "' is NOT well-formed for module "
										+ mod.getName()
										+ ". Filetype was presumed to be "
										+ infc.getFormat()
										+ ", but the according module "
										+ "couldn't finish", file);
							}

							if (this.errorIfSigmatchDiffers) {
								// TODO: Make option available to only ignore
								// sigmatch errors in HTML files, as they
								// are never valid!

								return false;
							}
						}

						// -----------------------------------
						// JHOVE Comment:
						// We want to know what modules matched the
						// signature, so we force the sigMatch
						// property to be persistent.
						info.setSigMatch(infc.getSigMatch());
					}
					catch (Exception e) {
						/*
						 * JHOVE Comment: The assumption is that in trying to
						 * analyze the wrong type of file, the module may go off
						 * its track and throw an exception, so we just continue
						 * on to the next module.
						 */
						continue;
					}
				}
			}

			// Finally write all the gathered information into the LmerFile
			// objects.
			info.show(handler);

			// NOTE: If you get strange XML errors here (e.g. not allowed XML
			// chars), please check the validity of the techMD section and then
			// extend the correctUmlautsInMixSections() method in HTMLUtils with
			// the faulty char.
			JhoveDocument testJhove = JhoveDocument.Factory.parse(HTMLUtils
					.correctUmlautsInMixSections(handler.getXmlData()));
			Uof uof = (Uof) this.processData.getMetadata();

			// Add the file to the UOF if no file ID is existing right now. Just
			// update the file data, if a file ID already exists (the file
			// section then already was created by the PrepareMigration module).
			try {
				uof.getFileId(file);
				uof.updateFileData(file, testJhove.toString());
			}
			catch (Exception e) {
				uof.addFile(file, testJhove.toString());
			}

			LmerFile lmerfile = uof.getLmerFileCopy(uof.getFileTechMdId(file));

			// Check the file extension and go to error state if the extension
			// is unknown to DIAS or if there is none existing at all.
			// Extensions from the acceptedUnknownFileFormats list are excepted
			// in every case.
			if (lmerfile.getFormatArray(0).getStringValue().equalsIgnoreCase(
					FormatRegistry.UNKNOWN_FILETYPE)) {
				String formatstring = info.getFormat();
				if (info.getVersion() != null) {
					formatstring += " " + info.getVersion();
				}

				// Check if the file actually HAS a file extension.
				int hasExtension = file.getName().lastIndexOf('.');
				String fileExtension = "";

				if (hasExtension != -1) {
					fileExtension = file.getName().substring(hasExtension);
				}

				// No file extension at all!
				if (!this.acceptedUnknownFileFormats.contains("ALL")
						&& fileExtension.equals("")) {
					this.step.setStatus(Status.ERROR, "SIP contains "
							+ "file format unknown to DIAS with NO "
							+ "file extension", file);
					return false;
				}

				// Unaccepted file extension!
				if (!this.acceptedUnknownFileFormats.contains("ALL")
						&& !this.acceptedUnknownFileFormats
								.contains(fileExtension)) {
					this.step.setStatus(Status.ERROR, "SIP contains "
							+ "file format unknown to DIAS with "
							+ "file extension " + fileExtension, file);
					return false;
				}

				// Accepted file extension!
				this.step.setStatus(Status.RUNNING, "SIP contains "
						+ "file format unknown to DIAS, but "
						+ "file extension " + fileExtension
						+ " is permitted to be ingested");
				return true;
			}
		}

		return true;
	}

	/***************************************************************************
	 * Original JHOVE Comment: Processes the file. Returns false if aborted or
	 * if the module is incapable of validation. This shouldn't be called if the
	 * module doesn't have the validation feature.
	 * 
	 * @author Matthias Neubauer, German National Library
	 * @version 20050809
	 * @since 20050802
	 * @param module - The Module to use
	 * @param verbose - Used verbosity, true if maximum, false if minimum
	 * @param file - The file to process
	 * @param info - RepInfo object to hold the information
	 * @exception Exception
	 * @return false if aborted, or if the module is incapable of validation.
	 **************************************************************************/
	protected boolean processFile(Module module, boolean verbose, File file,
			RepInfo info) throws Exception {

		// Abort if the selected module is incapable of validation
		if (!module.hasFeature("edu.harvard.hul.ois.jhove.canValidate")) {
			return false;
		}

		// -Creating a JhoveBase an App, because the modules need them------
		JhoveBase je = new JhoveBase();
		je.setTempDirectory(this.tempDir);
		je.setChecksumFlag(this._checksum);
		je.setShowRawFlag(this._showRaw);
		je.setSignatureFlag(this._signature);

		module.setBase(je);

		App app = new App(this.NAM, this.REL, MetadataGenerator.DATE, this.USE,
				this.RIGHT);
		module.setApp(app);
		// ----------------------------------------------------------------

		module.setVerbosity(verbose ? Module.MAXIMUM_VERBOSITY
				: Module.MINIMUM_VERBOSITY);
		module.applyDefaultParams();
		if (module.isRandomAccess()) {

			/* Module needs random access input. */

			RandomAccessFile raf = new RandomAccessFile(file, "r");
			if (this._signature) {
				module.checkSignatures(file, raf, info);
			}
			else {
				module.parse(raf, info);
			}
			raf.close();
		}
		else {

			/* Module accepts stream input. */

			InputStream stream = new FileInputStream(file);
			if (this._signature) {
				module.checkSignatures(file, stream, info);
			}
			else {
				int parseIndex = module.parse(stream, info, 0);
				/*
				 * If parse returns non-zero, reparse with a fresh stream.
				 */
				while (parseIndex != 0) {
					stream.close();
					stream = new FileInputStream(file);
					parseIndex = module.parse(stream, info, parseIndex);
				}
			}
			stream.close();
		}
		return true; // Successful processing
	}

	/***************************************************************************
	 * Log the Jhove Info- and ErrorMessages.
	 * 
	 * @author Matthias Neubauer, German National Library
	 * @param messages - The list of messages
	 **************************************************************************/
	protected void logJhoveMessages(List messages) {
		ListIterator iter = messages.listIterator();

		while (iter.hasNext()) {
			Message m = (Message) iter.next();

			if (m.getClass().getName().endsWith("ErrorMessage")) {
				this.step.setStatus(Status.WARNING, "Jhove ErrorMessage: "
						+ m.getMessage()
						+ (m.getSubMessage() != null ? ", Submessage: "
								+ m.getSubMessage() : "")
						+ (m.getOffset() != -1 ? ", Offset: " + m.getOffset()
								: ""));
			}
			else if (this.jhoveIsVerbose) {
				this.step.setStatus(Status.RUNNING, "Jhove InfoMessage: "
						+ m.getMessage()
						+ (m.getSubMessage() != null ? ", Submessage: "
								+ m.getSubMessage() : "")
						+ (m.getOffset() != -1 ? ", Offset: " + m.getOffset()
								: ""));
			}
		}
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @return Returns the showPdfPages.
	 **************************************************************************/
	public boolean isShowPdfPages() {
		return this.showPdfPages;
	}

	/***************************************************************************
	 * @param _showPdfPages The showPdfPages to set.
	 **************************************************************************/
	public void setShowPdfPages(boolean _showPdfPages) {
		this.showPdfPages = _showPdfPages;
	}

	/***************************************************************************
	 * @param _showGifGlobalColorTable The showGifGlobalColorTable to set.
	 **************************************************************************/
	public void setShowGifGlobalColorTable(boolean _showGifGlobalColorTable) {
		this.showGifGlobalColorTable = _showGifGlobalColorTable;
	}

	/***************************************************************************
	 * @param _showPdfImages The showPdfImages to set.
	 **************************************************************************/
	public void setShowPdfImages(boolean _showPdfImages) {
		this.showPdfImages = _showPdfImages;
	}

	/***************************************************************************
	 * @param _showPdfFonts The showPdfOutlines to set.
	 **************************************************************************/
	public void setShowPdfFonts(boolean _showPdfFonts) {
		this.showPdfFonts = _showPdfFonts;
	}

	/***************************************************************************
	 * @param _showPdfOutlines The showPdfOutlines to set.
	 **************************************************************************/
	public void setShowPdfOutlines(boolean _showPdfOutlines) {
		this.showPdfOutlines = _showPdfOutlines;
	}

	/***************************************************************************
	 * @param _showHtmlImages The showHtmlImages to set.
	 **************************************************************************/
	public void setShowHtmlImages(boolean _showHtmlImages) {
		this.showHtmlImages = _showHtmlImages;
	}

	/***************************************************************************
	 * @param _showHtmlLinks The showHtmlLinks to set.
	 **************************************************************************/
	public void setShowHtmlLinks(boolean _showHtmlLinks) {
		this.showHtmlLinks = _showHtmlLinks;
	}

	/***************************************************************************
	 * @param _showTiffMixColorMap The showTiffMixColorMap to set.
	 **************************************************************************/
	public void setShowTiffMixColorMap(boolean _showTiffMixColorMap) {
		this.showTiffMixColorMap = _showTiffMixColorMap;
	}

	/***************************************************************************
	 * @param _showTiffPhotoshopProperties The showTiffPhotoshopProperties to
	 *            set.
	 **************************************************************************/
	public void setShowTiffPhotoshopProperties(
			boolean _showTiffPhotoshopProperties) {
		this.showTiffPhotoshopProperties = _showTiffPhotoshopProperties;
	}

	/***************************************************************************
	 * @param _showWaveAESAudioMetadata The showWaveAESAudioMetadata to set.
	 **************************************************************************/
	public void setShowWaveAESAudioMetadata(boolean _showWaveAESAudioMetadata) {
		this.showWaveAESAudioMetadata = _showWaveAESAudioMetadata;
	}

	/***************************************************************************
	 * @param _showIso9660FileStructure The showIso9660FileStructure to set.
	 **************************************************************************/
	public void setShowIso9660FileStructure(boolean _showIso9660FileStructure) {
		this.showIso9660FileStructure = _showIso9660FileStructure;
	}

	/***************************************************************************
	 * @param _tempDir The tempDir to set.
	 **************************************************************************/
	public void setTempDir(String _tempDir) {
		this.tempDir = _tempDir;
	}

	/***************************************************************************
	 * @param _processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData _processData) {
		this.processData = _processData;
	}

	/***************************************************************************
	 * @param _step The step to set.
	 **************************************************************************/
	public void setStep(Step _step) {
		this.step = _step;
	}

	/***************************************************************************
	 * @param _jhoveConfigFile The jhoveConfigFile to set.
	 **************************************************************************/
	public void setJhoveConfigFile(String _jhoveConfigFile) {
		this.jhoveConfigFile = _jhoveConfigFile;
	}

	/***************************************************************************
	 * Adds each element of the list used in the main config file to the list of
	 * unknown but accepted file formats.
	 * 
	 * @author Stefan Funk, SUB Göttingen 16.01.2007
	 * @param element
	 **************************************************************************/
	public void setAcceptedUnknownFileFormats(String element) {
		this.acceptedUnknownFileFormats.add(element);
	}

	/***************************************************************************
	 * If set to TRUE, JHOVE info messages will be logged, too.
	 * 
	 * @param verbose The verbose to set.
	 * @author Stefan Funk, SUB Göttingen, 15.02.2007
	 **************************************************************************/
	public void setJhoveIsVerbose(boolean verbose) {
		this.jhoveIsVerbose = verbose;
	}

	/***************************************************************************
	 * If set to TRUE, go to error state if JHOVE detects a file with a sigmatch
	 * that doesn't fit to the file type matching module.
	 * 
	 * @param _errorIfSigmatchDiffers The errorIfSigmatchDiffers to set.
	 * @author Stefan Funk, SUB Göttingen, 21.02.2007
	 **************************************************************************/
	public void setErrorIfSigmatchDiffers(boolean _errorIfSigmatchDiffers) {
		this.errorIfSigmatchDiffers = _errorIfSigmatchDiffers;
	}

	/***************************************************************************
	 * OutputHandler for JHOVE specialized for structure, inherited from
	 * edu.harvard.hul.ois.jhove.handler.XmlHandler
	 * 
	 * @author Matthias Neubauer, German National Library
	 * @version 20051028
	 * @see edu.harvard.hul.ois.jhove.handler.XmlHandler
	 * @see de.langzeitarchivierung.kopal.actionmodule.MetadataGenerator
	 * @since 20050729
	 **************************************************************************/
	protected class KopalHandler extends
			edu.harvard.hul.ois.jhove.handler.XmlHandler
	{

		// CONSTANTS *****************************************************

		protected final String			EOL				= System
																.getProperty("line.separator");

		protected static final String	NAME			= "KOPAL";

		// Version is set to 1.3 to be consistent with the version of the
		// XMLHandler, which is relevant to the path of the XML Schemas
		protected static final String	RELEASE			= "1.3";
		protected static final String	NOTE			= "OutputHandler to retrieve "
																+ "Metadata";
		protected static final String	RIGHTS			= "Copyright 2005-2007 by "
																+ "Project KOPAL http://"
																+ "kopal.langzeit"
																+ "archivierung.de/";

		// CLASS VARIABLES (Static variables) *****************************

		// STATE (Instance variables) ************************************

		protected String				sourceFileType	= "";

		protected StringBuffer			xmldata			= new StringBuffer("");

		protected File					contentfile;

		// CREATION (Constructors, factory methods, static/inst init) *****

		/***********************************************************************
		 * Constructor with a given LmerSIP-Object
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050824
		 * @since 20050729
		 **********************************************************************/
		public KopalHandler() {
			super(NAME, RELEASE, MetadataGenerator.DATE, NOTE, RIGHTS);
			Agent agent = new Agent("German National Library",
					AgentType.EDUCATIONAL);
			agent.setAddress("Adickesallee 1, " + "60322 Frankfurt am Main, "
					+ "Germany");
			agent.setTelephone("+49 (69) 1525-0");
			agent.setEmail("postfach@dbf.dbb.de");
			this._vendor = agent;
		}

		// MANIPULATION (Manipulation - what the object does) *************

		/***********************************************************************
		 * Method to retrieve the relevant metadata out of a RepInfo object and
		 * store it into a LmerData object
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20051027
		 * @since 20050729
		 * @param info - The RepInfo that holds the metadata
		 **********************************************************************/
		@Override
		public void show(RepInfo info) {

			String margin = getIndent(++this._level);
			String margn2 = margin + " ";
			String margn3 = margn2 + " ";

			this.sourceFileType = info.getFormat();

			// empty StringBuffer before writing new entries
			this.xmldata = new StringBuffer("");

			// Write the XML header
			showHeader();

			// Write module specific informations
			Module module = info.getModule();
			if (module != null) {
				String[][] attr = { { "release", module.getRelease() },
						{ "date", date.format(module.getDate()) } };
				this.xmldata.append(margn3
						+ element("reportingModule", attr, module.getName())
						+ System.getProperty("line.separator"));
			}

			String format = info.getFormat();
			String version = info.getVersion();

			// If the format is textbased, change it to "TXT" to match Kopal
			// purposes
			if (format.equals("ASCII") || format.equals("UTF-8")) {
				info.setFormat("TXT File");
				format = info.getFormat();
			}

			// If the format is JPEG 2000, change it to "JPX" or "JP2" depending
			// on its profile to match Kopal purposes
			if (format.equals("JPEG 2000")) {
				String profile = (String) info.getProfile().get(0);
				if (profile.equalsIgnoreCase("jpx")) {
					info.setFormat("JPX");
				}
				if (profile.equalsIgnoreCase("jp2")) {
					info.setFormat("JP2");
				}
				format = info.getFormat();
			}

			// Write the format
			if (format != null) {
				this.xmldata.append(margn3 + element("format", format)
						+ System.getProperty("line.separator"));
			}
			// Write the version
			if (version != null) {
				this.xmldata.append(margn3 + element("version", version)
						+ System.getProperty("line.separator"));
			}

			// Write the profile if format is PS
			if (format.equals("PS")) {
				List profiles = info.getProfile();
				if (!profiles.isEmpty()) {
					String prof = info.getProfile().get(0).toString();
					if (prof != null) {
						this.xmldata.append(margn3 + element("profile", prof)
								+ System.getProperty("line.separator"));
					}
				}
			}

			// Write the status of the file
			String wfStr;
			switch (info.getWellFormed()) {
				case RepInfo.TRUE:
					wfStr = "Well-Formed";
					break;

				case RepInfo.FALSE:
					wfStr = "Not well-formed";
					break;

				default:
					wfStr = "Unknown";
					break;
			}
			// If it's well-formed, append validity information
			if (info.getWellFormed() == RepInfo.TRUE) {
				switch (info.getValid()) {
					case RepInfo.TRUE:
						wfStr += " and valid";
						break;

					case RepInfo.FALSE:
						wfStr += ", but not valid";
						break;

					// case UNDETERMINED: add nothing
				}
			}
			this.xmldata.append(margn3 + element("status", wfStr)
					+ System.getProperty("line.separator"));

			// ------------------------------------------------------------
			List list = info.getSigMatch();
			int n = list.size();
			if (n > 0) {
				this.xmldata.append(margn3 + elementStart("sigMatch"));
				this._level++;
				for (int i = 0; i < n; i++) {
					this.xmldata.append(margn3
							+ element("module", (String) list.get(i)));
				}
				this._level--;
				this.xmldata.append(margn3 + elementEnd("sigMatch"));
			}

			list = info.getMessage();
			n = list.size();
			if (n > 0) {
				this.xmldata.append(margn3 + elementStart("messages"));
				this._level++;
				for (int i = 0; i < n; i++) {
					showMessage((Message) list.get(i));
				}
				this._level--;
				this.xmldata.append(margn3 + elementEnd("messages"));
			}
			// -------------------------------------------------------------

			String mime = info.getMimeType();
			if (mime != null) {
				this.xmldata.append(margn3 + element("mimeType", mime));
			}

			Map map = info.getProperty();

			// parse links in HTML files seperately for LMER linkedTo fields
			if (format != null
					&& (format.equals("HTML") || format.equals("XHTML"))) {
				if (map != null) {
					if (map.size() > 0) {
						Iterator iter = map.keySet().iterator();
						while (iter.hasNext()) {
							String key = (String) iter.next();
							Property property = info.getProperty(key);
							getMetadataFromHtml(property);
						}
					}
				}
			}

			// write the property tags of the file
			if (map != null) {

				if (map.size() > 0) {
					this.xmldata.append(margn3 + elementStart("properties")
							+ this.EOL);
					Iterator iter = map.keySet().iterator();
					while (iter.hasNext()) {
						String key = (String) iter.next();
						Property property = info.getProperty(key);
						showProperty(property);
					}
					this.xmldata.append(margn3 + elementEnd("properties")
							+ this.EOL);
				}
			}

			// Show XML footer
			showFooter();

			// Erase references to memory filling objects,
			// so GC can collect them
			map = null;
			this.contentfile = null;

			// reduce the margin-level again
			--this._level;
		}

		/***********************************************************************
		 * Method to represent property info, in form of xml, into a string for
		 * this.xmldata part
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050905
		 * @since 20050830
		 **********************************************************************/
		@Override
		public void showHeader() {
			String margin = getIndent(++this._level);

			String[][] attrs = {
					{ "xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance" },
					{ "xmlns", "http://hul.harvard.edu/ois/xml/ns/jhove" },
					{
							"xsi:schemaLocation",
							"http://hul.harvard.edu/ois/xml/ns/jhove "
									+ "http://hul.harvard.edu/ois/xml/xsd/jhove/"
									+ RELEASE + "/jhove.xsd" },
					{ "name", NAME }, { "release", RELEASE },
					{ "date", HandlerBase.date.format(this._date) } };
			this.xmldata.append(elementStart("jhove", attrs)
					+ System.getProperty("line.separator"));
			this.xmldata.append(margin
					+ element("date", toDateTime(new Date()))
					+ System.getProperty("line.separator"));
			this.xmldata.append(margin + elementStart("repInfo")
					+ System.getProperty("line.separator"));
		}

		/***********************************************************************
		 * Method to represent property info, in form of xml, into a string for
		 * this.xmldata part
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050831
		 * @since 20050830
		 **********************************************************************/
		@Override
		public void showFooter() {
			String margin = getIndent(this._level--);
			this.xmldata.append(margin + elementEnd("repInfo")
					+ System.getProperty("line.separator"));
			this.xmldata.append(elementEnd("jhove")
					+ System.getProperty("line.separator"));
		}

		/***********************************************************************
		 * Method to represent property info, in form of xml, into a string for
		 * this.xmldata part
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050831
		 * @since 20050801
		 * @param property - The property to be represented
		 **********************************************************************/
		@Override
		protected void showProperty(Property property) {
			String margin = getIndent(++this._level);
			String margn2 = margin + " ";
			String margn3 = margn2 + " ";
			String margn4 = margn3 + " ";

			// --------------------------------------
			// Switch some XML blocks on or off

			String propName = property.getName();
			if (propName.equals("GlobalColorTable")
					&& !MetadataGenerator.this.showGifGlobalColorTable
					&& this.sourceFileType.equalsIgnoreCase("GIF")) {
				--this._level;
				return;
			}
			if (propName.equals("Pages")
					&& !MetadataGenerator.this.showPdfPages
					&& this.sourceFileType.equalsIgnoreCase("PDF")) {
				// ++this._level;
				buildPageStatistics(property);
				--this._level;
				return;
			}
			if (propName.equals("Images")
					&& !MetadataGenerator.this.showPdfImages
					&& this.sourceFileType.equalsIgnoreCase("PDF")) {
				buildImageStatistics(property);
				--this._level;
				return;
			}
			if (propName.equals("Fonts")
					&& !MetadataGenerator.this.showPdfFonts
					&& this.sourceFileType.equalsIgnoreCase("PDF")) {
				buildFontStatistics(property);
				--this._level;
				return;
			}
			if (propName.equals("Outlines")
					&& !MetadataGenerator.this.showPdfOutlines
					&& this.sourceFileType.equalsIgnoreCase("PDF")) {
				--this._level;
				return;
			}
			if (propName.equals("PhotoshopProperties")
					&& !MetadataGenerator.this.showTiffPhotoshopProperties
					&& this.sourceFileType.equalsIgnoreCase("TIFF")) {
				--this._level;
				return;
			}
			if (propName.equals("Links")
					&& !MetadataGenerator.this.showHtmlLinks
					&& this.sourceFileType.equalsIgnoreCase("HTML")) {
				--this._level;
				return;
			}
			if (propName.equals("Images")
					&& !MetadataGenerator.this.showHtmlImages
					&& this.sourceFileType.equalsIgnoreCase("HTML")) {
				--this._level;
				return;
			}
			if (propName.equals("File Structure")
					&& !MetadataGenerator.this.showIso9660FileStructure
					&& this.sourceFileType.equalsIgnoreCase("ISO9660")) {
				buildFileStructureStatistics(property);
				--this._level;
				return;
			}

			// ---------------------------------------------

			PropertyArity arity = property.getArity();
			PropertyType type = property.getType();
			boolean valueIsProperty = (type.equals(PropertyType.PROPERTY));
			boolean valueIsNiso = (type.equals(PropertyType.NISOIMAGEMETADATA));
			boolean valueIsAes = (type.equals(PropertyType.AESAUDIOMETADATA));

			String[][] propAttrs = new String[2][];
			propAttrs[0] = new String[] { "arity", arity.toString() };
			propAttrs[1] = new String[] { "type", type.toString() };
			this.xmldata.append(margn2 + elementStart("property") + this.EOL);
			this.xmldata.append(margn3 + element("name", property.getName())
					+ this.EOL);
			this.xmldata.append(margn3 + elementStart("values", propAttrs)
					+ this.EOL);
			if (arity.equals(PropertyArity.SCALAR)) {
				/* Just a single value */
				if (valueIsProperty) {
					showProperty((Property) property.getValue());
				}
				else if (valueIsNiso) {
					this.xmldata.append(margn4 + elementStart("value")
							+ this.EOL);
					showNisoImageMetadata((NisoImageMetadata) property
							.getValue(), margn4);
					this.xmldata
							.append(margn4 + elementEnd("value") + this.EOL);
				}
				else if (valueIsAes
						&& MetadataGenerator.this.showWaveAESAudioMetadata) {
					this.xmldata.append(margn4 + elementStart("value")
							+ this.EOL);
					showAESAudioMetadata(
							(AESAudioMetadata) property.getValue(), margn4);
					this.xmldata
							.append(margn4 + elementEnd("value") + this.EOL);
				}
				else {
					this.xmldata.append(margn4
							+ element("value", property.getValue().toString())
							+ this.EOL);
				}
			}
			else if (arity.equals(PropertyArity.LIST)) {
				List propList = (List) property.getValue();
				ListIterator iter = propList.listIterator();
				while (iter.hasNext()) {
					Object val = iter.next();
					if (valueIsProperty) {
						showProperty((Property) val);
					}
					else if (valueIsNiso) {
						this.xmldata.append(margn4 + elementStart("value")
								+ this.EOL);
						showNisoImageMetadata((NisoImageMetadata) property
								.getValue(), margn4);
						this.xmldata.append(margn4 + elementEnd("value")
								+ this.EOL);
					}
					else if (valueIsAes
							&& MetadataGenerator.this.showWaveAESAudioMetadata) {
						this.xmldata.append(margn4 + elementStart("value")
								+ this.EOL);
						showAESAudioMetadata((AESAudioMetadata) property
								.getValue(), margn4);
						this.xmldata.append(margn4 + elementEnd("value")
								+ this.EOL);
					}
					else {
						this.xmldata.append(margn4
								+ element("value", valueToString(val))
								+ this.EOL);
					}
				}
			}
			else if (arity.equals(PropertyArity.MAP)) {
				/*
				 * For a map, the key is the "key" attribute of its
				 * corresponding value
				 */
				Map propMap = (Map) property.getValue();
				Iterator keyIter = propMap.keySet().iterator();
				while (keyIter.hasNext()) {
					Object key = keyIter.next();
					String keystr = key.toString();
					Object val = propMap.get(key);
					String[][] attrs = new String[1][];
					String[] keyAttr = new String[2];
					keyAttr[0] = "key";
					keyAttr[1] = keystr;
					attrs[0] = keyAttr;
					if (valueIsProperty) {

						Property pval = (Property) val;
						// If the key equals the property name,
						// suppress the key

						if (pval.getName().equals(keystr)) {
							this.xmldata.append(margn4 + elementStart("value"));
						}
						else {
							this.xmldata.append(margn4
									+ elementStart("value", attrs));
						}
						showProperty(pval);
						this.xmldata.append(margn4 + elementEnd("value")
								+ this.EOL);

					}
					else if (valueIsNiso) {
						this.xmldata.append(margn4 + elementStart("value")
								+ this.EOL);
						showNisoImageMetadata((NisoImageMetadata) val, margn4);
						this.xmldata.append(margn4 + elementEnd("value")
								+ this.EOL);
					}
					else if (valueIsAes
							&& MetadataGenerator.this.showWaveAESAudioMetadata) {
						this.xmldata.append(margn4 + elementStart("value")
								+ this.EOL);
						showAESAudioMetadata((AESAudioMetadata) val, margn4);
						this.xmldata.append(margn4 + elementEnd("value")
								+ this.EOL);
					}
					else {
						this.xmldata.append(margn4
								+ element("value", attrs, valueToString(val))
								+ this.EOL);
					}
				}
			}
			else if (arity.equals(PropertyArity.SET)) {
				Set propSet = (Set) property.getValue();
				Iterator iter = propSet.iterator();
				while (iter.hasNext()) {
					Object val = iter.next();
					if (valueIsProperty) {
						showProperty((Property) val);
					}
					else {
						this.xmldata.append(margn4
								+ element("value", valueToString(val))
								+ this.EOL);
					}
				}
			}
			else if (arity.equals(PropertyArity.ARRAY)) {
				showArrayProperty(property, margn4);
			}
			this.xmldata.append(margn3 + elementEnd("values") + this.EOL);
			this.xmldata.append(margn2 + elementEnd("property") + this.EOL);
			--this._level;
		}


		/***********************************************************************
		 * Original JHOVE Comment: The array property has so many special cases
		 * of its own that we break it out of showProperty
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050831
		 * @since 20050801
		 * @param property - The property to be represented
		 * @param margin - The margin of the XML-Tags
		 **********************************************************************/
		@Override
		protected void showArrayProperty(Property property, String margin) {
			boolean[] boolArray = null;
			byte[] byteArray = null;
			char[] charArray = null;
			java.util.Date[] dateArray = null;
			double[] doubleArray = null;
			float[] floatArray = null;
			int[] intArray = null;
			long[] longArray = null;
			Object[] objArray = null;
			Property[] propArray = null;
			short[] shortArray = null;
			String[] stringArray = null;
			Rational[] rationalArray = null;
			NisoImageMetadata[] nisoArray = null;
			AESAudioMetadata[] aesArray = null;
			int n = 0;

			PropertyType propType = property.getType();
			if (PropertyType.BOOLEAN.equals(propType)) {
				boolArray = (boolean[]) property.getValue();
				n = boolArray.length;
			}
			else if (PropertyType.BYTE.equals(propType)) {
				byteArray = (byte[]) property.getValue();
				n = byteArray.length;
			}
			else if (PropertyType.CHARACTER.equals(propType)) {
				charArray = (char[]) property.getValue();
				n = charArray.length;
			}
			else if (PropertyType.DATE.equals(propType)) {
				dateArray = (java.util.Date[]) property.getValue();
				n = dateArray.length;
			}
			else if (PropertyType.DOUBLE.equals(propType)) {
				doubleArray = (double[]) property.getValue();
				n = doubleArray.length;
			}
			else if (PropertyType.FLOAT.equals(propType)) {
				floatArray = (float[]) property.getValue();
				n = floatArray.length;
			}
			else if (PropertyType.INTEGER.equals(propType)) {
				intArray = (int[]) property.getValue();
				n = intArray.length;
			}
			else if (PropertyType.LONG.equals(propType)) {
				longArray = (long[]) property.getValue();
				n = longArray.length;
			}
			else if (PropertyType.OBJECT.equals(propType)) {
				objArray = (Object[]) property.getValue();
				n = objArray.length;
			}
			else if (PropertyType.SHORT.equals(propType)) {
				shortArray = (short[]) property.getValue();
				n = shortArray.length;
			}
			else if (PropertyType.STRING.equals(propType)) {
				stringArray = (String[]) property.getValue();
				n = stringArray.length;
			}
			else if (PropertyType.RATIONAL.equals(propType)) {
				rationalArray = (Rational[]) property.getValue();
				n = rationalArray.length;
			}
			else if (PropertyType.PROPERTY.equals(propType)) {
				propArray = (Property[]) property.getValue();
				n = propArray.length;
			}
			else if (PropertyType.NISOIMAGEMETADATA.equals(propType)) {
				nisoArray = (NisoImageMetadata[]) property.getValue();
				n = nisoArray.length;
			}
			else if (PropertyType.AESAUDIOMETADATA.equals(propType)) {
				aesArray = (AESAudioMetadata[]) property.getValue();
				n = nisoArray.length;
			}

			for (int i = 0; i < n; i++) {
				String elem;
				if (PropertyType.BOOLEAN.equals(propType)) {
					elem = String.valueOf(boolArray[i]);
				}
				else if (PropertyType.BYTE.equals(propType)) {
					elem = String.valueOf(byteArray[i]);
				}
				else if (PropertyType.CHARACTER.equals(propType)) {
					elem = String.valueOf(charArray[i]);
				}
				else if (PropertyType.DATE.equals(propType)) {
					elem = dateArray[i].toString();
				}
				else if (PropertyType.DOUBLE.equals(propType)) {
					elem = String.valueOf(doubleArray[i]);
				}
				else if (PropertyType.FLOAT.equals(propType)) {
					elem = String.valueOf(floatArray[i]);
				}
				else if (PropertyType.INTEGER.equals(propType)) {
					elem = String.valueOf(intArray[i]);
				}
				else if (PropertyType.LONG.equals(propType)) {
					elem = String.valueOf(longArray[i]);
				}
				else if (PropertyType.OBJECT.equals(propType)) {
					elem = valueToString(objArray[i]);
				}
				else if (PropertyType.SHORT.equals(propType)) {
					elem = String.valueOf(shortArray[i]);
				}
				else if (PropertyType.STRING.equals(propType)) {
					elem = stringArray[i];
				}
				else if (PropertyType.RATIONAL.equals(propType)) {
					elem = rationalArray[i].toString();
				}
				else if (PropertyType.PROPERTY.equals(propType)) {
					showProperty(propArray[i]);
					continue;
				}
				else if (PropertyType.NISOIMAGEMETADATA.equals(propType)) {
					showNisoImageMetadata(nisoArray[i], margin);
					continue;
				}
				else if (PropertyType.AESAUDIOMETADATA.equals(propType)) {
					showAESAudioMetadata(aesArray[i], margin);
					continue;
				}
				else
					elem = "<error>";
				this.xmldata.append(margin + element("value", elem) + this.EOL);
			}
		}

		/***********************************************************************
		 * Original JHOVE Comment: Display the NISO image metadata formatted
		 * according to the MIX schema.
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050831
		 * @since 20050801
		 * @param niso - NISO image metadata
		 * @param margn - The margin for the mix-block
		 **********************************************************************/
		protected void showNisoImageMetadata(NisoImageMetadata niso,
				String margn) {
			String margin = " " + margn;

			String[][] attrs = {
					{ "xmlns:mix", "http://www.loc.gov/mix/" },
					{ "xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance" },
					{
							"xsi:schemaLocation",
							"http://www.loc.gov/mix/"
									+ " http://www.loc.gov/mix/mix.xsd" } };
			this.xmldata.append(margin + elementStart("mix:mix", attrs)
					+ this.EOL);

			showNisoBasicImageParameters(niso, margin);
			showNisoImageCreation(niso, margin);
			showNisoImagingPerformanceAssessment(niso, margin);
			showNisoChangeHistory(niso, margin);

			this.xmldata.append(margin + elementEnd("mix:mix") + this.EOL);

			this._level--;
		}

		/***********************************************************************
		 * Original JHOVE Comment: The NISO Metadata output is split into
		 * multiple functions so that they're merely outrageously big rather
		 * than disgustingly big Top level element 1 of 4: BasicImageParameters
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050801
		 * @since 20050801
		 * @param niso NISO image metadata
		 * @param margin - The margin of the XML-Tags
		 **********************************************************************/
		@Override
		protected void showNisoBasicImageParameters(NisoImageMetadata niso,
				String margin) {
			String margn2 = margin + " ";
			String margn3 = margn2 + " ";
			String margn4 = margn3 + " ";
			String margn5 = margn4 + " ";
			String margn6 = margn5 + " ";

			this.xmldata.append(margn2
					+ elementStart("mix:" + "BasicImageParameters") + this.EOL);

			// Here things get rather deeply nested, and we want to be
			// able to back out if it turns out that we really have no
			// content to deliver. So rather than writing directly to
			// the writer, we write to a StringBuffer that may or may
			// not get used.
			StringBuffer fBuf = new StringBuffer(margn3
					+ elementStart("mix:Format") + this.EOL);
			boolean useFBuf = false; // set to true if some data showup
			String s = niso.getMimeType();
			if (s != null) {
				fBuf.append(margn4 + element("mix:MIMEType", s) + this.EOL);
				useFBuf = true;
			}
			if ((s = niso.getByteOrder()) != null) {
				fBuf.append(margn4 + element("mix:ByteOrder", s) + this.EOL);
				useFBuf = true;
			}
			int comp = niso.getCompressionScheme();
			int level = niso.getCompressionLevel();
			if (comp != NisoImageMetadata.NULL
					|| level != NisoImageMetadata.NULL) {
				fBuf
						.append(margn4 + elementStart("mix:Compression")
								+ this.EOL);
				if (comp != NisoImageMetadata.NULL) {
					fBuf.append(margn5
							+ element("mix:CompressionScheme", Integer
									.toString(comp)) + this.EOL);
				}
				if (level != NisoImageMetadata.NULL) {
					fBuf.append(margn5
							+ element("mix:CompressionLevel", Integer
									.toString(level)) + this.EOL);
				}
				fBuf.append(margn4 + elementEnd("mix:Compression") + this.EOL);
				useFBuf = true;
			}
			// Nest photometric interpretation tentative buffer in fBuf
			StringBuffer piBuf = new StringBuffer(margn4
					+ elementStart("mix:PhotometricInterpretation") + this.EOL);
			boolean usePIBuf = false;
			int n = niso.getColorSpace();
			if (n != NisoImageMetadata.NULL) {
				piBuf.append(margn5
						+ element("mix:ColorSpace", Integer.toString(n))
						+ this.EOL);
				usePIBuf = true;
			}
			String s2 = niso.getProfileURL();
			if ((s = niso.getProfileName()) != null || s2 != null) {
				piBuf
						.append(margn5 + elementStart("mix:ICCProfile")
								+ this.EOL);
				if (s != null) {
					piBuf.append(margn6 + element("mix:ProfileName", s)
							+ this.EOL);
				}
				if (s2 != null) {
					piBuf.append(margn6 + element("mix:ProfileURL", s)
							+ this.EOL);
				}
				piBuf.append(margn5 + elementEnd("mix:ICCProfile"));
				usePIBuf = true;
			}
			int[] iarray = niso.getYCbCrSubSampling();
			if (iarray != null) {
				piBuf.append(margn5
						+ element("mix:YCbCrSubSampling", integerArray(iarray))
						+ this.EOL);
				usePIBuf = true;
			}
			if ((n = niso.getYCbCrPositioning()) != NisoImageMetadata.NULL) {
				piBuf.append(margn5
						+ element("mix:YCbCrPositioning", Integer.toString(n))
						+ this.EOL);
				usePIBuf = true;
			}
			Rational[] rarray = niso.getYCbCrCoefficients();
			if (rarray != null) {
				piBuf.append(margn5
						+ element("mix:YCbCrCoefficients",
								rationalArray(rarray)) + this.EOL);
				usePIBuf = true;
			}
			if ((rarray = niso.getReferenceBlackWhite()) != null) {
				piBuf.append(margn5
						+ element("mix:ReferenceBlackWhite",
								rationalArray(rarray)) + this.EOL);
				usePIBuf = true;
			}
			piBuf.append(margn4
					+ elementEnd("mix:" + "PhotometricInterpretation")
					+ this.EOL);
			if (usePIBuf) {
				fBuf.append(piBuf);
				useFBuf = true;
			}

			// Now a tentative buffer for the Segments element
			StringBuffer segBuf = new StringBuffer(margn4
					+ elementStart("mix:Segments") + this.EOL);
			boolean useSegBuf = false;
			if ((n = niso.getSegmentType()) != NisoImageMetadata.NULL) {
				segBuf.append(margn5
						+ element("mix:SegmentType", Integer.toString(n))
						+ this.EOL);
				useSegBuf = true;
			}
			long[] larray = niso.getStripOffsets();
			if (larray != null) {
				segBuf.append(margn5
						+ element("mix:StripOffsets", longArray(larray))
						+ this.EOL);
				useSegBuf = true;
			}
			long ln = niso.getRowsPerStrip();
			if (ln != NisoImageMetadata.NULL) {
				segBuf.append(margn5
						+ element("mix:RowsPerStrip", Long.toString(ln))
						+ this.EOL);
				useSegBuf = true;
			}
			if ((larray = niso.getStripByteCounts()) != null) {
				segBuf.append(margn5
						+ element("mix:StripByteCounts", longArray(larray))
						+ this.EOL);
				useSegBuf = true;
			}
			if ((ln = niso.getTileWidth()) != NisoImageMetadata.NULL) {
				segBuf.append(margn5
						+ element("mix:TileWidth", Long.toString(ln))
						+ this.EOL);
				useSegBuf = true;
			}
			if ((ln = niso.getTileLength()) != NisoImageMetadata.NULL) {
				segBuf.append(margn5
						+ element("mix:TileLength", Long.toString(ln))
						+ this.EOL);
				useSegBuf = true;
			}
			if ((larray = niso.getTileOffsets()) != null) {
				segBuf.append(margn5
						+ element("mix:TileOffsets", longArray(larray))
						+ this.EOL);
				useSegBuf = true;
			}
			if ((larray = niso.getTileByteCounts()) != null) {
				segBuf.append(margn5
						+ element("mix:TileByteCounts", longArray(larray))
						+ this.EOL);
				useSegBuf = true;
			}
			segBuf.append(margn4 + elementEnd("mix:Segments") + this.EOL);
			if (useSegBuf) {
				fBuf.append(segBuf);
				useFBuf = true;
			}
			if ((n = niso.getPlanarConfiguration()) != NisoImageMetadata.NULL) {
				fBuf.append(margn4
						+ element("mix:PlanarConfiguration", Integer
								.toString(n)) + this.EOL);
			}
			fBuf.append(margn3 + elementEnd("mix:Format") + this.EOL);
			if (useFBuf) {
				this.xmldata.append(fBuf.toString());
			}

			// Now a tentative buffer for the File element. It's actually pretty
			// safe that this will have some content, but I'd rather be
			// consistent. Besides, it's a good test of the pattern.
			StringBuffer fileBuf = new StringBuffer(margn3
					+ elementStart("mix:File") + this.EOL);
			boolean useFileBuf = false;
			s = niso.getImageIdentifier();
			if (s != null) {
				fileBuf.append(margn4 + element("mix:ImageIdentifier", s)
						+ this.EOL);
				useFileBuf = true;
			}
			ln = niso.getFileSize();
			if (ln != NisoImageMetadata.NULL) {
				fileBuf
						.append(margn4
								+ element("mix:FileSize", Long.toString(ln))
								+ this.EOL);
				useFileBuf = true;
			}
			n = niso.getChecksumMethod();
			s = niso.getChecksumValue();
			if (n != NisoImageMetadata.NULL || s != null) {
				fileBuf
						.append(margn4 + elementStart("mix:Checksum")
								+ this.EOL);
				if (n != NisoImageMetadata.NULL) {
					fileBuf
							.append(margn5
									+ element("mix:ChecksumMethod", Integer
											.toString(n)) + this.EOL);
				}
				if (s != null) {
					fileBuf.append(margn5 + element("mix:ChecksumValue", s)
							+ this.EOL);
				}
				fileBuf.append(margn4 + elementEnd("mix:Checksum") + this.EOL);
				useFileBuf = true;
			}
			n = niso.getOrientation();
			if (n != NisoImageMetadata.NULL) {
				fileBuf.append(margn4
						+ element("mix:Orientation", Integer.toString(n))
						+ this.EOL);
				useFileBuf = true;
			}
			n = niso.getDisplayOrientation();
			if (n != NisoImageMetadata.NULL) {
				fileBuf
						.append(margn4
								+ element("mix:DisplayOrientation", Integer
										.toString(n)) + this.EOL);
				useFileBuf = true;
			}
			ln = niso.getXTargetedDisplayAR();
			long ln2 = niso.getYTargetedDisplayAR();
			if (ln != NisoImageMetadata.NULL || ln2 != NisoImageMetadata.NULL) {
				fileBuf.append(margn4 + elementStart("mix:TargetedDisplayAR")
						+ this.EOL);
				if (ln != NisoImageMetadata.NULL) {
					fileBuf.append(margn5
							+ element("mix:XTargetedDisplayAR", Long
									.toString(ln)) + this.EOL);
				}
				if (ln2 != NisoImageMetadata.NULL) {
					fileBuf.append(margn5
							+ element("mix:YTargetedDisplayAR", Long
									.toString(ln2)) + this.EOL);
				}
				fileBuf.append(margn4
						+ elementEnd("mix:" + "TargetedDisplayAR") + this.EOL);
				useFileBuf = true;
			}

			fileBuf.append(margn3 + elementEnd("mix:File") + this.EOL);
			if (useFileBuf) {
				this.xmldata.append(fileBuf.toString());
			}
			this.xmldata.append(margn2
					+ elementEnd("mix:" + "BasicImageParameters") + this.EOL);

		}

		/***********************************************************************
		 * Original JHOVE Comment: The NISO Metadata output is split into
		 * multiple functions so that they're merely outrageously big rather
		 * than disgustingly big Top level element 2 of 4: ImageCreation
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050801
		 * @since 20050801
		 * @param niso NISO image metadata
		 * @param margin - The margin of the XML-Tags
		 **********************************************************************/
		@Override
		protected void showNisoImageCreation(NisoImageMetadata niso,
				String margin) {
			String margn2 = margin + " ";
			String margn3 = margn2 + " ";
			String margn4 = margn3 + " ";
			String margn5 = margn4 + " ";
			String margn6 = margn5 + " ";

			this.xmldata.append(margn2 + elementStart("mix:" + "ImageCreation")
					+ this.EOL);
			String s = niso.getSourceType();
			if (s != null) {
				this.xmldata.append(margn3 + element("mix:SourceType", s)
						+ this.EOL);
			}
			s = niso.getSourceID();
			if (s != null) {
				this.xmldata.append(margn3 + element("mix:SourceID", s)
						+ this.EOL);
			}
			s = niso.getImageProducer();
			if (s != null) {
				this.xmldata.append(margn3
						+ element("mix:" + "ImageProducer", s) + this.EOL);
			}

			StringBuffer hostBuf = new StringBuffer(margn3
					+ elementStart("mix:Host") + this.EOL);
			boolean useHostBuf = false;
			s = niso.getHostComputer();
			if (s != null) {
				hostBuf.append(margn4 + element("mix:HostComputer", s)
						+ this.EOL);
				useHostBuf = true;
			}
			s = niso.getOS();
			if (s != null) {
				hostBuf.append(margn4 + element("mix:OperatingSystem", s)
						+ this.EOL);
				useHostBuf = true;
			}
			s = niso.getOSVersion();
			if (s != null) {
				hostBuf.append(margn4 + element("mix:OSVersion", s) + this.EOL);
				useHostBuf = true;
			}
			hostBuf.append(margn3 + elementEnd("mix:Host") + this.EOL);
			if (useHostBuf) {
				this.xmldata.append(hostBuf);
			}

			s = niso.getDeviceSource();
			if (s != null) {
				this.xmldata.append(margn3 + element("mix:DeviceSource", s)
						+ this.EOL);
			}

			// Here things get rather deeply nested, and we want to be
			// able to back out if it turns out that we really have no
			// content to deliver. So rather than writing directly to
			// the writer, we write to a StringBuffer that may or may
			// not get used.
			StringBuffer sscBuf = new StringBuffer(margn3
					+ elementStart("mix:ScanningSystemCapture") + this.EOL);
			boolean useSSCBuf = false; // set to true if we find
			// interesting data.
			StringBuffer sshBuf = new StringBuffer(margn4
					+ elementStart("mix:ScanningSystemHardware") + this.EOL);
			boolean useSSHBuf = false;
			s = niso.getScannerManufacturer();
			if (s != null) {
				sshBuf.append(margn5 + element("mix:ScannerManufacturer", s)
						+ this.EOL);
				useSSHBuf = true;
			}
			s = niso.getScannerModelName();
			String s1 = niso.getScannerModelNumber();
			String s2 = niso.getScannerModelSerialNo();
			if (s != null || s1 != null || s2 != null) {
				sshBuf.append(margn5 + elementStart("mix:ScannerModel")
						+ this.EOL);
				if (s != null) {
					sshBuf.append(margn6 + element("mix:ScannerModelName", s)
							+ this.EOL);
					useSSHBuf = true;
				}
				if (s1 != null) {
					sshBuf.append(margn6
							+ element("mix:" + "ScannerModelNumber", s1)
							+ this.EOL);
					useSSHBuf = true;
				}
				if (s2 != null) {
					sshBuf.append(margn6
							+ element("mix:" + "ScannerModelSerialNo ", s2)
							+ this.EOL);
					useSSHBuf = true;
				}
				sshBuf.append(margn5 + elementEnd("mix:ScannerModel")
						+ this.EOL);
			}
			sshBuf.append(margn4
					+ elementEnd("mix:" + "ScanningSystemHardware") + this.EOL);
			if (useSSHBuf) {
				// There's some ScanningSystemHardware content, and
				// therefore some ScanningSystemCapture content; keep
				// the element.
				sscBuf.append(sshBuf);
				useSSCBuf = true;
			}

			// Build a tentative buffer for scanning system software
			StringBuffer sssBuf = new StringBuffer(margn4
					+ elementStart("mix:ScanningSystemSoftware") + this.EOL);
			boolean useSSSBuf = false;
			s = niso.getScanningSoftware();
			if (s != null) {
				sssBuf.append(margn5 + element("mix:ScanningSoftware", s)
						+ this.EOL);
				useSSSBuf = true;
			}
			s = niso.getScanningSoftwareVersionNo();
			if (s != null) {
				sssBuf.append(margn5
						+ element("mix:" + "ScanningSoftwareVersionNo", s)
						+ this.EOL);
				useSSSBuf = true;
			}
			sssBuf.append(margn4
					+ elementEnd("mix:" + "ScanningSystemSoftware") + this.EOL);
			if (useSSSBuf) {
				// There's some ScanningSystemSoftware content, and
				// therefore some ScanningSystemCapture content; keep
				// the element.
				sscBuf.append(sssBuf);
				useSSCBuf = true;
			}

			// Build a tentative buffer for scanner capture settings
			StringBuffer scsBuf = new StringBuffer(margn4
					+ elementStart("mix:ScannerCaptureSettings") + this.EOL);
			boolean useSCSBuf = false;
			double d = niso.getPixelSize();
			if (d != NisoImageMetadata.NILL) {
				scsBuf.append(margn5
						+ element("mix:PixelSize", Double.toString(d))
						+ this.EOL);
				useSCSBuf = true;
			}
			d = niso.getXPhysScanResolution();
			double d1 = niso.getYPhysScanResolution();
			if (d != NisoImageMetadata.NILL || d1 != NisoImageMetadata.NILL) {
				scsBuf.append(margn5
						+ elementStart("mix:" + "PhysScanResolution")
						+ this.EOL);
				if (d != NisoImageMetadata.NILL) {
					scsBuf.append(margn6
							+ element("mix:" + "XphysScanResolution", Double
									.toString(d)) + this.EOL);
				}
				if (d1 != NisoImageMetadata.NILL) {
					scsBuf.append(margn6
							+ element("mix:" + "YphysScanResolution", Double
									.toString(d1)) + this.EOL);
				}
				scsBuf.append(margn5
						+ elementEnd("mix:" + "PhysScanResolution") + this.EOL);
				useSCSBuf = true;
			}
			scsBuf.append(margn4
					+ elementEnd("mix:" + "ScannerCaptureSettings") + this.EOL);
			if (useSCSBuf) {
				sscBuf.append(scsBuf);
				useSSCBuf = true;
			}
			sscBuf.append(margn3 + elementEnd("mix:" + "ScanningSystemCapture")
					+ this.EOL);

			// Finally! Do we use any of this stuff we just went through?
			if (useSSCBuf) {
				this.xmldata.append(sscBuf.toString());
			}


			// Same deal for digital camera capture; put the element in a
			// tentative StringBuffer and then decide if it's non-trivial.
			StringBuffer dccBuf = new StringBuffer(margn3
					+ elementStart("mix:DigitalCameraCapture") + this.EOL);
			boolean useDCCBuf = false;
			s = niso.getDigitalCameraManufacturer();
			if (s != null) {
				dccBuf.append(margn4
						+ element("mix:" + "DigitalCameraManufacturer", s)
						+ this.EOL);
				useDCCBuf = true;
			}
			s = niso.getDigitalCameraModel();
			if (s != null) {
				dccBuf.append(margn4
						+ element("mix:" + "DigitalCameraModel", s) + this.EOL);
				useDCCBuf = true;
			}

			dccBuf.append(margn3 + elementEnd("mix:" + "DigitalCameraCapture")
					+ this.EOL);
			if (useDCCBuf) {
				this.xmldata.append(dccBuf.toString());
			}


			// Same tentative buffer deal for camera capture settings.
			StringBuffer ccsBuf = new StringBuffer(margn3
					+ elementStart("mix:CameraCaptureSettings") + this.EOL);
			boolean useCCSBuf = false;
			d = niso.getFNumber();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf
						.append(margn4
								+ element("mix:FNumber", Double.toString(d))
								+ this.EOL);
				useCCSBuf = true;
			}
			d = niso.getExposureTime();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ element("mix:ExposureTime", Double.toString(d))
						+ this.EOL);
				useCCSBuf = true;
			}
			d = niso.getBrightness();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ element("mix:Brightness", Double.toString(d))
						+ this.EOL);
				useCCSBuf = true;
			}
			d = niso.getExposureBias();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ element("mix:ExposureBias", Double.toString(d))
						+ this.EOL);
				useCCSBuf = true;
			}
			double[] darray = niso.getSubjectDistance();
			if (darray != null) {
				ccsBuf.append(margn4
						+ element("mix:SubjectDistance", doubleArray(darray))
						+ this.EOL);
				useCCSBuf = true;
			}
			int n = niso.getMeteringMode();
			if (n != NisoImageMetadata.NULL) {
				ccsBuf.append(margn4
						+ element("mix:MeteringMode", Integer.toString(n))
						+ this.EOL);
				useCCSBuf = true;
			}
			n = niso.getSceneIlluminant();
			if (n != NisoImageMetadata.NULL) {
				ccsBuf.append(margn4
						+ element("mix:SceneIlluminant", Integer.toString(n))
						+ this.EOL);
				useCCSBuf = true;
			}
			d = niso.getColorTemp();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ element("mix:ColorTemp", Double.toString(d))
						+ this.EOL);
				useCCSBuf = true;
			}
			d = niso.getFocalLength();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ element("mix:FocalLength", Double.toString(d))
						+ this.EOL);
				useCCSBuf = true;
			}
			n = niso.getFlash();
			if (n != NisoImageMetadata.NULL) {
				ccsBuf.append(margn4
						+ element("mix:Flash", Integer.toString(n)) + this.EOL);
				useCCSBuf = true;
			}
			d = niso.getFlashEnergy();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ element("mix:FlashEnergy", Double.toString(d))
						+ this.EOL);
				useCCSBuf = true;
			}
			n = niso.getFlashReturn();
			if (n != NisoImageMetadata.NULL) {
				ccsBuf.append(margn4
						+ element("mix:FlashReturn", Integer.toString(n))
						+ this.EOL);
				useCCSBuf = true;
			}
			n = niso.getBackLight();
			if (n != NisoImageMetadata.NULL) {
				ccsBuf.append(margn4
						+ element("mix:BackLight", Integer.toString(n))
						+ this.EOL);
				useCCSBuf = true;
			}
			d = niso.getExposureIndex();
			if (d != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ element("mix:ExposureIndex", Double.toString(d))
						+ this.EOL);
				useCCSBuf = true;
			}
			n = niso.getAutoFocus();
			if (n != NisoImageMetadata.NULL) {
				ccsBuf.append(margn4
						+ element("mix:AutoFocus", Integer.toString(n))
						+ this.EOL);
				useCCSBuf = true;
			}
			d = niso.getXPrintAspectRatio();
			d1 = niso.getYPrintAspectRatio();
			if (d != NisoImageMetadata.NILL || d1 != NisoImageMetadata.NILL) {
				ccsBuf.append(margn4
						+ elementStart("mix:" + "PrintAspectRatio") + this.EOL);
				if (d != NisoImageMetadata.NILL) {
					ccsBuf.append(margn5
							+ element("mix:" + "XPrintAspectRatio", Double
									.toString(d)) + this.EOL);
				}
				if (d1 != NisoImageMetadata.NILL) {
					ccsBuf.append(margn5
							+ element("mix:" + "YPrintAspectRatio", Double
									.toString(d1)) + this.EOL);
					ccsBuf.append(margn4
							+ elementEnd("mix:" + "PrintAspectRatio")
							+ this.EOL);
					useCCSBuf = true;
				}
			}
			ccsBuf.append(margn3 + elementEnd("mix:" + "CameraCaptureSettings")
					+ this.EOL);
			if (useCCSBuf) {
				this.xmldata.append(ccsBuf.toString());
			}

			// Finally we get a relative breather without having to check
			// nested elements.
			n = niso.getSensor();
			if (n != NisoImageMetadata.NULL) {
				this.xmldata
						.append(margn3
								+ element("mix:Sensor", Integer.toString(n))
								+ this.EOL);
			}
			s = niso.getDateTimeCreated();
			if (s != null) {
				this.xmldata.append(margn3
						+ element("mix:" + "DateTimeCreated", s));
			}
			s = niso.getMethodology();
			if (s != null) {
				this.xmldata.append(margn3 + element("mix:" + "Methodology", s)
						+ this.EOL);
			}

			this.xmldata.append(margn2 + elementEnd("mix:" + "ImageCreation")
					+ this.EOL);
		}

		/***********************************************************************
		 * Original JHOVE Comment: The NISO Metadata output is split into
		 * multiple functions so that they're merely outrageously big rather
		 * than disgustingly big Top level element 3 of 4:
		 * ImagingPerformanceAssessment
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050801
		 * @since 20050801
		 * @param niso NISO image metadata
		 * @param margin - The margin of the XML-Tags
		 **********************************************************************/
		@Override
		protected void showNisoImagingPerformanceAssessment(
				NisoImageMetadata niso, String margin) {
			String margn2 = margin + " ";
			String margn3 = margn2 + " ";
			String margn4 = margn3 + " ";
			String margn5 = margn4 + " ";
			String margn6 = margn5 + " ";
			String margn7 = margn6 + " ";

			StringBuffer ipaBuf = new StringBuffer(margn3
					+ elementStart("mix:" + "ImagingPerformanceAssessment")
					+ this.EOL);
			boolean useIPABuf = false; // set to true if we find
			// interesting data.
			StringBuffer smBuf = new StringBuffer(margn4
					+ elementStart("mix:SpatialMetrics") + this.EOL);
			boolean useSMBuf = false;

			int n = niso.getSamplingFrequencyPlane();
			if (n != NisoImageMetadata.NULL) {
				smBuf.append(margn5
						+ element("mix:" + "SamplingFrequencyPlane", Integer
								.toString(n)) + this.EOL);
				useSMBuf = true;
			}
			n = niso.getSamplingFrequencyUnit();
			if (n != NisoImageMetadata.NULL) {
				smBuf.append(margn5
						+ element("mix:" + "SamplingFrequencyUnit", Integer
								.toString(n)) + this.EOL);
				useSMBuf = true;
			}
			Rational r = niso.getXSamplingFrequency();
			if (r != null) {
				smBuf.append(margn5
						+ element("mix:XSamplingFrequency", Long.toString(r
								.toLong())) + this.EOL);
				useSMBuf = true;
			}
			r = niso.getYSamplingFrequency();
			if (r != null) {
				smBuf.append(margn5
						+ element("mix:YSamplingFrequency", Long.toString(r
								.toLong())) + this.EOL);
				useSMBuf = true;
			}
			long ln = niso.getImageWidth();
			if (ln != NisoImageMetadata.NULL) {
				smBuf.append(margn5
						+ element("mix:ImageWidth", Long.toString(ln))
						+ this.EOL);
				useSMBuf = true;
			}
			ln = niso.getImageLength();
			if (ln != NisoImageMetadata.NULL) {
				smBuf.append(margn5
						+ element("mix:ImageLength", Long.toString(ln))
						+ this.EOL);
				useSMBuf = true;
			}

			double d = niso.getSourceXDimension();
			n = niso.getSourceXDimensionUnit();
			if (d != NisoImageMetadata.NILL || n != NisoImageMetadata.NULL) {
				smBuf.append(margn5 + elementStart("mix:Source_X") + this.EOL);
				if (d != NisoImageMetadata.NILL) {
					smBuf.append(margn6
							+ element("mix:Source_Xdimension", Double
									.toString(d)) + this.EOL);
				}
				if (n != NisoImageMetadata.NULL) {
					smBuf.append(margn6
							+ element("mix:" + "Source_XdimensionUnit", Integer
									.toString(n)) + this.EOL);
				}
				smBuf.append(margn5 + elementEnd("mix:Source_X") + this.EOL);
				useSMBuf = true;
			}

			d = niso.getSourceYDimension();
			n = niso.getSourceYDimensionUnit();
			if (d != NisoImageMetadata.NILL || n != NisoImageMetadata.NULL) {
				smBuf.append(margn4 + elementStart("mix:Source_Y") + this.EOL);
				if (d != NisoImageMetadata.NILL) {
					smBuf.append(margn5
							+ element("mix:Source_Ydimension", Double
									.toString(d)) + this.EOL);
				}
				if (n != NisoImageMetadata.NULL) {
					smBuf.append(margn5
							+ element("mix:" + "Source_YdimensionUnit", Integer
									.toString(n)) + this.EOL);
				}
				smBuf.append(margn4 + elementEnd("mix:Source_Y") + this.EOL);
				useSMBuf = true;
			}
			smBuf.append(margn3 + elementEnd("mix:SpatialMetrics") + this.EOL);
			if (useSMBuf) {
				ipaBuf.append(smBuf);
				useIPABuf = true;
			}

			// Now a tentative buffer for the Energetics element
			StringBuffer eBuf = new StringBuffer(margn3
					+ elementStart("mix:Energetics") + this.EOL);
			boolean useEBuf = false;
			int[] iarray = niso.getBitsPerSample();
			if (iarray != null) {
				eBuf.append(margn4
						+ element("mix:BitsPerSample",
								integerArray(iarray, ',')) + this.EOL);
				useEBuf = true;
			}
			n = niso.getSamplesPerPixel();
			if (n != NisoImageMetadata.NULL) {
				eBuf.append(margn4
						+ element("mix:SamplesPerPixel", Integer.toString(n))
						+ this.EOL);
				useEBuf = true;
			}
			iarray = niso.getExtraSamples();
			if (iarray != null) {
				eBuf.append(margn4
						+ element("mix:ExtraSamples", integerArray(iarray))
						+ this.EOL);
				useEBuf = true;
			}
			// Tentative buffer for colormap element within Energetics
			// Show only if the flag ist true!
			if (MetadataGenerator.this.showTiffMixColorMap) {
				StringBuffer cmBuf = new StringBuffer(margn4
						+ elementStart("mix:Colormap") + this.EOL);
				boolean useCMBuf = false;
				String str = niso.getColormapReference();
				if (str != null) {
					cmBuf.append(margn5 + element("mix:Reference", str)
							+ this.EOL);
					useCMBuf = true;
				}
				iarray = niso.getColormapRedValue();
				if (iarray != null) {
					cmBuf.append(margn5 + elementStart("mix:Wrap"));
					// If the red array is there, assume the others are too,
					// and are equal in length.
					int[] bcarray = niso.getColormapBitCodeValue();
					int[] garray = niso.getColormapGreenValue();
					int[] barray = niso.getColormapBlueValue();
					try {
						for (int i = 0; i < iarray.length; i++) {
							cmBuf.append(margn6 + elementStart("mix:Color")
									+ this.EOL);
							n = bcarray[i];
							if (n != NisoImageMetadata.NULL) {
								cmBuf
										.append(margn7
												+ element("mix:"
														+ "BitCodeValue",
														Integer.toString(n))
												+ this.EOL);
							}
							n = iarray[i];
							if (n != NisoImageMetadata.NULL) {
								cmBuf.append(margn7
										+ element("mix:RedValue", Integer
												.toString(n)) + this.EOL);
							}
							n = garray[i];
							if (n != NisoImageMetadata.NULL) {
								cmBuf.append(margn7
										+ element("mix:GreenValue", Integer
												.toString(n)) + this.EOL);
							}
							n = barray[i];
							if (n != NisoImageMetadata.NULL) {
								cmBuf.append(margn7
										+ element("mix:BlueValue", Integer
												.toString(n)) + this.EOL);
							}

							cmBuf.append(margn6 + elementEnd("mix:Color")
									+ this.EOL);
						}
					}
					catch (Exception e) {
						// If the assumption mentioned above is wrong,
						// we'll get broken XML, but at least won't die here.
					}
					cmBuf.append(margn5 + elementEnd("mix:Wrap") + this.EOL);
					useCMBuf = true;
				}
				cmBuf.append(margn4 + elementEnd("mix:Colormap"));
				if (useCMBuf) {
					eBuf.append(cmBuf);
					useEBuf = true;
				}
			}
			iarray = niso.getGrayResponseCurve();
			n = niso.getGrayResponseUnit();
			if (iarray != null || n != NisoImageMetadata.NULL) {
				eBuf.append(margn4 + elementStart("mix:GrayResponse")
						+ this.EOL);
				if (iarray != null) {
					eBuf.append(margn5
							+ element("mix:GrayResponseCurve",
									integerArray(iarray)) + this.EOL);
				}
				if (n != NisoImageMetadata.NULL) {
					eBuf.append(margn5
							+ element("mix:GrayResponseUnit", Integer
									.toString(n)) + this.EOL);
				}
				eBuf.append(margn4 + elementEnd("mix:GrayResponse") + this.EOL);
				useEBuf = true;
			}
			r = niso.getWhitePointXValue();
			Rational r1 = niso.getWhitePointYValue();
			if (r != null || r1 != null) {
				// These are specified in the scheme as CIExyType. All the
				// schema tells us about this type is, that it's a string.
				eBuf.append(margn4 + elementStart("mix:WhitePoint") + this.EOL);
				if (r != null) {
					eBuf.append(margn5
							+ element("mix:WhitePoint_Xvalue", r.toString())
							+ this.EOL);
				}
				if (r1 != null) {
					eBuf.append(margn5
							+ element("mix:WhitePoint_Yvalue", r1.toString())
							+ this.EOL);
				}
				eBuf.append(margn4 + elementEnd("mix:WhitePoint") + this.EOL);
				useEBuf = true;
			}

			r = niso.getPrimaryChromaticitiesRedX();
			// For simplicity, we check only the red x in deciding whether
			// to incorporate this element. A partial set of
			// chromaticities would be meaningless anyway.
			if (r != null) {
				eBuf.append(margn4
						+ elementStart("mix:" + "PrimaryChromaticities")
						+ this.EOL);
				eBuf.append(margn5
						+ element("mix:" + "PrimaryChromaticities_RedX", r
								.toString()) + this.EOL);
				r = niso.getPrimaryChromaticitiesRedY();
				if (r != null) {
					eBuf.append(margn5
							+ element("mix:" + "PrimaryChromaticities_RedY", r
									.toString()) + this.EOL);
				}
				r = niso.getPrimaryChromaticitiesGreenX();
				if (r != null) {
					eBuf.append(margn5
							+ element("mix:" + "PrimaryChromaticities_GreenX",
									r.toString()) + this.EOL);
				}
				r = niso.getPrimaryChromaticitiesGreenY();
				if (r != null) {
					eBuf.append(margn5
							+ element("mix:" + "PrimaryChromaticities_GreenY",
									r.toString()) + this.EOL);
				}
				r = niso.getPrimaryChromaticitiesBlueX();
				if (r != null) {
					eBuf.append(margn5
							+ element("mix:" + "PrimaryChromaticities_BlueX", r
									.toString()) + this.EOL);
				}
				r = niso.getPrimaryChromaticitiesBlueY();
				if (r != null) {
					eBuf.append(margn5
							+ element("mix:" + "PrimaryChromaticities_BlueY", r
									.toString()) + this.EOL);
				}
				eBuf.append(margn4
						+ elementEnd("mix:" + "PrimaryChromaticities")
						+ this.EOL);
				useEBuf = true;
			}

			eBuf.append(margn3 + elementEnd("mix:Energetics") + this.EOL);
			if (useEBuf) {
				ipaBuf.append(eBuf);
				useIPABuf = true;
			}


			// Another tentative buffer for TargetData
			StringBuffer tdBuf = new StringBuffer(margn3
					+ elementStart("mix:TargetData") + this.EOL);
			boolean useTDBuf = false;
			n = niso.getTargetType();
			if (n != NisoImageMetadata.NULL) {
				tdBuf.append(margn4
						+ element("mix:TargetType", Integer.toString(n))
						+ this.EOL);
				useTDBuf = true;
			}

			// Nest a TargetID tentative buffer in the TargetData buffer
			StringBuffer tiBuf = new StringBuffer(margn4
					+ elementStart("mix:TargetID") + this.EOL);
			boolean useTIBuf = false;
			String s = niso.getTargetIDManufacturer();
			if (s != null) {
				tiBuf.append(margn5 + element("mix:TargetIDManufacturer", s)
						+ this.EOL);
				useTIBuf = true;
			}
			s = niso.getTargetIDName();
			if (s != null) {
				tiBuf
						.append(margn5 + element("mix:TargetIDName", s)
								+ this.EOL);
				useTIBuf = true;
			}
			s = niso.getTargetIDNo();
			if (s != null) {
				tiBuf.append(margn5 + element("mix:TargetIDNo", s) + this.EOL);
				useTIBuf = true;
			}
			s = niso.getTargetIDMedia();
			if (s != null) {
				tiBuf.append(margn5 + element("mix:TargetIDMedia", s)
						+ this.EOL);
				useTIBuf = true;
			}
			tiBuf.append(margn4 + elementEnd("mix:TargetID") + this.EOL);
			if (useTIBuf) {
				tdBuf.append(tiBuf);
				useTDBuf = true;
			}

			s = niso.getImageData();
			if (s != null) {
				tdBuf.append(margn5 + element("mix:ImageData", s) + this.EOL);
				useTDBuf = true;
			}
			s = niso.getPerformanceData();
			if (s != null) {
				tdBuf.append(margn5 + element("mix:PerformanceData", s)
						+ this.EOL);
				useTDBuf = true;
			}
			s = niso.getProfiles();
			if (s != null) {
				tdBuf.append(margn5 + element("mix:Profiles", s) + this.EOL);
				useTDBuf = true;
			}
			tdBuf.append(margn3 + elementEnd("mix:TargetData") + this.EOL);
			if (useTDBuf) {
				ipaBuf.append(tdBuf);
				useIPABuf = true;
			}

			ipaBuf.append(margn2
					+ elementEnd("mix:" + "ImagingPerformanceAssessment")
					+ this.EOL);
			if (useIPABuf) {
				this.xmldata.append(ipaBuf.toString());
			}
		}

		/***********************************************************************
		 * Original JHOVE Comment: The NISO Metadata output is split into
		 * multiple functions so that they're merely outrageously big rather
		 * than disgustingly big Top level element 4 of 4: ChangeHistory
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050801
		 * @since 20050801
		 * @param niso NISO image metadata
		 * @param margin - The margin of the XML-Tags
		 **********************************************************************/
		@Override
		protected void showNisoChangeHistory(NisoImageMetadata niso,
				String margin) {
			String margn2 = margin + " ";
			String margn3 = margn2 + " ";
			String margn4 = margn3 + " ";
			String margn5 = margn4 + " ";

			// Yet again, build elements in tentative buffers and
			// throw them away if they prove trivial.
			StringBuffer chBuf = new StringBuffer(margn2
					+ elementStart("mix:ChangeHistory") + this.EOL);
			boolean useCHBuf = false;
			StringBuffer ipBuf = new StringBuffer(margn3
					+ elementStart("mix:ImageProcessing") + this.EOL);
			boolean useIPBuf = false;

			String s = niso.getDateTimeProcessed();
			if (s != null) {
				ipBuf.append(margn4 + element("DateTimeProcessed", s)
						+ this.EOL);
				useIPBuf = true;
			}
			s = niso.getSourceData();
			if (s != null) {
				ipBuf.append(margn4 + element("SourceData", s) + this.EOL);
				useIPBuf = true;
			}
			s = niso.getProcessingAgency();
			if (s != null) {
				ipBuf
						.append(margn4 + element("ProcessingAgency", s)
								+ this.EOL);
				useIPBuf = true;
			}

			// Third-level nesting of tentative buffer!
			StringBuffer psBuf = new StringBuffer(margn4
					+ elementStart("ProcessingSoftware") + this.EOL);
			boolean usePSBuf = false;
			s = niso.getProcessingSoftwareName();
			if (s != null) {
				psBuf.append(margn5 + element("ProcessingSoftwareName", s)
						+ this.EOL);
				usePSBuf = true;
			}
			s = niso.getProcessingSoftwareVersion();
			if (s != null) {
				psBuf.append(margn5 + element("ProcessingSoftwareVersion", s)
						+ this.EOL);
				usePSBuf = true;
			}
			psBuf.append(margn4 + elementEnd("ProcessingSoftware") + this.EOL);
			if (usePSBuf) {
				ipBuf.append(psBuf);
				useIPBuf = true;
			}

			// Hard to say, but I think the intent is that there be
			// one ProcessingActions element per processing action.
			String[] sarray = niso.getProcessingActions();
			if (sarray != null) {
				for (int i = 0; i < sarray.length; i++) {
					ipBuf.append(margn4
							+ element("ProcessingActions", sarray[i])
							+ this.EOL);
				}
				useIPBuf = true;
			}
			ipBuf.append(margn3 + elementEnd("mix:ImageProcessing") + this.EOL);
			if (useIPBuf) {
				chBuf.append(ipBuf);
				useCHBuf = true;
			}
			chBuf.append(margn2 + elementEnd("mix:ChangeHistory") + this.EOL);
			if (useCHBuf) {
				this.xmldata.append(chBuf.toString());
			}
		}

		/***********************************************************************
		 * Original JHOVE Comment: Display the NISO image metadata formatted
		 * according to the AES schema.
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050831
		 * @since 20050801
		 * @param aes AES audio metadata
		 * @param margn - The margin for the aes-block
		 **********************************************************************/
		protected void showAESAudioMetadata(AESAudioMetadata aes, String margn) {
			// this._level += 4;
			// final String margin = getIndent(this._level);
			final String margin = " " + margn;
			final String margn2 = margin + " ";
			final String margn3 = margn2 + " ";
			final String margn4 = margn3 + " ";
			final String margn5 = margn4 + " ";
			// final String margn6 = margn5 + " ";

			// ID strings. These are arbitrary, but must be unique
			// within the document.
			final String formatRegionID = "J1";
			final String faceRegionID = "J2";
			final String faceID = "J3";
			final String audioObjectID = "J4";
			final String streamIDBase = "J9";

			final String[][] attrs = {
					{ "xmlns:aes", "http://www.aes.org/audioObject" },
					{ "xmlns:tcf", "http://www.aes.org/tcf" },
					{ "xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance" },
					{ "ID", audioObjectID },
					{ "analogDigitalFlag", aes.getAnalogDigitalFlag() },
					{ "disposition", "Validated by JHOVE" },
					{ "schemaVersion", "1.02b" } };
			this.xmldata.append(margin + elementStart("aes:audioObject", attrs)
					+ this.EOL);
			String s = aes.getFormat();
			if (s != null) {
				String v = aes.getSpecificationVersion();
				String[][] fmattrs = new String[1][2];
				fmattrs[0][0] = "specificationVersion";
				if (v != null) {
					fmattrs[0][1] = v;
				}
				else {
					// Shouldn't happen
					fmattrs[0][1] = "";
				}
				this.xmldata.append(margn2 + element("aes:format", fmattrs, s)
						+ this.EOL);
			}
			s = aes.getAppSpecificData();
			if (s != null) {
				this.xmldata.append(margn2 + element("aes:appSpecificData", s)
						+ this.EOL);
			}
			s = aes.getAudioDataEncoding();
			if (s != null) {
				this.xmldata.append(margn2
						+ element("aes:audioDataEncoding", s) + this.EOL);
			}
			int in = aes.getByteOrder();
			if (in != AESAudioMetadata.NULL) {
				this.xmldata.append(margn2
						+ element("aes:byteOrder",
								in == AESAudioMetadata.BIG_ENDIAN
										? "BIG_ENDIAN" : "LITTLE_ENDIAN")
						+ this.EOL);
			}
			long lin = aes.getFirstSampleOffset();
			if (lin != AESAudioMetadata.NULL) {
				this.xmldata.append(margn2
						+ element("aes:" + "firstSampleOffset", Long
								.toString(lin)) + this.EOL);
			}
			String[] use = aes.getUse();
			if (use != null) {
				String[][] uattrs = new String[][] { { "useType", use[0] },
						{ "otherType", use[1] } };
				this.xmldata.append(margn2 + element("aes:use", uattrs)
						+ this.EOL);
			}
			s = aes.getPrimaryIdentifier();
			if (s != null) {
				String t = aes.getPrimaryIdentifierType();
				String[][] idattrs = new String[1][2];
				idattrs[0][0] = "identifierType";
				if (t != null) {
					idattrs[0][1] = t;
				}
				else {
					// Shouldn't happen
					idattrs[0][1] = "";
				}
				this.xmldata.append(margn2
						+ element("aes:primaryIdentifier", idattrs, s)
						+ this.EOL);
			}

			// Add the face information, which is mostly filler.
			// In the general case, it can contain multiple Faces;
			// this isn't supported yet.
			List facelist = aes.getFaceList();
			if (!facelist.isEmpty()) {
				final String[][] faceRegionAttrs = { { "ID", faceRegionID },
						{ "formatRef", formatRegionID },
						{ "faceRef", "faceID" }, { "label", "BuiltByJHOVE" } };
				final String[][] faceAttrs = { { "direction", null },
						{ "ID", faceID }, { "audioObjectRef", audioObjectID },
						{ "label", "Face" } };
				AESAudioMetadata.Face f = (AESAudioMetadata.Face) facelist
						.get(0);
				faceAttrs[0][1] = f.getDirection();
				this.xmldata.append(margn2
						+ elementStart("aes:face", faceAttrs) + this.EOL);
				// Fill in a minimal time range.
				AESAudioMetadata.TimeDesc startTime = f.getStartTime();
				if (startTime != null) {
					this.xmldata.append(margn3
							+ elementStart("aes:" + "timeline") + this.EOL);
					writeAESTimeRange(margn3, startTime, f.getDuration());
					this.xmldata.append(margn3
							+ elementEnd("aes:" + "timeline") + this.EOL);
				}

				// For the present, assume just one face region
				AESAudioMetadata.FaceRegion facergn = f.getFaceRegion(0);
				this.xmldata.append(margn3
						+ elementStart("aes:region", faceRegionAttrs)
						+ this.EOL);
				this.xmldata.append(margn4 + elementStart("aes:" + "timeRange")
						+ this.EOL);
				writeAESTimeRange(margn3, facergn.getStartTime(), facergn
						.getDuration());
				this.xmldata.append(margn4 + elementEnd("aes:" + "timeRange")
						+ this.EOL);
				int nchan = aes.getNumChannels();
				if (nchan != AESAudioMetadata.NULL) {
					this.xmldata.append(margn4
							+ element("aes:" + "numChannels", Integer
									.toString(nchan)) + this.EOL);
				}
				String[] locs = aes.getMapLocations();
				for (int ch = 0; ch < nchan; ch++) {
					// write a stream element for each channel
					String[][] streamAttrs = {
							{ "ID", streamIDBase + Integer.toString(ch) },
							{ "label", "JHOVE" },
							{ "faceRegionRef", faceRegionID }

					};
					this.xmldata.append(margn4
							+ elementStart("aes:" + "stream", streamAttrs)
							+ this.EOL);
					String[][] chanAttrs = {
							{ "channelNum", Integer.toString(ch) },
							{ "mapLocation", locs[ch] } };
					this.xmldata.append(margn5
							+ element("aes:" + "channelAssignment", chanAttrs)
							+ this.EOL);
					this.xmldata.append(margn4 + elementEnd("aes:" + "stream")
							+ this.EOL);
				}
				this.xmldata.append(margn3 + elementEnd("aes:region")
						+ this.EOL);
				this.xmldata.append(margn2 + elementEnd("aes:face") + this.EOL);
			}

			// In the general case, a FormatList can contain multiple
			// FormatRegions. This doesn't happen with any of the current
			// modules; if it's needed in the future, simply set up an
			// iteration loop on formatList.
			List flist = aes.getFormatList();
			if (!flist.isEmpty()) {
				AESAudioMetadata.FormatRegion rgn = (AESAudioMetadata.FormatRegion) flist
						.get(0);
				int bitDepth = rgn.getBitDepth();
				double sampleRate = rgn.getSampleRate();
				int wordSize = rgn.getWordSize();
				String[] bitRed = rgn.getBitrateReduction();
				// Build a FormatRegion subtree if at least one piece
				// of data that goes into it is present.
				if (bitDepth != AESAudioMetadata.NULL
						|| sampleRate != AESAudioMetadata.NILL
						|| wordSize != AESAudioMetadata.NULL) {
					this.xmldata.append(margn2
							+ elementStart("aes:" + "formatList") + this.EOL);
					String[][] frAttr = { { "ID", formatRegionID } };
					this.xmldata.append(margn3
							+ elementStart("aes:" + "formatRegion", frAttr)
							+ this.EOL);
					if (bitDepth != AESAudioMetadata.NULL) {
						this.xmldata.append(margn4
								+ element("aes:" + "bitDepth", Integer
										.toString(bitDepth)) + this.EOL);
					}
					if (sampleRate != AESAudioMetadata.NILL) {
						this.xmldata.append(margn4
								+ element("aes:" + "sampleRate", Double
										.toString(sampleRate)) + this.EOL);
					}
					if (wordSize != AESAudioMetadata.NULL) {
						this.xmldata.append(margn4
								+ element("aes:" + "wordSize", Integer
										.toString(wordSize)) + this.EOL);
					}
					if (bitRed != null) {
						this.xmldata.append(margn4
								+ elementStart("aes:" + "bitrateReduction")
								+ this.EOL);
						this.xmldata.append(margn5
								+ element("aes:codecName", bitRed[0])
								+ this.EOL);
						this.xmldata.append(margn5
								+ element("aes:codecNameVersion", bitRed[1])
								+ this.EOL);
						this.xmldata.append(margn5
								+ element("aes:codecCreatorApplication",
										bitRed[2]) + this.EOL);
						this.xmldata.append(margn5
								+ element("aes:codecCreatorApplicationVersion",
										bitRed[3]) + this.EOL);
						this.xmldata.append(margn5
								+ element("aes:codecQuality", bitRed[4])
								+ this.EOL);
						this.xmldata
								.append(margn5
										+ element("aes:dataRate", bitRed[5])
										+ this.EOL);
						this.xmldata.append(margn5
								+ element("aes:dataRateMode", bitRed[6])
								+ this.EOL);
						this.xmldata.append(margn4
								+ elementEnd("aes:" + "bitrateReduction")
								+ this.EOL);
					}
					this.xmldata.append(margn3
							+ elementEnd("aes:" + "formatRegion") + this.EOL);
					this.xmldata.append(margn2
							+ elementEnd("aes:" + "formatList") + this.EOL);
				}
			}
			/* This should go somewhere, but where? */
			// int nchan = aes.getNumChannels();
			// if(nchan != AESAudioMetadata.NULL) {
			// this.xmldata.append(margn2 + element("aes:numChannels",
			// Integer.toString(nchan)));
			// }
			this.xmldata.append(margin + elementEnd("aes:audioObject")
					+ this.EOL);

			// this._level -= 3;
		}

		/*
		 * Break out the writing of a timeRangeType element. This always gives a
		 * start time of 0. This is all FAKE DATA for the moment.
		 */
		protected void writeAESTimeRange(String baseIndent,
				AESAudioMetadata.TimeDesc start,
				AESAudioMetadata.TimeDesc duration) {
			final String margn1 = baseIndent + " ";
			final String margn2 = margn1 + " ";
			final String margn3 = margn2 + " ";
			final String[][] attrs = { { "tcf:frameCount", "30" },
					{ "tcf:timeBase", "1000" },
					{ "tcf:videoField", "FIELD_1" },
					{ "tcf:countingMode", "NTSC_NON_DROP_FRAME" } };
			final String[][] ffAttrs = { { "tcf:framing", "NOT_APPLICABLE" },
					{ "xsi:type", "tcf:ntscFilmFramingType" } };
			this.xmldata.append(margn1 + elementStart("tcf:startTime", attrs)
					+ this.EOL);
			this.xmldata.append(margn2
					+ element("tcf:hours", Integer.toString(start.getHours()))
					+ this.EOL);
			this.xmldata.append(margn2
					+ element("tcf:minutes", Integer.toString(start
							.getMinutes())) + this.EOL);
			this.xmldata.append(margn2
					+ element("tcf:seconds", Integer.toString(start
							.getSeconds())) + this.EOL);
			this.xmldata
					.append(margn2
							+ element("tcf:frames", Integer.toString(start
									.getFrames())) + this.EOL);
			String[][] sampleAttrs = { { "tcf:sampleRate", "" } };
			sampleAttrs[0][1] = Integer.toString((int) start.getSampleRate());
			this.xmldata.append(margn2
					+ elementStart("tcf:samples", sampleAttrs) + this.EOL);
			this.xmldata.append(margn3
					+ element("tcf:numberOfSamples", Integer.toString(start
							.getSamples())) + this.EOL);
			this.xmldata.append(margn2 + elementEnd("tcf:samples") + this.EOL);
			// this.xmldata.append(margn2 + element
			// ("tcf:samples", Integer
			// .toString(start.getSamples()) ));
			this.xmldata.append(margn2 + element("tcf:filmFraming", ffAttrs)
					+ this.EOL);
			this.xmldata
					.append(margn1 + elementEnd("tcf:startTime") + this.EOL);

			if (duration != null) {
				this.xmldata.append(margn1
						+ elementStart("tcf:" + "duration", attrs) + this.EOL);
				this.xmldata.append(margn2
						+ element("tcf:hours", Integer.toString(duration
								.getHours())) + this.EOL);
				this.xmldata.append(margn2
						+ element("tcf:minutes", Integer.toString(duration
								.getMinutes())) + this.EOL);
				this.xmldata.append(margn2
						+ element("tcf:seconds", (Integer.toString(duration
								.getSeconds()))) + this.EOL);
				this.xmldata.append(margn2
						+ element("tcf:frames", Integer.toString(duration
								.getFrames())) + this.EOL);
				sampleAttrs[0][1] = Integer.toString((int) duration
						.getSampleRate());
				this.xmldata.append(margn2
						+ elementStart("tcf:" + "samples", sampleAttrs)
						+ this.EOL);
				this.xmldata.append(margn3
						+ element("tcf:numberOfSamples", Integer
								.toString(duration.getSamples())) + this.EOL);
				this.xmldata.append(margn2 + elementEnd("tcf:samples")
						+ this.EOL);
				// this.xmldata.append(margn2 + element
				// ("tcf:samples", Integer
				// .toString(duration.getSamples()) ));
				this.xmldata.append(margn2
						+ element("tcf:" + "filmFraming", ffAttrs) + this.EOL);
				this.xmldata.append(margn1 + elementEnd("tcf:" + "duration")
						+ this.EOL);
			}
		}

		@Override
		protected void showMessage(Message message) {
			String margin = getIndent(++this._level);
			String[][] attrs = new String[3][];
			boolean hasAttr = false;
			attrs[0] = new String[] { "subMessage", null };
			attrs[1] = new String[] { "offset", null };
			attrs[2] = new String[] { "severity", null };

			String submsg = message.getSubMessage();
			if (submsg != null) {
				attrs[0][1] = submsg;
				hasAttr = true;
			}
			long offset = message.getOffset();
			if (offset > -1) {
				attrs[1][1] = Long.toString(offset);
				hasAttr = true;
			}
			if (message instanceof ErrorMessage) {
				attrs[2][1] = "error";
				hasAttr = true;
			}
			else if (message instanceof InfoMessage) {
				attrs[2][1] = "info";
				hasAttr = true;
			}
			if (hasAttr) {
				this.xmldata.append(margin
						+ element("message", attrs, message.getMessage()));
			}
			else {
				this.xmldata.append(margin
						+ element("message", message.getMessage()));
			}
			this._level--;
		}

		/***********************************************************************
		 * Searches a corresponding file ID to a given relative link
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20051028
		 * @since 20051027
		 * @param link - The link, read from the HTML file
		 * @return the file ID of the linked file
		 **********************************************************************/
		protected File getFileforLink(String link) {
			String linkPath = "";
			String linkFile = "";
			// convert MS Windows --> *NIX
			link = link.replace('\\', '/');
			// cut protocol prefix if any
			int idx = link.indexOf("://");
			if (idx > 0) {
				String urlprefix = link.substring(0, idx);
				link = link.substring(idx + 3);
				if (!urlprefix.equalsIgnoreCase("file")) {
					return null;
				}
			}
			// prepare path and file name
			idx = link.lastIndexOf("/");
			if (idx > -1) {
				linkPath = link.substring(0, idx + 1);
				if (linkPath.startsWith("./")) {
					linkPath = linkPath.substring(2);
				}
				else if (linkPath.startsWith("/")) {
					linkPath = linkPath.substring(1);
				}
				linkFile = link.substring(idx + 1);
			}
			else {
				linkFile = link;
			}

			// Erase Anchors if any
			if (linkFile.indexOf('#') > -1) {
				linkFile = linkFile.split("#")[0];
			}

			// Check for "../" and erase it if existent
			int upcount = 0;

			while (linkPath.lastIndexOf("../") > -1) {
				linkPath = linkPath.substring(linkPath.indexOf("../") + 3);
				upcount++; // keep track of how many up-steps there were
			}

			List fl = MetadataGenerator.this.processData.getFileList();

			String root = this.contentfile.getPath();

			root = root.substring((MetadataGenerator.this.processData
					.getPathToContentFiles().length()), (root
					.lastIndexOf(System.getProperty("file.separator")) + 1));

			root = root.replace('\\', '/');

			// if there were up-steps, correct the root-path
			if (upcount > 0) {

				String[] paths = root.split("/");
				String tempstr = "";
				for (int j = 0; j < paths.length - upcount; j++) {
					tempstr += paths[j] + "/";
				}
				root = tempstr;
			}

			// go through the filelist and search for a corresponding fileID
			for (int i = 0; i < fl.size(); i++) {
				File file = (File) fl.get(i);
				String filePath = file.getPath();
				filePath = filePath.substring(
						(MetadataGenerator.this.processData
								.getPathToContentFiles().length()), (filePath
								.lastIndexOf(System
										.getProperty("file.separator")) + 1));

				filePath = filePath.replace('\\', '/');

				String fileName = file.getName();

				if (fileName.equals(linkFile)
						&& !fileName.equals(this.contentfile.getName())) {
					if (filePath.equals(root + linkPath)) {
						return file;
					}
				}
			}
			return null;
		}

		/***********************************************************************
		 * Method to get linked files from an HTML file into a LmerFile object
		 * 
		 * Walks through the properties and picks out "Links"- and
		 * "Src"-properties. Calls itself recursively for deeper properties
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20050822
		 * @since 20050822
		 * @param property - The property holding the data
		 **********************************************************************/
		protected void getMetadataFromHtml(Property property) {
			// Search for linked files
			String propName = property.getName();
			if (propName.equals("Links")) {
				List propList = (List) property.getValue();
				ListIterator iter = propList.listIterator();
				while (iter.hasNext()) {
					Object val = iter.next();
					File linkedFile = getFileforLink((String) val);
					if (linkedFile != null) {

						List<File> links = MetadataGenerator.this.linkedToList
								.get(this.contentfile);

						// if there is a list of linked files already...
						if (links != null) {
							links.add(linkedFile);
						}
						// if there is no entry for this contentfile yet
						else {
							List<File> templist = new ArrayList<File>();
							templist.add(linkedFile);
							MetadataGenerator.this.linkedToList.put(
									this.contentfile, templist);
						}
					}
				}
				return;
			}

			// Search for files, linked as resources
			if (propName.equals("Src")) {
				File linkedFile = getFileforLink((String) property.getValue());
				if (linkedFile != null) {

					List<File> links = MetadataGenerator.this.linkedToList
							.get(this.contentfile);

					// if there is a list of linked files already...
					if (links != null) {
						links.add(linkedFile);
					}
					// if there is no entry for this contentfile yet
					else {
						List<File> templist = new ArrayList<File>();
						templist.add(linkedFile);
						MetadataGenerator.this.linkedToList.put(
								this.contentfile, templist);
					}
				}
				return;
			}

			// -Walk throgh the other properties the standard way-----------
			PropertyArity arity = property.getArity();
			PropertyType type = property.getType();
			boolean valueIsProperty = (type.equals(PropertyType.PROPERTY));

			if (arity.equals(PropertyArity.SCALAR)) {
				/* Just a single value */
				if (valueIsProperty) {
					getMetadataFromHtml((Property) property.getValue());
				}
			}
			else if (arity.equals(PropertyArity.LIST)) {
				List propList = (List) property.getValue();
				ListIterator iter = propList.listIterator();
				while (iter.hasNext()) {
					Object val = iter.next();
					if (valueIsProperty) {
						getMetadataFromHtml((Property) val);
					}
				}
			}
			else if (arity.equals(PropertyArity.MAP)) {
				/*
				 * For a map, the key is the "key" attribute of its
				 * corresponding value
				 */
				Map propMap = (Map) property.getValue();
				Iterator keyIter = propMap.keySet().iterator();
				while (keyIter.hasNext()) {
					Object objkey = keyIter.next();
					String keystr = objkey.toString();
					Object val = propMap.get(objkey);
					String[][] attrs = new String[1][];
					String[] keyAttr = new String[2];
					keyAttr[0] = "key";
					keyAttr[1] = keystr;
					attrs[0] = keyAttr;
					if (valueIsProperty) {

						Property pval = (Property) val;
						// If the key equals the property name,
						// suppress the key
						getMetadataFromHtml(pval);
					}
				}
			}
			else if (arity.equals(PropertyArity.SET)) {
				Set propSet = (Set) property.getValue();
				Iterator iter = propSet.iterator();
				while (iter.hasNext()) {
					Object val = iter.next();
					if (valueIsProperty) {
						getMetadataFromHtml((Property) val);
					}
				}
			}
			// -------------------------------------------------------------*/
		}

		/***********************************************************************
		 * Method to parse an "Images" property and print out its most relevant
		 * data as statistics
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20060802
		 * @since 20060802
		 * @param property - The property holding the data
		 **********************************************************************/
		protected void buildImageStatistics(Property property) {
			ArrayList<int[]> comprSchemes = new ArrayList<int[]>();

			List imagePropList = (List) property.getValue();
			ListIterator imageIter = imagePropList.listIterator();
			while (imageIter.hasNext()) {
				Property tempimageprop = (Property) imageIter.next();

				List imageValuesPropList = (List) tempimageprop.getValue();
				ListIterator imageValuesIter = imageValuesPropList
						.listIterator();
				boolean nisoFound = false;
				while (imageValuesIter.hasNext() && !nisoFound) {
					Property tempvalueprop = (Property) imageValuesIter.next();

					if (tempvalueprop.getType().equals(
							PropertyType.NISOIMAGEMETADATA)) {
						nisoFound = true;

						NisoImageMetadata niso = (NisoImageMetadata) tempvalueprop
								.getValue();

						// Read and count CompressionSchemes
						int tempScheme = niso.getCompressionScheme();

						ListIterator schemeIter = comprSchemes.listIterator();
						boolean found = false;
						while (schemeIter.hasNext() && !found) {
							int[] tempint = (int[]) schemeIter.next();
							if (tempint[0] == tempScheme) {
								tempint[1]++;
								found = true;
							}
						}
						if (!found) {
							int[] temp = { tempScheme, 1 };
							comprSchemes.add(temp);
						}
					}
				}
			}

			// Build JHOVE xml representation of gathered image statistics

			if (!comprSchemes.isEmpty()) {
				StringBuffer imagesXml = new StringBuffer("");

				// Number of Images

				imagesXml.append("<property> <name>ImageStatistics</name>"
						+ "<values arity=\"List\"" + " type=\"Property\">");

				imagesXml.append("<property> <name>TotalNumberOfImages</name>"
						+ " <values arity=\"Scalar\" "
						+ " type=\"Integer\"> <value>" + imagePropList.size()
						+ "</value>");
				imagesXml.append("</values></property>");

				imagesXml.append("<property> <name>CompressionSchemes</name>"
						+ "<values arity=\"List\"" + " type=\"Property\">");

				ListIterator comprSchemesIter = comprSchemes.listIterator();
				while (comprSchemesIter.hasNext()) {
					int[] tempComprScheme = (int[]) comprSchemesIter.next();

					imagesXml
							.append("<property> <name>CompressionScheme</name>"
									+ " <values arity=\"List\""
									+ " type=\"Property\">");

					// Print Compression Scheme Code
					imagesXml.append("<property> <name>SchemeCode</name>"
							+ " <values arity=\"Scalar\" "
							+ " type=\"Integer\"> <value>" + tempComprScheme[0]
							+ "</value>");
					imagesXml.append("</values></property>");

					// Print Compression Scheme Name
					imagesXml.append("<property> <name>SchemeName</name>"
							+ " <values arity=\"Scalar\" "
							+ " type=\"String\"> <value>"
							+ compressionSchemeIntegerRepresentation(
									tempComprScheme[0],
									NisoImageMetadata.COMPRESSION_SCHEME,
									NisoImageMetadata.COMPRESSION_SCHEME_INDEX)
							+ "</value>");
					imagesXml.append("</values></property>");

					// Print Count
					imagesXml.append("<property> <name>NumberOfImages</name>"
							+ " <values arity=\"Scalar\" "
							+ " type=\"Integer\"> <value>" + tempComprScheme[1]
							+ "</value>");
					imagesXml.append("</values></property>");

					imagesXml.append("</values></property>");
				}

				imagesXml.append("</values></property>");
				imagesXml.append("</values></property>");

				this.xmldata.append(imagesXml);
			}
		}

		/***********************************************************************
		 * Method to find the text representation of a cmpression scheme number
		 * representation
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20060802
		 * @since 20060802
		 * @param n - The value to find
		 * @param labels - The String array with text values
		 * @param index - The int array with the int values
		 * @return String - The representation of the searched value
		 **********************************************************************/
		private String compressionSchemeIntegerRepresentation(int n,
				String[] labels, int[] index) {
			int idx = -1;
			for (int i = 0; i < index.length; i++) {
				if (n == index[i]) {
					idx = i;
					break;
				}
			}
			if (idx > -1) {
				return labels[idx];
			}

			// If we don't get a match, or do get an exception
			return "Unknown";
		}

		/***********************************************************************
		 * Method to parse an "Pages" property and print out its most relevant
		 * data as statistics.
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20060802
		 * @since 20060802
		 * @param property - The property holding the data
		 **********************************************************************/
		protected void buildPageStatistics(Property property) {


			List pagesList = (List) property.getValue();

			// Build JHOVE xml representation of gathered page statistics

			if (!pagesList.isEmpty()) {
				StringBuffer pagesXml = new StringBuffer("");

				// Number of Pages

				pagesXml.append("<property> <name>TotalNumberOfPages</name>"
						+ " <values arity=\"Scalar\" "
						+ " type=\"Integer\"> <value>" + pagesList.size()
						+ "</value></values></property>");

				this.xmldata.append(pagesXml);
			}
		}

		/***********************************************************************
		 * Method to parse an "Fonts" property and print out its most relevant
		 * data as statistics.
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20060802
		 * @since 20060802
		 * @param property - The property holding the data
		 **********************************************************************/
		protected void buildFontStatistics(Property property) {

			List fontTypesList = (List) property.getValue();
			ListIterator fontTypesIter = fontTypesList.listIterator();

			StringBuffer fontsXml = new StringBuffer(
					"<property> <name>FontStatistics</name><values arity="
							+ "\"List\" type=\"Property\">");

			while (fontTypesIter.hasNext()) {
				Property tempfonttype = (Property) fontTypesIter.next();

				ArrayList fonts = getFontList(tempfonttype);

				fontsXml.append("<property> <name>" + tempfonttype.getName()
						+ "</name><values arity=\"List\""
						+ " type=\"Property\">");

				for (int i = 0; i < fonts.size(); i++) {
					fontsXml
							.append("<property> <name>Font</name><values arity="
									+ "\"List\" type=\"Property\">");
					fontsXml.append("<property> <name>Name</name> <values "
							+ "arity=\"Scalar\" type=\"String\"> <value>"
							+ ((String[]) fonts.get(i))[0]
							+ "</value></values></property>");
					fontsXml.append("<property> <name>Embedded</name>"
							+ " <values arity=\"Scalar\" "
							+ " type=\"Boolean\"> <value>"
							+ ((String[]) fonts.get(i))[1]
							+ "</value></values></property>");
					fontsXml.append("</values></property>");

				}
				fontsXml.append("</values></property>");
			}

			fontsXml.append("</values></property>");

			this.xmldata.append(fontsXml);

		}

		/***********************************************************************
		 * Method to get a list of all "Fonts" for a "Fonttype"
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20060802
		 * @since 20060802
		 * @return ArrayList - The List of all fonts belonging to the given type
		 * @param fonttype - The property holding the data
		 **********************************************************************/
		private ArrayList getFontList(Property fonttype) {

			ArrayList<String[]> fontList = new ArrayList<String[]>();

			List tempFontsList = (List) fonttype.getValue();

			ListIterator FontsIter = tempFontsList.listIterator();

			while (FontsIter.hasNext()) {
				Property tempFont = (Property) FontsIter.next();

				String baseFont = null;
				boolean embedded = false;

				Property tempProperty = tempFont.getByName("BaseFont");

				/*
				 * A String array contains the gathered statistic infos. [0]:
				 * Fontname [1]: Embedded - true or false
				 */

				if (tempProperty != null) {
					baseFont = (String) tempProperty.getValue();

					int lio = baseFont.lastIndexOf('+');

					String temp = baseFont.substring(lio + 1);

					if (temp.length() > 0) {
						baseFont = temp;
					}

					String[] temp2 = baseFont.split("~");
					if (temp2.length > 1) {
						baseFont = temp2[0];
					}
				}

				tempProperty = tempFont.getByName("FontFile");

				if (tempProperty != null) {
					embedded = ((Boolean) tempProperty.getValue())
							.booleanValue();
				}
				else {
					tempProperty = tempFont.getByName("FontFile2");
					if (tempProperty != null) {
						embedded = ((Boolean) tempProperty.getValue())
								.booleanValue();
					}
					else {
						tempProperty = tempFont.getByName("FontFile3");
						if (tempProperty != null) {
							embedded = ((Boolean) tempProperty.getValue())
									.booleanValue();
						}
					}
				}

				if (baseFont != null) {

					ListIterator fontIter = fontList.listIterator();
					boolean found = false;
					while (fontIter.hasNext() && !found) {
						String[] tempstring = (String[]) fontIter.next();
						if (tempstring[0].equalsIgnoreCase(baseFont)) {
							found = true;
						}
					}
					if (!found) {
						if (embedded) {
							String[] temp = { baseFont, "true" };
							fontList.add(temp);
						}
						else {
							String[] temp = { baseFont, "false" };
							fontList.add(temp);
						}

					}
				}
			}
			return fontList;
		}

		/***********************************************************************
		 * Method to parse a "File Structure" property and print out its most
		 * relevant data as statistics.
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20061222
		 * @since 20061222
		 * @param property - The property holding the data
		 **********************************************************************/
		protected void buildFileStructureStatistics(Property property) {
			HashMap<String, Integer> contents = new HashMap<String, Integer>();

			List fsPropList = (List) property.getValue();
			ListIterator fsIter = fsPropList.listIterator();
			// Go through the list of file entries...
			while (fsIter.hasNext()) {
				Property tempFileEntryprop = (Property) fsIter.next();

				// if it is a directory, gather informations for all
				// contained files
				if (tempFileEntryprop.getByName("Type").getValue().equals(
						"Directory")) {
					recurseContentParser(tempFileEntryprop, contents);
				}
				else { // if it is a file...

					String type = tempFileEntryprop.getName();
					// ... get its file extension
					try {
						type = type.substring(type.lastIndexOf('.') + 1);
					}
					catch (IndexOutOfBoundsException e) {
						type = "Without file extension";
					}
					// and count its appearances.
					Integer appearances = contents.get(type);

					if (appearances == null) {
						contents.put(type, new Integer(1));
					}
					else {
						contents.put(type, new Integer(
								appearances.intValue() + 1));
					}

				}
			}

			// Build JHOVE xml representation of gathered file structure
			// statistics

			if (!contents.isEmpty()) {
				StringBuffer fsXml = new StringBuffer("");

				// Menge der Bilder

				fsXml.append("<property> <name>FileStructureStatistics</name>"
						+ "<values arity=\"List\"" + " type=\"Property\">");
				fsXml.append("<property> <name>Type Appearances</name>"
						+ "<values arity=\"List\"" + " type=\"Property\">");



				Iterator typesIter = contents.keySet().iterator();
				while (typesIter.hasNext()) {
					String type = (String) typesIter.next();

					fsXml.append("<property> <name>" + type + "</name>"
							+ " <values arity=\"Scalar\""
							+ " type=\"Integer\">" + "<value>"
							+ contents.get(type).intValue() + "</value>");
					fsXml.append("</values></property>");
				}

				fsXml.append("</values></property>");
				fsXml.append("</values></property>");

				this.xmldata.append(fsXml);
			}
		}

		/***********************************************************************
		 * Method to parse a "Directory" property and add its contents to the
		 * main contents list
		 * 
		 * @author Matthias Neubauer, German National Library
		 * @version 20061222
		 * @since 20061222
		 * @param property - The property holding the data
		 * @param contents - the list of contents
		 **********************************************************************/
		private void recurseContentParser(Property property,
				HashMap<String, Integer> contents) {

			Integer dirs = contents.get("Directories");

			if (dirs == null) {
				contents.put("Directories", new Integer(1));
			}
			else {
				contents.put("Directories", new Integer(dirs.intValue() + 1));
			}

			Property contentsProp = property.getByName("Content");

			if (contentsProp != null) {

				List subcontents = (List) contentsProp.getValue();
				ListIterator subIter = subcontents.listIterator();

				while (subIter.hasNext()) {
					Property tempFileEntryprop = (Property) subIter.next();

					if (tempFileEntryprop.getByName("Type").getValue().equals(
							"Directory")) {
						recurseContentParser(tempFileEntryprop, contents);
					}
					else {

						String type = tempFileEntryprop.getName();

						try {
							type = type.substring(type.lastIndexOf('.') + 1);
						}
						catch (IndexOutOfBoundsException e) {
							type = "Without file extension";
						}

						Integer appearances = contents.get(type);

						if (appearances == null) {
							contents.put(type, new Integer(1));
						}
						else {
							contents.put(type, new Integer(appearances
									.intValue() + 1));
						}
					}
				}
			}
		}

		/***********************************************************************
		 * @return Returns the xmlData.
		 **********************************************************************/
		public String getXmlData() {
			return this.xmldata.toString();
		}

		/***********************************************************************
		 * @param _contentfile - The contetnfile to set
		 **********************************************************************/
		public void setContentFile(File _contentfile) {
			this.contentfile = _contentfile;
		}
	}

}
