/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / MigrateFiles.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 **********************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import gov.loc.mets.MetsType;

import java.io.File;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import de.ddb.lmeRprocess.LmerProcess;
import de.ddb.lmeRprocess.LmerProcess.Steps;
import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.formats.Uof;

/*******************************************************************************
 * This class converts all files of the given DIAS file format ID to another
 * format and fills the files' provenance metadata.
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20070625
 * @since 20070504
 ******************************************************************************/

public class MigrateFiles implements ActionModule
{

	// CONSTANTS **********************************************************

	public static final String	MIGRATION_FILE_TABLE			= "MIGRATION_FILE_TABLE";

	// STATE (Instance variables) ****************************************

	ProcessData					processData;
	Step						step;

	List<String>				migrationToolExecutionCommand	= new ArrayList<String>();
	String						runtimeOutput;
	String						runtimeErrorOutput;

	String						sourceFileName					= null;
	String						destFileName					= null;

	public void setDestFileName(String destFileName) {
		this.destFileName = destFileName;
	}

	public void setSourceFileName(String sourceFileName) {
		this.sourceFileName = sourceFileName;
	}

	/***************************************************************************
	 * Default constructor.
	 **************************************************************************/
	public MigrateFiles() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ****************

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 */
	public void go() {
		// Get the UOF metadata object.
		Uof uof = null;
		try {
			uof = (Uof) this.processData.getMetadata();
		}
		catch (Exception e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Error accessing metadata");
		}

		File sourceFile = new File(this.sourceFileName);
		File destFile = new File(this.destFileName);

		// Get a copy of the LMER file and its ID.
		String lmerFileId = null;
		try {
			lmerFileId = uof.getFileTechMdId(sourceFile);
		}
		catch (Exception e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Error extracting LMER "
					+ "file data", sourceFile);
			return;
		}

		// Process the argument list.
		String args = "";

		Iterator argIter = this.migrationToolExecutionCommand.iterator();
		while (argIter.hasNext()) {
			String arg = (String) argIter.next();

			// Replace "SOURCEFILE" with the input file name and "DESTFILE" with
			// the output file name.
			// arg = arg.replace("SOURCEFILE", "\"" +
			// sourceFile.getAbsolutePath() + "\"");
			// arg = arg.replace("DESTFILE", "\"" + destFile.getAbsolutePath() +
			// "\"");
			arg = arg.replace("SOURCEFILE", sourceFile.getAbsolutePath());
			arg = arg.replace("DESTFILE", destFile.getAbsolutePath());

			// TODO Make quotes work for different operating systems!
			args += arg + " ";
		}

		// String[] args = new String[0];
		// args = this.migrationToolExecutionCommand.toArray(args);

		System.out.println(args);

		// Call the executer here!
		Runtime rt = Runtime.getRuntime();
		Process proc = null;

		try {
			// TODO Better submit the arguments as an array of strings.
			proc = rt.exec(args);

			// TODO What to do with the error messages?
			// InputStream processErrorOutput = new BufferedInputStream(
			// runningProcess.getErrorStream());
		}
		catch (Exception e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Error executing command line: "
					+ args);
		}

		try {
			// Wait for the process to exit and print status.
			switch (proc.waitFor()) {
				case 0:
					// Success. Put source and destination file name into hash
					// table.
					HashMap<String, String> filenames = (HashMap<String, String>) this.processData
							.getCustomData().get(
									MigrateFiles.MIGRATION_FILE_TABLE);
					if (filenames == null) {
						filenames = new HashMap<String, String>();
					}
					filenames.put(sourceFile.getAbsolutePath(), destFile
							.getAbsolutePath());
					this.processData.getCustomData().put(
							MigrateFiles.MIGRATION_FILE_TABLE, filenames);
					break;
				default:
					this.step.setStatus(Status.ERROR, "Exit value of "
							+ "the migrating tool is " + proc.exitValue());
					return;
			}
		}
		catch (InterruptedException e) {
			this.step.setStatus(Status.ERROR, "Interrupt error while "
					+ "executing command line");
			return;
		}

		// Change the file name in the file sec.
		MetsType mets = (MetsType) uof.getMetadataImplementation();
		gov.loc.mets.FileGrpType.File fileArray[] = mets.getFileSec()
				.getFileGrpArray(0).getFileArray();
		for (int i = 0; i < fileArray.length; i++) {
			String lastTechMdId = (String) fileArray[i].getADMID().get(
					fileArray[i].getADMID().size() - 1);
			if (lmerFileId.equals(lastTechMdId)) {
				fileArray[i].getFLocatArray(0).setHref(
						Uof.FILE_LOCAT_HREF_PREFIX
								+ destFile.toURI().toString().substring(
										uof.getBasePath().toString().length()));
			}
		}

		// Create an LmerProcess object for each file.
		LmerProcess provMd = LmerProcess.Factory.newInstance();

		// Add the files' information of the migration.
		// TODO Where do we get some of the data from? Maybe from the config
		// file? Maybe from the migration tool used? Which can we take from the
		// conversion process?
		provMd.setComments("files - comments");
		Calendar cal = Calendar.getInstance();
		cal.setTimeInMillis(System.currentTimeMillis());
		provMd.setCompletionDate(cal);
		provMd.setPermission("files - permission");
		cal = Calendar.getInstance();
		cal.setTimeInMillis(System.currentTimeMillis());
		provMd.setPermissionDate(cal);
		provMd.setProcessCreator("files - creator");
		provMd.setPurpose("files - purpose");
		provMd.setResult("files - result");
		Steps _steps = provMd.addNewSteps();
		de.ddb.lmeRprocess.LmerProcess.Steps.Step _step = _steps.addNewStep();
		_step.setNUMBER(BigInteger.valueOf(1));
		_step.setStringValue("files - step");

		// Add the newly created metadata object to the UOF.
		try {
			uof.addProvMd(destFile, provMd);
		}
		catch (Exception e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Error adding file "
					+ "provenance metadata to the UOF", destFile);
			return;
		}

		this.processData.setMetadata(uof);

		// Finally delete the migrated original file.
		if (!sourceFile.delete()) {
			this.step.setStatus(Status.ERROR, "File deletion failed",
					sourceFile);
			return;
		}

		// Re-set the path to the content files
		this.processData.setPathToContentFiles(this.processData
				.getPathToContentFiles());

		this.step.setStatus(Status.DONE, "File migration complete");
	}

	// GET & SET METHODS *******************************************************

	/***************************************************************************
	 * @param processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData processData) {
		this.processData = processData;
	}

	/***************************************************************************
	 * @param step The step to set.
	 **************************************************************************/
	public void setStep(Step step) {
		this.step = step;
	}

	/***************************************************************************
	 * @param migrationToolExecutionCommand The string to add to the CLI arg
	 *            list.
	 **************************************************************************/
	public void setMigrationToolExecutionCommand(
			String migrationToolExecutionCommand) {
		this.migrationToolExecutionCommand.add(migrationToolExecutionCommand);
	}

}
