/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / PrepareMigration.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 **********************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import gov.loc.mets.MdSecType;
import gov.loc.mets.MetsType;

import java.io.File;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;

import de.ddb.lmeRfile.LmerFile;
import de.ddb.lmeRobject.LmerObject;
import de.ddb.lmeRprocess.LmerProcess;
import de.ddb.lmeRprocess.LmerProcess.Steps;
import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.formats.Uof;
import de.langzeitarchivierung.kopal.retrieval.DiasAccess;
import de.langzeitarchivierung.kopal.retrieval.DiasAccessResponse;
import de.langzeitarchivierung.kopal.util.Configurator;

/*******************************************************************************
 * Create a new Uof object and tranfer all needed metadata from the old METS
 * file.
 * 
 * TODO At the moment we only copy the sections (techMd and digiprovMd) that are
 * referenced both from the fileGroup and the files. Mustn't we copy ALL
 * sections without checking any references?
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20070710
 * @since 20070507
 ******************************************************************************/

public class PrepareMigration implements ActionModule
{

	// STATE (Instance variables) ****************************************

	ProcessData		processData;
	Step			step;

	Uof				uof;
	MetsType		mets;
	Uof				lastUof;
	MetsType		lastMets;
	String			migrateMetadataRecordCreator;

	String			provmdComments		= "";
	String			provmdPermission	= "";
	String			provmdCreator		= "";
	String			provmdPurpose		= "";
	String			provmdResult		= "";
	List<String>	provmdSteps			= new ArrayList<String>();

	/***************************************************************************
	 * Default constructor.
	 **************************************************************************/
	public PrepareMigration() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ****************

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 */
	public void go() {
		try {
			// Get the current Uof object and its METS implementation.
			this.uof = (Uof) this.processData.getMetadata();
			this.uof.setBasePath(new File(this.processData
					.getPathToContentFiles()));
			this.mets = (MetsType) this.uof.getMetadataImplementation();

			// Load the existing mets.xml file as a template in new Uof object
			// and get its METS implementation.
			File metsFilename = new File(this.processData
					.getPathToContentFiles(), "mets.xml");
			if (!metsFilename.exists()) {
				this.step.setStatus(Status.ERROR, "No mets.xml found");
				return;
			}
			this.lastUof = new Uof();
			this.lastUof.useTemplate(metsFilename);
			this.lastUof.setBasePath(new File(this.processData
					.getPathToContentFiles()));
			this.lastMets = (MetsType) this.lastUof.getMetadataImplementation();

			// Put all needed data into the current Uof object.
			// NOTE: Depends on the METS template used for creating the new Uof
			// object, e.g. if there is no lmerObject defined in the template,
			// you have to define it here!
			this.step.setStatus(Status.RUNNING, "Transferring metadata from "
					+ "the existing METS file");

			// Transfer all descriptive metadata sections.
			this.mets.setDmdSecArray(this.lastMets.getDmdSecArray());

			// Transfer some lmerObject data.
			transferObjectTechMd();

			// Tranfer data from the file group.
			this.mets.getFileSec().getFileGrpArray(0).setADMID(
					this.lastMets.getFileSec().getFileGrpArray(0).getADMID());
			this.mets.getAmdSecArray(0).getTechMDArray(0).setID(
					this.lastUof.getObjectTechMdId());

			// Transfer object's provenance metadata.
			transferObjectDigiprovMd();

			// Transfer all file data, techMd and digiprovMd.
			transferFileMd();

			// Transfer the structMap data.
			this.mets.setStructMapArray(this.lastMets.getStructMapArray());

			this.step.setStatus(Status.RUNNING, "Metadata transfer complete");

			// Set new object's provenance metadata.
			this.step.setStatus(Status.RUNNING, "Updating provenance "
					+ "metadata section");

			// Check the metadata record creator (must be KOPAL DIAS for a DIAS
			// migration).
			if (!this.lastUof.getLmerObjectCopy().getMetadataRecordCreator()
					.equals(this.migrateMetadataRecordCreator)) {
				this.step.setStatus(Status.ERROR, "The metadata record "
						+ "creator of the given METS file is not equal to "
						+ "the expected '" + this.migrateMetadataRecordCreator
						+ "'");
				return;
			}

			// Add an lmerProcess digiprovMd section for the object.
			createDigiprovMd();

			this.step.setStatus(Status.RUNNING, "Update of the provenance "
					+ "metadata section complete");

			// At last delete the old mets.xml file and remove the METS file
			// from the LMER file list.
			if (metsFilename.delete()) {
				this.processData.getFileList().remove(metsFilename);
				this.step.setStatus(Status.RUNNING, "METS file deletion "
						+ "complete");
			}
			else {
				this.step.setStatus(Status.ERROR, "METS file deletion "
						+ "failed", metsFilename);
				return;
			}

			this.step.setStatus(Status.DONE, "Migration preparation complete");
		}
		catch (Exception e) {
			e.printStackTrace();
			this.step.setStatus(Status.ERROR, "Migration preparation failed. "
					+ "System message: " + e.getMessage());
		}
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ********

	/***************************************************************************
	 * Transfer some lmerObject information. The DIP's status and its object
	 * version doesn't need to be set for a SIP's ingestion. So we leave it out
	 * here.
	 * 
	 * NOTE: The object version must be checked by the Migration Manager. We
	 * want each SIP of one migration tree to have a unique asset version!
	 * 
	 * @throws Exception
	 **************************************************************************/
	private void transferObjectTechMd() throws Exception {
		LmerObject lastLmerObject = this.lastUof.getLmerObjectCopy();
		LmerObject lmerObject = this.uof.getLmerObjectCopy();

		if (lastLmerObject.getName() != null) {
			lmerObject.setName(lastLmerObject.getName());
		}
		if (lastLmerObject.getPersistentIdentifier() != null) {
			lmerObject.setPersistentIdentifier(lastLmerObject
					.getPersistentIdentifier());
		}
		if (lastLmerObject.getMasterCreationDate() != null) {
			lmerObject.setMasterCreationDate(lastLmerObject
					.getMasterCreationDate());
		}
		if (lastLmerObject.getStartFile() != null) {
			lmerObject.setStartFile(lastLmerObject.getStartFile());
		}

		this.uof.setLmerObject(lmerObject);
	}

	/***************************************************************************
	 * Copy all the object's lmerProcess provenance metadata sections referenced
	 * from the ADMID list of the file group.
	 **************************************************************************/
	private void transferObjectDigiprovMd() {
		List lastDigiprovObjectMdIdList = this.lastMets.getFileSec()
				.getFileGrpArray(0).getADMID();
		lastDigiprovObjectMdIdList = lastDigiprovObjectMdIdList.subList(0,
				lastDigiprovObjectMdIdList.size() - 1);

		MdSecType lastDigiprovObjectMdArray[] = this.lastMets.getAmdSecArray(0)
				.getDigiprovMDArray();

		for (int i = 0; i < lastDigiprovObjectMdArray.length; i++) {
			if (lastDigiprovObjectMdIdList
					.contains(lastDigiprovObjectMdArray[i].getID())) {
				this.mets.getAmdSecArray(0).addNewDigiprovMD().set(
						lastDigiprovObjectMdArray[i]);
			}
		}
	}

	/***************************************************************************
	 * Transfers all data concerning each file to the new Uof object.
	 * 
	 * @throws Exception
	 **************************************************************************/
	private void transferFileMd() throws Exception {
		gov.loc.mets.FileGrpType.File lastFiles[] = this.lastMets.getFileSec()
				.getFileGrpArray(0).getFileArray();
		MdSecType lastTechMdArray[] = this.lastMets.getAmdSecArray(0)
				.getTechMDArray();

		for (int i = 0; i < lastFiles.length; i++) {
			// Set some values from the file group.
			gov.loc.mets.FileGrpType.File files = this.mets.getFileSec()
					.getFileGrpArray(0).addNewFile();
			files.setADMID(lastFiles[i].getADMID());
			files.setID(lastFiles[i].getID());
			files.setFLocatArray(lastFiles[i].getFLocatArray());

			// Create the techMd section and add the lmerFile:format tag.
			String lastTechMdId = (String) lastFiles[i].getADMID().get(
					lastFiles[i].getADMID().size() - 1);

			// Find the techMd section with the correct techMd ID.
			for (int j = 0; j < lastTechMdArray.length; j++) {
				if (lastTechMdArray[j].getID().equals(lastTechMdId)) {
					MdSecType techMd = this.mets.getAmdSecArray(0)
							.addNewTechMD();
					techMd.setID(lastTechMdId);
					techMd.setMdWrap(lastTechMdArray[j].getMdWrap());
					if (techMd.getMdWrap().isSetBinData()) {
						techMd.getMdWrap().unsetBinData();
					}
					if (techMd.getMdWrap().isSetLABEL()) {
						techMd.getMdWrap().unsetLABEL();
					}
					if (techMd.getMdWrap().isSetMIMETYPE()) {
						techMd.getMdWrap().unsetMIMETYPE();
					}
					if (techMd.getMdWrap().isSetOTHERMDTYPE()) {
						techMd.getMdWrap().unsetOTHERMDTYPE();
					}
					if (techMd.getMdWrap().isSetXmlData()) {
						LmerFile lmerFile = LmerFile.Factory.newInstance();
						lmerFile.setFormatArray(this.lastUof.getLmerFileCopy(
								lastTechMdId).getFormatArray());
						techMd.getMdWrap().getXmlData().set(lmerFile);
					}
				}
			}

			// Copy all lmerProcess provenance metadata sections referenced
			// from the ADMID list of each file.
			List lastDigiprovFileMdIdList = lastFiles[i].getADMID();
			lastDigiprovFileMdIdList = lastDigiprovFileMdIdList.subList(0,
					lastDigiprovFileMdIdList.size() - 1);
			MdSecType lastDigiprovFileMdArray[] = this.lastMets.getAmdSecArray(
					0).getDigiprovMDArray();
			for (int j = 0; j < lastDigiprovFileMdArray.length; j++) {
				if (lastDigiprovFileMdIdList
						.contains(lastDigiprovFileMdArray[j].getID())) {
					this.mets.getAmdSecArray(0).addNewDigiprovMD().set(
							lastDigiprovFileMdArray[j]);
				}
			}
		}
	}

	/***************************************************************************
	 * Create a new lmerProcess object and add it to the new Uof.
	 * 
	 * @throws Exception
	 **************************************************************************/
	private void createDigiprovMd() throws Exception {
		// Create an LmerProcess object, put the internal DIAS ID to
		// oldObjectIdentifier and add the oldMetadataRecordCreator.
		LmerProcess provMd = LmerProcess.Factory.newInstance();
		provMd.setOldMetadataRecordCreator(this.lastUof.getLmerObjectCopy()
				.getMetadataRecordCreator());
		provMd.setOldObjectIdentifier(this.lastMets.getOBJID());

		// TODO Change the following code block to the one string allocation if
		// the DIAS PRD system still returns no internal DIAS ID in the METS
		// file.
		// Replacement START
		// String oldObjectVersion = "1";
		// Replacement END

		// Replacement START
		DiasAccess da = new DiasAccess();
		Configurator.configure(da);
		DiasAccessResponse dar = DiasAccess.parseDiasResponse(da
				.requestShortMdByIntId(provMd.getOldObjectIdentifier(),
						DiasAccess.RESPONSE_FORMAT_XML));
		String oldObjectVersion = dar.getMetadata().get(0).getAssetVersion();
		// Replacement END

		if (oldObjectVersion == null || oldObjectVersion.equals("")) {
			throw new Exception("No old object version given in the "
					+ "custom data");
		}
		provMd.setOldObjectVersion(oldObjectVersion);

		// To add the general information of the migration, please configure
		// either in the config or in the policies file.
		provMd.setComments(this.provmdComments);
		Calendar cal = Calendar.getInstance();
		cal.setTimeInMillis(System.currentTimeMillis());
		provMd.setCompletionDate(cal);
		provMd.setPermission(this.provmdPermission);
		cal = Calendar.getInstance();
		cal.setTimeInMillis(System.currentTimeMillis());
		provMd.setPermissionDate(cal);
		provMd.setProcessCreator(this.provmdCreator);
		provMd.setPurpose(this.provmdPurpose);
		provMd.setResult(this.provmdResult);
		// Add all the steps from the config file.
		if (!this.provmdSteps.isEmpty()) {
			Steps _steps = provMd.addNewSteps();
			Iterator iter = this.provmdSteps.iterator();
			int count = 0;
			while (iter.hasNext()) {
				de.ddb.lmeRprocess.LmerProcess.Steps.Step _step = _steps
						.addNewStep();
				_step.setNUMBER(BigInteger.valueOf(count));
				_step.setStringValue((String) iter.next());
				count++;
			}
		}

		// Create new UOF object, add the newly created objects and put it
		// in the processData.
		this.uof.addProvMd(provMd);
		this.processData.setMetadata(this.uof);
	}

	// GET & SET METHODS *******************************************************

	/***************************************************************************
	 * @param processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData processData) {
		this.processData = processData;
	}

	/***************************************************************************
	 * @param step The step to set.
	 **************************************************************************/
	public void setStep(Step step) {
		this.step = step;
	}

	/***************************************************************************
	 * @return Returns the migrateMetadataRecordCreator.
	 **************************************************************************/
	public String getMigrateMetadataRecordCreator() {
		return this.migrateMetadataRecordCreator;
	}

	/***************************************************************************
	 * @param migrateMetadataRecordCreator The migrateMetadataRecordCreator to
	 *            set.
	 **************************************************************************/
	public void setMigrateMetadataRecordCreator(
			String _migrateMetadataRecordCreator) {
		this.migrateMetadataRecordCreator = _migrateMetadataRecordCreator;
	}

	/***************************************************************************
	 * @param provmdComments The provmdComments to set.
	 **************************************************************************/
	public void setProvmdComments(String provmdComments) {
		this.provmdComments = provmdComments;
	}

	/***************************************************************************
	 * @param provmdCreator The provmdCreator to set.
	 **************************************************************************/
	public void setProvmdCreator(String provmdCreator) {
		this.provmdCreator = provmdCreator;
	}

	/***************************************************************************
	 * @param provmdPermission The provmdPermission to set.
	 **************************************************************************/
	public void setProvmdPermission(String provmdPermission) {
		this.provmdPermission = provmdPermission;
	}

	/***************************************************************************
	 * @param provmdPurpose The provmdPurpose to set.
	 **************************************************************************/
	public void setProvmdPurpose(String provmdPurpose) {
		this.provmdPurpose = provmdPurpose;
	}

	/***************************************************************************
	 * @param provmdResult The provmdResult to set.
	 **************************************************************************/
	public void setProvmdResult(String provmdResult) {
		this.provmdResult = provmdResult;
	}

	/***************************************************************************
	 * @param provmdSteps The provmdSteps to set.
	 **************************************************************************/
	public void setProvmdSteps(String _provmdStep) {
		this.provmdSteps.add(_provmdStep);
	}

}
