/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / SubmitSipToClientLoader.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.text.DecimalFormat;
import java.util.Timer;
import java.util.TimerTask;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;

/*******************************************************************************
 * This class submits an asset with custom protocol to the client loader, which
 * is another instance of the WorkflowTool.class.
 * 
 * PRECONDITION: A Server must be running to start a ClientLoaderServerThread as
 * a ServerThread to exchange data (in this case the assets filename, the used
 * policy and the asset itself).
 * 
 * RESULT: The asset is submitted to the client loader and stored locally on a
 * client loader's file location.
 * 
 * TODO Use the java.nio class for faster transmission or maybe use SSH for
 * security reasons.
 * 
 * @author Stefan Funk, SUB Göttingen
 * @version 20070129
 * @see de.langzeitarchivierung.kopal.processstarter
 * @since 20050628
 ******************************************************************************/

public class SubmitSipToClientLoader implements ActionModule
{

	// STATE (Instance variables) ******************************************

	protected ProcessData			processData;
	protected Step					step;

	protected long					totalTime;
	protected int					bytesSubmitted;
	protected int					fileLength;

	private String					clientLoaderServer;
	private int						clientLoaderPort;
	private Socket					outputServerSocket;
	private BufferedOutputStream	fileWriter;
	private String					submitPolicyName			= "";
	private File					sourceFile					= null;
	private long					fileSubmitNotificationTime	= 500;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Default constructor
	 **************************************************************************/
	public SubmitSipToClientLoader() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 **************************************************************************/
	public void go() {
		// Get the assets path from the SIP.
		this.sourceFile = this.processData.getSipLocation();

		try {
			prepareConnectionToServer(this.clientLoaderServer,
					this.clientLoaderPort);
			connectToServer(this.clientLoaderServer, this.clientLoaderPort);
			submitSipToServer(this.sourceFile, this.submitPolicyName);
			closeConnectionToServer();
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, e.getMessage());
			e.printStackTrace();
			return;
		}

		// All steps have been processed correctly.
		this.step.setStatus(Status.DONE, "Module has been processed correctly");
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ****

	/***************************************************************************
	 * The connection to the client loader will be prepared here, which means
	 * that the ClientLoaderServerThread is asked for a port to submit one asset
	 * and it's ingest policy.
	 * 
	 * @param serverAddress The client loader's server address as a string.
	 * @param port The port of the client loader server.
	 * @throws IOException
	 **************************************************************************/
	private void prepareConnectionToServer(String serverAddress, int port)
			throws IOException {
		String errorMessage = "Unable to contact the client loader server";

		try {
			// Contact the interaction manager.
			Socket negotiatePortSocket = new Socket(serverAddress, port);
			DataInputStream negotiatePortStream = new DataInputStream(
					negotiatePortSocket.getInputStream());

			// Get the port needed.
			this.clientLoaderPort = negotiatePortStream.readInt();
			negotiatePortStream.readInt();
			negotiatePortStream.close();
		}
		catch (Exception e) {
			e.printStackTrace();
			throw new IOException(errorMessage);
		}
	}

	/***************************************************************************
	 * The connection between the thread that awaits the SIP (WorkflowTool
	 * acting as a client loader) and the WorkflowTool acting as an assetBuilder
	 * is made up using the received port.
	 * 
	 * @param serverAddress - The client loader's server address.
	 * @param serverPort - The client loader's port to exchange data.
	 * @throws IOException
	 **************************************************************************/
	private void connectToServer(String serverAddress, int serverPort)
			throws IOException {
		String errorMessage = "Unable to connect to the client loader";

		try {
			this.step.setStatus(Status.RUNNING, "Connecting to the client "
					+ "loader '" + serverAddress + "'");

			this.outputServerSocket = new Socket(serverAddress, serverPort);

			this.step.setStatus(Status.RUNNING, "Connected to the client "
					+ "loader");
		}
		catch (Exception e) {
			e.printStackTrace();
			throw new IOException(errorMessage);
		}
	}

	/***************************************************************************
	 * The asset and the ingest policy is submitted to the client loader using a
	 * socket.
	 * 
	 * @param file - The file object of the asset to submit.
	 * @param tempIngestPolicyName - The ingest policy of the asset to submit.
	 * @throws IOException
	 **************************************************************************/
	private void submitSipToServer(File file, String tempIngestPolicyName)
			throws IOException {
		String errorMessage = "Could not submit the asset '" + file.getName()
				+ "' to the client loader";
		long startTime = System.currentTimeMillis();
		long endTime;

		try {
			ObjectOutputStream outputStream = new ObjectOutputStream(
					new BufferedOutputStream(this.outputServerSocket
							.getOutputStream()));

			this.step.setStatus(Status.RUNNING, "Submitting file name '"
					+ file.getName() + "'");

			outputStream.writeObject(file.getName());

			this.step.setStatus(Status.RUNNING, "Submitting ingest policy "
					+ "name '" + tempIngestPolicyName + "'");

			outputStream.writeObject(tempIngestPolicyName);
			outputStream.flush();

			// Create a buffered file input stream and submit the file.
			BufferedInputStream fis = new BufferedInputStream(
					new FileInputStream(file));
			this.fileLength = fis.available();
			this.fileWriter = new BufferedOutputStream(this.outputServerSocket
					.getOutputStream());

			this.step.setStatus(Status.RUNNING, "Submitting file");

			final int BUFFER_LENGTH = 1048576;
			byte[] buffer = new byte[BUFFER_LENGTH];

			// Show a notification every fileCopyNotificationTime seconds, until
			// status is set to DONE.
			Timer timer = new Timer();
			timer.schedule(new ProgressNotificationTask(), 0,
					this.fileSubmitNotificationTime);

			// Start submitting.
			int bytesRead = fis.read(buffer);
			while (bytesRead > 0) {
				this.fileWriter.write(buffer, 0, bytesRead);
				this.bytesSubmitted += bytesRead;
				endTime = System.currentTimeMillis();
				this.totalTime = (endTime - startTime);
				bytesRead = fis.read(buffer);
			}

			// End submission.
			timer.cancel();
			fis.close();
			this.fileWriter.flush();

			this.step.setStatus(Status.RUNNING, "Submitted: 100%. "
					+ "Total time: "
					+ KopalTimestamp.getDurationInHours(System
							.currentTimeMillis()
							- startTime));
		}

		catch (Exception e) {
			e.printStackTrace();
			throw new IOException(errorMessage);
		}
	}

	/***************************************************************************
	 * The connection to the client loader's server thread is closed.
	 * 
	 * @throws IOException
	 **************************************************************************/
	private void closeConnectionToServer() throws IOException {
		String errorMessage = "Unable to close connection to client loader";

		try {
			this.step.setStatus(Status.RUNNING, "Closing connection to "
					+ "client loader");
			this.outputServerSocket.close();
			this.step.setStatus(Status.RUNNING, "Connection closed");

		}
		catch (Exception e) {
			throw new IOException(errorMessage);
		}
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param _clientLoaderPort The clientLoaderPort to set.
	 **************************************************************************/
	public void setClientLoaderPort(int _clientLoaderPort) {
		this.clientLoaderPort = _clientLoaderPort;
	}

	/***************************************************************************
	 * @param _clientLoaderServer The clientLoaderServer to set.
	 **************************************************************************/
	public void setClientLoaderServer(String _clientLoaderServer) {
		this.clientLoaderServer = _clientLoaderServer;
	}

	/***************************************************************************
	 * @param _processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData _processData) {
		this.processData = _processData;
	}

	/***************************************************************************
	 * @param _step The step to set.
	 **************************************************************************/
	public void setStep(Step _step) {
		this.step = _step;
	}

	/***************************************************************************
	 * @return String The submitPolicyName.
	 **************************************************************************/
	public String getSubmitPolicyName() {
		return this.submitPolicyName;
	}

	/***************************************************************************
	 * @param _submitPolicyName The policy name to set.
	 **************************************************************************/
	public void setSubmitPolicyName(String _submitPolicyName) {
		this.submitPolicyName = _submitPolicyName;
	}

	/***************************************************************************
	 * @param _fileSubmitNotificationTime The file submit notification time to
	 *            set.
	 **************************************************************************/
	public void setFileSubmitNotificationTime(long _fileSubmitNotificationTime) {
		this.fileSubmitNotificationTime = _fileSubmitNotificationTime;
	}

	/***************************************************************************
	 * Just gives notification log output.
	 * 
	 * @author funk
	 **************************************************************************/
	class ProgressNotificationTask extends TimerTask
	{
		/**
		 * @see java.lang.Runnable#run()
		 */
		@Override
		public synchronized void run() {
			DecimalFormat df = new DecimalFormat("0");

			int submitted = SubmitSipToClientLoader.this.bytesSubmitted;
			int length = SubmitSipToClientLoader.this.fileLength;
			int time = (int) SubmitSipToClientLoader.this.totalTime;

			int percent = new Float((float) submitted / (float) length * 100)
					.intValue();
			int kbSec = submitted / (time < 1 ? 1 : time);

			SubmitSipToClientLoader.this.step.setStatus(Status.RUNNING,
					"Submitted: " + percent + "% ("
							+ df.format(submitted / (1024 * 1024)) + " of "
							+ df.format(length / (1024 * 1024)) + " MB -> "
							+ kbSec + " KB/sec)");
		}
	}

}
