/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / SubmitSipToDias.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.File;
import java.util.HashMap;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.ingest.DiasIngest;
import de.langzeitarchivierung.kopal.ingest.DiasIngestResponse;
import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;
import de.langzeitarchivierung.kopal.util.db.MappToHib;

/*******************************************************************************
 * Actionmodule that transfers the created SIP to the DIAS via FTP.
 * 
 * PRECONDITION: A completely created DIAS conform SIP (.zip etc.) with its path
 * stored in the according parameter of the LmerSIP object.
 * 
 * RESULT: The LmerSIP has been transfered to the DIAS system. The ticket ID is
 * stored in the database.
 * 
 * @author Matthias Neubauer, German National Library
 * @version 20070219
 * @see de.langzeitarchivierung.kopal.ingest.DiasIngest
 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule
 * @since 20050907
 ******************************************************************************/

public class SubmitSipToDias implements ActionModule
{

	// STATE (Instance variables) ******************************************

	private ProcessData	processData;
	private Step		step;

	protected String	preloadArea;
	protected boolean	useDatabase	= false;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * constructor
	 **************************************************************************/

	public SubmitSipToDias() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/**
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 */
	public void go() {
		DiasIngestResponse dir = new DiasIngestResponse();
		String ticketId;
		String alert;

		// Get the SIP's name (including the institutional path).
		String sipIngestName = new File(this.preloadArea).getName() + "/"
				+ this.processData.getSipLocation().getName();

		// Get an ingest ticket from the DIAS and parse it.
		this.step.setStatus(Status.RUNNING, "Requesting a SIP ingest "
				+ "ticket ID from DIAS for SIP '" + sipIngestName + "'");

		// Configure DiasIngest.
		DiasIngest di = new DiasIngest();
		try {
			Configurator.configure(di);
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error configuring DiasIngest"
					+ " object. System message: " + e.getMessage());
			e.printStackTrace();
			return;
		}

		String diasResponse = "";
		try {
			diasResponse = di.requestSipIngest(sipIngestName);
			dir = DiasIngest.parseDiasResponse(diasResponse);
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error requesting a SIP"
					+ " ingest ticket from the DIAS. System message: "
					+ e.getMessage());
			e.printStackTrace();
			return;
		}

		ticketId = dir.getMetadata().get(0).getSipIngestTicketNumber();
		alert = dir.getMetadata().get(0).getAlert();

		// Check for errors.
		if (!alert.equals(DiasIngestResponse.NOT_IN_DIAS_RESPONSE)) {
			this.step.setStatus(Status.ERROR, "SIP ingest ticket ID "
					+ "request failed. DIAS alert: " + alert);
			return;
		}
		else if (ticketId.equals(DiasIngestResponse.NOT_IN_DIAS_RESPONSE)) {
			this.step.setStatus(Status.ERROR, "SIP ingest ticket ID "
					+ "request failed. No ticket ID available in "
					+ "DIAS response.");
			return;
		}

		this.step.setStatus(Status.RUNNING, "SIP ingest ticket ID is '"
				+ ticketId + "'");

		// Set the ticket ID in custom ID hash map.
		HashMap<Object, Object> customIds = this.processData.getCustomIds();
		customIds.put(MappToHib.ID_DIASTICKETID, ticketId);

		// Start sumitting the SIP to DIAS.
		this.step.setStatus(Status.RUNNING, "Submitting SIP to the DIAS");
		long submitStartTime = System.currentTimeMillis();

		// TODO Set the FTP parameters from the DiasIngestResponse?
		try {
			di.initializeFtpParameters();
			di.submitFileToDias(this.processData.getSipLocation(),
					this.preloadArea);
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Transfer failed. "
					+ "System message: " + e.getMessage());
			e.printStackTrace();
			return;
		}

		// TODO If we use a restricted shell, please use the following:
		//
		// ingest.submitFileToDias(new File(sipListElement.getLmerSIP()
		// .getAssetFileLocation()), null);

		long totalSubmissionTime = System.currentTimeMillis() - submitStartTime;
		this.step
				.setStatus(
						Status.DONE,
						"SIP '"
								+ this.processData.getSipLocation().getName()
								+ "' has been sucessfully submitted in "
								+ KopalTimestamp
										.getDurationInHours(totalSubmissionTime)
								+ " ("
								+ (this.processData.getSipLocation().length() / totalSubmissionTime)
								+ " kb/s)");
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param _processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData _processData) {
		this.processData = _processData;
	}

	/***************************************************************************
	 * @param _step The step to set.
	 **************************************************************************/
	public void setStep(Step _step) {
		this.step = _step;
	}

	/***************************************************************************
	 * @param _preloadArea The preloadArea to set.
	 **************************************************************************/
	public void setPreloadArea(String _preloadArea) {
		this.preloadArea = _preloadArea;
	}

	/***************************************************************************
	 * @param _useDatabase The useDatabase to set.
	 **************************************************************************/
	public void setUseDatabase(boolean _useDatabase) {
		this.useDatabase = _useDatabase;
	}

}
