/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / Unzip.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;

/*******************************************************************************
 * Decompresses all files from a .zip file stored under SipLocation in a
 * destination directory.
 * 
 * PRECONDITION: The file under SipLocation does exist and is a .zip file.
 * 
 * RESULT: The Unzipped DIP will be existing in a subdirectory of the temporary
 * directory, and the PathToContentFiles in processData will be set.
 * 
 * @author Funk
 * @version 20070503
 * @since 20070502
 ******************************************************************************/

public class Unzip implements ActionModule
{

	// STATE (Instance variables) *****************************************

	ProcessData	processData;
	Step		step;

	File		sourceFile;
	String		tempDir;
	boolean		deleteZipOnExit	= false;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Default Constructor
	 **************************************************************************/
	public Unzip() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 **************************************************************************/
	public void go() {
		long startTime = System.currentTimeMillis();
		ZipInputStream zipStream = null;

		try {
			this.sourceFile = this.processData.getSipLocation();

			// Check if really a ZIP file.
			if (!this.sourceFile.getName().endsWith(".zip")
					&& !this.sourceFile.getName().endsWith(".ZIP")) {
				this.step.setStatus(Status.ERROR, "File is not a ZIP file "
						+ "or at least has no ZIP file extension",
						this.sourceFile);
				return;
			}

			File destFolder = new File(this.sourceFile.getParent(), "unpacked_"
					+ this.sourceFile.getName());

			destFolder.mkdir();

			this.step.setStatus(Status.RUNNING, "Decompressing DIP");

			zipStream = new ZipInputStream(new BufferedInputStream(
					new FileInputStream(this.sourceFile)));

			storeZipContent(zipStream, destFolder);
			zipStream.close();

			// Set the content file's path.
			this.processData
					.setPathToContentFiles(destFolder.getAbsolutePath());

		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "ZIP decompression failed. "
					+ "System message: " + e.getMessage());
			e.printStackTrace();
			return;
		}
		finally {
			if (zipStream != null) {
				try {
					zipStream.close();
				}
				catch (Exception closeE) {
					this.step.setStatus(Status.ERROR, "Error closing stream. "
							+ "System message: " + closeE.getMessage());
				}
			}
		}

		// Delete ZIP if wanted.
		if (this.deleteZipOnExit) {
			boolean deleted = this.sourceFile.delete();
			if (deleted) {
				this.step.setStatus(Status.RUNNING, "Original ZIP file "
						+ "deleted");
			}
			else {
				this.step.setStatus(Status.WARNING, "Original ZIP file "
						+ "could not be deleted");
			}
		}

		this.step.setStatus(Status.DONE, "DIP decompressed in "
				+ KopalTimestamp.getDurationInHours(System.currentTimeMillis()
						- startTime));
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ****

	/***************************************************************************
	 * Goes throught all the ZIP file entries and stores all of them including
	 * all subdirectories.
	 * 
	 * @param zipStream
	 * @param destFolder
	 * @throws IOException
	 **************************************************************************/
	private void storeZipContent(ZipInputStream zipStream, File destFolder)
			throws IOException {
		// Go throught all the ZIP entries.
		ZipEntry metsEntry = zipStream.getNextEntry();

		while (metsEntry != null) {
			if (metsEntry.isDirectory()) {
				File newDir = new File(destFolder, metsEntry.getName());
				newDir.mkdir();
			}
			else {
				storeZipFile(zipStream, new File(destFolder, metsEntry
						.getName()));
			}

			metsEntry = zipStream.getNextEntry();
		}
	}

	/***************************************************************************
	 * Stores a single file from the ZIP.
	 * 
	 * @param zipStream
	 * @param destinationFile
	 **************************************************************************/
	private void storeZipFile(ZipInputStream zipStream, File destinationFile) {
		BufferedOutputStream tempMetsStream = null;

		try {
			tempMetsStream = new BufferedOutputStream(new FileOutputStream(
					destinationFile));

			// Decompress now.
			final int BUFFER_LENGTH = 256 * 1024;
			byte[] buffer = new byte[BUFFER_LENGTH];
			int bytesRead = zipStream.read(buffer);
			while (bytesRead > 0) {
				tempMetsStream.write(buffer, 0, bytesRead);
				bytesRead = zipStream.read(buffer);
			}

			tempMetsStream.flush();
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Extraction failed. "
					+ "System message: " + e.getMessage());
		}
		finally {
			if (tempMetsStream != null) {
				try {
					tempMetsStream.close();
				}
				catch (Exception closeE) {
					this.step.setStatus(Status.ERROR, "Error closing stream: "
							+ closeE.getMessage());
				}
			}
		}
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param _tempDir The tempDir to set.
	 **************************************************************************/
	public void setTempDir(String _tempDir) {
		this.tempDir = _tempDir;
	}

	/***************************************************************************
	 * @param _processData The processData to set.
	 **************************************************************************/
	public void setProcessData(ProcessData _processData) {
		this.processData = _processData;
	}

	/***************************************************************************
	 * @param _step The step to set.
	 **************************************************************************/
	public void setStep(Step _step) {
		this.step = _step;
	}

	/***************************************************************************
	 * @param _deleteZipOnExit The deleteZipOnExit to set.
	 **************************************************************************/
	public void setDeleteZipOnExit(boolean _deleteZipOnExit) {
		this.deleteZipOnExit = _deleteZipOnExit;
	}

}
