/*******************************************************************************
 * de.langzeitarchivierung.kopal.actionmodule / XorFileChecksums.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.actionmodule;

import java.io.File;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Status;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.util.ChecksumGenerator;
import de.langzeitarchivierung.kopal.util.db.MappToHib;

/*******************************************************************************
 * Adds the XORed file checksums to the database.
 * 
 * @author ludwig
 ******************************************************************************/
public class XorFileChecksums implements ActionModule
{

	// STATE (Instance variables) *****************************************

	ProcessData	processData;
	Step		step;

	boolean		useDatabase				= false;
	boolean		errorIfAlreadyIngested	= true;
	String		defaultOwner			= "";

	// MANIPULATION (Manipulation - what the object does) ******************

	/**
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#go()
	 */
	public void go() {
		if (!this.useDatabase) {
			this.step.setStatus(Status.WARNING, "Configuration says "
					+ "useDatabase = false");
			this.step.setStatus(Status.DONE, "Added no checksum to database");
			return;
		}

		// Calculate checksums.
		String xorFileChecksum = null;
		try {
			this.step.setStatus(Status.RUNNING, "Calculating XORed file "
					+ "checksum");
			xorFileChecksum = ChecksumGenerator.getXorFileChecksums(new File(
					this.processData.getPathToContentFiles()));
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error calculating XORed file "
					+ "checksum. System message: " + e.getMessage());
			e.printStackTrace();
			return;
		}

		// Check if the calculated checksum already exists in the database.
		try {
			if (MappToHib.alreadyReceivedInput(xorFileChecksum,
					this.defaultOwner)) {
				String message = "A SIP with the calculated XORed file checksum "
						+ "already is existing in the DIAS for owner '"
						+ this.defaultOwner
						+ "' or another ingest is in progress";

				if (this.errorIfAlreadyIngested) {
					this.step.setStatus(Status.ERROR, message);
					return;
				}

				this.step.setStatus(Status.WARNING, message);
			}
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error accessing database: "
					+ e.getMessage());
			e.printStackTrace();
			return;
		}

		// Add the checksum to the database if not existing.
		try {
			this.step.setStatus(Status.RUNNING, "Adding XORed file checksum");
			MappToHib.addInputChecksum(this.processData, xorFileChecksum);
		}
		catch (Exception e) {
			this.step.setStatus(Status.ERROR, "Error accessing database. "
					+ "System message: " + e.getMessage());
			e.printStackTrace();
			return;
		}

		this.step.setStatus(Status.DONE, "Checksum calculation complete");
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @return boolean - the errorIfAlreadyIngested
	 **************************************************************************/
	public boolean isErrorIfAlreadyIngested() {
		return this.errorIfAlreadyIngested;
	}

	/***************************************************************************
	 * @param _errorIfAlreadyIngested
	 **************************************************************************/
	public void setErrorIfAlreadyIngested(boolean _errorIfAlreadyIngested) {
		this.errorIfAlreadyIngested = _errorIfAlreadyIngested;
	}

	/***************************************************************************
	 * @param _processData
	 **************************************************************************/
	public void setProcessData(ProcessData _processData) {
		this.processData = _processData;
	}

	/***************************************************************************
	 * @param _useDatabase
	 **************************************************************************/
	public void setUseDatabase(boolean _useDatabase) {
		this.useDatabase = _useDatabase;
	}

	/***************************************************************************
	 * @see de.langzeitarchivierung.kopal.actionmodule.ActionModule#setStep(de.langzeitarchivierung.kopal.Step)
	 * 
	 * @param _step
	 **************************************************************************/
	public void setStep(Step _step) {
		this.step = _step;
	}

	/***************************************************************************
	 * @param defaultOwner The defaultOwner to set.
	 * @author Stefan Funk, SUB Göttingen 03.07.2007
	 **************************************************************************/
	public void setDefaultOwner(String defaultOwner) {
		this.defaultOwner = defaultOwner;
	}

}
