/*******************************************************************************
 * de.langzeitarchivierung.kopal.administration / DiasAdmin.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.administration;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;

import org.apache.commons.digester.Digester;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.GetMethod;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import de.langzeitarchivierung.kopal.util.KopalXMLParser;

/*******************************************************************************
 * Class for using the administration interface of DIAS according to the
 * "Detailed Specifications for Administration and Search Interfaces to
 * DIAS-Core" Version 1.0 by IBM Nederland N.V. 2000, 2006
 * 
 * @author Matthias Neubauer, Deutsche Nationalbibliothek
 * @version 20070427
 * @since 20060925
 ******************************************************************************/

public class DiasAdmin
{
	// CONSTANTS **********************************************************

	private static final int	INPUT_BUFFER_LENGTH					= 1024 * 1024;
	private static final String	ADMIN_URL							= "/dias/administration?";
	private static final String	SEARCH_URL							= "/dias/search?";
	private static final String	INT_ID_PARAM						= "internalassetid=";
	private static final String	USER_PARAM							= "&username=";
	private static final String	PASSWORD_PARAM						= "&password=";
	private static final String	REQUEST_PARAM						= "request=";
	private static final String	AMP_REQUEST_PARAM					= "&request=";
	private static final String	SCHEMA_URL_PARAM					= "schemaURL=";
	private static final String	SCHEMA_NAME_PARAM					= "&schemaName=";
	private static final String	SCHEMA_DESCR_PARAM					= "&schemaDescription=";
	private static final String	SCHEMA_DATE_TIME_PARAM				= "&schemaDateTime=";
	private static final String	SCHEMA_ID_PARAM						= "schemaID=";
	private static final String	MODIFICATION_PARAM					= "modification=";
	private static final String	FROM_PARAM							= "&from=";
	private static final String	TO_PARAM							= "&to=";
	private static final String	FILETYPE_NAME_PARAM					= "fileTypeName=";
	private static final String	FILETYPE_VER_PARAM					= "&fileTypeVersion=";
	private static final String	FILETYPE_MIME_PARAM					= "&fileTypeMimeType=";

	/* TODO Name not consistent in specification, change paper. */
	private static final String	FILETYPE_EXT_PARAM					= "&fileTypeExtensions=";

	private static final String	FILETYPE_STATE_PARAM				= "&fileTypeState=";
	private static final String	FILETYPE_BASECODING_TYPE_PARAM		= "&fileTypeBaseCoding=";
	private static final String	FILETYPE_DESCR_PARAM				= "&fileTypeDescription=";

	/* TODO Specification change paper is not clear about if this is required */
	private static final String	FILETYPE_ID_PARAM					= "fileTypeID=";

	/* TODO Name not consistant in specification, change paper. */
	private static final String	ADD_FILETYPE_EXT_PARAM				= "&addFileTypeExtensions=";		// TODO:

	private static final String	VERSION_PARAM						= "version=";
	private static final String	QUERY_PARAM							= "&query=";
	private static final String	STARTRECORD_PARAM					= "&startRecord=";
	private static final String	MAXIMUMRECORDS_PARAM				= "&maximumRecords=";
	private static final String	RESULTSET_TTL_PARAM					= "&resultSetTTL=";
	private static final String	SORTKEYS_PARAM						= "&sortKeys=";
	private static final String	X_ADDITIONAL_DATA_ELEMENTS_PARAM	= "&x-additionalDataElements=";
	private static final String	X_LIST_AFFECTED_CONTENT_FILES_PARAM	= "&x-listAffectedContentFiles";
	private static final String	X_INDEXINFO_PARAM					= "&x-indexInfo";
	private static final String	X_RECORDINFO_PARAM					= "&x-recordInfo";
	private static final String	OPERATION_PARAM						= "&operation=";

	private static final String	DELETE_ASSET_REQUEST				= "deleteAsset";
	private static final String	DELETE_AIP_REQUEST					= "deleteAIP";
	private static final String	CHECK_CONSISTENCY_REQUEST			= "checkConsistency";
	private static final String	STORE_SCHEMA_REQUEST				= "storeSchema";
	private static final String	GET_SCHEMA_REQUEST					= "getSchema";
	private static final String	GET_SCHEMA_LIST_REQUEST				= "getSchemaList";
	private static final String	GET_MODIFIED_AIPS_REQUEST			= "getModifiedAIPs";
	private static final String	MODIFY_FILETYPE_REQUEST				= "modifyFileType";
	private static final String	STORE_FILETYPE_REQUEST				= "storeFileType";
	private static final String	EXPLAIN_OPERATION					= "explain";
	private static final String	SEARCH_RETRIEVE_OPERATION			= "searchRetrieve";

	// STATE (Instance variables) *****************************************

	// Resources.
	private HttpClient			diasClient;

	private String				adminServer							= null;
	private int					adminPort							= 80;
	private String				adminUser							= "";
	private String				adminPwd							= "";
	private String				keyStoreFile;
	private boolean				useHttps							= true;

	// SRU parameters.
	private String				sruVersion							= "1.1";
	private long				sruStartRecord						= 0;
	private long				sruMaximumRecords					= 0;
	private long				sruResultSetTTL						= 0;
	private String				sruSortKeys							= "";
	private String				sruX_additionalDataElements			= "";
	private boolean				sruX_listAffectedContentFiles		= false;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Constructor.
	 **************************************************************************/
	public DiasAdmin() {
		super();
		this.diasClient = new HttpClient();
	}

	/***************************************************************************
	 * Constructor.
	 * 
	 * @param server The server address
	 **************************************************************************/
	public DiasAdmin(String server) {
		super();
		this.adminServer = server;
	}

	/***************************************************************************
	 * Constructor.
	 * 
	 * @param server The server address
	 * @param p The port number
	 **************************************************************************/
	public DiasAdmin(String server, int p) {
		super();
		this.adminServer = server;
		this.adminPort = p;
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * Deletes an asset from DIAS. An asset refers to the AIP (Data Package) as
	 * well as the database entries that belong to an internal asset ID.
	 * 
	 * @param internalAssetID The internal asset id.
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String deleteAssetByInternalAssetID(String internalAssetID)
			throws IOException {
		String requestUrl = ADMIN_URL + INT_ID_PARAM + internalAssetID
				+ AMP_REQUEST_PARAM + DELETE_ASSET_REQUEST + USER_PARAM
				+ this.adminUser + PASSWORD_PARAM + this.adminPwd;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Deletes an asset from DIAS and returns DIAS response as a NodeList. An
	 * asset refers to the AIP (Data Package) as well as the database entries
	 * that belong to an internal asset ID.
	 * 
	 * @param internalAssetID The internal asset id.
	 * @return The DIAS response as a NodeList
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList deleteAssetByInternalAssetIDXml(String internalAssetID)
			throws IOException {
		String diasResponse = deleteAssetByInternalAssetID(internalAssetID);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Deletes an AIP from DIAS. An AIP refers only to the Data Package AND NOT
	 * to the the database entries that belong to an internal asset ID.
	 * 
	 * @param internalAssetID The internal asset id.
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String deleteAIPByInternalAssetID(String internalAssetID)
			throws IOException {
		String requestUrl = ADMIN_URL + INT_ID_PARAM + internalAssetID
				+ AMP_REQUEST_PARAM + DELETE_AIP_REQUEST + USER_PARAM
				+ this.adminUser + PASSWORD_PARAM + this.adminPwd;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Deletes an AIP from DIAS and returns DIAS response as a NodeList. An AIP
	 * refers only to the Data Package AND NOT to the the database entries that
	 * belong to an internal asset ID.
	 * 
	 * @param internalAssetID The internal asset id.
	 * @return The DIAS response as a NodeList
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList deleteAIPByInternalAssetIDXml(String internalAssetID)
			throws IOException {
		String diasResponse = deleteAIPByInternalAssetID(internalAssetID);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Checks the consistency for a stored asset
	 * 
	 * @param internalAssetID The internal asset id.
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String checkConsistencyForInternalAssetID(String internalAssetID)
			throws IOException {
		String requestUrl = ADMIN_URL + INT_ID_PARAM + internalAssetID
				+ AMP_REQUEST_PARAM + CHECK_CONSISTENCY_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Checks the consistency of a stored asset and returns DIAS response as a
	 * NodeList
	 * 
	 * @param internalAssetID The internal asset id.
	 * @return The DIAS response as a NodeList
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList checkConsistencyForInternalAssetIDXml(String internalAssetID)
			throws IOException {
		String diasResponse = checkConsistencyForInternalAssetID(internalAssetID);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Requests a result set of AIPs that apply to a certain modification during
	 * the given time frame
	 * 
	 * @param modification The type of modification ("created", "AIPdeleted", or
	 *            "assetDeleted")
	 * @param from The lower boundary of the time frame to search
	 * @param to The upper boundary of the time frame to search
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String getModifiedAIPs(String modification, String from, String to)
			throws IOException {
		String requestUrl = ADMIN_URL + MODIFICATION_PARAM + modification
				+ FROM_PARAM + from + TO_PARAM + to + AMP_REQUEST_PARAM
				+ GET_MODIFIED_AIPS_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Requests a result set of AIPs that apply to a certain modification during
	 * the given time frame and returns DIAS response as a NodeList
	 * 
	 * @param modification The type of modification
	 * @param from The lower boundary of the time frame to search
	 * @param to The upper boundary of the time frame to search
	 * @return The DIAS response as a NodeList
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList getModifiedAIPsXml(String modification, String from,
			String to) throws IOException {
		String diasResponse = getModifiedAIPs(modification, from, to);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Stores a new file type definition within DIAS
	 * 
	 * @param fileTypeName - The name of the file type *required*
	 * @param fileTypeVersion - The version of the file type *required*
	 * @param fileTypeMimeType - The mime type of the file type *required*
	 * @param fileTypeExtensions - A list of all extensions that exist for the
	 *            file type *required*
	 * @param fileTypeState - One of the following allowed states: "defined",
	 *            "active" or "obsolete". If not provided, "defined" will be the
	 *            default
	 * @param fileTypeDescription - A free text description of the file type
	 *            *optional* To store NO description, provide as empty String or
	 *            null
	 * @param fileTypeBaseCoding - The base coding of the file type *optional*
	 *            To store NO base coding, provide as empty String or null
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String storeFileType(String fileTypeName, String fileTypeVersion,
			String fileTypeMimeType, String[] fileTypeExtensions,
			String fileTypeState, String fileTypeDescription,
			String fileTypeBaseCoding) throws IOException {
		String requestUrl = ADMIN_URL + FILETYPE_NAME_PARAM + fileTypeName
				+ FILETYPE_VER_PARAM + fileTypeVersion + FILETYPE_MIME_PARAM
				+ fileTypeMimeType;

		if (fileTypeExtensions.length > 0) {
			requestUrl += FILETYPE_EXT_PARAM;

			for (int i = 0; i < fileTypeExtensions.length; i++) {
				requestUrl += fileTypeExtensions[i];
				if (i != fileTypeExtensions.length - 1) {
					requestUrl += ", ";
				}
			}
		}

		requestUrl += USER_PARAM + this.adminUser + PASSWORD_PARAM
				+ this.adminPwd;

		if (fileTypeState != null && !fileTypeState.equals("")) {
			requestUrl += FILETYPE_STATE_PARAM + fileTypeState;
		}
		else {
			requestUrl += FILETYPE_STATE_PARAM + "defined";
		}

		if (fileTypeDescription != null && !fileTypeDescription.equals("")) {
			requestUrl += FILETYPE_DESCR_PARAM + fileTypeDescription;
		}

		if (fileTypeBaseCoding != null && !fileTypeBaseCoding.equals("")) {
			requestUrl += FILETYPE_BASECODING_TYPE_PARAM + fileTypeBaseCoding;
		}

		requestUrl += AMP_REQUEST_PARAM + STORE_FILETYPE_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Stores a new file type definition within DIAS and returns DIAS response
	 * as a NodeList
	 * 
	 * @param fileTypeName - The name of the file type *required*
	 * @param fileTypeVersion - The version of the file type *required*
	 * @param fileTypeMimeType - The mime type of the file type *required*
	 * @param fileTypeExtensions - A list of all extensions that exist for the
	 *            file type *required*
	 * @param fileTypeState - One of the following allowed states: "defined",
	 *            "active" or "obsolete". If not provided, "defined" will be the
	 *            default
	 * @param fileTypeDescription - A free text description of the file type
	 *            *optional* To store NO description, provide as empty String or
	 *            null
	 * @param fileTypeBaseCoding - The base coding of the file type *optional*
	 *            To store NO base coding, provide as empty String or null
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList storeFileTypeXml(String fileTypeName,
			String fileTypeVersion, String fileTypeMimeType,
			String[] fileTypeExtensions, String fileTypeState,
			String fileTypeDescription, String fileTypeBaseCoding)
			throws IOException {
		String diasResponse = storeFileType(fileTypeName, fileTypeVersion,
				fileTypeMimeType, fileTypeExtensions, fileTypeState,
				fileTypeDescription, fileTypeBaseCoding);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Modifies a file type definition within DIAS
	 * 
	 * @param fileTypeID - The ID of the file type to modify *required*
	 * @param fileTypeExtensions - A list of extensions to add for the file type
	 *            *optional* To NOT add extensions, provide as empty array or
	 *            null
	 * @param fileTypeState - One of the following allowed states: "defined",
	 *            "active" or "obsolete". *optional* To NOT modify the state,
	 *            provide as empty String or null
	 * @param fileTypeDescription - A free text description of the file type
	 *            *optional* To NOT modify or add the description, provide as
	 *            empty String or null
	 * @param fileTypeBaseCoding - The base coding of the file type *optional*
	 *            To NOT modify or add the base coding, provide as empty String
	 *            or null
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String modifyFileType(String fileTypeID,
			String[] fileTypeExtensions, String fileTypeState,
			String fileTypeDescription, String fileTypeBaseCoding)
			throws IOException {
		String requestUrl = ADMIN_URL + FILETYPE_ID_PARAM + fileTypeID
				+ USER_PARAM + this.adminUser + PASSWORD_PARAM + this.adminPwd;

		if (fileTypeExtensions != null) {
			if (fileTypeExtensions.length > 0) {
				requestUrl += ADD_FILETYPE_EXT_PARAM;

				for (int i = 0; i < fileTypeExtensions.length; i++) {
					requestUrl += fileTypeExtensions[i];
					if (i != fileTypeExtensions.length - 1) {
						requestUrl += ", ";
					}
				}
			}
		}

		if (fileTypeState != null && !fileTypeState.equals("")) {
			requestUrl += FILETYPE_STATE_PARAM + fileTypeState;
		}

		if (fileTypeDescription != null && !fileTypeDescription.equals("")) {
			requestUrl += FILETYPE_DESCR_PARAM + fileTypeDescription;
		}

		if (fileTypeBaseCoding != null && !fileTypeBaseCoding.equals("")) {
			requestUrl += FILETYPE_BASECODING_TYPE_PARAM + fileTypeBaseCoding;
		}

		requestUrl += AMP_REQUEST_PARAM + MODIFY_FILETYPE_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Modifies a file type definition within DIAS and returns DIAS response as
	 * a NodeList
	 * 
	 * @param fileTypeID - The ID of the file type to modify *required*
	 * @param fileTypeExtensions - A list of extensions to add for the file type
	 *            *optional* To NOT add extensions, provide as empty array or
	 *            null
	 * @param fileTypeState - One of the following allowed states: "defined",
	 *            "active" or "obsolete". *optional* To NOT modify the state,
	 *            provide as empty String or null
	 * @param fileTypeDescription - A free text description of the file type
	 *            *optional* To NOT modify or add the description, provide as
	 *            empty String or null
	 * @param fileTypeBaseCoding - The base coding of the file type *optional*
	 *            To NOT modify or add the base coding, provide as empty String
	 *            or null
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList modifyFileTypeXml(String fileTypeID,
			String[] fileTypeExtensions, String fileTypeState,
			String fileTypeDescription, String fileTypeBaseCoding)
			throws IOException {
		String diasResponse = modifyFileType(fileTypeID, fileTypeExtensions,
				fileTypeState, fileTypeDescription, fileTypeBaseCoding);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Stores a new XML Schema within DIAS with a schema description
	 * 
	 * @param schemaURI - The URI for the source location of the XML schema that
	 *            must be stored
	 * @param schemaName - The namespace of the schema that must be stored
	 * @param schemaDescription - The description of the schema that must be
	 *            stored
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String storeXMLSchemaWithDescription(String schemaURI,
			String schemaName, String schemaDescription) throws IOException {
		String requestUrl = ADMIN_URL + SCHEMA_URL_PARAM + schemaURI
				+ SCHEMA_NAME_PARAM + schemaName + USER_PARAM + this.adminUser
				+ PASSWORD_PARAM + this.adminPwd + SCHEMA_DESCR_PARAM
				+ schemaDescription + AMP_REQUEST_PARAM + STORE_SCHEMA_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Stores a new XML Schema within DIAS with a schema description and returns
	 * DIAS response as a NodeList
	 * 
	 * @param schemaURI - The URI for the source location of the XML schema that
	 *            must be stored
	 * @param schemaName - The namespace of the schema that must be stored
	 * @param schemaDescription - The description of the schema that must be
	 *            stored
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList storeXMLSchemaWithDescriptionXml(String schemaURI,
			String schemaName, String schemaDescription) throws IOException {
		String diasResponse = storeXMLSchemaWithDescription(schemaURI,
				schemaName, schemaDescription);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Stores a new XML Schema within DIAS
	 * 
	 * @param schemaURI - The URI for the source location of the XML schema that
	 *            must be stored
	 * @param schemaName - The namespace of the schema that must be stored
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String storeXMLSchema(String schemaURI, String schemaName)
			throws IOException {
		String requestUrl = ADMIN_URL + SCHEMA_URL_PARAM + schemaURI
				+ SCHEMA_NAME_PARAM + schemaName + USER_PARAM + this.adminUser
				+ PASSWORD_PARAM + this.adminPwd + AMP_REQUEST_PARAM
				+ STORE_SCHEMA_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Stores a new XML Schema within DIAS and returns DIAS response as a
	 * NodeList
	 * 
	 * @param schemaURI - The URI for the source location of the XML schema that
	 *            must be stored
	 * @param schemaName - The namespace of the schema that must be stored
	 * @return The DIAS response as a NodeList
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList storeXMLSchemaXml(String schemaURI, String schemaName)
			throws IOException {
		String diasResponse = storeXMLSchema(schemaURI, schemaName);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Retrieves a XML schema that is stored within DIAS by schema URI
	 * 
	 * @param schemaURI - The URI for the source location of the XML schema that
	 *            is to be retrieved
	 * @param schemaDateTime - The ISO date/time stamp for which the schema must
	 *            be retrieved
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String getXMLSchemaBySchemaURI(String schemaURI,
			String schemaDateTime) throws IOException {
		String requestUrl = ADMIN_URL + SCHEMA_URL_PARAM + schemaURI
				+ SCHEMA_DATE_TIME_PARAM + schemaDateTime + AMP_REQUEST_PARAM
				+ GET_SCHEMA_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Retrieves a XML schema that is stored within DIAS by schema URI and
	 * returns DIAS response as a NodeList
	 * 
	 * @param schemaURI - The URI for the source location of the XML schema that
	 *            is to be retrieved
	 * @param schemaDateTime - The ISO date/time stamp for which the schema must
	 *            be retrieved
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList getXMLSchemaBySchemaURIXml(String schemaURI,
			String schemaDateTime) throws IOException {
		String diasResponse = getXMLSchemaBySchemaURI(schemaURI, schemaDateTime);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Retrieves a XML schema that is stored within DIAS by schema ID
	 * 
	 * @param schemaID - The internal ID of the schema requested
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String getXMLSchemaBySchemaID(String schemaID) throws IOException {
		String requestUrl = ADMIN_URL + SCHEMA_ID_PARAM + schemaID
				+ AMP_REQUEST_PARAM + GET_SCHEMA_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Retrieves a XML schema that is stored within DIAS by schema ID and
	 * returns DIAS response as a NodeList
	 * 
	 * @param schemaID - The internal ID of the schema requested
	 * @return The DIAS response as NodeList
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList getXMLSchemaBySchemaIDXml(String schemaID)
			throws IOException {
		String diasResponse = getXMLSchemaBySchemaID(schemaID);

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Retrieves a list of all available XML schemas that are stored with DIAS
	 * 
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String getXMLSchemaList() throws IOException {
		String requestUrl = ADMIN_URL + REQUEST_PARAM + GET_SCHEMA_LIST_REQUEST;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Retrieves a list of all available XML schemas that are stored with DIAS
	 * and returns the DIAS response as a NodeList
	 * 
	 * @return The DIAS response as a NodeList
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public NodeList getXMLSchemaListXml() throws IOException {
		String diasResponse = getXMLSchemaList();

		Document responseDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);

		return responseDocument.getElementsByTagName("DIASADMINRESPONSE");
	}

	/***************************************************************************
	 * Builds a SRUOperation from a given CQL String and sends it to the SRU
	 * server (DIAS).
	 * 
	 * @return The List of AIPs that apply to the search phrase together with
	 *         relevant SRU data wrapped in a DiasAdminSRUResponse object
	 * @param cqlQuery - The search phrase formed in CQL
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public DiasAdminSRUResponse searchAIPsByCQL(String cqlQuery)
			throws IOException {

		String response = applySRUOperation(buildSRUSearchRetrieveOperation(
				this.sruVersion, cqlQuery, this.sruStartRecord,
				this.sruMaximumRecords, this.sruResultSetTTL, this.sruSortKeys,
				this.sruX_additionalDataElements,
				this.sruX_listAffectedContentFiles));

		return parseDiasSRUResponse(response);
	}

	/***************************************************************************
	 * Builds a SRUOperation from a given CQL String and sends it to the SRU
	 * server (DIAS).
	 * 
	 * @return The whole SRU response as a String
	 * @param cqlQuery - The search phrase formed in CQL
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String searchAIPsByCQLStringResult(String cqlQuery)
			throws IOException {

		return applySRUOperation(buildSRUSearchRetrieveOperation(
				this.sruVersion, cqlQuery, this.sruStartRecord,
				this.sruMaximumRecords, this.sruResultSetTTL, this.sruSortKeys,
				this.sruX_additionalDataElements,
				this.sruX_listAffectedContentFiles));
	}

	/***************************************************************************
	 * Builds a SRU 'explain' operation with the given parameter values
	 * 
	 * @param version - The SRU version (default 1.1)
	 * @param x_indexInfo - To indicate that <indexInfo> must be included in the
	 *            explain response (See Specification point MDS.gen.05)
	 * @param x_recordInfo - To indicate that <recordInfo> must be included in
	 *            the explain response (See Specification point MDS.gen.05)
	 * 
	 * @return String - The SRU 'explain' operation as a String
	 * 
	 **************************************************************************/
	public String buildSRUExplainOperation(String version, boolean x_indexInfo,
			boolean x_recordInfo) {

		if (version == null || version.equals("")) {
			version = "1.1";
		}

		String sruUrl = VERSION_PARAM + version + OPERATION_PARAM
				+ EXPLAIN_OPERATION;

		if (x_indexInfo) {
			sruUrl += X_INDEXINFO_PARAM + "=true";
		}

		if (x_recordInfo) {
			sruUrl += X_RECORDINFO_PARAM + "=true";
		}

		return sruUrl;
	}

	/***************************************************************************
	 * Builds a SRU 'searchRetrieve' operation with the given parameter values
	 * 
	 * @param version - The SRU version (default 1.1)
	 * @param query - A CQL query
	 * @param startRecord - To specify the first record of the result set that
	 *            should be returned in the response message
	 * @param maximumRecords - To specify the maximum number of records to be
	 *            returned in a single response message
	 * @param resultSetTTL - To specify the number of seconds that a result set
	 *            should be maintained before it will be deleted by the server.
	 *            Note that this may be overruled by the server
	 * @param sortKeys - Sorting options for the result set. If no sortKeys
	 *            provided, the default sorting order will be ascending
	 *            ExternalAssetID, InternalAssetID Sortkeys must be provided as
	 *            a single String, the different keys seperated by commata.
	 * @param x_additionalDataElements - As many DIAS Standard or Custom Data
	 *            Elements may be repeated as needed, spereated by commas
	 * @param x_listAffectedContentFiles - Including this parameter indicates
	 *            that a list of FileIDs of all affected content files (files
	 *            that meet the search criteria) must be included in the
	 *            response record.
	 * @return String - The SRU 'searchRetrieve' operation as a String
	 * 
	 **************************************************************************/
	public String buildSRUSearchRetrieveOperation(String version, String query,
			long startRecord, long maximumRecords, long resultSetTTL,
			String sortKeys, String x_additionalDataElements,
			boolean x_listAffectedContentFiles) {

		if (version == null || version.equals("")) {
			version = "1.1";
		}

		String sruUrl = VERSION_PARAM + version + OPERATION_PARAM
				+ SEARCH_RETRIEVE_OPERATION;

		sruUrl += QUERY_PARAM + query;

		if (startRecord > 0) {
			sruUrl += STARTRECORD_PARAM + startRecord;
		}
		if (maximumRecords > 0) {
			sruUrl += MAXIMUMRECORDS_PARAM + maximumRecords;
		}
		if (resultSetTTL > 0) {
			sruUrl += RESULTSET_TTL_PARAM + resultSetTTL;
		}
		if (sortKeys != null && !sortKeys.equals("")) {
			sruUrl += SORTKEYS_PARAM + sortKeys;
		}
		if (x_additionalDataElements != null
				&& !x_additionalDataElements.equals("")) {
			sruUrl += X_ADDITIONAL_DATA_ELEMENTS_PARAM
					+ x_additionalDataElements;
		}

		if (x_listAffectedContentFiles) {
			sruUrl += X_LIST_AFFECTED_CONTENT_FILES_PARAM + "=true";
		}

		return sruUrl;
	}

	/***************************************************************************
	 * Applies a given SRU operation to the DIAS SRU server
	 * 
	 * @param sruOperation - A valid SRU operation
	 * @return The DIAS response as String
	 * @throws IOException thrown if the connection to DIAS failed
	 **************************************************************************/
	public String applySRUOperation(String sruOperation) throws IOException {

		String requestUrl = SEARCH_URL + sruOperation;

		return retrieveDataFromDIAS(requestUrl);
	}

	/***************************************************************************
	 * Parses the SRU response of the DIAS Admin SRU server. Uses
	 * parseDiasSRUResponse(Reader diasResponse)
	 * 
	 * @param diasSRUResponse
	 * @return a DiasAdminSRUResponse object which contains the values of the
	 *         response, or NULL, if there was a parse error or the response was
	 *         not according to the specifications
	 **************************************************************************/
	public static DiasAdminSRUResponse parseDiasSRUResponse(
			String diasSRUResponse) {
		return DiasAdmin
				.parseDiasSRUResponse(new StringReader(diasSRUResponse));
	}

	/***************************************************************************
	 * Parses the SRU response of the DIAS Admin SRU server. Uses Jakarta
	 * Commons digester.
	 * 
	 * @param diasSRUResponse
	 * @return a DiasAdminSRUResponse object which contains the values of the
	 *         response, or NULL, if there was a parse error or the response was
	 *         not according to the specifications
	 **************************************************************************/
	public static DiasAdminSRUResponse parseDiasSRUResponse(
			Reader diasSRUResponse) {
		Digester responseParser = new Digester();

		responseParser.setClassLoader(ClassLoader.getSystemClassLoader());

		// Create a DiasAccessResponse object for the dias response and map
		// the properties.
		responseParser.addObjectCreate("searchRetrieveResponse",
				DiasAdminSRUResponse.class);
		// Note: It is necessary to repeat the property in lower cases.
		responseParser.addBeanPropertySetter("searchRetrieveResponse/version",
				"version");
		responseParser.addBeanPropertySetter(
				"searchRetrieveResponse/numberOfRecords", "numberOfRecords");
		responseParser.addBeanPropertySetter(
				"searchRetrieveResponse/resultSetId", "resultSetId");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/resultSetIdleTime",
						"resultSetIdleTime");

		// Create a SRURecord object for the corresponding part
		// in the response and map the properties.
		responseParser.addObjectCreate("searchRetrieveResponse/records/record",
				"de.langzeitarchivierung.kopal.administration.SRURecord");

		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/ExternalAssetID",
						"externalAssetID");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/InternalAssetID",
						"internalAssetID");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/AssetVersion",
						"assetVersion");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/AssetDescription",
						"assetDescription");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/AIPIsDeleted",
						"AIPIsDeleted");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/AIPDeletionTimestamp",
						"AIPDeletionTimestamp");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/AssetCreationTimestamp",
						"assetCreationTimestamp");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/AssetIngestTimestamp",
						"assetIngestTimestamp");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/CustomMigration.oldObjectIdentifier",
						"CMOldObjectIdentifier");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/CustomMigration.oldObjectVersion",
						"CMOldObjectVersion");
		responseParser
				.addBeanPropertySetter(
						"searchRetrieveResponse/records/record/recordData/Asset/TotalAffectedContentFiles",
						"totalAffectedContentFiles");

		responseParser.addSetNext("searchRetrieveResponse/records/record",
				"addRecord");

		try {
			return (DiasAdminSRUResponse) responseParser.parse(diasSRUResponse);
		}
		catch (SAXException e) {
			e.printStackTrace();
			System.err.println("Dias response corrupt!");
		}
		catch (IOException ie) {
			System.out.println(ie.getMessage());
			System.err.println("Could not read source for dias response!");
		}

		return null;
	}

	/***************************************************************************
	 * Parses the admin response of the DIAS. Uses DOM.
	 * 
	 * @param diasResponse
	 * @return a DiasAdminResponse object which contains the values of the
	 *         response.
	 **************************************************************************/
	public static DiasAdminResponse parseDiasResponse(String diasResponse) {
		Document doc = KopalXMLParser.getDocumentFromString(diasResponse);

		// Create an object for the DIAS XML Response and fill its fields
		DiasAdminResponse response = new DiasAdminResponse();

		response.setVersion(KopalXMLParser.getFirstValue(doc, "version"));
		response.setRequest(KopalXMLParser.getFirstValue(doc, "request"));
		response.setDatetime(KopalXMLParser.getFirstValue(doc, "datetime"));
		response.setOriginator(KopalXMLParser.getFirstValue(doc, "originator"));

		// Check for an "error" tag
		NodeList typelist = doc.getElementsByTagName("error");

		if (typelist.getLength() > 0) {
			DiasAdminError error = new DiasAdminError();

			error.setErrortype(KopalXMLParser.getFirstValue(doc, "errortype"));
			error.setErrorcode(KopalXMLParser.getFirstValue(doc, "errorcode"));
			error.setErrormessage(KopalXMLParser.getFirstValue(doc,
					"errormessage"));
			error.setErrordetail(KopalXMLParser.getFirstValue(doc,
					"errordetail"));

			response.setData(error);
		}

		// Check for a "result" tag
		typelist = doc.getElementsByTagName("result");

		if (typelist.getLength() > 0) {
			DiasAdminResult result = new DiasAdminResult();

			result.setResultcode(typelist.item(0).getAttributes().item(0)
					.getNodeValue());

			result.setResultString(typelist.item(0).getTextContent());
			// Check for a result according to the "storeFileType" request
			NodeList datalist = doc.getElementsByTagName("fileTypeID");

			if (datalist.getLength() > 0) {
				StoreFileTypeResult resultdata = new StoreFileTypeResult();

				resultdata.setFileTypeID(KopalXMLParser.getFirstValue(doc,
						"fileTypeID"));
				resultdata.setFileTypeCreationDate(KopalXMLParser
						.getFirstValue(doc, "fileTypeCreationDate"));

				result.setResultXML(resultdata);
			}

			// Check for a result according to the "getModifiedAIPs" request
			datalist = doc.getElementsByTagName("modifiedAIP");

			if (datalist.getLength() > 0) {
				GetModifiedAIPsResult resultdata = new GetModifiedAIPsResult();

				for (int i = 0; i < datalist.getLength(); i++) {
					AIPData aip = new AIPData();
					NodeList aipdata = datalist.item(i).getChildNodes();
					for (int j = 0; j < aipdata.getLength(); j++) {
						Node node = aipdata.item(j);

						if (node.getNodeName().equals("internalAssetID")) {
							aip.setInternalAssetID(node.getTextContent());
						}

						if (node.getNodeName().equals("modificationTimeStamp")) {
							aip.setModificationTimeStamp(node.getTextContent());
						}
					}
					resultdata.addAIP(aip);
				}
				result.setResultXML(resultdata);
			}

			// Check for a result according to the "getSchema" request
			datalist = doc.getElementsByTagName("xmlSchemaDefinition");

			if (datalist.getLength() == 1) {
				GetSchemaResult resultdata = new GetSchemaResult();

				resultdata.setXmlSchemaDefinition(datalist.item(0));

				result.setResultXML(resultdata);
			}

			// Check for a result according to the "getSchemaList" request
			datalist = doc.getElementsByTagName("xmlSchema");

			boolean isGetSchemaListRequest = false;

			if (datalist.getLength() > 0) {
				isGetSchemaListRequest = true;

				GetSchemaListResult resultdata = new GetSchemaListResult();

				for (int i = 0; i < datalist.getLength(); i++) {
					XmlSchemaData schema = new XmlSchemaData();
					NodeList schemadata = datalist.item(i).getChildNodes();
					for (int j = 0; j < schemadata.getLength(); j++) {
						Node node = schemadata.item(j);

						if (node.getNodeName().equals("xmlSchemaID")) {
							schema.setXmlSchemaID(node.getTextContent());
						}

						// TODO XML or xml?
						if (node.getNodeName().equals("xmlSchemaName")) {
							schema.setXmlSchemaName(node.getTextContent());
						}
						// TODO XML or xml?
						if (node.getNodeName().equals("xmlSchemaDescription")) {
							schema.setXmlSchemaDescription(node
									.getTextContent());
						}
						// TODO XML or xml?
						if (node.getNodeName().equals("xmlSchemaURI")) {
							schema.setXmlSchemaURI(node.getTextContent());
						}
						// TODO XML or xml?
						if (node.getNodeName().equals("xmlSchemaVersion")) {
							schema.setXmlSchemaVersion(node.getTextContent());
						}
					}
					resultdata.addXmlSchema(schema);
				}
				result.setResultXML(resultdata);
			}

			// Check for a result according to the "getSchemaList" request
			datalist = doc.getElementsByTagName("XMLSchemaID");

			if (datalist.getLength() > 0 && !isGetSchemaListRequest) {
				StoreSchemaResult resultdata = new StoreSchemaResult();

				// TODO XML or xml?
				resultdata.setXmlSchemaID(KopalXMLParser.getFirstValue(doc,
						"XMLSchemaID"));
				// TODO XML or xml?
				resultdata.setXmlSchemaVersion(KopalXMLParser.getFirstValue(
						doc, "XMLSchemaVersion"));

				result.setResultXML(resultdata);
			}

			// Store result data
			response.setData(result);
		}

		return response;
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ****

	/***************************************************************************
	 * Gets the data of the specified url from the server. If setUseHttps(false)
	 * has been called, http is used, else (or if setUseHttps(true) has been
	 * called) https is used.
	 * 
	 * @param url The path of the data to retrieve (without serveraddress and
	 *            port but with leading slash)
	 * @return The retrieved data as a string
	 * @throws IOException thrown if connection failed
	 **************************************************************************/
	private String retrieveDataFromDIAS(String url) throws IOException {
		String protocol = "https";
		if (!this.useHttps) {
			protocol = "http";
		}

		GetMethod diasRequest = new GetMethod(protocol + "://"
				+ this.adminServer + ":" + this.adminPort + url);

		try {
			this.diasClient.executeMethod(diasRequest);
			StringBuffer result = new StringBuffer();
			InputStream responseStream = new BufferedInputStream(diasRequest
					.getResponseBodyAsStream());

			byte[] buffer = new byte[INPUT_BUFFER_LENGTH];

			int read = responseStream.read(buffer);
			while (read > 0) {
				result.append(new String(buffer, 0, read));
				read = responseStream.read(buffer);
			}

			return result.toString();
		}
		finally {
			diasRequest.releaseConnection();
		}
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @return Returns the useHttps.
	 **************************************************************************/
	public boolean isUseHttps() {
		return this.useHttps;
	}

	/***************************************************************************
	 * @param Https The useHttps to set.
	 **************************************************************************/
	public void setUseHttps(boolean Https) {
		this.useHttps = Https;
	}

	/***************************************************************************
	 * @return Returns the adminPort.
	 **************************************************************************/
	public int getAdminPort() {
		return this.adminPort;
	}

	/***************************************************************************
	 * @param p The port to set.
	 **************************************************************************/
	public void setAdminPort(int p) {
		this.adminPort = p;
	}

	/***************************************************************************
	 * @return Returns the adminServer.
	 **************************************************************************/
	public String getAdminServer() {
		return this.adminServer;
	}

	/***************************************************************************
	 * @param adminAddr The serverAddress to set.
	 **************************************************************************/
	public void setAdminServer(String adminAddr) {
		this.adminServer = adminAddr;
	}

	/***************************************************************************
	 * @return Returns the adminPwd.
	 **************************************************************************/
	public String getAdminPwd() {
		return this.adminPwd;
	}

	/***************************************************************************
	 * @param _adminPwd The adminPwd to set.
	 **************************************************************************/
	public void setAdminPwd(String _adminPwd) {
		this.adminPwd = _adminPwd;
	}

	/***************************************************************************
	 * @return Returns the adminUser.
	 **************************************************************************/
	public String getAdminUser() {
		return this.adminUser;
	}

	/***************************************************************************
	 * @param _adminUser The serverUser to set.
	 **************************************************************************/
	public void setAdminUser(String _adminUser) {
		this.adminUser = _adminUser;
	}

	/***************************************************************************
	 * @return Returns the keystoreFile.
	 **************************************************************************/
	public String getKeyStoreFile() {
		return this.keyStoreFile;
	}

	/***************************************************************************
	 * @param keyStore The keyStoreFile to set.
	 **************************************************************************/
	public void setKeyStoreFile(String keyStore) {
		System.setProperty("javax.net.ssl.trustStore", keyStore);
		this.keyStoreFile = keyStore;
	}

	/***************************************************************************
	 * @return Returns the SRU maximumRecords
	 **************************************************************************/
	public long getSruMaximumRecords() {
		return this.sruMaximumRecords;
	}

	/***************************************************************************
	 * @param _sruMaximumRecords The SRU maximumRecords to set.
	 **************************************************************************/
	public void setSruMaximumRecords(long _sruMaximumRecords) {
		this.sruMaximumRecords = _sruMaximumRecords;
	}

	/***************************************************************************
	 * @return Returns the SRU resultSetTTL
	 **************************************************************************/
	public long getSruResultSetTTL() {
		return this.sruResultSetTTL;
	}

	/***************************************************************************
	 * @param _sruResultSetTTL The SRU resultSetTTL to set.
	 **************************************************************************/
	public void setSruResultSetTTL(long _sruResultSetTTL) {
		this.sruResultSetTTL = _sruResultSetTTL;
	}

	/***************************************************************************
	 * @return Returns the SRU sortKeys
	 **************************************************************************/
	public String getSruSortKeys() {
		return this.sruSortKeys;
	}

	/***************************************************************************
	 * @param _sruSortKeys The SRU sortKeys to set.
	 **************************************************************************/
	public void setSruSortKeys(String _sruSortKeys) {
		this.sruSortKeys = _sruSortKeys;
	}

	/***************************************************************************
	 * @return Returns the SRU startRecord
	 **************************************************************************/
	public long getSruStartRecord() {
		return this.sruStartRecord;
	}

	/***************************************************************************
	 * @param _sruStartRecord The SRU startRecord to set.
	 **************************************************************************/
	public void setSruStartRecord(long _sruStartRecord) {
		this.sruStartRecord = _sruStartRecord;
	}

	/***************************************************************************
	 * @return Returns the SRU version
	 **************************************************************************/
	public String getSruVersion() {
		return this.sruVersion;
	}

	/***************************************************************************
	 * @param _sruVersion The SRU version to set.
	 **************************************************************************/
	public void setSruVersion(String _sruVersion) {
		this.sruVersion = _sruVersion;
	}

	/***************************************************************************
	 * @return Returns the SRU x_additionalDataElements
	 **************************************************************************/
	public String getSruX_additionalDataElements() {
		return this.sruX_additionalDataElements;
	}

	/***************************************************************************
	 * @param _sruX_additionalDataElements The SRU x_additionalDataElements to
	 *            set.
	 **************************************************************************/
	public void setSruX_additionalDataElements(
			String _sruX_additionalDataElements) {
		this.sruX_additionalDataElements = _sruX_additionalDataElements;
	}

	/***************************************************************************
	 * @return Returns the SRU x_listAffectedContentFiles
	 **************************************************************************/
	public boolean isSruX_listAffectedContentFiles() {
		return this.sruX_listAffectedContentFiles;
	}

	/***************************************************************************
	 * @param _sruX_listAffectedContentFiles The SRU x_listAffectedContentFiles
	 *            to set.
	 **************************************************************************/
	public void setSruX_listAffectedContentFiles(
			boolean _sruX_listAffectedContentFiles) {
		this.sruX_listAffectedContentFiles = _sruX_listAffectedContentFiles;
	}

}
