/*******************************************************************************
 * de.langzeitarchivierung.kopal.formats / Uof.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.formats;

import gov.loc.mets.AmdSecType;
import gov.loc.mets.DivType;
import gov.loc.mets.FileGrpType;
import gov.loc.mets.MdSecType;
import gov.loc.mets.MetsDocument;
import gov.loc.mets.MetsType;
import gov.loc.mets.StructMapType;
import gov.loc.mets.MdSecType.MdWrap;
import gov.loc.mets.MdSecType.MdRef.MDTYPE;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.DataFormatException;

import javax.xml.namespace.QName;

import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;
import org.purl.dc.elements.x11.DcSetDocument;
import org.purl.dc.elements.x11.DcSetDocument.DcSet;

import de.ddb.lmeRfile.LmerFile;
import de.ddb.lmeRobject.LmerObject;
import de.ddb.lmeRobject.LmerObject.TransferChecksum;
import de.ddb.lmeRprocess.LmerProcess;
import de.langzeitarchivierung.kopal.util.ChecksumGenerator;
import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.FormatRegistry;
import edu.harvard.hul.ois.xml.ns.jhove.JhoveDocument;
import edu.harvard.hul.ois.xml.ns.jhove.JhoveDocument.Jhove;

/*******************************************************************************
 * The kopal Universal Object Format (UOF).
 * 
 * CODE SNIPPET if you have to create a METS doc:
 * 
 * <code>
 * MetsDocument mets = MetsDocument.Factory.newInstance();
 * uof = mets.addNewMets();
 * XmlCursor cursor = mets.newCursor();
 * 
 * if (cursor.toFirstChild()) {
 *     System.out.println("Setting attributes");
 *     cursor.setAttributeText(
 *       new QName("http://www.w3.org/2001/XMLSchema-instance", "schemaLocation"),
 *       "http://www.loc.gov/METS/ http://www.loc.gov/standards/mets/mets.xsd " +
 *       "http://www.ddb.de/LMERfile " +
 *       "http://www.ddb.de/standards/lmer/lmer-file.xsd " +
 *       "http://www.ddb.de/LMERobject " +
 *       "http://www.ddb.de/standards/lmer/lmer-object.xsd " +
 *       "http://www.ddb.de/LMERprocess " +
 *       "http://www.ddb.de/standards/lmer/lmer-process.xsd " +
 *       "http://purl.org/dc/elements/1.1/ " +
 *       "http://dublincore.org/schemas/xmls/qdc/2003/04/02/dc.xsd"
 *     );
 *  }
 *  
 *  cursor.dispose();
 *  </code>
 * 
 * @author ludwig
 * @see http://kopal.langzeitarchivierung.de/index_objektspezifikation.php.en
 ******************************************************************************/

public class Uof implements MetadataFormat, Serializable
{

	// CONSTANTS **********************************************************

	private static final long		serialVersionUID		= 1L;
	private static final String		LMER_OBJECT_ID			= "TECHMD-OBJECT";
	private static final String		LMER_PROCESS_OBJECT_ID	= "DIGIPROVMD-OBJECT";

	public static final String		FILE_LOCAT_HREF_PREFIX	= "file://./";

	// CLASS VARIABLES (Static variables) **********************************

	private static Logger			defaultLogger			= Logger
																	.getLogger("de.langzeitarchivierung.kopal");

	/** Limits for DIAS */
	public static int				diasDcDescriptionLimit	= 1024;
	public static int				diasDcIdentifierLimit	= 255;
	public static int				diasDcTypeLimit			= 100;

	/** Prefixes used in the METS file */
	private static String			dmdPrefix				= "DMD";
	private static String			techMdPrefix			= "TECHMD";
	private static String			digiprovMdPrefix		= "DIGIPROVMD";
	private static String			mdWrapPostfix			= "MDWRAP";
	/** prefixSeparator can be the empty string */
	private static String			prefixSeparator			= "-";
	private static FormatRegistry	fRegistry				= new FormatRegistry();

	// STATE (Instance variables) *****************************************

	private int						dmdCounter				= 0;
	private int						fileCounter				= 0;

	/***************************************************************************
	 * The first part of the content file path which is omitted in the mets
	 * file. The default path is not very useful and should be changed.
	 **************************************************************************/
	private URI						basePathUri				= new File(
																	System
																			.getProperty("user.dir"))
																	.toURI();
	private boolean					noWhiteSpacesInTechMd	= false;
	private long					maxTechMDLength			= 262144;

	/** This contains the complete XmlBeans metadata classes */
	private MetsType				uof;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Constructor. Object configures itself.
	 **************************************************************************/
	public Uof() {
		super();

		try {
			Configurator.configure(this);
		}
		catch (Exception e) {
			e.printStackTrace();
			defaultLogger.log(Level.SEVERE, "Error configuring class "
					+ this.getClass().getName() + ". System message: "
					+ e.getMessage());
		}
	}

	// MANIPULATION (Manipulation - Implemented Methods) ******************

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getObjectId()
	 */
	public String getObjectId() throws XmlException {
		return getLmerObjectCopy().getPersistentIdentifier();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#setObjectId(java.lang.String)
	 */
	public void setObjectId(String id) throws XmlException, DataFormatException {
		LmerObject lobj = getLmerObjectCopy();
		lobj.setPersistentIdentifier(id);
		setLmerObject(lobj);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#addDmd(java.lang.String,
	 *      java.lang.String)
	 */
	public void addDescMd(String dmd, String mdType) throws XmlException {
		MdWrap dmdwrap = getNewDmdWrap();
		dmdwrap.setMIMETYPE("text/xml");
		MdWrap.MDTYPE.Enum metsMdTypes = MdWrap.MDTYPE.OTHER;

		if (mdType.equalsIgnoreCase("dc")) {
			metsMdTypes = MdWrap.MDTYPE.DC;
		}
		else if (mdType.equalsIgnoreCase("ddi")) {
			metsMdTypes = MdWrap.MDTYPE.DDI;
		}
		else if (mdType.equalsIgnoreCase("ead")) {
			metsMdTypes = MdWrap.MDTYPE.EAD;
		}
		else if (mdType.equalsIgnoreCase("fdgc")) {
			metsMdTypes = MdWrap.MDTYPE.FGDC;
		}
		else if (mdType.equalsIgnoreCase("lc_av")) {
			metsMdTypes = MdWrap.MDTYPE.LC_AV;
		}
		else if (mdType.equalsIgnoreCase("lom")) {
			metsMdTypes = MdWrap.MDTYPE.LOM;
		}
		else if (mdType.equalsIgnoreCase("marc")) {
			metsMdTypes = MdWrap.MDTYPE.MARC;
		}
		else if (mdType.equalsIgnoreCase("mods")) {
			metsMdTypes = MdWrap.MDTYPE.MODS;
		}
		else if (mdType.equalsIgnoreCase("nisoimg")) {
			metsMdTypes = MdWrap.MDTYPE.NISOIMG;
		}
		else if (mdType.equalsIgnoreCase("teihdr")) {
			metsMdTypes = MdWrap.MDTYPE.TEIHDR;
		}
		else if (mdType.equalsIgnoreCase("vra")) {
			metsMdTypes = MdWrap.MDTYPE.VRA;
		}
		else {
			// MdType is "other", add the MdType now.
			dmdwrap.setOTHERMDTYPE(mdType);
		}

		dmdwrap.setMDTYPE(metsMdTypes);

		XmlCursor dmdCursor = dmdwrap.addNewXmlData().newCursor();
		dmdCursor.toEndToken();

		XmlCursor sourceCursor = null;

		// Ensure the DC descriptive metadata is not too long.
		if (metsMdTypes == MdWrap.MDTYPE.DC) {
			XmlOptions opts = new XmlOptions();
			// Tricky! Because we get only an XML fragment of a dcSet we have to
			// deceive the factory parser.
			opts.setLoadReplaceDocumentElement(new QName(
					"http://purl.org/dc/elements/1.1/", "dcSet", "dc"));
			DcSet dcDmd = this.getCorrectedFieldLength(DcSetDocument.Factory
					.parse(dmd, opts).getDcSet());
			sourceCursor = dcDmd.newCursor();
		}
		else {
			sourceCursor = XmlObject.Factory.parse(dmd).newCursor();
		}

		sourceCursor.toFirstContentToken();

		while (sourceCursor.toNextSibling()) {
			sourceCursor.copyXml(dmdCursor);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#addFile(java.io.File,
	 *      java.lang.String)
	 */
	public void addFile(File contentFile, String techMd) throws Exception {
		// Check the techMd first and get a Jhove object from the techMd.
		Jhove jhoveInfo = checkXmlData(techMd);

		// Check the base path.
		if (!contentFile.toURI().toString().startsWith(
				this.basePathUri.toString())) {
			throw new Exception("File is not in the base path "
					+ this.basePathUri.toString());
		}

		// Assemble the filename.
		String fileNameToAdd = contentFile.toURI().toString().substring(
				this.basePathUri.toString().length() - 6);

		// Is this file already in the METS?
		FileGrpType.File files[] = this.uof.getFileSec().getFileGrpArray(0)
				.getFileArray();
		for (int i = 0; i < files.length; i++) {
			String fileNameInFileSec = files[i].getFLocatArray(0).getHref()
					.substring(FILE_LOCAT_HREF_PREFIX.length());
			if (fileNameToAdd.equals(fileNameInFileSec)) {
				throw new Exception("Duplicate file in the METS");
			}
		}

		// Create a new lmer file object and update some lmerFile information.
		LmerFile lmerFile = LmerFile.Factory.newInstance();
		lmerFile.addNewFormat();
		lmerFile = updateLmerFileData(lmerFile, jhoveInfo);

		// Add new tech md and mdwrap.
		AmdSecType[] amdSecs = this.uof.getAmdSecArray();
		MdSecType tMd = amdSecs[0].addNewTechMD();
		String techMdId = techMdPrefix + prefixSeparator + this.fileCounter;
		tMd.setID(techMdId);

		MdWrap fileMdWrap = tMd.addNewMdWrap();
		fileMdWrap.setID(techMdId + prefixSeparator + mdWrapPostfix);
		fileMdWrap.setMIMETYPE("text/xml");
		fileMdWrap.setMDTYPE(MdWrap.MDTYPE.OTHER);
		fileMdWrap.setOTHERMDTYPE("lmerFile");
		// fileMdWrap.setLABEL("lmerFile");
		fileMdWrap.addNewXmlData();

		// Add lmer file sec.
		fileMdWrap.getXmlData().set(lmerFile);

		// Add file entry.
		FileGrpType.File fileEntry = this.uof.getFileSec().getFileGrpArray(0)
				.addNewFile();

		// Create and set file ID.
		String fileId = "FILE" + prefixSeparator + this.fileCounter;
		fileEntry.setID(fileId);

		// Create and set the ADMIDs.
		List<String> admIds = new ArrayList<String>();
		admIds.add(techMdId);
		fileEntry.setADMID(admIds);

		// Add and set the file location.
		FileGrpType.File.FLocat fileLocat = fileEntry.addNewFLocat();
		fileLocat.setLOCTYPE(FileGrpType.File.FLocat.LOCTYPE.URL);
		fileLocat.setType("simple");

		// Update the other file data.
		fileEntry = updateFileSecFileData(contentFile, fileEntry, jhoveInfo);

		// Add filepointer to struct map.
		getFirstAssetDiv().addNewFptr().setFILEID(fileId);

		// Update the lmer object section.
		this.fileCounter++;
		LmerObject lobj = getLmerObjectCopy();
		lobj.setNumberOfFiles(new BigInteger(new Integer(this.fileCounter)
				.toString()));
		setLmerObject(lobj);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getFiles()
	 */
	public File[] getFiles() throws URISyntaxException {
		FileGrpType.File files[] = this.uof.getFileSec().getFileGrpArray(0)
				.getFileArray();
		File result[] = new File[files.length];

		for (int i = 0; i < files.length; i++) {
			String fileNameInFileSec = files[i].getFLocatArray(0).getHref()
					.substring(FILE_LOCAT_HREF_PREFIX.length());
			result[i] = new File(new URI(this.basePathUri.toString()
					+ fileNameInFileSec));
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getFileMd(java.io.File)
	 */
	public String getFileMd(File contentFile) throws Exception {
		FileGrpType.File files[] = this.uof.getFileSec().getFileGrpArray(0)
				.getFileArray();

		for (int i = 0; i < files.length; i++) {
			String absoluteFileSecPath = this.basePathUri
					+ files[i].getFLocatArray(0).getHref().substring(
							FILE_LOCAT_HREF_PREFIX.length());
			if (contentFile.toURI().toString().equals(absoluteFileSecPath)) {
				return files[i].toString();
			}
		}

		throw new FileNotFoundException("File not found in UOF");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getTechMd(java.io.File)
	 */
	public String getTechMd(File contentFile) throws FileNotFoundException,
			XmlException {
		return getLmerFileCopy(getFileTechMdId(contentFile)).getXmlDataArray(0)
				.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#addMigration(java.lang.String)
	 */
	public void addProvMd(Object provMd) throws Exception {
		String digiprovMdId = "";

		// Cast the java object to LmerProcess.
		LmerProcess lmerProcessFragment = (LmerProcess) provMd;

		// Get the list of ADMIDs from the file group.
		List admIdList = this.uof.getFileSec().getFileGrpArray(0).getADMID();

		// Check the old one and use the amount of IDs as new counter. Use the
		// list size as new counter: list size - 1 (TECHMD ID entry) + 1 (new
		// digiprov entry) - 1 (we computer scientists start counting at 0 :-).
		if (admIdList.size() > 0) {
			digiprovMdId = LMER_PROCESS_OBJECT_ID + prefixSeparator
					+ (admIdList.size() - 1);
		}

		// Add the file into the digiprov section.
		this.addDigiprovSec(lmerProcessFragment, digiprovMdId);

		// Set the digiprov ID in the file group's admid. At first add the newly
		// created migration ID, afterwards add all other IDs with the TechMdId
		// at the end.
		List<String> newAdmIdList = new ArrayList<String>();
		newAdmIdList.add(digiprovMdId);
		newAdmIdList
				.addAll(this.uof.getFileSec().getFileGrpArray(0).getADMID());
		this.uof.getFileSec().getFileGrpArray(0).setADMID(newAdmIdList);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#addMigration(java.io.File,
	 *      java.lang.String)
	 */
	public void addProvMd(File contentFile, Object provMd) throws Exception {
		String digiprovMdId = "";

		// Cast the java object to LmerProcess.
		LmerProcess lmerProcessFragment = (LmerProcess) provMd;

		// Get the FILEID of the current file.
		String fileId = getFileId(contentFile);

		// Get the ADMID from the last existing migration.
		List admIdList = getFileMdMets(contentFile).getADMID();

		// Check the old one and use the amount of IDs as new counter. Use the
		// list size as new counter: list size - 1 (TECHMD ID entry) + 1 (new
		// digiprov entry) - 1 (we computer scientists start counting at 0 :-).
		if (admIdList.size() > 0) {
			digiprovMdId = digiprovMdPrefix + prefixSeparator + fileId
					+ prefixSeparator + (admIdList.size() - 1);
		}

		// Add the file into the digiprov section.
		this.addDigiprovSec(lmerProcessFragment, digiprovMdId);

		// Add the digiprov ID in the file's admid. At first add the newly
		// created migration ID, afterwards add all other IDs with the TechMdId
		// at the end.
		List<String> newAdmIdList = new ArrayList<String>();
		newAdmIdList.add(digiprovMdId);
		newAdmIdList.addAll(getFileMdMets(contentFile).getADMID());
		getFileMdMets(contentFile).setADMID(newAdmIdList);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getProvMd(java.io.File)
	 */
	public List<String> getProvMd(File contentFile) throws Exception {
		List<String> result = new ArrayList<String>();

		// Get the list of the file's ADMIDs.
		List<String> admIdList = getFileMdMets(contentFile).getADMID();
		List<LmerProcess> lmerProcessList = this.getLmerProcessCopy(admIdList);

		Iterator iter = lmerProcessList.iterator();
		while (iter.hasNext()) {
			result.add(((LmerProcess) iter.next()).toString());
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getProvMd()
	 */
	public List<String> getProvMd() throws Exception {
		List<String> result = new ArrayList<String>();

		// Get the list of ADMIDs from the file group.
		List<String> admIdList = this.uof.getFileSec().getFileGrpArray(0)
				.getADMID();
		List<LmerProcess> lmerProcessList = this.getLmerProcessCopy(admIdList);

		Iterator iter = lmerProcessList.iterator();
		while (iter.hasNext()) {
			result.add(((LmerProcess) iter.next()).toString());
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getDescMd()
	 */
	public List<String> getDescMd() throws Exception {
		List<String> result = new ArrayList<String>();

		// Get the list of DMDIDs from the structmap.
		List<String> dmdIdList = this.uof.getStructMapArray(0).getDiv()
				.getDMDID();
		List<MdSecType> dmdSecList = this.getDmdSecCopy(dmdIdList);

		Iterator iter = dmdSecList.iterator();
		while (iter.hasNext()) {
			result.add(((MdSecType) iter.next()).toString());
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#useTemplate(java.io.File)
	 */
	public void useTemplate(File templateFile) throws Exception {
		MetsDocument xml = MetsDocument.Factory.parse(templateFile);
		this.uof = xml.getMets();

		defaultLogger.log(Level.FINEST, "METS template file "
				+ "sucessfully loaded");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#saveMetadataFormat(java.io.File)
	 */
	public void saveMetadataFormat(File outputFile) throws IOException,
			DataFormatException, XmlException {
		XmlOptions savingOpts = new XmlOptions();

		// This prefixes already are in the template file:
		//
		// HashMap suggestedPrefixes = new HashMap();
		// suggestedPrefixes.put("http://www.loc.gov/METS/", "");
		// suggestedPrefixes.put("http://purl.org/dc/elements/1.1/",
		// "dc");
		// suggestedPrefixes.put("http://www.w3.org/2001/XMLSchema-instance",
		// "xsi");
		// suggestedPrefixes.put("http://www.ddb.de/LMERfile",
		// "lmerFile");
		// suggestedPrefixes.put("http://www.ddb.de/LMERobject",
		// "lmerObject");
		// suggestedPrefixes.put("http://www.ddb.de/LMERprocess",
		// "lmerProcess");
		// suggestedPrefixes.put("http://www.w3.org/1999/xlink",
		// "xlink");
		// opts.setSaveSuggestedPrefixes(suggestedPrefixes);

		savingOpts.setSaveAggressiveNamespaces();
		// Want to pretty print the XML file?? Try this:
		// savingOpts.setSavePrettyPrint();
		savingOpts.setSaveOuter();

		updateUof();

		this.uof.save(outputFile, savingOpts);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.formats.MetadataFormat#getMetadataImplementation()
	 */
	public Object getMetadataImplementation() {
		return this.uof;
	}

	// MANIPULATION (Manipulation - further methods) ******************

	/***************************************************************************
	 * Rewrites all file data, that could have been altered within a file
	 * migration. If you already have the content file's file data in the METS
	 * metadata object, use it to update some file data for migrational use.
	 * 
	 * @param contentFile
	 * @param techMd
	 * @throws Exception
	 **************************************************************************/
	public void updateFileData(File contentFile, String techMd)
			throws Exception {
		Jhove jhoveInfo = checkXmlData(techMd);
		String fileId = getFileId(contentFile);

		// Update the file data in the file sec.
		gov.loc.mets.FileGrpType.File fileArray[] = this.uof.getFileSec()
				.getFileGrpArray(0).getFileArray();

		for (int i = 0; i < fileArray.length; i++) {
			if (fileArray[i].getID().equals(fileId)) {
				fileArray[i] = updateFileSecFileData(contentFile, fileArray[i],
						jhoveInfo);
			}
		}

		// Update the file data in the techMd.
		String techMdId = getFileTechMdId(contentFile);
		LmerFile lmerFile = getLmerFileCopy(techMdId);

		// Update the lmerFile data.
		lmerFile = updateLmerFileData(lmerFile, jhoveInfo);

		// Set the lmerFile data.
		gov.loc.mets.MdSecType techMdArray[] = this.uof.getAmdSecArray(0)
				.getTechMDArray();
		for (int i = 0; i < techMdArray.length; i++) {
			if (techMdArray[i].getID().equals(techMdId)) {
				techMdArray[i].getMdWrap().getXmlData().set(lmerFile);
			}
		}
	}

	/***************************************************************************
	 * LMER specific add-method for "linked-to" fields.
	 * 
	 * @param contentFile - The content file
	 * @param dependencies - The list of dependencies
	 * @throws FileNotFoundException
	 * @throws Exception
	 **************************************************************************/
	public void addLmerFileLinkedTo(File contentFile, List dependencies)
			throws FileNotFoundException, Exception {
		LmerFile lFile = getLmerFileCopy(contentFile);

		for (int i = 0; i < dependencies.size(); i++) {
			lFile.addLinkedTo(getFileId((File) dependencies.get(i)));
		}

		setLmerFile(lFile, getFileTechMdId(contentFile));
	}

	/***************************************************************************
	 * Gets a COPY of the LmerObject. Please do not forget to set the lmerObject
	 * again afterwards.
	 * 
	 * @return LmerObject - A COPY of the LmerObject.
	 * @throws XmlException
	 **************************************************************************/
	public LmerObject getLmerObjectCopy() throws XmlException {
		return LmerObject.Factory.parse(getTechMdXmlData(getObjectTechMdId())
				.toString());
	}

	/***************************************************************************
	 * Sets the LmerObject and recalculates the transfer checksum. The transfer
	 * checksum is the XORed file checksums.
	 * 
	 * @param lObj the LmerObject to set.
	 * @throws XmlException
	 * @throws DataFormatException
	 **************************************************************************/
	public void setLmerObject(LmerObject lObj) throws XmlException,
			DataFormatException {
		// Update the checksums.
		getTechMdXmlData(getObjectTechMdId()).set(
				recalculateObjectChecksum(lObj));
	}

	/***************************************************************************
	 * Gets a COPY of the LmerFile of the specified file.
	 * 
	 * @param contentFile The file for which we want the lmerFile.
	 * @return LmerFile - A COPY of the LmerFile or null if the contentFile does
	 *         not exist.
	 * @throws FileNotFoundException
	 * @throws XmlException
	 **************************************************************************/
	public LmerFile getLmerFileCopy(File contentFile)
			throws FileNotFoundException, XmlException {
		String techMdId = getFileTechMdId(contentFile);

		if (techMdId != null) {
			return LmerFile.Factory
					.parse(getTechMdXmlData(techMdId).toString());
		}

		return null;
	}

	/***************************************************************************
	 * Gets a COPY of the LmerFile of the specified techMdId.
	 * 
	 * @param techMdId The techMdId for which we want the lmerFile.
	 * @return LmerFile - A COPY of the LmerFile or null if the techMdId does
	 *         not exist.
	 * @throws XmlException
	 **************************************************************************/
	public LmerFile getLmerFileCopy(String techMdId) throws XmlException {
		return LmerFile.Factory.parse(getTechMdXmlData(techMdId).toString());
	}

	/***************************************************************************
	 * Gets a copy of all the descriptive metadata objects from the dmdSec of
	 * the specified identifier list.
	 * 
	 * @param dmdIdList
	 * @return List<MdSecType> - A List containing all DMD sections.
	 * @throws Exception
	 **************************************************************************/
	public List<MdSecType> getDmdSecCopy(List<String> dmdIdList)
			throws Exception {
		// Create a list and add all the descriptive metadata blocks concerning
		// the given DMDID.
		ArrayList<MdSecType> result = new ArrayList<MdSecType>();

		// Iterate over all the DMDIDs and add the descriptive metadata sections
		// to the result list.
		Iterator iter = dmdIdList.iterator();
		while (iter.hasNext()) {
			String dmdId = (String) iter.next();
			MdSecType dmdSec = getDmdSecCopy(dmdId);
			if (dmdSec != null) {
				result.add(dmdSec);
			}
		}

		return result;
	}

	/***************************************************************************
	 * Gets a copy of the descriptive metadata object of the specified ID.
	 * 
	 * @param dmdId
	 * @return MdSecType - A COPY of the specified DMD section.
	 * @throws Exception
	 **************************************************************************/
	public MdSecType getDmdSecCopy(String dmdId) throws Exception {
		// Get the array of dmd sections.
		MdSecType dmdArray[] = this.uof.getDmdSecArray();

		// Go throught all the dmd sections and return the one that fits to the
		// given ID.
		for (int i = 0; i < dmdArray.length; i++) {
			if (dmdArray[i].getID().equals(dmdId)) {
				return MdSecType.Factory.parse(dmdArray[i].getMdWrap()
						.getXmlData().toString());
			}
		}

		return null;
	}

	/***************************************************************************
	 * Gets a copy of all the LmerProcess objects of the specified identifier
	 * list.
	 * 
	 * @param admIdList
	 * @return List<LmerProcess> - A list of all provenance metadata sections
	 *         specified by the given ADMID list.
	 * @throws Exception
	 **************************************************************************/
	public List<LmerProcess> getLmerProcessCopy(List<String> admIdList)
			throws Exception {
		// Create a list and add all the provenance metadata blocks concerning
		// the given ADMID, starting with the first (the last migration
		// metadata).
		ArrayList<LmerProcess> result = new ArrayList<LmerProcess>();

		// Iterate over all the ADMIDs and add the provenance metadata sections
		// to the result list. We can iterate over ALL the IDs, because the
		// getLmerProcessCopy() method only iterates over the digiprovMd
		// sections.
		Iterator iter = admIdList.iterator();
		while (iter.hasNext()) {
			String admId = (String) iter.next();
			LmerProcess lmerProcess = getLmerProcessCopy(admId);
			if (lmerProcess != null) {
				result.add(lmerProcess);
			}
		}

		return result;
	}

	/***************************************************************************
	 * Gets a copy of the LmerProcess object of the specified digiprovMdId.
	 * 
	 * @param digiprovMdId
	 * @return LmerProcess - A COPY of the provenance metadata section specified
	 *         by the digiprovMdId.
	 **************************************************************************/
	public LmerProcess getLmerProcessCopy(String digiprovMdId) throws Exception {
		// Get the array of digiprov sections.
		MdSecType digiprovArray[] = this.uof.getAmdSecArray(0)
				.getDigiprovMDArray();

		// Go throught all the digiprov sections and return the one that fits to
		// the given ID.
		for (int i = 0; i < digiprovArray.length; i++) {
			if (digiprovArray[i].getID().equals(digiprovMdId)) {
				return LmerProcess.Factory.parse(digiprovArray[i].getMdWrap()
						.getXmlData().toString());
			}
		}

		return null;
	}

	/***************************************************************************
	 * Returns the Jhove technical metadata for a content file.
	 * 
	 * @param contentFile The file you want the jhove data for.
	 * @return Jhove - The jhove object of the given content file.
	 * @throws XmlException if the technical metadata for the content file can
	 *             not be parsed.
	 * @throws FileNotFoundException if the contentFile is not in the mets.
	 **************************************************************************/
	public Jhove getJhoveTechMd(File contentFile) throws XmlException,
			FileNotFoundException {
		return getJhoveTechMd(getFileTechMdId(contentFile));
	}

	/***************************************************************************
	 * Returns the jhove technical metadata for a techMdId in the mets.
	 * 
	 * @param techMdId The ID of the technical Metadata.
	 * @return Jhove - The jhove object of the given ID.
	 * @throws XmlException if the techMdId does not exist
	 **************************************************************************/
	public Jhove getJhoveTechMd(String techMdId) throws XmlException {
		return JhoveDocument.Factory.parse(
				getLmerFileCopy(techMdId).getXmlDataArray(0).toString())
				.getJhove();
	}

	/***************************************************************************
	 * Returns a copy of the first dublin core descriptive metadata section in
	 * the mets.
	 * 
	 * @return DcSet - The dcSet object of the first dublin core descriptive
	 *         metadata section in the mets or null if none exists.
	 * @throws XmlException
	 **************************************************************************/
	public DcSet getDcCopy() throws XmlException {
		MdSecType dmd[] = this.uof.getDmdSecArray();

		for (int i = 0; i < dmd.length; i++) {
			if (dmd[i].getMdWrap().getMDTYPE().toString().equals(
					MDTYPE.DC.toString())) {
				XmlOptions opts = new XmlOptions();
				// Tricky! Because we get only an xmlfragment of a dcSet we
				// have to deceive the Factory parser.
				opts.setLoadReplaceDocumentElement(new QName(
						"http://purl.org/dc/elements/1.1/", "dcSet", "dc"));
				return DcSetDocument.Factory.parse(
						dmd[i].getMdWrap().getXmlData().toString(), opts)
						.getDcSet();
			}
		}

		return null;
	}

	/***************************************************************************
	 * Set the lmerFile block under the amdSec.techMD.mdWrap.XmlData hierarchy
	 * with the given techMdID.
	 * 
	 * @param lFile
	 * @param techMdId
	 * @throws Exception
	 **************************************************************************/
	public void setLmerFile(LmerFile lFile, String techMdId) throws Exception {
		getTechMdXmlData(techMdId).set(lFile);
		// Update checksums.
		setLmerObject(getLmerObjectCopy());
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ****

	/***************************************************************************
	 * Corrects the length of the dublin core metadata fields if any of them is
	 * too long.
	 * 
	 * @param dc The dcSet to correct
	 * @return DcSet - The corrected dcSet
	 **************************************************************************/
	private DcSet getCorrectedFieldLength(DcSet dc) {
		dc.setContributorArray(getCorrectedArray(dc.getContributorArray(),
				diasDcTypeLimit));
		dc.setTitleArray(getCorrectedArray(dc.getTitleArray(),
				diasDcIdentifierLimit));
		dc.setCreatorArray(getCorrectedArray(dc.getCreatorArray(),
				diasDcTypeLimit));
		dc.setSubjectArray(getCorrectedArray(dc.getSubjectArray(),
				diasDcIdentifierLimit));
		dc.setDescriptionArray(getCorrectedArray(dc.getDescriptionArray(),
				diasDcDescriptionLimit));
		dc.setPublisherArray(getCorrectedArray(dc.getPublisherArray(),
				diasDcTypeLimit));
		dc.setDateArray(getCorrectedArray(dc.getDateArray(), diasDcTypeLimit));
		dc.setTypeArray(getCorrectedArray(dc.getTypeArray(), diasDcTypeLimit));
		dc.setFormatArray(getCorrectedArray(dc.getFormatArray(),
				diasDcTypeLimit));
		dc.setIdentifierArray(getCorrectedArray(dc.getIdentifierArray(),
				diasDcIdentifierLimit));
		dc.setSourceArray(getCorrectedArray(dc.getSourceArray(),
				diasDcIdentifierLimit));
		dc.setLanguageArray(getCorrectedArray(dc.getLanguageArray(),
				diasDcTypeLimit));
		dc.setRelationArray(getCorrectedArray(dc.getRelationArray(),
				diasDcIdentifierLimit));
		dc.setCoverageArray(getCorrectedArray(dc.getCoverageArray(),
				diasDcTypeLimit));
		dc.setRightsArray(getCorrectedArray(dc.getRightsArray(),
				diasDcIdentifierLimit));

		return dc;
	}

	/***************************************************************************
	 * Divides the given string array. The result strings will not use more
	 * memory than maxLengthInBytes bytes. We now divide in strings of a certain
	 * BYTE amount instead of a certain CHAR amount.
	 * 
	 * @param toCorrect
	 * @param maxLengthInBytes
	 * @return String[] - The corrected Array.
	 **************************************************************************/
	private String[] getCorrectedArray(String toCorrect[], int maxLengthInBytes) {
		ArrayList<String> correctedStrings = new ArrayList<String>();
		int amountOfCharacters = maxLengthInBytes;

		// DEBUG OUTPUPT!
		//
		// for (int j = 0; j < toCorrect.length; j++) {
		// System.out.println("dc:value[" + j + "]: '" + toCorrect[j] + "'");
		// }
		//
		// DEBUG OUTPUT!

		for (int i = 0; i < toCorrect.length; i++) {
			try {
				while (toCorrect[i].getBytes("UTF-8").length > maxLengthInBytes) {
					// Look for the correct position to divide the string.
					// We must find the amount of characters, which fits into
					// maxLengthInBytes UTF-8 bytes.
					while (toCorrect[i].substring(0, amountOfCharacters)
							.getBytes("UTF-8").length > maxLengthInBytes) {
						amountOfCharacters--;
					}
					// Divide the string at position of the computed character
					// value.
					correctedStrings.add(toCorrect[i].substring(0,
							amountOfCharacters));
					toCorrect[i] = toCorrect[i].substring(amountOfCharacters);
				}
			}
			catch (UnsupportedEncodingException e) {
				// Doesn't occur here, because we always chose the correct
				// encoding :-)
				e.printStackTrace();
			}
			correctedStrings.add(toCorrect[i]);
		}

		// Create a new string array and fill it.
		String result[] = new String[correctedStrings.size()];

		Iterator strIter = correctedStrings.iterator();
		for (int i = 0; i < result.length; i++) {
			result[i] = (String) strIter.next();
		}

		// DEBUG OUTPUPT!
		//
		// for (int j = 0; j < result.length; j++) {
		// System.out.println("dc:result[" + j + "]: '" + result[j] + "'");
		// }
		// System.out.println();
		//
		// DEBUG OUTPUPT!

		return result;
	}

	/***************************************************************************
	 * Adds a digiprov metadata section to the Uof.
	 * 
	 * @param digiprovInfo - The lmerProcess data to add.
	 * @param digiprovMdId - The digiprov metadata ID of the lkmerProcessData to
	 *            set.
	 **************************************************************************/
	private void addDigiprovSec(LmerProcess digiprovInfo, String digiprovMdId) {
		// Add new digiprov md section to the amd sec.
		AmdSecType[] amdSecs = this.uof.getAmdSecArray();
		MdSecType digiprovMd = amdSecs[0].addNewDigiprovMD();
		digiprovMd.setID(digiprovMdId);

		// Create a new mdwrap section.
		MdWrap digiprovMdWrap = digiprovMd.addNewMdWrap();
		digiprovMdWrap.setID(digiprovMdId + prefixSeparator + mdWrapPostfix);
		digiprovMdWrap.setMIMETYPE("text/xml");
		digiprovMdWrap.setMDTYPE(MdWrap.MDTYPE.OTHER);
		digiprovMdWrap.setOTHERMDTYPE("lmerProcess");
		// digiprovMdWrap.setLABEL("lmerProcess");

		// Add the provenance metadata into the xmldata.
		digiprovMdWrap.addNewXmlData();
		digiprovMdWrap.getXmlData().set(digiprovInfo);
	}

	/***************************************************************************
	 * Updates the Metadata with correct creation dates.
	 * 
	 * @throws XmlException
	 * @throws DataFormatException
	 **************************************************************************/
	private void updateUof() throws XmlException, DataFormatException {
		this.uof.getMetsHdr().setCREATEDATE(Calendar.getInstance());
		LmerObject lObj = getLmerObjectCopy();
		lObj.setMasterCreationDate(Calendar.getInstance());
		lObj.setMetadataCreationDate(Calendar.getInstance());
		setLmerObject(lObj);
	}

	/***************************************************************************
	 * Returns an lmerObject with the updated XOR file checksums. This
	 * calculation of the object checksum uses the already calculated checksum
	 * in the lmerFile sections.
	 * 
	 * @param lObj The lmerObject for which the checksum has to be calculated.
	 * @return LmerObject - The updated lmerObject.
	 * @throws DataFormatException If something goes wrong while calculatin the
	 *             checksums.
	 **************************************************************************/
	private LmerObject recalculateObjectChecksum(LmerObject lObj)
			throws DataFormatException {
		TransferChecksum xorFileChecksum = null;

		if (lObj.sizeOfTransferChecksumArray() == 0) {
			xorFileChecksum = lObj.addNewTransferChecksum();
		}
		else {
			xorFileChecksum = TransferChecksum.Factory.newInstance();
		}

		xorFileChecksum
				.setCHECKSUMTYPE(ChecksumGenerator.XOR_OF_FILE_CHECKSUMS);

		FileGrpType.File fileEntries[] = this.uof.getFileSec().getFileGrpArray(
				0).getFileArray();

		// The first transfer checksum is the checksum of the first existing
		// file.
		if (fileEntries.length > 0) {
			xorFileChecksum.setStringValue(fileEntries[0].getCHECKSUM());
		}

		// Later transfer checksum are xored with the checksums of files.
		for (int i = 1; i < fileEntries.length; i++) {
			xorFileChecksum.setStringValue(ChecksumGenerator.xorHexStrings(
					xorFileChecksum.getStringValue(), fileEntries[i]
							.getCHECKSUM()));
		}

		lObj.setTransferChecksumArray(0, xorFileChecksum);
		return lObj;
	}

	/***************************************************************************
	 * Creates and references a new DMD sec and DMD wrap.
	 * 
	 * @return MdWrap - The new DMD wrap of the new DMD sec.
	 **************************************************************************/
	private MdWrap getNewDmdWrap() {
		String dmdId = dmdPrefix + prefixSeparator + this.dmdCounter;
		MdSecType dmdSec = this.uof.addNewDmdSec();
		dmdSec.setID(dmdId);

		// Adds the id of the dmd sec in the div sec.
		List<String> dmdList = new ArrayList<String>();
		dmdList.add(dmdId);
		DivType assetDiv = getFirstAssetDiv();
		if (assetDiv.getDMDID() != null) {
			// TODO: Make schemas.jar Java 1.5 compliant
			dmdList.addAll(assetDiv.getDMDID());
		}
		assetDiv.setDMDID(dmdList);

		this.dmdCounter++;
		return dmdSec.addNewMdWrap();
	}

	/***************************************************************************
	 * Returns the XmlData-block in the amdSec.techMD.mdWrap.XmlData hierarchy
	 * for the given techMD-id.
	 * 
	 * @param id
	 * @return XmlData
	 * @throws XmlException
	 **************************************************************************/
	private MdSecType.MdWrap.XmlData getTechMdXmlData(String id)
			throws XmlException {
		AmdSecType amdSecs[] = this.uof.getAmdSecArray();

		for (int i = 0; i < amdSecs.length; i++) {
			MdSecType mdSecs[] = amdSecs[i].getTechMDArray();

			for (int j = 0; j < mdSecs.length; j++) {
				if (mdSecs[j].getID().equals(id)) {
					mdSecs[j].getMdWrap().getXmlData();
					return mdSecs[j].getMdWrap().getXmlData();
				}
			}
		}

		throw new XmlException("mdWrap xmlData block with ID " + id + " "
				+ "not found");
	}

	/***************************************************************************
	 * Returns the first div section of type asset or null if none exist.
	 * 
	 * @return DivType
	 **************************************************************************/
	private DivType getFirstAssetDiv() {
		StructMapType[] sm = this.uof.getStructMapArray();

		for (int i = 0; i < sm.length; i++) {
			if (sm[i].getDiv().getTYPE().equalsIgnoreCase("ASSET")) {
				return sm[i].getDiv();
			}
		}

		return null;
	}

	/***************************************************************************
	 * Returns for a specified content file the file id in the file section.
	 * This works only if contentFile has the same absolute path as basePathUri +
	 * fileSecEntry.
	 * 
	 * @param contentFile
	 * @return String
	 * @throws FileNotFoundException
	 **************************************************************************/
	public String getFileId(File contentFile) throws FileNotFoundException {
		FileGrpType.File files[] = this.uof.getFileSec().getFileGrpArray(0)
				.getFileArray();

		for (int i = 0; i < files.length; i++) {
			String absoluteFileSecPath = this.basePathUri
					+ files[i].getFLocatArray(0).getHref().substring(
							FILE_LOCAT_HREF_PREFIX.length());

			if (contentFile.toURI().toString().equals(absoluteFileSecPath)) {
				return files[i].getID();
			}
		}

		throw new FileNotFoundException("File not found in UOF");
	}

	/***************************************************************************
	 * Returns for the object the referenced admId for the technical metadata in
	 * the file section (techMd).
	 * 
	 * @return String - The object's ID which references the techMd section if
	 *         set, empty string otherwise.
	 **************************************************************************/
	public String getObjectTechMdId() {
		String result = "";

		// Get the list of ADMIDs from the file group.
		List<String> admIdList = this.uof.getFileSec().getFileGrpArray(0)
				.getADMID();

		// Return the TECHMD identifier (the LAST identifier in the list of
		// IDs!!), if existing.
		if (admIdList.size() > 0) {
			result = admIdList.get(admIdList.size() - 1);
		}

		return result;
	}

	/***************************************************************************
	 * Returns for a specified content file the referenced admId (in UOF it's
	 * called techMdId) in the file section. This works only if contentFile has
	 * the same absolute path as basePathUri + fileSecEntry.
	 * 
	 * @param contentFile
	 * @return String
	 * @throws FileNotFoundException
	 **************************************************************************/
	public String getFileTechMdId(File contentFile)
			throws FileNotFoundException {
		FileGrpType.File files[] = this.uof.getFileSec().getFileGrpArray(0)
				.getFileArray();

		// Go throught all files.
		for (int i = 0; i < files.length; i++) {
			String absoluteFileSecPath = this.basePathUri
					+ files[i].getFLocatArray(0).getHref().substring(
							FILE_LOCAT_HREF_PREFIX.length());

			// It the two pathes are equal, return the TECHMD identifier (the
			// LAST identifier in the list of IDs!!).
			if (contentFile.toURI().toString().equals(absoluteFileSecPath)) {
				return (String) files[i].getADMID().get(
						files[i].getADMID().size() - 1);
			}
		}

		throw new FileNotFoundException("File not found in UOF");
	}

	/***************************************************************************
	 * Returns for a specified content file the file metadata from the file
	 * section. This works only if contentFile has the same absolute path as
	 * basePathUri + fileSecEntry.
	 * 
	 * @param contentFile
	 * @return FileGrpType.File
	 * @throws Exception
	 **************************************************************************/
	public FileGrpType.File getFileMdMets(File contentFile) throws Exception {
		FileGrpType.File files[] = this.uof.getFileSec().getFileGrpArray(0)
				.getFileArray();

		for (int i = 0; i < files.length; i++) {
			String absoluteFileSecPath = this.basePathUri
					+ files[i].getFLocatArray(0).getHref().substring(
							FILE_LOCAT_HREF_PREFIX.length());
			if (contentFile.toURI().toString().equals(absoluteFileSecPath)) {
				return files[i];
			}
		}

		throw new FileNotFoundException("File not found in UOF");
	}

	/***************************************************************************
	 * Returns for a specified content file the referenced digiprovId from the
	 * file section. This works only if contentFile has the same absolute path
	 * as basePathUri + fileSecEntry.
	 * 
	 * @param contentFile
	 * @return String
	 * @throws FileNotFoundException
	 **************************************************************************/
	public List<String> getFileProvMdIdList(File contentFile)
			throws FileNotFoundException {
		FileGrpType.File files[] = this.uof.getFileSec().getFileGrpArray(0)
				.getFileArray();

		// Go throught all files.
		for (int i = 0; i < files.length; i++) {
			String absoluteFileSecPath = this.basePathUri
					+ files[i].getFLocatArray(0).getHref().substring(
							FILE_LOCAT_HREF_PREFIX.length());

			// It the two pathes are equal, return the list of the DIGIPROVMD
			// identifier (all except the last).
			if (contentFile.toURI().toString().equals(absoluteFileSecPath)) {
				return files[i].getADMID().subList(0,
						files[i].getADMID().size() - 1);
			}
		}

		throw new FileNotFoundException("File not found in UOF");
	}

	/***************************************************************************
	 * Update the lmer File data. Used from both addFile() and updateFileData().
	 * 
	 * @param lmerFile
	 * @param jhoveInfo
	 * @return
	 **************************************************************************/
	private LmerFile updateLmerFileData(LmerFile lmerFile, Jhove jhoveInfo) {
		// Mandatory lmer file.
		String format = jhoveInfo.getRepInfoArray(0).getFormat();
		String version = jhoveInfo.getRepInfoArray(0).getVersion();
		if (format == null) {
			format = "Not specified!";
		}

		fRegistry.getFormat(format, version, lmerFile);

		// Optional lmer file data.
		//
		// lFile.setCategory(Category.BINARY);
		// lFile.setFormatInfos("testFormatInfo");
		// lFile.setCreatorApplication("");
		// lFile.setViewerApplication("");
		// lFile.setLinkedTo("testLinkedTo");
		// lFile.setComments("Comments2");
		//
		// These are already in mets so we don't use them.
		//
		// lFile.setFileIdentifier("");
		// lFile.setPath("");
		// lFile.setName("");
		// lFile.setSize(new BigInteger("0"));
		// lFile.setFileDateTime(Calendar.getInstance());
		// FileChecksum fileCheck = FileChecksum.Factory.newInstance();
		// fileCheck.setCHECKSUMTYPE("SHA-1");
		// fileCheck.setStringValue("qwertz");
		// lFile.setFileChecksum(fileCheck);
		// lFile.setMimeType("text/xml");

		// Add jhove metadata to lmer file.
		lmerFile.addNewXmlData();
		lmerFile.getXmlDataArray(0).setMDTYPE("JHOVE");
		XmlCursor xmlDataCursor = lmerFile.getXmlDataArray(0).newCursor();
		xmlDataCursor.toEndToken();
		jhoveInfo.newCursor().copyXml(xmlDataCursor);

		return lmerFile;
	}

	/***************************************************************************
	 * Cheack the techMd String for whitespaces and maximum length. Used from
	 * both addFile() and updateFileData().
	 * 
	 * @param techMd
	 * @return A JhoveInfo object.
	 * @throws Exception
	 **************************************************************************/
	private Jhove checkXmlData(String techMd) throws Exception {
		// Strip techMd from superfluous white spaces.
		XmlOptions noWhiteSpaces = new XmlOptions();

		if (this.noWhiteSpacesInTechMd) {
			noWhiteSpaces.setLoadStripWhitespace();
		}

		Jhove jhoveInfo = JhoveDocument.Factory.parse(techMd, noWhiteSpaces)
				.getJhove();

		if (jhoveInfo.xmlText().length() > this.maxTechMDLength) {
			throw new Exception("techMD length exceeds DIAS limit of "
					+ this.maxTechMDLength + " bytes");
		}

		return jhoveInfo;
	}

	/***************************************************************************
	 * Updates the file data of one file in the file group. Used from both
	 * addFile() and updateFileData().
	 * 
	 * @param file
	 * @return
	 **************************************************************************/
	private gov.loc.mets.FileGrpType.File updateFileSecFileData(
			File contentFile, gov.loc.mets.FileGrpType.File file,
			Jhove jhoveInfo) throws Exception {
		Calendar createDate = Calendar.getInstance();
		createDate.setTimeInMillis(contentFile.lastModified());

		file.setCREATED(createDate);
		file.setSIZE(contentFile.length());
		file.setCHECKSUM(ChecksumGenerator.getSHA1(new FileInputStream(
				contentFile)));
		file.setCHECKSUMTYPE(FileGrpType.File.CHECKSUMTYPE.SHA_1);
		file.setMIMETYPE(jhoveInfo.getRepInfoArray(0).getMimeType());
		file.getFLocatArray(0).setHref(
				FILE_LOCAT_HREF_PREFIX
						+ contentFile.toURI().toString().substring(
								this.basePathUri.toString().length()));

		return file;
	}

	/***************************************************************************
	 * Returns true if the file has migration data (a reference to an
	 * lmerProcess section from the file section), false otherwise.
	 * 
	 * @param contentFile
	 * @return boolean - true or false.
	 * @throws FileNotFoundException
	 **************************************************************************/
	public boolean hasMigrationFileData(File contentFile)
			throws FileNotFoundException {
		// Return true if the digiprovMdIdList is not empty for this file.
		if (!getFileProvMdIdList(contentFile).isEmpty()) {
			return true;
		}

		return false;
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * Sets the base path which is omitted in the file locat entry.
	 * 
	 * @param basePath
	 **************************************************************************/
	public void setBasePath(File basePath) {
		this.basePathUri = basePath.toURI();
	}

	/***************************************************************************
	 * @return The base path URI
	 **************************************************************************/
	public URI getBasePath() {
		return this.basePathUri;
	}

	// public String[] getFileNames() throws URISyntaxException{
	// FileGrpType.File files[] =
	// uof.getFileSec().getFileGrpArray(0).getFileArray();
	// String result[] = new [files.length];
	// for (int i = 0; i < files.length; i++) {
	// String fileNameInFileSec =
	// files[i].getFLocatArray(0).getHref()
	// .substring(FILE_LOCAT_HREF_PREFIX.length());
	// result[i] = new File (new URI(basePathUri + fileNameInFileSec));
	// }
	//        
	// return result;
	// }

	/***************************************************************************
	 * @param _dmdPrefix The dmdPrefix to set.
	 **************************************************************************/
	public static void setDmdPrefix(String _dmdPrefix) {
		Uof.dmdPrefix = _dmdPrefix;
	}

	/***************************************************************************
	 * @param _mdWrapPostfix The mdWrapPostfix to set.
	 **************************************************************************/
	public static void setMdWrapPostfix(String _mdWrapPostfix) {
		Uof.mdWrapPostfix = _mdWrapPostfix;
	}

	/***************************************************************************
	 * @param _techMdPrefix The techMdPrefix to set.
	 **************************************************************************/
	public static void setTechMdPrefix(String _techMdPrefix) {
		Uof.techMdPrefix = _techMdPrefix;
	}

	/***************************************************************************
	 * @return The noWhiteSpacesInTechMd
	 **************************************************************************/
	public boolean isNoWhiteSpacesInTechMd() {
		return this.noWhiteSpacesInTechMd;
	}

	/***************************************************************************
	 * @param _noWhiteSpacesInTechMd The noWhiteSpInTechMd to set.
	 **************************************************************************/
	public void setNoWhiteSpacesInTechMd(boolean _noWhiteSpacesInTechMd) {
		this.noWhiteSpacesInTechMd = _noWhiteSpacesInTechMd;
	}

	/***************************************************************************
	 * @return The maxTechMDLength
	 **************************************************************************/
	public long getMaxTechMDLength() {
		return this.maxTechMDLength;
	}

	/***************************************************************************
	 * @param _maxTechMDLength The noWhiteSpInTechMd to set.
	 **************************************************************************/
	public void setMaxTechMDLength(long _maxTechMDLength) {
		this.maxTechMDLength = _maxTechMDLength;
	}

	/***************************************************************************
	 * @param _diasDcDescriptionLimit
	 **************************************************************************/
	public static void setDiasDcDescriptionLimit(int _diasDcDescriptionLimit) {
		Uof.diasDcDescriptionLimit = _diasDcDescriptionLimit;
	}

	/***************************************************************************
	 * @param _diasDcIdentifierLimit
	 **************************************************************************/
	public static void setDiasDcIdentifierLimit(int _diasDcIdentifierLimit) {
		Uof.diasDcIdentifierLimit = _diasDcIdentifierLimit;
	}

	/***************************************************************************
	 * @param _diasDcTypeLimit
	 **************************************************************************/
	public static void setDiasDcTypeLimit(int _diasDcTypeLimit) {
		Uof.diasDcTypeLimit = _diasDcTypeLimit;
	}

}
