/*******************************************************************************
 * de.langzeitarchivierung.kopal.ingest / DiasIngest.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.ingest;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.digester.Digester;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPReply;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.jcraft.jsch.ChannelSftp;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.SftpException;

import de.langzeitarchivierung.kopal.util.FormatInfo;
import de.langzeitarchivierung.kopal.util.KopalXMLParser;

/*******************************************************************************
 * Class for ingesting SIPs to Dias according to the SIP Interface Specification
 * 2.7 (including Ticketing System).
 * 
 * TODO Use digester to return response beans?
 * 
 * TODO Improve error handling!
 * 
 * @author Ludwig
 * @since 20060227
 * @version 20070214
 ******************************************************************************/

public class DiasIngest
{

	// CONSTANTS **********************************************************

	private static final int	INPUT_BUFFER_LENGTH							= 1024 * 1024;

	// URLs.
	private static final String	ACCESS_MANAGER_COMMAND_URL					= "/AccessManager/AccessManager?cmd=";
	private static final String	INGEST_STATUS_COMMAND_URL					= "/IngestStatus/Ingest?";
	private static final String	SOURCEID_URL								= "&destId=amgr&sourceId=";
	private static final String	USER_URL									= "&CMUser=";
	private static final String	PASSWORD_URL								= "&CMPasswd=";
	private static final String	SIP_NAME_URL								= "&SipName=";
	private static final String	TICKET_ID_URL								= "&TicketId=";

	// Request types.
	private static final String	REQUEST_FILE_TYPES_COMMAND					= "filetypes";
	private static final String	REQUEST_STORAGE_RULES_COMMAND				= "storagerules";
	private static final String	REQUEST_SIP_TRANSPORT_PARAMETERS_COMMAND	= "FTPInfo";

	// STATE (Instance variables) *****************************************

	private String				keyStoreFile								= "";
	private String				sourceId									= "";
	private String				cmUser										= "";
	private String				cmPwd										= "";
	private String				knownHostsFile								= "";
	private String				ingestServer								= null;

	private int					ingestPort									= 80;

	private boolean				strictHostKeyChecking						= true;
	private boolean				useHttps									= true;

	// CLASS VARIABLES (Static variables) ***********************************

	private static Logger		defaultLogger								= Logger
																					.getLogger("de.langzeitarchivierung.kopal");

	// Resources.
	private HttpClient			diasClient;

	// FTP transport parameters as returned from DIAS.
	private static String		ftpUser										= "";
	private static String		sipPreloadarea								= "";
	private static String		ftpServer									= "";
	private static String		ftpPassword									= "";

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Default Constructor.
	 **************************************************************************/
	public DiasIngest() {
		super();
		this.diasClient = new HttpClient();
	}

	/***************************************************************************
	 * Constructor with the server to connect.
	 * 
	 * @param serverReqAddress host name or ip address
	 * @param serverReqPort the port number
	 **************************************************************************/
	public DiasIngest(String serverReqAddress, int serverReqPort) {
		super();
		this.ingestServer = serverReqAddress;
		this.ingestPort = serverReqPort;
		this.diasClient = new HttpClient();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * Configures the command line options. Using Jakarata Commons Cli.
	 * 
	 * @return the command line options object.
	 **************************************************************************/
	public static Options cliOptions() {
		Options cliOptions = new Options();

		// Short, long, hasArg, desc (request modifier options).
		cliOptions.addOption("u", "user", true, "The submitting CMUser");
		cliOptions.addOption("p", "password", true, "The CMPassword of the "
				+ "CMUser");
		cliOptions.addOption("a", "address", true, "The server address");
		cliOptions.addOption("n", "port", true, "The server port");
		cliOptions.addOption("f", "file", true, "The file to submit");
		cliOptions.addOption("t", "testdias", false, "Print dias responses");
		cliOptions.addOption("hp", "show-properties", false, "Print the "
				+ "system properties and continue");
		cliOptions.addOption("h", "help", false, "Print this dialog and exit");

		return cliOptions;
	}

	/***************************************************************************
	 * Main method for testing purposes.
	 * 
	 * @param args the command line.
	 **************************************************************************/
	public static void main(String[] args) {
		// Parse the command line options.
		CommandLineParser cliParser = new PosixParser();
		Options cliOptions = cliOptions();
		CommandLine cLine = null;

		try {
			cLine = cliParser.parse(cliOptions, args);
		}
		catch (ParseException e) {
			e.printStackTrace();
			System.err.println("Commandline and its arguments are invalid!");
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("DiasIngest", cliOptions);
			System.exit(1);
		}

		// Print system properties and continue.
		if (cLine.hasOption("hp")) {
			Properties p = System.getProperties();
			for (Enumeration e = p.propertyNames(); e.hasMoreElements();) {
				Object po = e.nextElement();
				System.out.println("Name:" + po + " Value:"
						+ p.getProperty((String) po));
			}
		}

		// Print help dialog.
		if (cLine.hasOption("h")) {
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("DiasIngest", cliOptions);
			System.exit(0);
		}

		// Get host name, port, user name and password from command line and set
		// them.
		DiasIngest ingest = new DiasIngest(cLine.getOptionValue("a"), Integer
				.parseInt(cLine.getOptionValue("n")));
		ingest.setCmUser(cLine.getOptionValue("u"));
		ingest.setCmPwd(cLine.getOptionValue("p"));

		// Print dias responses and exit.
		if (cLine.hasOption("t")) {
			ingest.testDias();
			System.exit(0);
		}

		// Get the file name to upload.
		File asset = new File(cLine.getOptionValue("f"));

		// If it isn't the absolute path try relative.
		if (!asset.exists()) {
			asset = new File(System.getProperty("user.dir"), cLine
					.getOptionValue("f"));
		}
		if (!asset.exists()) {
			System.err.println("File " + cLine.getOptionValue("f")
					+ " does not exist");
		}

		long startTime = System.currentTimeMillis();

		try {
			ingest.initializeFtpParameters();
			// Upload in home dir.
			// ingest.submitFileToDias(asset, null);

			// Upload in DIAS defined dir.
			ingest.submitFileToDias(asset, DiasIngest.sipPreloadarea);
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}

		System.out // Well, this ugly line wrapping is Eclipse's fault
				.println((asset.length() / 1024)
						+ " KB transmitted to Dias in "
						+ ((System.currentTimeMillis() - startTime) / 1000)
						+ " secs = "
						+ ((asset.length() / 1024) / ((System
								.currentTimeMillis() - startTime) / 1000))
						+ " KB/sec");

		System.exit(0);
	}

	/***************************************************************************
	 * Retrieves the list of supported file types from dias and returns it as
	 * xml.
	 * 
	 * @return A NodeList with nodes as specified in the LmerSIP Interface
	 *         Specification IFI_SIP002A
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	public NodeList requestFileTypesXml() throws Exception {
		String diasResponse = requestFileTypes();
		Document fileTypesDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);
		return fileTypesDocument.getElementsByTagName("Item");
	}

	/***************************************************************************
	 * Requests from DIAS the supported file types
	 * 
	 * @return XML as String with the supported fileTypes
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	public String requestFileTypes() throws Exception {
		return retrieveDataFromDIAS(ACCESS_MANAGER_COMMAND_URL
				+ REQUEST_FILE_TYPES_COMMAND + SOURCEID_URL + this.sourceId);
	}

	/***************************************************************************
	 * Retrieves the list of supported storage rules from dias and returns it as
	 * xml.
	 * 
	 * @return A NodeList with nodes as specified in the LmerSIP Interface
	 *         Specification IFI_SIP011A
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	public NodeList requestStorageRulesXml() throws Exception {
		String diasResponse = requestStorageRules();
		Document fileTypesDocument = KopalXMLParser
				.getDocumentFromString(diasResponse);
		return fileTypesDocument.getElementsByTagName("Item");
	}

	/***************************************************************************
	 * Requests the list of supported storage rules from dias.
	 * 
	 * @return XML as string with the supported storage rules
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	private String requestStorageRules() throws Exception {
		return retrieveDataFromDIAS(ACCESS_MANAGER_COMMAND_URL
				+ REQUEST_STORAGE_RULES_COMMAND + SOURCEID_URL + this.sourceId);
	}

	/***************************************************************************
	 * Prints the dias responses for requests of supported filetypes,
	 * storagerules, transport parameters and ingest status.
	 **************************************************************************/
	public void testDias() {
		try {
			System.out.println("FILETYPES:");
			System.out.println(requestFileTypes());
			System.out.println();
			System.out.println("STORAGERULES:");
			System.out.println(requestStorageRules());
			System.out.println();
			System.out.println("SIPTRANSPORTPARAMETERS:");
			System.out.println(requestSipTransportParameters());
			System.out.println();
			System.out.println("REQUESTINGEST:");
			System.out.println(requestSipIngest("UHU.ZIP"));
			System.out.println();
			System.out.println("REQUESTINGESTSTATUS:");
			System.out.println(requestSipIngestStatus("THE_TEST_ID"));
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/***************************************************************************
	 * Initializes/updates variables with ftp transport parameters from DIAS if
	 * we haven't initialized them already.
	 * 
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	public void initializeFtpParameters() throws Exception {
		// If we have no FTP parameters.
		if (ftpServer.equals("")) {
			forceInitializeFtpParameters();
		}
	}

	/***************************************************************************
	 * Requests the ftp transport parameters from dias.
	 * 
	 * @return XML as string with the ftp transport parameters
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	public String requestSipTransportParameters() throws Exception {
		return retrieveDataFromDIAS(ACCESS_MANAGER_COMMAND_URL
				+ REQUEST_SIP_TRANSPORT_PARAMETERS_COMMAND + SOURCEID_URL
				+ this.sourceId + USER_URL + this.cmUser + PASSWORD_URL
				+ this.cmPwd);
	}

	/***************************************************************************
	 * Retrieves the list of ftp transport parameters from dias and returns it
	 * as xml.
	 * 
	 * @return A NodeList with nodes as specified in the LmerSIP Interface
	 *         Specification IFI_SIP021A
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	public NodeList requestSipTransportParametersXml() throws Exception {
		Document fileTypesDocument;

		String diasResponse = requestSipTransportParameters();

		if (!diasResponse.equals("")) {
			fileTypesDocument = KopalXMLParser
					.getDocumentFromString(diasResponse);
		}
		else {
			throw new IOException("XML response from DIAS is empty,"
					+ " please check the request parameters");
		}

		return fileTypesDocument.getElementsByTagName("FTPInfoItem");
	}

	/***************************************************************************
	 * Transfers asset to an ftp server specified with address, port, user and
	 * password in config file and not as returned from dias
	 * 
	 * @param serverDir The directory where to save the asset. If null the
	 *            sipPreloadarea as set with initializeFtpParameters() is used.
	 *            If initializeFtpParameters() hasn't been called the current
	 *            directory (usually the home dir) is used.
	 * @param asset
	 * @throws IOException
	 **************************************************************************/
	public void submitFileToFtp(File asset, String serverDir)
			throws IOException {

		FTPClient ftp = new FTPClient();
		FileInputStream fis = null;

		try {

			// System.out.println("Connecting to : " + this.ingestServer
			// + this.ingestPort);

			ftp.connect(this.ingestServer, this.ingestPort);

			// System.out.println("Successful connected to : "
			// + this.ingestServer + this.ingestPort);

			// TODO Is this if statement necessary? the finally block below
			// disconnects too...
			if (!FTPReply.isPositiveCompletion(ftp.getReplyCode())) {
				ftp.disconnect();
				throw new IOException("Server refused FTP connection. "
						+ "System message: " + ftp.getReplyCode());
			}

			// Try login, if it fails try again with new FTP parameters.
			if (!ftp.login(this.cmUser, this.cmPwd)) {
				throw new IOException("Server refused FTP user and password. "
						+ "System message: " + ftp.getReplyCode());
			}

			ftp.enterLocalPassiveMode();
			ftp.setFileType(org.apache.commons.net.ftp.FTP.BINARY_FILE_TYPE);

			if (serverDir != null) {
				// Try changing to work dir.
				if (!ftp.changeWorkingDirectory(serverDir)) {
					throw new IOException("Changing to directory " + serverDir
							+ "failed. System message: " + ftp.getReplyCode());
				}
			}
			else if (!sipPreloadarea.equals("")) {
				// Try changing to work dir.
				if (!ftp.changeWorkingDirectory(sipPreloadarea)) {
					throw new IOException("Changing to directory " + serverDir
							+ "failed. System message: " + ftp.getReplyCode());
				}
			}

			fis = new FileInputStream(asset);
			String filename = asset.getName() + ".tmp";
			if (!ftp.storeFile(filename, fis)) {
				throw new IOException("Uploading file failed: "
						+ ftp.getReplyCode());
			}

			if (!ftp.rename(filename, asset.getName())) {
				throw new IOException("Renaming asset failed! "
						+ ftp.getReplyCode());
			}

			ftp.logout();

		}
		finally {
			if (ftp.isConnected()) {
				ftp.disconnect();
			}
			if (fis != null) {
				fis.close();
			}
		}
	}

	/***************************************************************************
	 * Returns a ticket number from DIAS for a given SIP ingest name as XML.
	 * 
	 * @return
	 * @throws Exception
	 **************************************************************************/
	public NodeList requestSipIngestXml(String _sipName) throws Exception {
		Document sipTicketNumberDocument;

		String diasResponse = requestSipIngest(_sipName);

		if (!diasResponse.equals("")) {
			sipTicketNumberDocument = KopalXMLParser
					.getDocumentFromString(diasResponse);
		}
		else {
			throw new Exception("XML response from DIAS is empty,"
					+ " please check the request parameters");
		}

		return sipTicketNumberDocument.getElementsByTagName("METADATABLOCK");
	}


	/***************************************************************************
	 * Returns a ticket number from DIAS for a given SIP ingest name.
	 * 
	 * @return
	 * @throws Exception
	 **************************************************************************/
	public String requestSipIngest(String _sipName) throws Exception {
		return retrieveDataFromDIAS(INGEST_STATUS_COMMAND_URL + USER_URL
				+ this.cmUser + PASSWORD_URL + this.cmPwd + SIP_NAME_URL
				+ _sipName);
	}

	/***************************************************************************
	 * Returns a ticket number from DIAS for a given SIP ingest name as XML.
	 * 
	 * @return
	 * @throws Exception
	 **************************************************************************/
	public NodeList requestSipIngestStatusXml(String _theTicket)
			throws Exception {
		Document sipIngestStatusDocument;

		String diasResponse = requestSipIngestStatus(_theTicket);

		if (!diasResponse.equals("")) {
			sipIngestStatusDocument = KopalXMLParser
					.getDocumentFromString(diasResponse);
		}
		else {
			throw new Exception("XML response from DIAS is empty,"
					+ " please check your request parameters");
		}

		return sipIngestStatusDocument.getElementsByTagName("METADATABLOCK");
	}

	/***************************************************************************
	 * Returns an ingest status from DIAS for a given SIP ingest name.
	 * 
	 * @return
	 * @throws Exception
	 **************************************************************************/
	public String requestSipIngestStatus(String _theTicket) throws Exception {
		return retrieveDataFromDIAS(INGEST_STATUS_COMMAND_URL + TICKET_ID_URL
				+ _theTicket);
	}

	/***************************************************************************
	 * Parses the ingest response of the DIAS. Uses parseDiasResponse(Reader
	 * diasResponse)
	 * 
	 * @param diasResponse
	 * @return a DiasIngestResponse object which contains the values of the
	 *         response.
	 * @throws Exception
	 **************************************************************************/
	public static DiasIngestResponse parseDiasResponse(String diasResponse)
			throws Exception {
		return DiasIngest.parseDiasResponse(new StringReader(diasResponse));
	}

	/***************************************************************************
	 * Parses the ingest response of the DIAS. Uses Jakarta Commons digester.
	 * 
	 * @param diasResponse A DIAS response as a NodeList taken from a
	 *            requestSomethingXml() method.
	 * @return A DiasIngestResponse object filled with useful information.
	 * @throws Exception
	 **************************************************************************/
	public static DiasIngestResponse parseDiasResponse(Reader diasResponse)
			throws Exception {
		Digester responseParser = new Digester();

		// TODO Strange bugs are the reason for setting the class loader!
		// CHECK IF WE CAN DO WITHOUT IT!
		responseParser.setClassLoader(ClassLoader.getSystemClassLoader());
		// Validating requires inline DTD in DIAS response.
		responseParser.setValidating(true);

		// Create a DiasIngestResponse object for the DIAS response and map the
		// properties.
		responseParser.addObjectCreate("MetadataBlock",
				DiasIngestResponse.class);
		responseParser
				.addSetProperties("MetadataBlock", "type", "responseType");
		responseParser.addSetProperties("MetadataBlock", "version",
				"responseVersion");
		// Note: It is necessary to repeat the property in lower case.
		responseParser.addBeanPropertySetter("MetadataBlock/Description",
				"description");

		// Create a DiasIngestMetadataResponse object for the corresponding part
		// in the response and map the properties.
		responseParser.addObjectCreate("MetadataBlock/Metadata",
				"de.langzeitarchivierung.kopal.ingest."
						+ "DiasIngestMetadataResponse");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "SIP_Filename", "sipFilename");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "SIPIngestTicketNumber", "sipIngestTicketNumber");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "FTP_Password", "ftpPassword");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "SIP.preloadarea", "sipPreloadarea");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "FTP_Server", "ftpServer");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "FTP_User", "ftpUser");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "StartTimestamp", "startTimestamp");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "SourceID", "sourceId");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "State", "state");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "Substate", "substate");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "SupplementaryInformation", "supplementaryInformation");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "Timestamp", "timestamp");
		responseParser.addBeanPropertySetter("MetadataBlock/Metadata/"
				+ "Alert", "alert");

		// Advance to next block.
		responseParser.addSetNext("MetadataBlock/Metadata", "addMetadata");

		try {
			return (DiasIngestResponse) responseParser.parse(diasResponse);
		}
		catch (Exception e) {
			e.printStackTrace();
			throw new Exception("Unable to parse DIAS response. "
					+ "Response corrupt");
		}
	}

	/***************************************************************************
	 * Parses all attribute information out of a DIAS format response and sets
	 * all affected attributes in the returned DiasIngestResponse object using
	 * the apache.commons.digester.
	 * 
	 * @author Funk, SUB Goettingen
	 * @param formatResponse
	 * @return
	 * @throws SAXException
	 * @throws IOException
	 **************************************************************************/
	public static ArrayList<FormatInfo> parseDiasFormatResponse(
			String formatResponse) throws SAXException, IOException {
		ArrayList<FormatInfo> result = new ArrayList<FormatInfo>();

		// Use the commons.digester for XML parsing.
		Digester formatParser = new Digester();
		formatParser.setValidating(true);
		formatParser.addObjectCreate("FILETYPES", ArrayList.class);
		formatParser.addObjectCreate("FILETYPES/Item", FormatInfo.class);
		formatParser.addSetProperties("FILETYPES/Item");
		formatParser.addSetNext("FILETYPES/Item", "add");

		StringReader sr = new StringReader(formatResponse);
		result = (ArrayList<FormatInfo>) formatParser.parse(sr);

		return result;
	}

	/***************************************************************************
	 * Tests if the specified file exists in the specified directory. Useful for
	 * checking the error area.
	 * 
	 * @param asset the filename to check
	 * @param serverDir the directory in which to look for the file.
	 * @return true if the file exists, otherwise false
	 * @throws Exception
	 **************************************************************************/
	public boolean existsFile(File asset, String serverDir) throws Exception {
		ChannelSftp sftpChannel = null;
		Session sftpSession = null;

		try {
			JSch jsch = new JSch();
			jsch.setKnownHosts(new FileInputStream(this.knownHostsFile));

			sftpSession = jsch.getSession(ftpUser, ftpServer);
			sftpSession.setPassword(ftpPassword);

			if (!this.strictHostKeyChecking) {
				// This is insecure!
				Hashtable<String, String> config = new Hashtable<String, String>();
				config.put("StrictHostKeyChecking", "no");
				sftpSession.setConfig(config);
			}

			sftpSession.connect();

			sftpChannel = (ChannelSftp) sftpSession.openChannel("sftp");
			sftpChannel.connect();

			sftpChannel.cd(serverDir);

			sftpChannel.get(asset.getName());
		}
		catch (SftpException e) {
			// TODO Check if this the correct error code. It is unknown because
			// Jsch has no code documentation :-(
			if (e.id == ChannelSftp.SSH_FX_NO_SUCH_FILE) {
				return false;
			}
			// Log the Exception and throw it again!
			e.printStackTrace();
			throw e;
		}
		finally {
			if (sftpChannel != null && sftpChannel.isConnected()) {
				sftpChannel.disconnect();
			}
			if (sftpSession != null && sftpSession.isConnected()) {
				sftpSession.disconnect();
			}
		}
		return true;
	}

	/***************************************************************************
	 * Tests if the specified file name exists in the specified directory.
	 * Useful for checking the error area.
	 * 
	 * @param asset the filename to check
	 * @param serverDir the directory in which to look for the file.
	 * @return true if the file exists, otherwise false
	 * @throws Exception
	 * @author Kadir Karaca Koçer, German National Library
	 **************************************************************************/
	public boolean existsFile(String asset, String serverDir) throws Exception {
		JSch jsch = new JSch();
		Session sftpSession = null;
		ChannelSftp sftpChannel = null;
		java.util.Vector fileList = null;
		boolean exists = false;

		try {
			jsch.setKnownHosts(new FileInputStream(this.knownHostsFile));

			sftpSession = jsch.getSession(ftpUser, ftpServer);
			sftpSession.setPassword(ftpPassword);
			if (!this.strictHostKeyChecking) {
				// This is insecure!
				Hashtable<String, String> config = new Hashtable<String, String>();
				config.put("StrictHostKeyChecking", "no");
				sftpSession.setConfig(config);
			}
			sftpSession.connect();

			if (sftpSession != null) {
				sftpChannel = (ChannelSftp) sftpSession.openChannel("sftp");
				sftpChannel.connect();
				fileList = sftpChannel.ls(serverDir);
				com.jcraft.jsch.ChannelSftp.LsEntry le = null;
				for (int i = 0; i < fileList.size(); i++) {
					le = (com.jcraft.jsch.ChannelSftp.LsEntry) fileList.get(i);
					if (asset.equalsIgnoreCase(le.getFilename())) {
						exists = true;
						break;
					}
				}
				/*
				 * while (fileList.iterator().hasNext()) { le =
				 * (com.jcraft.jsch.ChannelSftp.LsEntry) fileList
				 * .iterator().next(); if
				 * (asset.equalsIgnoreCase(le.getFilename())) { exists = true;
				 * break; } }
				 */
			}
		}
		catch (Exception e) {
			// Log the Exception and throw it again!
			e.printStackTrace();
			throw e;
		}
		finally {
			// close() or exit() ???
			if (sftpChannel != null && sftpChannel.isConnected())
				sftpChannel.exit();
			if (sftpSession != null && sftpSession.isConnected())
				sftpSession.disconnect();
		}

		return exists;
	}

	/***************************************************************************
	 * Transfers a file to Dias via sftp.
	 * 
	 * @param asset The File to transfer.
	 * @param serverDir The directory where to save the asset. If null the
	 *            sipPreloadarea as set with initializeFtpParameters() is used.
	 *            If null and initializeFtpParameters() hasn't been called the
	 *            current directory (usually the home dir) is used.
	 * @throws Exception thrown if transfer to dias failed
	 **************************************************************************/
	public void submitFileToDias(File asset, String serverDir) throws Exception {
		ChannelSftp sftpChannel = null;
		Session sftpSession = null;
		BufferedInputStream fis = null;

		try {
			JSch jsch = new JSch();

			jsch.setKnownHosts(new FileInputStream(this.knownHostsFile));

			sftpSession = jsch.getSession(ftpUser, ftpServer);
			sftpSession.setPassword(ftpPassword);
			if (!this.strictHostKeyChecking) {
				// This is insecure!
				Hashtable<String, String> config = new Hashtable<String, String>();
				config.put("StrictHostKeyChecking", "no");
				sftpSession.setConfig(config);
			}
			sftpSession.connect();

			sftpChannel = (ChannelSftp) sftpSession.openChannel("sftp");
			sftpChannel.connect();

			if (serverDir != null) {
				sftpChannel.cd(serverDir);
			}
			else if (!sipPreloadarea.equals("")) {
				sftpChannel.cd(sipPreloadarea);
			}

			fis = new BufferedInputStream(new FileInputStream(asset));
			String filename = asset.getName() + ".tmp";
			sftpChannel.put(fis, filename);
			sftpChannel.rename(asset.getName() + ".tmp", asset.getName());
		}
		catch (Exception e) {
			throw e;
		}
		finally {
			if (sftpChannel != null && sftpChannel.isConnected()) {
				sftpChannel.disconnect();
			}
			if (sftpSession != null && sftpSession.isConnected()) {
				sftpSession.disconnect();
			}
			if (fis != null) {
				fis.close();
			}
		}
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ****

	/***************************************************************************
	 * Initializes/updates variables with ftp transport parameters from DIAS.
	 * 
	 * TODO check if this method should be synchronized. Or doesn't it matter if
	 * the values are initialized two times?
	 * 
	 * @throws Exception thrown if connection to dias failed
	 **************************************************************************/
	private void forceInitializeFtpParameters() throws Exception {
		NodeList ftpParameters = requestSipTransportParametersXml();

		// Dias returns only one FTP parameter set although more sets are
		// possible according to the specification.
		Element ftpParameter = (Element) ftpParameters.item(0);

		ftpUser = ftpParameter.getAttribute("FTP_User");
		sipPreloadarea = ftpParameter.getAttribute("SIP.Preloadarea");
		ftpServer = ftpParameter.getAttribute("FTP_Server");
		ftpPassword = ftpParameter.getAttribute("FTP_Password");
	}

	/***************************************************************************
	 * Gets the data of the specified url from the server. If setUseHttps(false)
	 * has been called, http is used, else (or if setUseHttps(true) has been
	 * called) https is used.
	 * 
	 * @param url The path of the data to retrieve (without serveraddress and
	 *            serverRequestPort but with leading slash)
	 * @return The retrieved data as a string
	 * @throws Exception thrown if connection failed
	 **************************************************************************/
	private String retrieveDataFromDIAS(String url) throws Exception {
		int start;
		int end;

		// Check the protocol.
		String protocol = "https";
		if (!this.useHttps) {
			protocol = "http";
		}

		// Assemble the complete URL.
		String completeUrl = protocol + "://" + this.ingestServer + ":"
				+ this.ingestPort + url;
		String urlToLog = completeUrl;

		// Test if some password information is included.
		if (completeUrl.indexOf(PASSWORD_URL) != -1) {
			start = completeUrl.indexOf(PASSWORD_URL) + PASSWORD_URL.length();
			if ((end = completeUrl.indexOf("&", start)) == -1) {
				end = completeUrl.length();
			}
			urlToLog = completeUrl.substring(0, start) + "*****"
					+ completeUrl.substring(end);
		}

		// Finally log the password clean URL.
		defaultLogger.log(Level.FINE, "DIAS request URL: " + urlToLog);

		// Start retrieving data from DIAS.
		GetMethod diasRequest = new GetMethod(completeUrl);

		try {
			this.diasClient.executeMethod(diasRequest);
			StringBuffer result = new StringBuffer();
			InputStream responseStream = new BufferedInputStream(diasRequest
					.getResponseBodyAsStream());
			byte[] buffer = new byte[INPUT_BUFFER_LENGTH];

			int read = responseStream.read(buffer);
			while (read > 0) {
				result.append(new String(buffer, 0, read));
				read = responseStream.read(buffer);
			}

			return result.toString();
		}
		catch (Exception e) {
			throw e;
		}
		finally {
			diasRequest.releaseConnection();
		}
	}

	// GET & SET METHODS *******************************************************

	/***************************************************************************
	 * Gets the password.
	 * 
	 * @return Returns the password.
	 **************************************************************************/
	public String getCmPwd() {
		return this.cmPwd;
	}

	/***************************************************************************
	 * Sets the password.
	 * 
	 * @param password The password to set.
	 **************************************************************************/
	public void setCmPwd(String password) {
		this.cmPwd = password;
	}

	/***************************************************************************
	 * Gets the sourceId.
	 * 
	 * @return Returns the sourceId.
	 **************************************************************************/
	public String getSourceId() {
		return this.sourceId;
	}

	/***************************************************************************
	 * Sets the sourceId.
	 * 
	 * @param _sourceId The sourceId to set.
	 **************************************************************************/
	public void setSourceId(String _sourceId) {
		this.sourceId = _sourceId;
	}

	/***************************************************************************
	 * Gets the user.
	 * 
	 * @return Returns the user.
	 **************************************************************************/
	public String getCmUser() {
		return this.cmUser;
	}

	/***************************************************************************
	 * Sets the user.
	 * 
	 * @param user The user to set.
	 **************************************************************************/
	public void setCmUser(String user) {
		this.cmUser = user;
	}

	/***************************************************************************
	 * Gets the server port.
	 * 
	 * @return Returns the serverRequestPort.
	 **************************************************************************/

	public int getIngestPort() {
		return this.ingestPort;
	}

	/***************************************************************************
	 * Sets the server port.
	 * 
	 * @param port The serverRequestPort to set.
	 **************************************************************************/
	public void setIngestPort(int port) {
		this.ingestPort = port;
	}

	/***************************************************************************
	 * Gets the server address.
	 * 
	 * @return Returns the serverRequestAddress.
	 **************************************************************************/
	public String getIngestServer() {
		return this.ingestServer;
	}

	/***************************************************************************
	 * Sets the server address.
	 * 
	 * @param serverAddress The serverRequestAddress to set.
	 **************************************************************************/
	public void setIngestServer(String serverAddress) {
		this.ingestServer = serverAddress;
	}

	/***************************************************************************
	 * @return Returns the useHttps.
	 **************************************************************************/

	public boolean isUseHttps() {
		return this.useHttps;
	}

	/***************************************************************************
	 * @param _useHttps The useHttps to set.
	 **************************************************************************/
	public void setUseHttps(boolean _useHttps) {
		this.useHttps = _useHttps;
	}

	/***************************************************************************
	 * @return Returns the knownHostsFile.
	 **************************************************************************/
	public String getKnownHostsFile() {
		return this.knownHostsFile;
	}

	/***************************************************************************
	 * @param _knownHostsFile The knownHostsFile to set.
	 **************************************************************************/
	public void setKnownHostsFile(String _knownHostsFile) {
		this.knownHostsFile = _knownHostsFile;
	}

	/***************************************************************************
	 * @param _strictHostKeyChecking The strictHostKeyChecking to set.
	 **************************************************************************/

	public void setStrictHostKeyChecking(boolean _strictHostKeyChecking) {
		this.strictHostKeyChecking = _strictHostKeyChecking;
	}

	/***************************************************************************
	 * @return the keyStoreFile
	 **************************************************************************/
	public String getKeyStoreFile() {
		return this.keyStoreFile;
	}

	/***************************************************************************
	 * @param _keyStoreFile The keyStoreFile to set.
	 **************************************************************************/
	public void setKeyStoreFile(String _keyStoreFile) {
		System.setProperty("javax.net.ssl.trustStore", _keyStoreFile);
		this.keyStoreFile = _keyStoreFile;
	}

}
