/**********************************************************************
 * de.langzeitarchivierung.kopal.ingest / IngestLayer.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.ingest;

import de.langzeitarchivierung.kopal.ws.wsdl2java.*;
import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.FileUtils;
import de.langzeitarchivierung.kopal.util.KopalXMLParser;
import de.langzeitarchivierung.kopal.util.WSUtils;

/*******************************************************************************
 * Class to provide web service access to DIAS-Core sytem. For more information
 * about how it works, and why, see the SIP Interface Specification provided by
 * IBM-Holland.
 * 
 * @author Kadir Karaca Koçer, German National Library
 * @version 20070126
 * @see DiasIngest
 * @since 20070126
 ******************************************************************************/

public class IngestLayer
{

	// CLASS VARIABLES (Static variables) **********************************

	// TODO: transfer those hard-coded strings to config file.
	private static String	ABSOLUTE_ROOT_DIR		= "c:/temp/";
	private static String	ABS_PATH_TO_FILE_LIST	= ABSOLUTE_ROOT_DIR
															+ "fileformatslist.xml";
	private static String	REL_PATH_TO_FILE_LIST	= "ftp://bernstein.d-nb.de/pub/kopal/fileformatslist.xml";

	private static boolean	isIngestActive			= false;

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * Returns the official list of the file formats that are supported by DIAS
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070606
	 **************************************************************************/
	public static ResponseElemDocument getFileFormatList(UserType user) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user,
				"getFileFormatList");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (!isIngestActive) {
			// Ingest is not running. Return an error message.
			response.getResponseElem().setStatusCode(560);
			response.getResponseElem().setResponseText(
					"Ingest is not running. Please consult the administrator.");
			return response;
		}

		if (new java.io.File(ABS_PATH_TO_FILE_LIST).exists()) {
			response.getResponseElem().setStatusCode(201);
			response.getResponseElem().setResponseText(
					"Operation successfull."
							+ " Please download the list from the given link.");
			response.getResponseElem().setLinkToFile(REL_PATH_TO_FILE_LIST);
		}
		else {
			response.getResponseElem().setStatusCode(501);
			response
					.getResponseElem()
					.setResponseText(
							"File format list file missing!"
									+ " Please call reloadFileFormatList to reload it from DIAS.");
		}
		return response;
	}

	/***************************************************************************
	 * Reloads from DIAS the official list of the file formats that are
	 * supported by DIAS
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070606
	 **************************************************************************/
	public static ResponseElemDocument reloadFileFormatList(UserType user) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user,
				"reloadFileFormatList");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (isIngestActive) {
			// Ingest is still running. It can be dangerous to change the list
			// during
			// other processes read it. Return an error message.
			response.getResponseElem().setStatusCode(561);
			response.getResponseElem().setResponseText(
					"Ingest is still running. Please block it first.");
			return response;
		}

		String xmlstring = "";

		java.io.File xmlfile = new java.io.File(ABS_PATH_TO_FILE_LIST);

		// Configure the DiasIngest object and get the file type response from
		// DIAS. If not successful, return the error
		try {
			// Read the config file and configure the connection to DIAS
			Configurator.setConfigDocument(KopalXMLParser
					.getDocument(WSUtils.CONFIG_FILE));
			// create a connection and retrieve the FormatRegistry from DIAS.
			DiasIngest dias = new DiasIngest();
			Configurator.configure(dias);
			xmlstring = dias.requestFileTypes();
			// Check if the answer valid
			DiasIngest.parseDiasFormatResponse(xmlstring);
			// save the file
			// TODO: solve the problem: two files with the same content becouse
			// of
			// file system access rights?
			FileUtils.storeFile(xmlfile, xmlstring);
			response.getResponseElem().setStatusCode(202);
			response
					.getResponseElem()
					.setResponseText(
							"The list of known file formats is succesfully downloaded and saved.");
		}
		catch (org.xml.sax.SAXException e) {
			response.getResponseElem().setStatusCode(562);
			response
					.getResponseElem()
					.setResponseText(
							"Parsing the DIAS format "
									+ "registry file failed. Please check the response");
			response.getResponseElem().setErrorMessage(
					"SAXException in IngestLayer.reloadFileFormatList()"
							+ e.getMessage());
		}
		catch (java.io.IOException e) {
			response.getResponseElem().setStatusCode(563);
			response.getResponseElem().setResponseText(
					"Request for DIAS format " + "registry file failed");
			response.getResponseElem().setErrorMessage(
					"IOException in IngestLayer.reloadFileFormatList() +"
							+ e.getMessage());
		}
		catch (Exception e) {
			response.getResponseElem().setStatusCode(564);
			response.getResponseElem().setResponseText(
					"Request for DIAS format " + "registry file failed");
			response.getResponseElem().setErrorMessage(
					"Exception in IngestLayer.reloadFileFormatList()"
							+ e.getMessage());
		}

		return response;
	}

	/***************************************************************************
	 * Builds a SIP package confirming SIP Interface Specification and ingests
	 * it into DIAS.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param policy Ingest policy for that digital object
	 * @param metadataArray One obligatory technical and five optional metadata
	 *            blocks desribing this digital object. At least the location of
	 *            the object
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070606
	 **************************************************************************/
	public static ResponseElemDocument ingestDocument(UserType user,
			String policy, CustomXMLType[] metadataArray) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "ingestSIP");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (!isIngestActive) {
			// Ingest is marked blocked by the administrator. Reject the ingest
			// request.
			response.getResponseElem().setStatusCode(560);
			response
					.getResponseElem()
					.setResponseText(
							"Ingest is blocked by the administrator. "
									+ "Please allow it using the allowIngestRequests service.");
			return response;
		}

		// TODO: parse the metadata array and activate ActionModules
		System.out.println("Ingesting the object with:\n Policy: " + policy
				+ "\nMetadata:\n" + metadataArray[0].xmlText());

		return response;
	}

	/***************************************************************************
	 * Ingests a already build SIP into DIAS.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param urn URN of this digital document. Its not known becouse the SIP is
	 *            ready build.
	 * @param linkToSIP URL to find the SIP -its most probably on an other
	 *            server.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070606
	 **************************************************************************/
	public static ResponseElemDocument ingestSIP(UserType user, String urn,
			String linkToSIP) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "ingestSIP");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (!isIngestActive) {
			// Ingest is marked blocked by the administrator. Reject the ingest
			// request.
			response.getResponseElem().setStatusCode(560);
			response
					.getResponseElem()
					.setResponseText(
							"Ingest is blocked by the administrator. "
									+ "Please allow it using the allowIngestRequests service.");
			return response;
		}

		// TODO: read the ready made SIP and copy it to DIAS load directory
		System.out.println(urn + "\n" + linkToSIP);

		return response;
	}

	/***************************************************************************
	 * Blocks the further coming ingest requests.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070607
	 * @since 20070126
	 **************************************************************************/
	public static ResponseElemDocument blockIngest(UserType user) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "blockIngest");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (isIngestActive) {
			// Stop Ingest
			isIngestActive = false;
			response.getResponseElem().setStatusCode(210);
			response
					.getResponseElem()
					.setResponseText(
							"Ingest blocked. Please check existing users before shut down.");
		}
		else {
			// Ingest is not running. No need to stop
			response.getResponseElem().setStatusCode(211);
			response.getResponseElem().setResponseText(
					"Ingest is already blocked. No need to block.");
		}

		return response;
	}

	/***************************************************************************
	 * Allows the further coming ingest requests to be processed again.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070126
	 * @since 20070126
	 **************************************************************************/
	public static ResponseElemDocument allowIngestRequests(UserType user) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "continueIngest");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (!isIngestActive) {
			// Allow Ingest
			isIngestActive = true;
			response.getResponseElem().setStatusCode(212);
			response.getResponseElem().setResponseText("Ingest allowed.");
		}
		else {
			// Ingest is not running. No need to stop
			response.getResponseElem().setStatusCode(213);
			response.getResponseElem().setResponseText(
					"Ingest is not blocked. No need to change status.");
		}
		return response;
	}

	/***************************************************************************
	 * Returns some statistical data about Ingest Web Service
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070607
	 * @since 20070124
	 **************************************************************************/
	public static ResponseElemDocument getIngestStatistics(UserType user) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user,
				"getIngestStatistics");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		// Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (isIngestActive) {
			// TODO: implement the algorithm

		}
		else {
			// Retrieval is not running. Return an error message.
			response.getResponseElem().setStatusCode(560);
			response.getResponseElem().setResponseText(
					"Ingest is not running. Please consult the administrator.");
		}
		return response;
	}

	/***************************************************************************
	 * Resets the statistical data about Ingest Web Service
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see ResponseElemDocument
	 * @version 20070607
	 * @since 20070124
	 **************************************************************************/
	public static ResponseElemDocument resetIngestlStatistics(UserType user) {
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user,
				"resetIngestStatistics");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		// Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (!isIngestActive) {
			// TODO: implement the algorithm
		}
		else {
			// Retrieval is still running. Return an error message.
			response.getResponseElem().setStatusCode(561);
			response
					.getResponseElem()
					.setResponseText(
							"Ingest is still running. Please block new ingest requests first.");
		}

		return response;
	}

}
