/*******************************************************************************
 * de.langzeitarchivierung.kopal.processstarter / ClientLoaderServerThread.java
 *  
 * Copyright 2005-2007 by Project kopal 
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.processstarter;

import java.net.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.io.*;

import de.langzeitarchivierung.kopal.Policy;
import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.ProcessQueue;

/*******************************************************************************
 * Awaits one SIP with custom protocol from a workflow tool, which submits the
 * SIP and the ingest policy using two ports it gets from the
 * ClientLoaderServerThread, who gets it from here.
 * 
 * PRECONDITION: Two free ports must be available from the system to submit and
 * receive data.
 * 
 * RESULT: This thread waits for SERVER_TIMEOUT seconds for a WorkflowTool
 * instance to commit a SIP.
 * 
 * TODO IN A LATER VERSION Use the nio class for faster transmission or maybe
 * use SSH for security reasons.
 * 
 * @author Funk
 * @version 20070131
 * @since 20050628
 ******************************************************************************/

public class ClientLoaderServerThread implements ServerThread
{

	// CLASS VARIABLES (Static variables) ***********************************

	private static Logger			defaultLogger	= Logger
															.getLogger("de.langzeitarchivierung.kopal");

	// STATE (Instance variables) *******************************************

	private int						serverTimeout	= 10000;
	private ServerSocket			inputServerSocket;
	private Socket					inputSocket;
	private ObjectInputStream		in;
	private BufferedInputStream		is;
	private BufferedOutputStream	fos;
	private String					destinationDir	= "";
	private String					fileName;
	private File					pathToReceivedSip;
	private String					policyName;
	private String					defaultPolicyName;
	private ProcessQueue			processQueue;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * @throws IOException
	 **************************************************************************/
	public ClientLoaderServerThread() throws IOException {
		this.inputServerSocket = new ServerSocket(0);
	}

	/***************************************************************************
	 * @see de.langzeitarchivierung.kopal.processstarter
	 *      .ProcessStarter#setProcessQueue(ProcessQueue processQueue)
	 * @param _processQueue - The queue of processes
	 **************************************************************************/
	public void setProcessQueue(ProcessQueue _processQueue) {
		this.processQueue = _processQueue;
	}

	/***************************************************************************
	 * @see de.langzeitarchivierung.kopal.processstarter.ProcessStarter#run()
	 **************************************************************************/
	public void run() {
		String errorMessage = "Unable to open socket";

		try {
			// There is a timeout after SERVER_TIMEOUT millisecs.
			this.inputServerSocket.setSoTimeout(this.serverTimeout);

			defaultLogger.log(Level.INFO, "ClientLoaderServerThread #"
					+ Thread.currentThread().getId()
					+ " is listening for one SIP on port "
					+ this.getInputPort());

			// Now wait for input.
			try {
				this.inputSocket = this.inputServerSocket.accept();
			}
			catch (InterruptedIOException e) {
				defaultLogger.log(Level.WARNING, "Thread Timeout! "
						+ "This thread no longer waits for a SIP.");
			}

			receiveFileNameAndPolicy();
			receiveFile();
			closeConnection();

			defaultLogger.log(Level.INFO, "Client Loader properly "
					+ "received SIP '" + this.fileName + "'");

			addSip();

			defaultLogger.log(Level.INFO, "SIP '" + this.fileName
					+ "' added to the process queue, thread terminates");
		}
		catch (Exception e) {
			defaultLogger.log(Level.SEVERE, errorMessage);
			e.printStackTrace();

			return;
		}
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ****

	/***************************************************************************
	 * ServerThread receives the SIP's filename and the SIP's ingest policy.
	 * 
	 * @throws IOException
	 **************************************************************************/
	private void receiveFileNameAndPolicy() throws IOException {
		String errorMessage = "Unable to receive filename and policy";

		try {
			defaultLogger.log(Level.INFO, "Receiving filename and policy");

			// Create input stream.
			this.in = new ObjectInputStream(new BufferedInputStream(
					this.inputSocket.getInputStream()));

			// Reveive filename and ingest policy strings from the asset
			// builder.
			this.fileName = (String) this.in.readObject();
			this.policyName = (String) this.in.readObject();

			defaultLogger.log(Level.INFO, "Filename '" + this.fileName
					+ "' and policy '" + this.policyName + "' received");

			// Take the default ingest policy from the main config file if the
			// received policy name does not exist in the policy configuration
			// file "policies.xml".
			if (!Policy.exists(this.policyName)) {
				defaultLogger.log(Level.WARNING, "Received ingest policy '"
						+ this.policyName
						+ "' does not exist in policy configuration file."
						+ " Using default policy '" + this.defaultPolicyName
						+ "'");
				this.policyName = this.defaultPolicyName;
			}
		}
		catch (Exception e) {
			throw new IOException(errorMessage);
		}
	}

	/***************************************************************************
	 * ServerThread receives the SIP as a file.
	 * 
	 * @throws IOException
	 **************************************************************************/
	private void receiveFile() throws IOException {
		String errorMessage = "Unable to receive file";

		try {
			defaultLogger.log(Level.INFO, "Receiving file '" + this.fileName
					+ "' from " + this.inputSocket.getInetAddress());

			// We determine the fileNameEnding and Beginning in case we need to
			// create a new filename because the old one already exists.
			String fileNameEnding = null;

			// Only this three extensions are allowed by DIAS. See IBM SIP
			// specification.
			if (this.fileName.endsWith(".zip")) {
				fileNameEnding = ".zip";
			}
			else if (this.fileName.endsWith(".tar")) {
				fileNameEnding = ".tar";
			}
			else if (this.fileName.endsWith(".tar.gz")) {
				fileNameEnding = ".tar.gz";
			}

			String fileNameBeginning = this.fileName.substring(0, this.fileName
					.length()
					- fileNameEnding.length());

			// Now try to create a new destination file and if the file already
			// exists we try again with a new filename.
			int writeAttempt = 1;
			File destFile = new File(this.destinationDir, this.fileName);

			while (!destFile.createNewFile()) {
				destFile = new File(this.destinationDir, fileNameBeginning
						+ "-" + writeAttempt++ + fileNameEnding);
				defaultLogger.log(Level.FINER, "File already exists, "
						+ "creating new destination filename '"
						+ destFile.getName() + "'");
			}

			defaultLogger.log(Level.INFO, "Writing file to '"
					+ destFile.getAbsolutePath() + "'");

			// Define the used in- and output streams.
			this.fos = new BufferedOutputStream(new FileOutputStream(destFile));
			this.is = new BufferedInputStream(this.inputSocket.getInputStream());

			// Recieve the SIP and write it to a file.
			final int BUFFER_LENGTH = 1048576;
			byte[] buffer = new byte[BUFFER_LENGTH];

			int bytesRead = this.is.read(buffer);
			while (bytesRead > 0) {
				this.fos.write(buffer, 0, bytesRead);
				bytesRead = this.is.read(buffer);
			}

			// Close the stream when ready.
			this.fos.flush();
			this.fos.close();

			// Set the SIPs location.
			this.fileName = destFile.getName();
			this.pathToReceivedSip = destFile;

			defaultLogger.log(Level.INFO, "File received");
		}
		catch (Exception e) {
			throw new IOException(errorMessage);
		}
	}

	/***************************************************************************
	 * Closes the connection.
	 * 
	 * @throws IOException
	 **************************************************************************/
	private void closeConnection() throws IOException {
		String errorMessage = "Unable to close the connection to"
				+ " the asset builder";

		try {
			defaultLogger.log(Level.INFO, "Closing connection");

			this.inputSocket.close();

			defaultLogger.log(Level.INFO, "Connection closed");

		}
		catch (Exception e) {
			throw new IOException(errorMessage + ": " + e);
		}
	}

	/***************************************************************************
	 * Adds the received SIP to the WorkflowTool's process queue.
	 **************************************************************************/
	private void addSip() {
		try {
			ProcessData pData = new ProcessData(this.policyName, this.fileName);
			pData.setSipLocation(this.pathToReceivedSip);

			this.processQueue.addElement(pData);

		}
		catch (Exception e) {
			e.printStackTrace();
			// TODO Handle exception
		}
	}

	/***************************************************************************
	 * @param _destinationDir The destinationDir to set.
	 **************************************************************************/
	public void setDestinationDir(String _destinationDir) {
		this.destinationDir = _destinationDir;
	}

	/***************************************************************************
	 * @param _defaultPolicyName The defaultPolicyName to set.
	 **************************************************************************/
	public void setDefaultPolicyName(String _defaultPolicyName) {
		this.defaultPolicyName = _defaultPolicyName;
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @see de.langzeitarchivierung.kopal.processstarter.ServerThread#getInputPort()
	 * 
	 * @return int
	 **************************************************************************/
	public int getInputPort() {
		return this.inputServerSocket.getLocalPort();
	}

	/***************************************************************************
	 * @see de.langzeitarchivierung.kopal.processstarter.ServerThread#getOutputPort()
	 * 
	 * @return int
	 **************************************************************************/
	public int getOutputPort() {
		return 0;
	}

	/***************************************************************************
	 * @return int
	 **************************************************************************/
	public int getServerTimeout() {
		return this.serverTimeout;
	}

	/***************************************************************************
	 * @param _serverTimeout
	 **************************************************************************/
	public void setServerTimeout(int _serverTimeout) {
		this.serverTimeout = _serverTimeout;
	}

}
