/*******************************************************************************
 * de.langzeitarchivierung.kopal.processstarter / MonitorHotfolderBase.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.processstarter;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.ProcessQueue;
import de.langzeitarchivierung.kopal.util.FileUtils;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;
import de.langzeitarchivierung.kopal.util.db.DBAccess;
import de.langzeitarchivierung.kopal.util.db.Various;

/*******************************************************************************
 * This is a general process starter, which looks up a hotfolder every X seconds
 * and if something new is in it, the folder (or file if wished) is added to the
 * process list. There are three abstrace classes to set the process name, the
 * custom identifier list and the persistent identifier of the object.
 * 
 * PRECONDITION: Needs a valid directory out of the -i flag of the WorkflowTool
 * or out of the main config file.
 * 
 * RESULT: Adds all directories (and if wished also files) out of the source
 * directory to the process list and sets the ID (taken from the directory name)
 * and the pathToContentFiles.
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20070713
 * @see de.langzeitarchivierung.kopal.processstarter
 * @since 20051023
 ******************************************************************************/

public abstract class MonitorHotfolderBase implements ProcessStarter
{

	// CLASS VARIABLES (Static variables) ***********************************

	protected static Logger					defaultLogger		= Logger
																		.getLogger("de.langzeitarchivierung.kopal");
	protected static ArrayList<File>		listOfFiles			= new ArrayList<File>();
	protected static HashMap<File, Long>	fileSizes			= new HashMap<File, Long>();

	/** Constants in the database for various data (varioustype) */
	public static final String				VAR_SIPSOURCEASURI	= "sipSourceAsUri";
	static boolean							useDatabase			= false;

	// STATE (Instance variables) ******************************************

	ProcessQueue							processQueue;
	String									defaultPolicyName;
	String									inputDirFromCli;
	File									hotfolderDir;
	long									startingOffset		= 0;
	long									checkingOffset		= 10000;
	long									addingOffset		= 3000;
	protected boolean						readDirectoriesOnly	= true;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Constructor
	 **************************************************************************/
	public MonitorHotfolderBase() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * @see java.lang.Runnable#run()
	 **************************************************************************/
	public void run() {
		// Take the hotfolder's path out of the commandline if existing.
		if (this.inputDirFromCli != null && !this.inputDirFromCli.equals("")) {
			this.hotfolderDir = new File(this.inputDirFromCli);
		}

		defaultLogger.log(Level.INFO, "Using policy '" + this.defaultPolicyName
				+ "'");
		defaultLogger.log(Level.INFO, "'" + this.hotfolderDir + "' is taken "
				+ "as input hotfolder, waiting for input");
		defaultLogger.log(Level.INFO, "Hotfolder scheduling starts in "
				+ KopalTimestamp.getDurationInHours(this.startingOffset)
				+ ", recheck every "
				+ KopalTimestamp.getDurationInHours(this.checkingOffset));

		try {
			Thread.sleep(this.startingOffset);

			// TODO Maybe add a posibility here to exit the process starter
			// without sending a CTRL-C?
			while (true) {
				monitorHotfolderLocation();
				Thread.sleep(this.checkingOffset);
			}
		}
		catch (InterruptedException e) {
			defaultLogger.log(Level.SEVERE, "ProcessStarter was interrupted. "
					+ "System message: " + e.getMessage());
			e.printStackTrace();
		}
	}

	// INTERNAL (Internal - implementation details, local classes, ...)

	/***************************************************************************
	 * Check the hotfolder location for new files and folders.
	 **************************************************************************/
	private void monitorHotfolderLocation() {
		File fileList[] = this.hotfolderDir.listFiles();
		File currentFile = null;

		// Check if the given directory exists.
		if (!this.hotfolderDir.exists() || !this.hotfolderDir.isDirectory()) {
			defaultLogger.log(Level.SEVERE, "The given hotfolder path "
					+ "does not exist or is not a directory");

			System.exit(6);
		}

		defaultLogger.log(Level.INFO, "Checking hotfolder "
				+ this.hotfolderDir.getAbsolutePath() + " for new content");

		// Go throught all the files in the given dir.
		int sipsToRun = fileList.length;

		for (int i = 0; i < sipsToRun; i++) {
			currentFile = fileList[i];

			// Process all files not contained in the list yet.
			if (!MonitorHotfolderBase.listOfFiles.contains(currentFile)) {
				MonitorHotfolderBase.listOfFiles.add(currentFile);

				// Check if file is restricted.
				if (sipIsRestricted(currentFile)) {
					// Add the file to the listOfFiles if so.
					defaultLogger.log(Level.FINEST, "File '" + currentFile
							+ "' is marked as restricted, skipping this file");
				}

				// If not, add the file to the scheduling list.
				else {
					// Calculate and save folder size.
					MonitorHotfolderBase.fileSizes.put(currentFile, new Long(
							FileUtils.getCompleteFileSize(currentFile)));

					// Give some scheduling log messages.
					Long filesize = MonitorHotfolderBase.fileSizes
							.get(currentFile);
					String timestamp = KopalTimestamp
							.getDurationInHours(MonitorHotfolderBase.this.addingOffset);
					defaultLogger.log(Level.INFO, "Scheduling file '"
							+ currentFile.getName() + "' (" + filesize
							+ " bytes) for adding to the process list, "
							+ "waiting " + timestamp);

					// Schedule the addition of the new SIP.
					Timer timer = new Timer();
					AddAssetTask aat = new AddAssetTask();
					aat.setCurrentFile(currentFile);
					timer.schedule(aat, MonitorHotfolderBase.this.addingOffset);
				}
			}
		}

		defaultLogger.log(Level.INFO, "All current files scheduled, "
				+ "waiting for more");
	}

	/***************************************************************************
	 * The name of the process to identify the file's process in the process
	 * list must be defined in an abstract class.
	 * 
	 * @param currentFile The process name for this file is created here.
	 * @return String The process name to return.
	 **************************************************************************/
	protected abstract String getProcessName(File currentFile);

	/***************************************************************************
	 * A custom identifier (if existing) must be defined in an abstract class.
	 * 
	 * @param currentFile The current file to set the custom identifier for.
	 * @return HashMap<String, String> The hashmap containing all the custom
	 *         IDs.
	 **************************************************************************/
	protected abstract HashMap<String, String> getCustomIdList(File currentFile);

	/***************************************************************************
	 * The <lmerObject:persistentIdentifier> must be defined in an abstract
	 * class.
	 * 
	 * @param currentFile The file to create a persistent ID for.
	 * @return String The persistent ID to return.
	 **************************************************************************/
	protected abstract String getPersistentIdentifier(File currentFile);

	/***************************************************************************
	 * Here you have the ability to restrict the added files somehow, e.g. only
	 * add SIPs to the process list that includes a mets.xml file.
	 * 
	 * @param currentFile
	 * @return TRUE if the file shall be added to the process list, FALSE if
	 *         not.
	 **************************************************************************/
	protected abstract boolean sipIsRestricted(File currentFile);

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * Sets the input directory taken from command line interface.
	 * 
	 * @see de.langzeitarchivierung.kopal.processstarter
	 *      .ProcessStarter#setInput(java.lang.Object)
	 **************************************************************************/
	public void setInput(String input) {
		this.inputDirFromCli = input;
	}

	/***************************************************************************
	 * Sets the process list.
	 * 
	 * @see de.langzeitarchivierung.kopal.processstarter
	 *      .ProcessStarter#setProcessQueue(
	 *      de.langzeitarchivierung.kopal.ProcessQueue)
	 * @param procList
	 **************************************************************************/
	public void setProcessQueue(ProcessQueue procList) {
		this.processQueue = procList;
	}

	/***************************************************************************
	 * @param defPolicyName The defaultPolicyName to set.
	 **************************************************************************/
	public void setDefaultPolicyName(String defPolicyName) {
		this.defaultPolicyName = defPolicyName;
	}

	/***************************************************************************
	 * @param _hotfolderDir The hotfolder directory to set.
	 **************************************************************************/
	public void setHotfolderDir(String _hotfolderDir) {
		this.hotfolderDir = new File(_hotfolderDir);
	}

	/***************************************************************************
	 * @param _readDirectoriesOnly Should only directories be processed as an
	 *            asset?
	 **************************************************************************/
	public void setReadDirectoriesOnly(boolean _readDirectoriesOnly) {
		this.readDirectoriesOnly = _readDirectoriesOnly;
	}

	/***************************************************************************
	 * @return Returns the addingOffset.
	 * @author Stefan Funk, SUB Goettingen 13.10.2006
	 **************************************************************************/
	public long getAddingOffset() {
		return this.addingOffset;
	}

	/***************************************************************************
	 * @param _addingOffset The addingOffset to set.
	 * @author Stefan Funk, SUB Goettingen 13.10.2006
	 **************************************************************************/
	public void setAddingOffset(long _addingOffset) {
		this.addingOffset = _addingOffset;
	}

	/***************************************************************************
	 * @return Returns the checkingOffset.
	 * @author Stefan Funk, SUB Goettingen 13.10.2006
	 **************************************************************************/
	public long getCheckingOffset() {
		return this.checkingOffset;
	}

	/***************************************************************************
	 * @param _checkingOffset The checkingOffset to set.
	 * @author Stefan Funk, SUB Goettingen 13.10.2006
	 **************************************************************************/
	public void setCheckingOffset(long _checkingOffset) {
		this.checkingOffset = _checkingOffset;
	}

	/***************************************************************************
	 * @return Returns the startingOffset.
	 * @author Stefan Funk, SUB Goettingen 13.10.2006
	 **************************************************************************/
	public long getStartingOffset() {
		return this.startingOffset;
	}

	/***************************************************************************
	 * @param _startingOffset The startingOffset to set.
	 * @author Stefan Funk, SUB Goettingen 13.10.2006
	 **************************************************************************/
	public void setStartingOffset(long _startingOffset) {
		this.startingOffset = _startingOffset;
	}

	/***************************************************************************
	 * @param useDatabase The useDatabase to set.
	 **************************************************************************/
	public static void setUseDatabase(boolean useDatabase) {
		MonitorHotfolderBase.useDatabase = useDatabase;
	}

	/***************************************************************************
	 * Adds a given directory or file including all necesarry things to the
	 * process list.
	 **************************************************************************/

	protected class AddAssetTask extends TimerTask
	{

		// STATE (Instance variables) ******************************************

		File	currentFile	= null;

		/**
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Runnable#run()
		 */
		@Override
		public void run() {
			long oldFileSize;

			try {
				// Get the complete file size.
				oldFileSize = (MonitorHotfolderBase.fileSizes
						.get(this.currentFile)).longValue();
				log(Level.FINE, "Recalculating complete file size");

				// If it's equal to the latest calculated size, process the
				// element.
				if (oldFileSize == FileUtils
						.getCompleteFileSize(this.currentFile)) {
					log(Level.FINE, "Complete file size does match the "
							+ "recently calculated size, prepare to process");

					// Remove the completeFileSize from the HashMap.
					MonitorHotfolderBase.fileSizes.remove(this.currentFile);

					// Check if the SIP's source already exist in our database.
					// Make this configurable? No, better delete the entry from
					// the database!
					if (checkSipSource()) {
						log(Level.WARNING, "A SIP with path "
								+ this.currentFile.getAbsolutePath()
								+ " has already been ingested, skipping");
						return;
					}

					// If it's not existing, add the SIP to the process queue.
					addSipToProcessQueue();
				}

				// If the file size is not equal to the latest calculated one,
				// retry later.
				else {
					// Remove file from fileList to schedule again later.
					MonitorHotfolderBase.listOfFiles.remove(this.currentFile);

					log(Level.INFO, "Scheduling of file delayed because "
							+ "of file size differences, maybe uploading is "
							+ "still in progress");
				}
			}
			catch (Exception e) {
				defaultLogger.log(Level.SEVERE, "Error adding file to the "
						+ "process list. System message: " + e.getMessage(),
						this.currentFile);
				e.printStackTrace();
			}
		}

		// INTERNAL (Internal - implementation details, local classes, ...)

		/***********************************************************************
		 * Add the new element to the process queue.
		 * 
		 * @throws Exception
		 **********************************************************************/
		private void addSipToProcessQueue() throws Exception {
			// Add the element using the default policy from the config file and
			// the directory name as a process name to identify the list
			// element.
			ProcessData pd = new ProcessData(
					MonitorHotfolderBase.this.defaultPolicyName,
					getProcessName(this.currentFile));

			// Set the LMER persistent identifier (URN).
			pd.getMetadata().setObjectId(
					getPersistentIdentifier(this.currentFile));

			// Set the path to the SIP's content files.
			log(Level.INFO, "Setting path to content files");
			pd.setPathToContentFiles(this.currentFile.getAbsolutePath());

			// Put it as SIP source in the custom data.
			String customSipSource = this.currentFile.toURI().toString();
			pd.getCustomData().put(MonitorHotfolderBase.VAR_SIPSOURCEASURI,
					customSipSource);
			log(Level.FINE, "Added custom date "
					+ MonitorHotfolderBase.VAR_SIPSOURCEASURI + " = "
					+ customSipSource);

			// If there are some custom IDs, set the custom IDs in newly
			// created ID items.
			if (getCustomIdList(this.currentFile) != null) {
				HashMap<String, String> customIdList = getCustomIdList(this.currentFile);
				pd.getCustomIds().putAll(customIdList);

				Iterator iter = customIdList.keySet().iterator();

				while (iter.hasNext()) {
					String name = (String) iter.next();
					String value = customIdList.get(name);
					log(Level.FINE, "Added custom ID " + name + " = " + value);
				}
			}

			// Add the SIP to the process list.
			MonitorHotfolderBase.this.processQueue.addElement(pd);

			if (getPersistentIdentifier(this.currentFile) == null) {
				log(Level.INFO, "Added to the process list");
				log(Level.WARNING, "Persistent identifier yet unknown");
			}
			else {
				log(Level.INFO, "Added to the process list, "
						+ "persistent identifier set to "
						+ getPersistentIdentifier(this.currentFile));
			}
		}

		/***********************************************************************
		 * Just a small loging method.
		 * 
		 * @param level
		 * @param message
		 **********************************************************************/
		private void log(Level level, String message) {
			defaultLogger.log(level, getProcessName(this.currentFile)
					+ "  -->  " + message);
		}

		/***********************************************************************
		 * Checks if the SIP's source (here the pathToContentFiles) already is
		 * existing in the database. Shall indicate the SIP already was
		 * processed and ingested, using an URI string.
		 * 
		 * @return
		 **********************************************************************/
		private boolean checkSipSource() {
			Various uri;

			try {
				if (MonitorHotfolderBase.useDatabase) {
					uri = (Various) DBAccess.getUniqueObject(Various.class,
							"type", MonitorHotfolderBase.VAR_SIPSOURCEASURI,
							"value", this.currentFile.toURI().toString());

					if (uri != null) {
						return true;
					}
				}
			}
			catch (Exception e) {
				log(Level.SEVERE, "Error accessing database. System message: "
						+ e.getLocalizedMessage());
				e.printStackTrace();
			}

			return false;
		}

		// GET & SET METHODS **************************************************

		/***********************************************************************
		 * @param _currentFile
		 **********************************************************************/
		public void setCurrentFile(File _currentFile) {
			this.currentFile = _currentFile;
		}
	}

}
