/*******************************************************************************
 * de.langzeitarchivierung.kopal.processstarter	/ MonitorHttpLocationBase.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.processstarter;

import java.io.File;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.httpclient.HttpClient;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.ProcessQueue;
import de.langzeitarchivierung.kopal.util.HTTPUtils;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;
import de.langzeitarchivierung.kopal.util.db.DBAccess;
import de.langzeitarchivierung.kopal.util.db.Various;

/*******************************************************************************
 * This process starter does check a HTTP location for objects to add to the
 * process queue. Every <code>checkingOffset<code> millies after finishing the
 * URL checking the folder will be checked for (new) folders/files. If there are
 * some, after <code>addingOffset</code> millies the checksum of all included
 * index files will be computed again and if they do match, the folder is added
 * to the process queue.
 * 
 * PRECONDITION: A valid HTTP location given in the config file.
 * 
 * RESULT: The process queue is filled with objects to process. The custom
 * process data element "httpLocation" now contains the path to the object's
 * content file(s) to be used for downloading by a following module (e.g.
 * actionmodule.FileCopyHttpBase).
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20070329
 * @see de.langzeitarchivierung.kopal.processstarter
 * @since 20061016
 ******************************************************************************/

public abstract class MonitorHttpLocationBase implements ProcessStarter
{

	// CLASS VARIABLES (Static variables) ***********************************

	protected static Logger			defaultLogger		= Logger
																.getLogger("de.langzeitarchivierung.kopal");
	static List<String>				listOfUrls			= new ArrayList<String>();
	static HashMap<String, String>	indexDirChecksums	= new HashMap<String, String>();

	/***************************************************************************
	 * Constants in the database for various data (varioustype), used in the
	 * class FileCopyHttpBase, too.
	 **************************************************************************/
	public static final String		VAR_SIPSOURCEASURI	= "sipSourceAsUri";
	static boolean					useDatabase			= false;

	// STATE (Instance variables) ******************************************

	ProcessQueue					processQueue;
	String							defaultPolicyName;
	String							inputDirFromCli;
	String							rootHtml;
	String							urlToMonitor;
	long							startingOffset		= 0;
	long							checkingOffset		= 10000;
	long							addingOffset		= 3000;
	protected boolean				readDirectoriesOnly	= true;

	// CREATION (Constructors, factory methods, static/inst init) **********

	/***************************************************************************
	 * Constructor
	 **************************************************************************/
	public MonitorHttpLocationBase() {
		super();
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * @see java.lang.Runnable#run()
	 **************************************************************************/
	public void run() {
		// Take the hotfolder's URL out of the command line if existing.
		if (this.inputDirFromCli != null && !this.inputDirFromCli.equals("")) {
			this.urlToMonitor = this.inputDirFromCli;
		}

		defaultLogger.log(Level.INFO, "Using policy '" + this.defaultPolicyName
				+ "'");
		defaultLogger.log(Level.INFO, this.urlToMonitor + " is taken as input "
				+ "HTTP location, waiting for input");
		defaultLogger.log(Level.INFO, "HTTP scheduling starts in "
				+ KopalTimestamp.getDurationInHours(this.startingOffset)
				+ ", recheck every "
				+ KopalTimestamp.getDurationInHours(this.checkingOffset));

		try {
			Thread.sleep(this.startingOffset);

			// TODO Maybe add a posibility here to exit the process starter
			// without sending a CTRL-C?
			while (true) {
				monitorHttpLocation();
				Thread.sleep(this.checkingOffset);
			}
		}
		catch (InterruptedException e) {
			defaultLogger.log(Level.SEVERE, "ProcessStarter was interrupted. "
					+ "System message: " + e.getMessage());
			e.printStackTrace();
		}
	}

	// INTERNAL (Internal - implementation details, local classes, ...)

	/***************************************************************************
	 * Check the HTTP location for new URLs.
	 **************************************************************************/
	private void monitorHttpLocation() {
		HttpClient client = new HttpClient();
		ArrayList hrefList = new ArrayList();
		String currentUrl;

		// Check if the given URL exist and if it is available and get the
		// root HTML file as a string.
		try {
			hrefList = HTTPUtils.getAllHrefs(client,
					MonitorHttpLocationBase.this.urlToMonitor.toString());
		}
		catch (Exception e) {
			defaultLogger.log(Level.SEVERE, "Error accessing "
					+ MonitorHttpLocationBase.this.urlToMonitor);
			e.printStackTrace();

			System.exit(15);
		}

		defaultLogger.log(Level.INFO, "Checking URL "
				+ MonitorHttpLocationBase.this.urlToMonitor
				+ " for new content");

		// Go throught all the files in the given directory.
		Iterator iter = hrefList.iterator();

		while (iter.hasNext()) {
			currentUrl = (String) iter.next();

			// Process all URLs not contained in the list yet.
			if (!MonitorHttpLocationBase.listOfUrls.contains(currentUrl)) {
				MonitorHttpLocationBase.listOfUrls.add(currentUrl);

				// Check if URL is restricted and add it to the list.
				if (sipIsRestricted(client, currentUrl)) {
					defaultLogger.log(Level.FINER, "The URL " + currentUrl
							+ " is marked as restricted, skipping this URL");
				}
				else {
					try {
						// Calculate and save the complete index dir checksum.
						String indexDirString = HTTPUtils
								.getCompleteIndexDirChecksum(client, currentUrl);

						MonitorHttpLocationBase.indexDirChecksums.put(
								currentUrl, indexDirString);
					}
					catch (Exception e) {
						defaultLogger.log(Level.SEVERE,
								"Error calculating the "
										+ "complete index directory checksum. "
										+ "System message: " + e.getMessage(),
								new File(currentUrl));
						e.printStackTrace();
					}

					// Give some scheduling log messages.
					String timestamp = KopalTimestamp
							.getDurationInHours(MonitorHttpLocationBase.this.addingOffset);
					defaultLogger.log(Level.INFO, "Scheduling URL "
							+ currentUrl + " for adding to the process list "
							+ "in " + timestamp);
					defaultLogger.log(Level.FINE, "Complete index directory "
							+ "checksum is "
							+ MonitorHttpLocationBase.indexDirChecksums
									.get(currentUrl));

					// Schedule the addition of the new SIP.
					Timer timer = new Timer();
					AddAssetTask aat = new AddAssetTask();
					aat.setCurrentUrl(currentUrl);
					timer.schedule(aat,
							MonitorHttpLocationBase.this.addingOffset);
				}
			}
		}

		defaultLogger.log(Level.INFO, "All current URLs scheduled, "
				+ "waiting for more");
	}

	/***************************************************************************
	 * The name of the process to identify the file's process in the process
	 * list must be defined in a abstract class.
	 * 
	 * @param currentUrl The process name for this URL is created here.
	 * @return String The process name to return.
	 **************************************************************************/
	protected abstract String getProcessName(String currentUrn);

	/***************************************************************************
	 * A custom identifier (if existing) must be defined in a abstract class.
	 * 
	 * @param currentUrl The current URL to set the custom identifier for.
	 * @return HashMap<String, String> The hashmap containing all the custom
	 *         IDs.
	 **************************************************************************/
	protected abstract HashMap<String, String> getCustomIdList(String currentUrn);

	/***************************************************************************
	 * The <lmerObject:persistentIdentifier> must be defined in a abstract
	 * class.
	 * 
	 * @param currentUrl The URL to create a persistent ID for.
	 * @return String The persistent ID to return.
	 **************************************************************************/
	protected abstract String getPersistentIdentifier(String currentUrn);

	/***************************************************************************
	 * Here you have the ability to restrict the added URL somehow, e.g. only
	 * add SIPs to the process list that includes a mets.xml file.
	 * 
	 * @param currentFile
	 * @return TRUE if the file shall be added to the process list, FALSE if
	 *         not.
	 **************************************************************************/
	protected abstract boolean sipIsRestricted(HttpClient client,
			String currentUrl);

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * Sets the input directory taken from command line interface.
	 * 
	 * @see de.langzeitarchivierung.kopal.processstarter
	 *      .ProcessStarter#setInput(String)
	 **************************************************************************/
	public void setInput(String input) {
		this.inputDirFromCli = input;
	}

	/***************************************************************************
	 * Sets the process list.
	 * 
	 * @see de.langzeitarchivierung.kopal.processstarter
	 *      .ProcessStarter#setProcessQueue(
	 *      de.langzeitarchivierung.kopal.ProcessQueue)
	 **************************************************************************/
	public void setProcessQueue(ProcessQueue _processQueue) {
		this.processQueue = _processQueue;
	}

	/***************************************************************************
	 * @param defPolicyName The defaultPolicyName to set.
	 **************************************************************************/
	public void setDefaultPolicyName(String defPolicyName) {
		this.defaultPolicyName = defPolicyName;
	}

	/***************************************************************************
	 * @param _readDirectoriesOnly Should only directories be processed as an
	 *            asset?
	 **************************************************************************/
	public void setReadDirectoriesOnly(boolean _readDirectoriesOnly) {
		this.readDirectoriesOnly = _readDirectoriesOnly;
	}

	/***************************************************************************
	 * @return Returns the addingOffset.
	 **************************************************************************/
	public long getAddingOffset() {
		return this.addingOffset;
	}

	/***************************************************************************
	 * @param _addingOffset The addingOffset to set.
	 **************************************************************************/
	public void setAddingOffset(long _addingOffset) {
		this.addingOffset = _addingOffset;
	}

	/***************************************************************************
	 * @return Returns the checkingOffset.
	 **************************************************************************/
	public long getCheckingOffset() {
		return this.checkingOffset;
	}

	/***************************************************************************
	 * @param _checkingOffset The checkingOffset to set.
	 **************************************************************************/
	public void setCheckingOffset(long _checkingOffset) {
		this.checkingOffset = _checkingOffset;
	}

	/***************************************************************************
	 * @return Returns the startingOffset.
	 **************************************************************************/
	public long getStartingOffset() {
		return this.startingOffset;
	}

	/***************************************************************************
	 * @param _startingOffset The startingOffset to set.
	 **************************************************************************/
	public void setStartingOffset(long _startingOffset) {
		this.startingOffset = _startingOffset;
	}

	/***************************************************************************
	 * @param _urlToMonitor The URL to monitor.
	 **************************************************************************/
	public void setUrlToMonitor(String _urlToMonitor) {
		this.urlToMonitor = _urlToMonitor;
	}

	/***************************************************************************
	 * @return Returns the urlToMonitor.
	 **************************************************************************/
	public String getUrlToMonitor() {
		return this.urlToMonitor.toString();
	}

	/***************************************************************************
	 * @param useDatabase The useDatabase to set.
	 **************************************************************************/
	public static void setUseDatabase(boolean useDatabase) {
		MonitorHttpLocationBase.useDatabase = useDatabase;
	}

	/***************************************************************************
	 * Adds a given directory or file including all necesarry things to the
	 * process list.
	 **************************************************************************/

	protected class AddAssetTask extends TimerTask
	{

		// STATE (Instance variables) ******************************************

		String	currentUrl	= null;

		/***********************************************************************
		 * @see java.lang.Runnable#run()
		 **********************************************************************/
		@Override
		public void run() {
			HttpClient client = new HttpClient();
			String oldCompleteIndexChecksum;

			try {
				// Get the complete index directory checksum.
				oldCompleteIndexChecksum = (MonitorHttpLocationBase.indexDirChecksums
						.get(this.currentUrl));
				log(Level.FINE, "Recalculating complete index directory "
						+ "checksum");

				// If it equals the latest calculated sum, process the element.
				if (oldCompleteIndexChecksum.equals(HTTPUtils
						.getCompleteIndexDirChecksum(client, this.currentUrl))) {
					log(Level.FINE, "Complete index directory checksum does "
							+ "match recently calculated checksum, prepare "
							+ "to process");

					// Remove the completeIndexDirChecksum from the HashMap.
					MonitorHttpLocationBase.indexDirChecksums
							.remove(this.currentUrl);

					// Check if the SIP's URL already exist in our database.
					// Make this configurable? No, better delete the entry from
					// the database!
					if (checkSipSource()) {
						log(Level.WARNING, "A SIP from the URL "
								+ this.currentUrl
								+ " has already been ingested, skipping");
						return;
					}

					// If it's not existing, add the SIP to the process queue.
					addSipToProcessQueue();
				}

				// If the checksum is not equal to the latest calculated one,
				// retry later.
				else {
					log(Level.INFO, "Scheduling delayed because of checksum "
							+ "differences, maybe uploading is still in "
							+ "progress");

					// Remove file from the fileList to schedule it again later.
					MonitorHttpLocationBase.listOfUrls.remove(this.currentUrl);
				}
			}
			catch (Exception e) {
				defaultLogger.log(Level.SEVERE, "Error adding URL. "
						+ "System message: " + e.getMessage(), this.currentUrl);
				e.printStackTrace();
			}
		}

		// INTERNAL (Internal - implementation details, local classes, ...)

		/***********************************************************************
		 * Add the new element to the process queue.
		 * 
		 * @throws Exception
		 **********************************************************************/
		private void addSipToProcessQueue() throws Exception {
			// Add the element using the default policy from the config file and
			// the name as a process name to identify the list element.
			ProcessData pd = new ProcessData(
					MonitorHttpLocationBase.this.defaultPolicyName,
					getProcessName(this.currentUrl));

			// Set the LMER persistent identifier (URN).
			pd.getMetadata().setObjectId(
					getPersistentIdentifier(this.currentUrl));

			// Set the path to the URL to download from as SIP source in the
			// custom data.
			String customSipSource = URI.create(this.currentUrl).toString();
			pd.getCustomData().put(MonitorHttpLocationBase.VAR_SIPSOURCEASURI,
					customSipSource);
			log(Level.FINE, "Added custom date "
					+ MonitorHttpLocationBase.VAR_SIPSOURCEASURI + " = "
					+ customSipSource);

			// If there are some custom IDs, set the custom IDs in newly created
			// ID items.
			if (getCustomIdList(this.currentUrl) != null) {
				HashMap<String, String> customIdList = getCustomIdList(this.currentUrl);
				pd.getCustomIds().putAll(customIdList);

				Iterator iter = customIdList.keySet().iterator();

				while (iter.hasNext()) {
					String name = (String) iter.next();
					String value = customIdList.get(name);
					log(Level.FINE, "Added custom ID " + name + " = " + value);
				}
			}

			// Add the SIP to the process list.
			MonitorHttpLocationBase.this.processQueue.addElement(pd);

			if (getPersistentIdentifier(this.currentUrl) == null) {
				log(Level.INFO, "Added to the process list");
				log(Level.WARNING, "Persistent identifier yet unknown");
			}
			else {
				log(Level.INFO, "Added to the process list, persistent "
						+ "identifier set to "
						+ getPersistentIdentifier(this.currentUrl));
			}
		}

		/***********************************************************************
		 * Just a small logging method.
		 * 
		 * @param level
		 * @param message
		 **********************************************************************/
		private void log(Level level, String message) {
			defaultLogger.log(level, getProcessName(this.currentUrl)
					+ "  -->  " + message);
		}

		/***********************************************************************
		 * Checks if the SIP's source (here the current URL) already is existing
		 * in the database. Shall indicate the SIP already was processed and
		 * ingested, using an URI string.
		 * 
		 * @return
		 **********************************************************************/
		private boolean checkSipSource() {
			Various uri;

			try {
				if (MonitorHttpLocationBase.useDatabase) {
					uri = (Various) DBAccess.getUniqueObject(Various.class,
							"type", MonitorHttpLocationBase.VAR_SIPSOURCEASURI,
							"value", URI.create(this.currentUrl).toString());

					if (uri != null) {
						return true;
					}
				}
			}
			catch (Exception e) {
				log(Level.SEVERE, "Error accessing database. System message: "
						+ e.getMessage());
				e.printStackTrace();
			}

			return false;
		}

		// GET & SET METHODS **************************************************

		/***********************************************************************
		 * @param _currentUrl
		 **********************************************************************/
		protected void setCurrentUrl(String _currentUrl) {
			this.currentUrl = _currentUrl;
		}
	}

}
