/*******************************************************************************
 * de.langzeitarchivierung.kopal.processstarter.migrationmanager / FileObjCharacterTranslator.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 *
 *
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.processstarter.migrationmanager;

import gov.loc.mets.MetsType;
import gov.loc.mets.DivType.Mptr;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;

import de.langzeitarchivierung.kopal.formats.Uof;

/*******************************************************************************
 * Queries a set of mets files for the characterised objects.
 * 
 * @author Ludwig
 * @version 0.5
 ******************************************************************************/

public class FileObjCharacterTranslator implements ObjCharacterTranslator
{

	// STATE (Instance variables) ******************************************

	private List<File>	metsFiles	= new ArrayList<File>();

	/***************************************************************************
	 * Default Constructor.
	 **************************************************************************/
	public FileObjCharacterTranslator() {
		super();
	}

	/***************************************************************************
	 * Constructor for querying the specified file or dir.
	 * 
	 * @param metsFileOrDirString The path of a mets file or directory which
	 *            contains mets files.
	 * @throws Exception if accessing the mets files fails
	 **************************************************************************/
	public FileObjCharacterTranslator(String metsFileOrDirString)
			throws Exception {
		super();
		File metsFileOrDir = new File(metsFileOrDirString);
		if (metsFileOrDir.isDirectory()) {
			this.setMetsFilesDir(metsFileOrDirString);
		}
		else if (metsFileOrDir.isFile()) {
			this.addMetsFile(metsFileOrDir);
		}
		else {
			throw new Exception("Neither file nor directory: "
					+ metsFileOrDirString);
		}
	}


	// MANIPULATION (Manipulation - what the object does) *****************

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.processstarter.migrationmanager.ObjCharacterTranslator#getCharacterisedObjectValue(java.lang.String,
	 *      de.langzeitarchivierung.kopal.processstarter.migrationmanager.ObjCharacter)
	 */
	public List<String> getCharacterisedObjectValue(String internalId,
			ObjCharacter objCharacter) throws Exception {
		List<String> results = new ArrayList<String>();
		XmlObject objToQuery = null;

		// if hits for all objects and not just for this internalId are
		// requested
		if (internalId == null || internalId.equals("")) {

			// get all results...
			List<XmlObject> xmlResults = queryResultsOnAllFiles(objCharacter
					.getCharacterisationQuery(), null);
			// and add the values without xml fragments to the result list
			for (Iterator xmlIter = xmlResults.iterator(); xmlIter.hasNext();) {
				XmlObject xmlHit = (XmlObject) xmlIter.next();
				results.add(xmlHit.getDomNode().getFirstChild().getNodeValue());
			}
		}
		else {
			objToQuery = getObject(internalId);
			try {
				XmlObject[] queryResults = objToQuery.selectPath(objCharacter
						.getCharacterisationQuery());
				for (int i = 0; i < queryResults.length; i++) {
					XmlObject hit = queryResults[i];

					results
							.add(hit.getDomNode().getFirstChild()
									.getNodeValue());
				}
			}
			catch (RuntimeException e) {
				throw new XmlException("XPath Syntax is invalid: "
						+ e.getMessage());
			}

		}

		return results;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.processstarter.migrationmanager.ObjCharacterTranslator#getCharacterisedObjects(de.langzeitarchivierung.kopal.processstarter.migrationmanager.ObjCharacter)
	 */
	public HashSet<String> getCharacterisedObjects(ObjCharacter objChr)
			throws Exception {
		HashSet<String> results = new HashSet<String>();

		// get the hits
		HashSet<XmlObject> xmlObjectResults = new HashSet<XmlObject>(
				queryUniqueResultsOnAllFiles(objChr.getCharacterisationQuery(),
						objChr.getCharacterisationValue()));

		// get the internal ids
		for (Iterator hitIter = xmlObjectResults.iterator(); hitIter.hasNext();) {
			XmlObject hit = (XmlObject) hitIter.next();
			results.add(FileObjCharacterTranslator.getInternalId(hit));
		}

		return results;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see de.langzeitarchivierung.kopal.processstarter.migrationmanager.ObjCharacterTranslator#isNotMigrated(java.lang.String)
	 */
	public boolean isNotMigrated(String internalId) throws Exception {
		if (!queryUniqueResultsOnAllFiles(ObjCharacter.MIGRATION_QUERY,
				internalId).isEmpty()) {
			return false;
		}
		return true;
	}

	/***************************************************************************
	 * Adds the specified mets file to the list of files to query for
	 * characterized objects.
	 * 
	 * @param metsFile the mets file to check for future queries
	 * @throws Exception
	 **************************************************************************/
	public void addMetsFile(File metsFile) throws Exception {
		if (!metsFile.isFile()) {
			throw new FileNotFoundException("This is not a file");
		}
		Uof metsToCheck = new Uof();
		// if the parsing succeeds it is indeed a mets file
		metsToCheck.useTemplate(metsFile);
		this.metsFiles.add(metsFile);
	}

	/***************************************************************************
	 * Adds all mets files in the the specified dir and its direct sub dirs to
	 * the list of files to query for characterized objects.
	 * 
	 * @param metsFilesDir directory containing mets files or direct sub
	 *            directories with mets files.
	 **************************************************************************/
	public void setMetsFilesDir(String metsFilesDirPath)
			throws FileNotFoundException, XmlException {
		File metsFilesDir = new File(metsFilesDirPath);
		if (!metsFilesDir.isDirectory()) {
			throw new FileNotFoundException("Not a directory: "
					+ metsFilesDir.getAbsolutePath());
		}

		File[] dirEntries = metsFilesDir.listFiles();
		for (int i = 0; i < dirEntries.length; i++) {
			if (dirEntries[i].isFile()
					&& dirEntries[i].getName().equalsIgnoreCase("mets.xml")) {
				try {
					addMetsFile(dirEntries[i]);
				}
				// the parsing failed so it wasn't really a mets file
				catch (Exception e) {
					throw new XmlException("File "
							+ dirEntries[i].getAbsolutePath()
							+ " is no real mets file: " + e.getMessage());
				}
			}
			else if (dirEntries[i].isDirectory()) {
				this.setMetsFilesDir(dirEntries[i].getAbsolutePath());
			}
		}
	}

	/***************************************************************************
	 * Returns a list of mets pointer objects for an object with the specified
	 * internal id.
	 * 
	 * @param intId internal id of the object to analyse
	 * @return List of xmlobjects of mets pointers
	 **************************************************************************/
	public List<XmlObject> getMptrs(String intId) throws Exception {
		try {
			return Arrays.asList(getObject(intId).selectPath(
					ObjCharacter.MPTR_QUERY));
		}
		catch (RuntimeException e) {
			throw new XmlException("XPath Syntax is invalid: " + e.getMessage());
		}
	}

	/***************************************************************************
	 * Returns a list of xmlobject hits for a specified object characterisation
	 * in the specified file.
	 * 
	 * @param toCheck the mets file to check
	 * @param objChr the object characterisation to look for
	 * @return list of xmlobject hits
	 * @throws Exception if parsing or quering of the mets file fails
	 **************************************************************************/
	public List<XmlObject> checkFileForHits(File toCheck, ObjCharacter objChr)
			throws Exception {
		Uof metsToCheck = new Uof();
		// parse metsfiles
		metsToCheck.useTemplate(toCheck);

		return FileObjCharacterTranslator.getQueryResults(
				(XmlObject) metsToCheck.getMetadataImplementation(), objChr);
	}

	/***************************************************************************
	 * Returns the file object of the mets file with the requested internal id
	 * or null if no such file exists.
	 * 
	 * @param intId internal id
	 * @return file corresponding to the internal id or null if no such file
	 *         exists
	 * @throws Exception if query fails or more than one
	 **************************************************************************/
	public File getFile(String intId) throws Exception {
		Set<File> result = queryFiles(new ObjCharacter(
				ObjCharacter.INT_ID_QUERY, intId));
		if (result.size() == 1) {
			return result.iterator().next();
		}
		if (result.size() > 1) {
			throw new Exception(
					"More than one object with the same internal id!");
		}

		return null;
	}

	/***************************************************************************
	 * Returns the href mets pointers of a list of mets pointer xml objects.
	 * 
	 * @param mptrs mets pointer xml objects
	 * @return list of href mets pointers
	 **************************************************************************/
	public static List<String> getMptrsString(List<XmlObject> mptrs) {
		ArrayList<String> mptrStrings = new ArrayList<String>();
		for (Iterator mptrIter = mptrs.iterator(); mptrIter.hasNext();) {
			Mptr mptr = (Mptr) mptrIter.next();
			mptrStrings.add(mptr.getHref());
		}

		return mptrStrings;
	}

	/***************************************************************************
	 * Returns the internal id of an xml object which is part of a mets file.
	 * 
	 * TODO move this method into FileObjCharacterTranslator or in another class
	 * because its signature depends on a specific object implementation
	 * (XmlBeans) ?
	 * 
	 * @param chrResult xml object representing part of a mets file
	 * @return internal id of null if none is present
	 **************************************************************************/
	public static String getInternalId(XmlObject chrResult) throws XmlException {
		XmlObject[] result = null;
		try {
			result = chrResult.selectPath(ObjCharacter.INT_ID_QUERY);
		}
		catch (RuntimeException e) {
			throw new XmlException("XPath Syntax is invalid: " + e.getMessage());
		}
		if (result.length > 0 && result[0].getDomNode().getFirstChild() != null) {
			return result[0].getDomNode().getFirstChild().getNodeValue();
		}

		return null;
	}

	/***************************************************************************
	 * Queries the xml beans representation of an uof for the specified object
	 * characterisation.
	 * 
	 * TODO move this method into FileObjCharacterTranslator or in another class
	 * because its signature depends on a specific object implementation
	 * (XmlBeans)?
	 * 
	 * @param uofToQuery xml beans representation of an uof
	 * @param objChr object characterisation to look for
	 * @return list of hits
	 **************************************************************************/
	public static List<XmlObject> getQueryResults(XmlObject uofToQuery,
			ObjCharacter objChr) throws XmlException {
		List<XmlObject> results = new ArrayList<XmlObject>();
		XmlObject[] queryResults = null;
		try {
			queryResults = uofToQuery.selectPath(objChr
					.getCharacterisationQuery());
		}
		catch (RuntimeException e) {
			throw new XmlException(
					"XPath Syntax of object characterisation is invalid: "
							+ e.getMessage());
		}

		for (int i = 0; i < queryResults.length; i++) {
			XmlObject possibleHit = queryResults[i];

			String queryResult = possibleHit.getDomNode().getFirstChild()
					.getNodeValue();
			// if all hits should be counted or just those with the specified
			// value
			if (objChr.getCharacterisationValue() == null
					|| queryResult.equals(objChr.getCharacterisationValue())) {
				if (objChr.getCharacterisationQuery().equalsIgnoreCase(
						ObjCharacter.FILE_LINKEDTO_QUERY)) {
				}
				results.add(possibleHit);
			}
		}

		return results;
	}

	// INTERNAL (Internal - implementation details, local classes, ...) **

	/***************************************************************************
	 * Returns the xml object representation of the uof containing the requested
	 * internal id.
	 * 
	 * @param intId internal id
	 * @return xml object representation of the uof or null if no such object
	 *         exists
	 * @throws Exception if query fails or more than one object with the same
	 *             internal id exists.
	 **************************************************************************/
	private XmlObject getObject(String intId) throws Exception {

		Set<XmlObject> result = queryUniqueResultsOnAllFiles(
				ObjCharacter.INT_ID_QUERY, intId);
		if (result.size() == 1) {
			return result.iterator().next();
		}
		if (result.size() > 1) {
			throw new XmlException(
					"More than one object with the same internal id!");
		}

		return null;
	}

	/***************************************************************************
	 * Returns the files which match the query
	 * 
	 * @param objChr object characterisation to query for
	 * @return set of mets files matching the query
	 * @throws Exception if parsing or quering of the mets file fails
	 **************************************************************************/
	private Set<File> queryFiles(ObjCharacter objChr) throws Exception {
		HashSet<File> results = new HashSet<File>();
		for (Iterator metsFileIter = this.metsFiles.iterator(); metsFileIter
				.hasNext();) {
			File metsFile = (File) metsFileIter.next();

			if (!checkFileForHits(metsFile, objChr).isEmpty()) {
				results.add(metsFile);
			}
		}

		return results;
	}

	/***************************************************************************
	 * Get all results for the specified xpath-query and value
	 * 
	 * @param query xpath-query
	 * @param value expected result of the xpath-query
	 * @return list of xml objects
	 * @throws Exception if parsing or quering of the mets files fails
	 **************************************************************************/
	private List<XmlObject> queryResultsOnAllFiles(String query, String value)
			throws Exception {
		List<XmlObject> results = new ArrayList<XmlObject>();
		for (Iterator metsFileIter = this.metsFiles.iterator(); metsFileIter
				.hasNext();) {
			File metsFile = (File) metsFileIter.next();

			Uof metsToCheck = new Uof();
			// parse metsfiles
			metsToCheck.useTemplate(metsFile);

			// execute query
			MetsType metsBeans = (MetsType) metsToCheck
					.getMetadataImplementation();
			results.addAll(FileObjCharacterTranslator.getQueryResults(
					metsBeans, new ObjCharacter(query, value)));
		}

		return results;
	}

	/***************************************************************************
	 * Get all unique results for the specified xpath-query and value
	 * 
	 * @param query xpath-query
	 * @param value expected result of the xpath-query
	 * @return set of xml objects
	 * @throws Exception if parsing or quering of the mets files fails
	 **************************************************************************/
	private Set<XmlObject> queryUniqueResultsOnAllFiles(String query,
			String value) throws Exception {
		HashSet<XmlObject> results = new HashSet<XmlObject>();
		results.addAll(queryResultsOnAllFiles(query, value));
		return results;
	}

}
