/*******************************************************************************
 * de.langzeitarchivierung.kopal.processstarter.migrationmanager / MigrationEventScheduler.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 *
 *
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.processstarter.migrationmanager;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.apache.xmlbeans.XmlException;

import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.ProcessQueue;
import de.langzeitarchivierung.kopal.processstarter.ProcessStarter;
import de.langzeitarchivierung.kopal.retrieval.DiasAccess;
import de.langzeitarchivierung.kopal.retrieval.DiasAccessResponse;
import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.FileUtils;
import de.langzeitarchivierung.kopal.util.KopalClassLoader;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;

/*******************************************************************************
 * ProcessStarter which processes migration events, checks the archive for
 * object where a migration is necessary and schedules migration processes.
 * 
 * @author Ludwig
 * @version 0.5
 ******************************************************************************/

public class MigrationEventScheduler implements ProcessStarter
{

	// CLASS VARIABLES (Static variables) *********************************

	protected static Logger				defaultLogger						= Logger
																					.getLogger("de.langzeitarchivierung.kopal");

	private static final String			responseFormat						= DiasAccess.RESPONSE_FORMAT_XML;
	private static final String			dipFormat							= DiasAccess.DIP_FORMAT_ZIP;
	private static final long			waitUntilNextTry					= 10000;

	// STATE (Instance variables) ******************************************

	private ProcessQueue				processQueue;
	private String						workDir;
	private DiasAccess					da									= null;

	private List<String>				commandLine							= null;

	private String						objChrTranslatorClassName			= null;
	private String						migrationEventListenerClassName		= null;

	private MigrationEventListener		migEventListener					= null;
	private ObjCharacterTranslator		objChrTrans							= null;
	private ObjectDependencyAnalysator	objDependAnalyser					= null;

	private boolean						downloadDipIfUnsupportedRemoteQuery	= false;
	private boolean						resolveObjectDependencies			= false;

	/***************************************************************************
	 * The constructor.
	 **************************************************************************/
	public MigrationEventScheduler() {
		super();
		this.commandLine = new ArrayList<String>();
	}

	// MANIPULATION (Manipulation - what the object does) *****************

	/***************************************************************************
	 * Configures the whole migration manager, waits for migration events and
	 * schedules migration processes.
	 * 
	 * TODO make different parts of the migration manger threaded? For example
	 * the MigrationEventListener?
	 **************************************************************************/
	public void run() {
		try {
			// configure the static policy builder class
			Configurator
					.configure(de.langzeitarchivierung.kopal.processstarter.migrationmanager.MigrationPolicyBuilder.class);
			initialize();
		}
		catch (Exception e) {
			defaultLogger.log(Level.SEVERE,
					"Initialisation of migration manger " + "failed! Reason: "
							+ e.getMessage());
			e.printStackTrace();
			return;
		}

		// wait for event and get source object charakterisation and destination
		// def
		this.migEventListener.run();
	}

	/***************************************************************************
	 * Evaluates and executes the consequences of the migration event for the
	 * archive.
	 * 
	 * @param srcObjCharacter the characterisation of the object to migrate
	 * @param dstObjCharacter the characterisation of the object to migrate to
	 **************************************************************************/
	public void startMigrationEventConsequences(ObjCharacter srcObjCharacter,
			ObjCharacter dstObjCharacter) {
		defaultLogger.log(Level.INFO, "Migration event source: "
				+ srcObjCharacter + " destination: " + dstObjCharacter);

		List<String> objsToMigrate = null;
		try {
			objsToMigrate = this.checkArchive(srcObjCharacter);
		}
		catch (Exception e) {
			defaultLogger.log(Level.WARNING, "Checking the archive for the"
					+ " source object characterization failed. Reason: "
					+ e.getMessage());
			return;
		}

		defaultLogger.log(Level.INFO, "Number of hits in archive: "
				+ objsToMigrate.size());

		while (objsToMigrate != null && !objsToMigrate.isEmpty()) {
			String migHit = objsToMigrate.get(0);

			// get objects which are connected to this one
			List<String> metaMigrationObject = new ArrayList<String>();
			metaMigrationObject.add(migHit);

			try {
				File migrationDir = new File(this.workDir, Long.toString(System
						.currentTimeMillis()));
				migrationDir.mkdir();

				// TODO dealing with object dependencies is still messy...
				if (this.resolveObjectDependencies) {
					try {
						metaMigrationObject.addAll(this.objDependAnalyser
								.getDependentObjects(migHit));
					}
					catch (UnsupportedOperationException e) {
						if (this.downloadDipIfUnsupportedRemoteQuery) {
							retrieveDip(migHit, migrationDir);
							FileObjCharacterTranslator localMetsTranslator = new FileObjCharacterTranslator(
									migrationDir.getAbsolutePath());
							ObjectDependencyAnalysator localAnalyser = new ObjectDependencyAnalysator();
							localAnalyser
									.setObjChrTranslator(localMetsTranslator);
							metaMigrationObject.addAll(localAnalyser
									.getDependentObjects(migHit));
							// TODO analyse it
						}
						else {
							throw e;
						}
					}
				}

				prepareMigrationObjects(metaMigrationObject, migrationDir);

				List<ProcessData> migrationProcesses = MigrationPolicyBuilder
						.buildPolicy(migrationDir, srcObjCharacter,
								dstObjCharacter);

				// add all migration processes to the process queue
				for (Iterator migProcIter = migrationProcesses.iterator(); migProcIter
						.hasNext();) {
					ProcessData migProc = (ProcessData) migProcIter.next();
					// migProc.getPolicy().printPolicyTree();
					this.processQueue.addElement(migProc);
				}
			}
			catch (Exception e) {
				defaultLogger
						.log(Level.WARNING, "Migration of the object with the "
								+ "internal ID " + migHit + " failed. Reason: "
								+ e.getMessage());
				e.printStackTrace();
			}

			// remove all object from migration schedule for which processes
			// were scheduled
			objsToMigrate.removeAll(metaMigrationObject);
		}

	}

	/***************************************************************************
	 * Returns a list of all internal ids of the characterized objects which
	 * haven't been migrated.
	 * 
	 * TODO extend to check if the migrations satisfy the destination object
	 * character so that we build a branch in the migration tree
	 * 
	 * @param objCharacter the characterisation of the objects to migrate
	 * @returna list of all internal of the characterized objects ids which
	 *          haven't been migrated
	 **************************************************************************/
	public List<String> checkArchive(ObjCharacter objCharacter)
			throws Exception {
		List<String> migrationObjects = new ArrayList<String>();
		Set<String> possibleMigrations = this.objChrTrans
				.getCharacterisedObjects(objCharacter);

		// check if any of these objects hasn't been migrated
		for (Iterator hitIter = possibleMigrations.iterator(); hitIter
				.hasNext();) {
			String hitIntId = (String) hitIter.next();
			if (this.objChrTrans.isNotMigrated(hitIntId)) {
				migrationObjects.add(hitIntId);
			}
		}

		return migrationObjects;
	}

	// INTERNAL (Internal - implementation details, local classes, ...) **

	/***************************************************************************
	 * Initalizes the migration manager.
	 * 
	 * @throws Exception
	 **************************************************************************/
	private void initialize() throws Exception {
		this.da = new DiasAccess();
		Configurator.configure(this.da);

		this.objChrTrans = loadCharacterTranslatorClass(this.objChrTranslatorClassName);
		Configurator.configure(this.objChrTrans);

		this.migEventListener = loadMigrationEventListenerClass(this.migrationEventListenerClassName);

		Configurator.configure(this.migEventListener, this);

		// TODO replace this with dynamically loaded, more flexible
		// and different ObjectDependencyAnalysators?
		this.objDependAnalyser = new ObjectDependencyAnalysator();
		this.objDependAnalyser.setObjChrTranslator(this.objChrTrans);

	}

	/***************************************************************************
	 * Downloads all connected objects which have to be migrated into a
	 * directory in the working directory.
	 * 
	 * @param migObject A list of internalIds
	 * @return the directory in which all sub directories contain the objects
	 *         which have to be migrated together
	 **************************************************************************/
	private void prepareMigrationObjects(List<String> migObject,
			File migrationDir) throws Exception {
		String[] alreadyDownloaded = migrationDir.list();


		for (Iterator objIter = migObject.iterator(); objIter.hasNext();) {
			String intId = (String) objIter.next();

			String alreadyDownloadedFilename = FileUtils
					.getFilenamePersistentIdentifier(intId)
					+ ".zip";
			for (int i = 0; i < alreadyDownloaded.length; i++) {
				if (!alreadyDownloaded[i].endsWith(alreadyDownloadedFilename)) {
					File dip = retrieveDip(intId, migrationDir);
					unzip(dip);
					dip.delete();
				}
			}
		}
	}

	/***************************************************************************
	 * Retrieves the dip with the specified internal id from dias.
	 * 
	 * TODO move function into more appropriate class
	 * 
	 * @param intId the internal id of the object to retrieve.
	 * @param destDir the destination directory where the dip should be saved.
	 **************************************************************************/
	private File retrieveDip(String intId, File destDir) throws Exception {
		DiasAccessResponse dar = null;

		defaultLogger.log(Level.FINE, "Requesting DIP with internal DIAS ID '"
				+ intId + "'");

		do {
			dar = DiasAccess.parseDiasResponse(this.da.requestAsset(intId,
					responseFormat, dipFormat));
			// TODO we need to do something with the asset version
			// dar.getMetadata().get(0).getAssetVersion();

			// Try again later if response type is busy.
			if (dar.getResponseType().equals(
					DiasAccessResponse.RESPONSE_TYPE_BUSY)) {
				defaultLogger.log(Level.FINE, "DIAS says response type is "
						+ DiasAccessResponse.RESPONSE_TYPE_BUSY);
				long time = Long.parseLong(dar.getMetadata().get(0)
						.getDipRetrieveWaitTime());
				String unit = dar.getMetadata().get(0).getTimeUnit();

				if (unit.equals("sec")) {
					defaultLogger.log(Level.FINE,
							"DIP will be ready for download in "
									+ KopalTimestamp
											.getDurationInHours(time * 1000));
					Thread.sleep(time * 1000);
				}
				else {
					defaultLogger.log(Level.FINE,
							"DIP will be ready to download in " + time + " "
									+ unit);
					Thread.sleep(waitUntilNextTry);
				}
			}
			else if (dar.getResponseType().equals(
					DiasAccessResponse.RESPONSE_TYPE_ERROR)) {

				throw new IOException(" DIAS response type is "
						+ DiasAccessResponse.RESPONSE_TYPE_ERROR + " for DIP "
						+ "with the internalID " + intId + ": '"
						+ dar.getDescription() + "'");
			}

		}
		while (!dar.getResponseType().equals(
				DiasAccessResponse.RESPONSE_TYPE_DIP));

		// TODO Check the DIP's size! Request a TAR if too large!

		// Request the DIP and save it in the temp folder using a unique
		// filename.
		File theDip = this.da.downloadDiasPackage(dar, destDir
				.getAbsolutePath(), System.currentTimeMillis() + "_"
				+ FileUtils.getFilenamePersistentIdentifier(intId) + ".zip");

		defaultLogger.log(Level.FINE, "Retrieving DIP. Writing to '"
				+ theDip.getName() + "'");

		return theDip;
	}

	/***************************************************************************
	 * 
	 * TODO move function into more appropriate class
	 * 
	 * @param sourceFile
	 * @throws Exception
	 **************************************************************************/
	private void unzip(File sourceFile) throws Exception {
		ZipInputStream zipStream = null;

		try {

			File destFolder = new File(sourceFile.getParent(), "unpacked_"
					+ sourceFile.getName());

			// TODO If the destination file already exists generate a new folder
			// name?
			// while (destFolder.exists()) {
			// destFolder = new File(this.sourceFile.getParent(), System
			// .currentTimeMillis()
			// + "_" + this.sourceFile.getName());
			// }

			destFolder.mkdir();

			defaultLogger.log(Level.FINE, "Decompressing DIP");

			zipStream = new ZipInputStream(new BufferedInputStream(
					new FileInputStream(sourceFile)));

			storeZipContent(zipStream, destFolder);
			zipStream.close();

		}
		finally {
			if (zipStream != null) {
				try {
					zipStream.close();
				}
				catch (Exception closeE) {
					defaultLogger.log(Level.FINE,
							"Error closing stream. System message: "
									+ closeE.getMessage());
				}
			}
		}
	}

	/***************************************************************************
	 * Goes through all the ZIP file entries and stores all of them including
	 * all subdirectories.
	 * 
	 * TODO move function into more appropriate class
	 * 
	 * @param zipStream
	 * @param destFolder
	 * @throws IOException
	 **************************************************************************/
	private void storeZipContent(ZipInputStream zipStream, File destFolder)
			throws Exception {
		// Go throught all the ZIP entries.
		ZipEntry metsEntry = zipStream.getNextEntry();

		while (metsEntry != null) {
			if (metsEntry.isDirectory()) {
				File newDir = new File(destFolder, metsEntry.getName());
				newDir.mkdir();
			}
			else {
				storeZipFile(zipStream, new File(destFolder, metsEntry
						.getName()));
			}

			metsEntry = zipStream.getNextEntry();
		}
	}

	/***************************************************************************
	 * Stores a single file from a ZIP.
	 * 
	 * TODO move function into more appropriate class
	 * 
	 * @param zipStream
	 * @param destinationFile
	 **************************************************************************/
	private void storeZipFile(ZipInputStream zipStream, File destinationFile)
			throws Exception {
		BufferedOutputStream tempMetsStream = null;

		try {
			tempMetsStream = new BufferedOutputStream(new FileOutputStream(
					destinationFile));

			// Decompress now.
			final int BUFFER_LENGTH = 256 * 1024;
			byte[] buffer = new byte[BUFFER_LENGTH];
			int bytesRead = zipStream.read(buffer);
			while (bytesRead > 0) {
				tempMetsStream.write(buffer, 0, bytesRead);
				bytesRead = zipStream.read(buffer);
			}

			tempMetsStream.flush();
		}
		catch (Exception e) {
			defaultLogger.log(Level.SEVERE,
					"Extraction failed. System message: " + e.getMessage());
			throw e;
		}
		finally {
			if (tempMetsStream != null) {
				try {
					tempMetsStream.close();
				}
				catch (Exception closeE) {
					defaultLogger.log(Level.SEVERE, "Error closing stream: "
							+ closeE.getMessage());
				}
			}
		}
	}

	/***************************************************************************
	 * Goes recursively through the diretories and invokes copy method for files
	 * 
	 * TODO remove this test utility method. it is just a copy from
	 * actionmodule.filecopybase
	 * 
	 * @param source - The folder containing the source data
	 * @param destination - The destination folder
	 * @throws Exception
	 **************************************************************************/
	private void recurseCopyFiles(File source, File destination)
			throws Exception {
		if (source.isDirectory()) {
			File[] filelist = source.listFiles();
			File newdir = new File(destination, source.getName());

			// Create a new directory in the destination folder.
			newdir.mkdir();
			for (int i = 0; i < filelist.length; i++) {
				// Recurse through the directory.
				recurseCopyFiles(filelist[i], newdir);
			}
		}
		if (source.isFile()) {
			File destfile = new File(destination, source.getName());

			copyFile(source, destfile);
		}
	}

	/***************************************************************************
	 * Copies a single file from a given source to a given destination
	 * 
	 * TODO remove this test utility method. it is just a copy from
	 * actionmodule.filecopybase
	 * 
	 * @param source - The source file
	 * @param destination - The destination folder
	 * @throws Exception
	 **************************************************************************/
	private void copyFile(File source, File destination)
			throws FileNotFoundException, IOException {
		// Open input- and output streams for the source and destination
		// file.
		FileInputStream fis = new FileInputStream(source);
		FileOutputStream fos = new FileOutputStream(destination);

		// Copy the source file to the destination file through a buffer.
		byte[] buf = new byte[1048576];
		int i = 0;
		while ((i = fis.read(buf)) != -1) {
			fos.write(buf, 0, i);
		}

		// Close the Streams.
		fis.close();
		fos.close();
	}

	/***************************************************************************
	 * Dynamically loads the requested object character translator class.
	 * 
	 * @param objChrTranslatorClassName the class name without package name
	 * @return an instance of the requested class
	 * @throws ClassNotFoundException
	 * @throws MalformedURLException
	 * @throws InvocationTargetException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 **************************************************************************/
	private ObjCharacterTranslator loadCharacterTranslatorClass(
			String _objChrTranslatorClassName) throws ClassNotFoundException,
			MalformedURLException, InvocationTargetException,
			InstantiationException, IllegalAccessException {
		return (ObjCharacterTranslator) loadClass(_objChrTranslatorClassName);
	}

	/***************************************************************************
	 * Dynamically loads the requested migration event listener class.
	 * 
	 * @param migrationEventListenerClassName the class name without package
	 *            name
	 * @return an instance of the requested class
	 * @throws ClassNotFoundException
	 * @throws MalformedURLException
	 * @throws InvocationTargetException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 **************************************************************************/
	private MigrationEventListener loadMigrationEventListenerClass(
			String _migrationEventListenerClassName)
			throws ClassNotFoundException, MalformedURLException,
			InvocationTargetException, InstantiationException,
			IllegalAccessException {
		return (MigrationEventListener) loadClass(_migrationEventListenerClassName);
	}

	/***************************************************************************
	 * Dynamically loads the requested object of the requested migrationmanger
	 * package.
	 * 
	 * @param objChrTranslatorClassName the class name without migrationmanger
	 *            package prefix
	 * @return an instance of the requested class
	 * @throws ClassNotFoundException
	 * @throws MalformedURLException
	 * @throws InvocationTargetException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 **************************************************************************/
	private Object loadClass(String className) throws ClassNotFoundException,
			MalformedURLException, InvocationTargetException,
			InstantiationException, IllegalAccessException {
		Class objChrTrslClass = KopalClassLoader
				.getProcessStarter("migrationmanager." + className);
		return KopalClassLoader.getInstanceOfClass(objChrTrslClass);
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * Sets the class name of the object character translator.
	 * 
	 * @param objChrTranslatorClassName class name of the object character
	 *            translator
	 **************************************************************************/
	public void setMigrationEventListenerClassName(
			String migrationEventListenerClassName) {
		this.migrationEventListenerClassName = migrationEventListenerClassName;
	}

	/***************************************************************************
	 * Sets the class name of the object character translator.
	 * 
	 * @param objChrTranslatorClassName class name of the object character
	 *            translator
	 **************************************************************************/
	public void setObjChrTranslatorClassName(String objChrTranslatorClassName) {
		this.objChrTranslatorClassName = objChrTranslatorClassName;
	}

	/***************************************************************************
	 * Gives the processStarter the process list to which new assets should be
	 * added.
	 * 
	 * @param processQueue the process list to which new assets should be added
	 **************************************************************************/
	public void setProcessQueue(ProcessQueue processQueue) {
		this.processQueue = processQueue;
	}

	/***************************************************************************
	 * Sets the directory path of the directory in which the migration is done.
	 * 
	 * @param workDir the absolute directory path
	 **************************************************************************/
	public void setWorkDir(String workDir) {
		this.workDir = workDir;
	}

	/***************************************************************************
	 * Gives the processStarter commandline arguments.
	 * 
	 * @param input the specified input (from the commandline) for this
	 *            processstarter
	 **************************************************************************/
	public void setInput(String input) {
		this.commandLine.add(input);
	}

	/***************************************************************************
	 * Returns the command line as a list of strings.
	 * 
	 * @return command line as a list of strings
	 **************************************************************************/
	public List<String> getCommandLine() {
		return this.commandLine;
	}


	/***************************************************************************
	 * Set if the archive packages should be downloaded for further analysing if
	 * this can't be done remote
	 * 
	 * @param downloadDipIfUnsupportedRemoteQuery
	 **************************************************************************/
	public void setDownloadDipIfUnsupportedRemoteQuery(
			boolean downloadDipIfUnsupportedRemoteQuery) {
		this.downloadDipIfUnsupportedRemoteQuery = downloadDipIfUnsupportedRemoteQuery;
	}

	/***************************************************************************
	 * Set if object dependencies should be resolved.
	 * 
	 * @param resolveObjectDependencies
	 **************************************************************************/
	public void setResolveObjectDependencies(boolean resolveObjectDependencies) {
		this.resolveObjectDependencies = resolveObjectDependencies;
	}

}
