/*******************************************************************************
 * de.langzeitarchivierung.kopal.processstarter.migrationmanager / MigrationPolicyBuilder.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 *
 *
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.processstarter.migrationmanager;

import gov.loc.mets.MdSecType;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;

import koLibRIMigrationPolicyTemplates.MigrationPolicyTemplateDocument.MigrationPolicyTemplate;
import koLibRIMigrationPolicyTemplates.PolicyFragmentDocument.PolicyFragment;
import koLibRIMigrationPolicyTemplates.PropertyDocument.Property;
import koLibRIMigrationServiceRegistry.ActionModuleDocument.ActionModule;
import koLibRIMigrationServiceRegistry.ActionModuleParameterDocument.ActionModuleParameter;
import koLibRIMigrationServiceRegistry.MigrationServiceDocument.MigrationService;
import koLibRIPolicies.ConfigDocument.Config;
import koLibRIPolicies.StepDocument.Step;

import de.langzeitarchivierung.kopal.Policy;
import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.formats.Uof;
import de.langzeitarchivierung.kopal.util.FormatRegistry;

/*******************************************************************************
 * Builds the migration policies according to defined policy templates.
 * 
 * @author Ludwig
 * @version 0.5
 ******************************************************************************/

public class MigrationPolicyBuilder
{

	// CONSTANTS **********************************************************

	private static final String					SOURCE_FILE										= "SOURCEFILE";
	private static final String					SOURCE_FILE_EXTENSION							= "SOURCEEXTENSION";
	private static final String					DESTINATIONFILE									= "DESTINATIONFILE";
	private static final String					DESTINATION_FILE_EXTENSION						= "DESTINATIONEXTENSION";

	private static final String					SOURCE_OBJECT_CHARACTER							= "SOURCE-OBJECT-CHARACTER";
	private static final String					PREVIOUSLY_MIGRATED_SOURCE_FILES_AND_LINKEDTO	= "MIGRATED-DEPENDENCIES";
	private static final String					LINKEDTO_DEPENDENCY_OF_SOURCEFILE				= "LINKEDTO-DEPENDENCIES";
	private static final String					MIGRATION_PROCESS_NAME_PREFIX					= "Migration";

	// CLASS VARIABLES (Static variables) *********************************

	protected static Logger						defaultLogger									= Logger
																										.getLogger("de.langzeitarchivierung.kopal");
	private static MigrationServiceRegistry		migServRegistry;
	private static MigrationPolicyTemplateStore	migPolTemplateStore;
	private static FormatRegistry				fRegistry										= new FormatRegistry();

	// MANIPULATION (Manipulation - what the object does) *****************

	/***************************************************************************
	 * Builds executable migration policies for the specific source and
	 * destination object characters and object package directories in the
	 * specified parent directory. All objects in the specified directory are
	 * regarded as probably depending on one another.
	 * 
	 * TODO The policy template for each object is chosen independently from the
	 * other objects. But to get the appropriate policy templates all depencies
	 * have to be considerd and analysed. Has
	 * migPolTemplateStore.getAppropriateMigPolTemplate() to be extended?
	 * 
	 * @param metaMigrationDir
	 * @param srcObjCharacter
	 * @param dstObjCharacter
	 * @return
	 * @throws Exception
	 **************************************************************************/
	public static List<ProcessData> buildPolicy(File metaMigrationDir,
			ObjCharacter srcObjCharacter, ObjCharacter dstObjCharacter)
			throws Exception {
		ArrayList<ProcessData> migProcs = new ArrayList<ProcessData>();

		// check all object directories
		File[] migrationDirs = metaMigrationDir.listFiles();
		for (int i = 0; i < migrationDirs.length; i++) {
			File migrationDir = migrationDirs[i];
			StringBuffer resultingPolicy = new StringBuffer();
			if (migrationDir.isDirectory()) {

				// get xmlBean of mets.xml
				Uof metsToCheck = new Uof();
				metsToCheck.useTemplate(new File(migrationDir, "mets.xml"));
				XmlObject migObject = (XmlObject) metsToCheck
						.getMetadataImplementation();

				// chose migration template
				MigrationPolicyTemplate migPolTempl = migPolTemplateStore
						.getAppropriateMigPolTemplate(migObject);
				if (migPolTempl == null) {
					// TODO make the exception object specific
					throw new Exception(
							"No appropriate migration policy template found");
				}
				defaultLogger.log(Level.INFO, "Policy template "
						+ migPolTempl.getName() + " choosen");

				// build the policy based on all policy fragments of this
				// template
				PolicyFragment[] polFrags = migPolTempl.getPolicy()
						.getPolicyFragmentArray();
				for (int j = 0; j < polFrags.length; j++) {
					PolicyFragment polFrag = polFrags[j];
					// if this policy fragment references a static policy
					if (polFrag.getType().equalsIgnoreCase(
							MigrationPolicyTemplateStore.POLICY_NAME)) {
						resultingPolicy = Policy.buildPolicyString(
								resultingPolicy, polFrag.getValueArray(0));
						// if we have to build a policy fragment depending on
						// the choosen migration service
					}
					else {
						// we look for an appropriate migration service
						MigrationService migServ = getMigrationService(polFrag,
								srcObjCharacter, dstObjCharacter);
						defaultLogger.log(Level.INFO,
								"MigrationService choosen: "
										+ migServ.getServiceName());

						// if the policy fragment has a "SOURCE-FILE" argument
						// we take it
						// to determine the files to migrate, otherwise we use
						// the source object character
						List<String> fileUrisToMigrate = new ArrayList<String>();
						String[] filesToMigrateQuery = getPolFragPropValue(
								polFrag.getConfig().getPropertyArray(),
								SOURCE_FILE);
						for (int index = 0; index < filesToMigrateQuery.length; index++) {
							String fileMigQuery = filesToMigrateQuery[index];
							fileUrisToMigrate.addAll(getFileUris(migObject,
									fileMigQuery));
						}
						// if the policy fragment had no "SOURCE-FILE" argument,
						if (fileUrisToMigrate.isEmpty()) {
							// we take the srcobjcharacter
							fileUrisToMigrate = getFileUris(migObject,
									srcObjCharacter);
						}


						defaultLogger.log(Level.INFO,
								"Number of files to migrate: "
										+ fileUrisToMigrate.size());

						// for each file to migrate we build the policy fragment
						// which calls the migration service
						for (Iterator fileNameIter = fileUrisToMigrate
								.iterator(); fileNameIter.hasNext();) {
							String fileUriToMigrate = (String) fileNameIter
									.next();
							resultingPolicy = Policy.buildPolicyString(
									resultingPolicy,
									buildMigrationServicePolicyFragment(
											migServ, migrationDir,
											fileUriToMigrate, srcObjCharacter,
											dstObjCharacter, migObject, polFrag
													.getConfig()
													.getPropertyArray()));
						}
					}
				}
			}
			// after processing one migration dir has been completed
			// TODO change migration process name to something more meaningful
			ProcessData newProc = new ProcessData(new Policy(resultingPolicy),
					MIGRATION_PROCESS_NAME_PREFIX + System.currentTimeMillis());
			newProc.setPathToContentFiles(migrationDir.getAbsolutePath());
			migProcs.add(newProc);
		}

		// after processing all migration dirs has been completed
		return migProcs;
	}

	// INTERNAL (Internal - implementation details, local classes, ...) **

	/***************************************************************************
	 * Returns a MigrationService which fits to the policy template and source
	 * and destination object character. It returns a MigrationService matching
	 * the requested service name, the service type or or the source and
	 * destination object character.
	 * 
	 * TODO At the moment the test is an OR condition. Maybe it should be AND?
	 * 
	 * @param polFrag
	 * @param srcObjCharacter
	 * @param dstObjCharacter
	 * @return
	 * @throws Exception
	 **************************************************************************/
	private static MigrationService getMigrationService(PolicyFragment polFrag,
			ObjCharacter srcObjCharacter, ObjCharacter dstObjCharacter)
			throws Exception {
		MigrationService migServ = null;
		// a specific service?
		if (polFrag.getType().equalsIgnoreCase(
				MigrationPolicyTemplateStore.SERVICE_NAME)) {
			migServ = migServRegistry
					.getServiceByName(polFrag.getValueArray(0));
			if (migServ == null) {
				throw new Exception("No migration service "
						+ polFrag.getValueArray(0) + " found");
			}
			// a specific service type?
		}
		else if (polFrag.getType().equalsIgnoreCase(
				MigrationPolicyTemplateStore.SERVICE_TYPE)) {

			if (polFrag.getValueArray().length == 0) {
				throw new Exception(
						"No migration service types specified in config file.");
			}
			String[] requiredTypes = polFrag.getValueArray();
			migServ = migServRegistry.getServiceByType(requiredTypes);
			if (migServ == null) {
				throw new Exception(
						"No migration service with the requested types found");
			}
			// a service matching the src and dst criteria?
		}
		else if (polFrag.getType().equalsIgnoreCase(
				MigrationPolicyTemplateStore.SERVICE_CRITERIA)) {
			migServ = migServRegistry.getServiceByCriteria(srcObjCharacter,
					dstObjCharacter);
			if (migServ == null) {
				throw new Exception("No migration service matching the source "
						+ "criteria " + srcObjCharacter
						+ " and the destination " + "criteria "
						+ dstObjCharacter + " was found");
			}
		}

		return migServ;
	}

	/***************************************************************************
	 * Builds a policy fragment which is configured to execute the migration of
	 * the specified path and destionation object character.
	 * 
	 * TODO this method signature is really ugly
	 * 
	 * @param migServ
	 * @param sourceFileUri
	 * @param dstObjCharacter
	 * @return
	 **************************************************************************/
	private static StringBuffer buildMigrationServicePolicyFragment(
			MigrationService migServ, File migrationDir, String sourceFileUri,
			ObjCharacter srcObjCharacter, ObjCharacter dstObjCharacter,
			XmlObject migObject, Property[] polFragProps) throws XmlException {
		ActionModule migServActModule = migServ.getActionModule();

		// build basic objects for call of migration service actionmodule
		Step parentStep = Step.Factory.newInstance();
		Step migrationStep = parentStep.addNewStep();
		migrationStep.setClass1(migServActModule.getActionModuleClassName());

		ActionModuleParameter[] migServParams = migServActModule
				.getActionModuleParameterArray();
		Config migStepConfig = null;

		// add config element if we need it for parameters
		if (migServParams.length > 0) {
			migStepConfig = migrationStep.addNewConfig();
		}

		// for each parameter
		for (int j = 0; j < migServParams.length; j++) {
			// add property
			koLibRIPolicies.PropertyDocument.Property migStepConfigProp = migStepConfig
					.addNewProperty();
			migStepConfigProp.setField(migServParams[j]
					.getActionModuleParameterField());

			// set all values of the just created migration step config property
			// TODO put this ugly for loop into its own method?
			String[] migStepConfigPropValue = migServParams[j]
					.getActionModuleParameterValueArray();
			for (int k = 0; k < migStepConfigPropValue.length; k++) {
				// if the config parameter field contains keywords replace them
				// with the appropriate
				if (migStepConfigPropValue[k].contains(SOURCE_FILE)) {
					String absFileName = getAbsoluteFileName(migrationDir,
							sourceFileUri);
					migStepConfigProp.addValue(absFileName);
					//
				}
				else if (migStepConfigPropValue[k].contains(DESTINATIONFILE)) {
					String targetFileExtension = null;
					String[] propValues = getPolFragPropValue(polFragProps,
							DESTINATIONFILE);
					// if it is defined in the policy fragement we interpret it
					// as one (!) file format identifier
					if (propValues != null && propValues.length != 0) {
						targetFileExtension = fRegistry.getFormatInfo(
								propValues[0]).getFileExtension();
						// if not the dst object character is used
					}
					else {
						targetFileExtension = fRegistry.getFormatInfo(
								dstObjCharacter.getCharacterisationValue())
								.getFileExtension();
					}
					// now add the absolute path as migration service argument
					String sourceFileName = getAbsoluteFileName(migrationDir,
							sourceFileUri);
					String fileNamePrefix = sourceFileName.substring(0,
							sourceFileName.length() - 3);

					// we must check if the destination file name collides with
					// an
					// already existing file
					int i = 0;
					String modifiedFilename = fileNamePrefix;
					while (new File(modifiedFilename + targetFileExtension)
							.exists()) {
						modifiedFilename = fileNamePrefix + i;
					}
					migStepConfigProp.addValue(modifiedFilename
							+ targetFileExtension);
					//
				}
				else if (migStepConfigPropValue[k]
						.contains(DESTINATION_FILE_EXTENSION)) {
					String[] propValues = getPolFragPropValue(polFragProps,
							DESTINATION_FILE_EXTENSION);
					// if file extension of the destination files are explicitly
					// defined
					if (propValues != null && propValues.length != 0) {
						for (int i = 0; i < propValues.length; i++) {
							migStepConfigProp.addValue(fRegistry.getFormatInfo(
									propValues[i]).getFileExtension());
						}
						// if not the destination object character is used
					}
					else {
						migStepConfigProp.addValue(fRegistry.getFormatInfo(
								dstObjCharacter.getCharacterisationValue())
								.getFileExtension());
					}
					//
				}
				else if (migStepConfigPropValue[k]
						.contains(SOURCE_FILE_EXTENSION)) {
					String[] propValues = getPolFragPropValue(polFragProps,
							SOURCE_FILE_EXTENSION);
					// if file extension of the destination files are explicitly
					// defined
					if (propValues != null && propValues.length != 0) {
						for (int i = 0; i < propValues.length; i++) {
							migStepConfigProp.addValue(fRegistry.getFormatInfo(
									propValues[i]).getFileExtension());
						}
						// if not the source object character is used
					}
					else {
						migStepConfigProp.addValue(fRegistry.getFormatInfo(
								srcObjCharacter.getCharacterisationValue())
								.getFileExtension());
					}

					// TODO not used at the moment
				}
				else if (migStepConfigPropValue[k]
						.contains(LINKEDTO_DEPENDENCY_OF_SOURCEFILE)) {
					System.out.println(LINKEDTO_DEPENDENCY_OF_SOURCEFILE);

					List<String> fileDependencies = getSourceFileLinkedToDependencies(
							migObject, sourceFileUri);
					for (Iterator iter = fileDependencies.iterator(); iter
							.hasNext();) {
						String fileDependency = (String) iter.next();
						System.out.println("fileDependency:" + fileDependency);
						migStepConfigProp.addValue(getAbsoluteFileName(
								migrationDir, fileDependency));
					}

					// TODO not used at the moment
				}
				else if (migStepConfigPropValue[k]
						.contains(PREVIOUSLY_MIGRATED_SOURCE_FILES_AND_LINKEDTO)) {

					System.out
							.println(PREVIOUSLY_MIGRATED_SOURCE_FILES_AND_LINKEDTO);
					List<String> fileDependencies = getSourceFileLinkedToDependencies(
							migObject, sourceFileUri);
					List<String> fileUris = new ArrayList<String>();
					String[] propValues = getPolFragPropValue(polFragProps,
							PREVIOUSLY_MIGRATED_SOURCE_FILES_AND_LINKEDTO);
					// if it is defined in the policy fragement
					if (propValues != null && propValues.length != 0) {
						for (int i = 0; i < propValues.length; i++) {
							fileUris.addAll(getFileUris(migObject,
									propValues[i]));
						}
						// if not the src object character is used
					}
					else {
						fileUris = getFileUris(migObject, srcObjCharacter);
					}

					// we take only those file uris which in both lists.
					fileUris.retainAll(fileDependencies);

					// now add all absolute paths as migration service argument
					for (Iterator iter = fileUris.iterator(); iter.hasNext();) {
						String fileUri = (String) iter.next();
						migStepConfigProp.addValue(getAbsoluteFileName(
								migrationDir, fileUri));
					}

					// TODO does this make sense?
				}
				else if (migStepConfigPropValue[k]
						.contains(SOURCE_OBJECT_CHARACTER)) {

					// the literal value
				}
				else {
					migStepConfigProp.addValue(migStepConfigPropValue[k]);
				}
			}
		}

		// insert placeholder for additional policy fragments
		migrationStep.addNewTemplate();

		return new StringBuffer(parentStep.toString());
	}

	/***************************************************************************
	 * Returns the property values of a policy fragment.
	 * 
	 * @param polFragProps
	 * @param propField
	 * @return
	 **************************************************************************/
	private static String[] getPolFragPropValue(Property[] polFragProps,
			String propField) {
		for (int i = 0; i < polFragProps.length; i++) {
			Property property = polFragProps[i];
			// if the config field of the migration service is defined in the
			// policy template
			if (property.getField().equalsIgnoreCase(propField)) {
				return property.getValueArray();
			}
		}

		return null;
	}

	/***************************************************************************
	 * Gets the uris of all files which are linked with the lmerfile:linkedto
	 * from the file with the specified uri.
	 * 
	 * @param migObject
	 * @param sourceFileUri
	 * @return
	 * @throws XmlException
	 **************************************************************************/
	private static List<String> getSourceFileLinkedToDependencies(
			XmlObject migObject, String sourceFileUri) throws XmlException {
		List<String> fileUrisOfDependencies = new ArrayList<String>();
		String techId = getTechId(migObject, sourceFileUri);
		System.out.println("sourceFileUri:" + sourceFileUri);
		System.out.println("techId:" + techId);


		XmlObject sourceFileTechNode = FileObjCharacterTranslator
				.getQueryResults(migObject,
						new ObjCharacter(ObjCharacter.TECHMD_ID_QUERY, techId))
				.get(0);
		System.out.println(sourceFileTechNode);


		List<XmlObject> fileResults = FileObjCharacterTranslator
				.getQueryResults(sourceFileTechNode, new ObjCharacter(
						ObjCharacter.FILE_LINKEDTO_QUERY, null));
		for (Iterator iter = fileResults.iterator(); iter.hasNext();) {
			XmlObject depObject = (XmlObject) iter.next();
			System.out.println(depObject.getDomNode().getFirstChild()
					.getNodeValue());
			String depUri = getFileUriForFileId(migObject, depObject
					.getDomNode().getFirstChild().getNodeValue());
			System.out.println("dependency:" + depUri);
			fileUrisOfDependencies.add(depUri);
		}

		return fileUrisOfDependencies;
	}

	/***************************************************************************
	 * Gets the fileId for the file the specified uri.
	 * 
	 * @param migObject
	 * @param sourceFileUri
	 * @return
	 * @throws XmlException
	 **************************************************************************/
	private static String getFileId(XmlObject migObject, String sourceFileUri)
			throws XmlException {
		List<XmlObject> fileResults = FileObjCharacterTranslator
				.getQueryResults(migObject, new ObjCharacter(
						ObjCharacter.FILES_WITHIN_FILEGRP_QUERY, null));
		for (Iterator iter = fileResults.iterator(); iter.hasNext();) {
			gov.loc.mets.FileGrpType.File fileGrpEntry = (gov.loc.mets.FileGrpType.File) iter
					.next();
			if (fileGrpEntry.getFLocatArray(0).getHref().equalsIgnoreCase(
					sourceFileUri)) {
				return fileGrpEntry.getID();
			}
		}

		return null;
	}

	/***************************************************************************
	 * Gets the techId for the file the specified uri.
	 * 
	 * @param migObject
	 * @param sourceFileUri
	 * @return
	 * @throws XmlException
	 **************************************************************************/
	private static String getTechId(XmlObject migObject, String sourceFileUri)
			throws XmlException {
		List<XmlObject> fileResults = FileObjCharacterTranslator
				.getQueryResults(migObject, new ObjCharacter(
						ObjCharacter.FILES_WITHIN_FILEGRP_QUERY, null));
		for (Iterator iter = fileResults.iterator(); iter.hasNext();) {
			gov.loc.mets.FileGrpType.File fileGrpEntry = (gov.loc.mets.FileGrpType.File) iter
					.next();
			if (fileGrpEntry.getFLocatArray(0).getHref().equalsIgnoreCase(
					sourceFileUri)) {
				return (String) fileGrpEntry.getADMID().get(0);
			}
		}

		return null;
	}

	/***************************************************************************
	 * Returns the absolute file name for a relative file uri of the uof.
	 * 
	 * @param migrationDir
	 * @param fileUri
	 * @return
	 **************************************************************************/
	private static String getAbsoluteFileName(File migrationDir, String fileUri) {
		return new File(migrationDir, fileUri
				.substring(Uof.FILE_LOCAT_HREF_PREFIX.length() - 1))
				.getAbsolutePath();
	}

	/***************************************************************************
	 * Returns a list of all file uris in the mets file for a query which
	 * results in a hit corresponding to a file. For example in the techmd.
	 * 
	 * @param migObject
	 * @param fileQuery
	 * @return
	 * @throws XmlException
	 **************************************************************************/
	private static List<String> getFileUris(XmlObject migObject,
			String fileQuery) throws XmlException {
		List<String> fileUrisToMigrate = new ArrayList<String>();

		// determine which files have to be migrated
		// TODO has this to be done somewhere else? this cannot be done by the
		// char query because it
		// doesn't necessarily return the files to migrate?
		XmlObject[] fileHits = null;
		try {
			fileHits = migObject.selectPath(fileQuery);
		}
		catch (RuntimeException e) {
			throw new XmlException(
					"XPath Syntax of policy template is invalid: "
							+ e.getMessage());
		}

		for (int i = 0; i < fileHits.length; i++) {
			fileUrisToMigrate.add(getFileUriOfAncestorNode(migObject,
					fileHits[i]));
		}

		return fileUrisToMigrate;
	}

	/***************************************************************************
	 * Returns the file uris of an mets file which match the file hits of the
	 * object characterisation.
	 * 
	 * @param migObject the mets to query
	 * @param fileQuery object characterisation which matches a file
	 * @return list of file uris
	 * @throws XmlException
	 **************************************************************************/
	private static List<String> getFileUris(XmlObject migObject,
			ObjCharacter fileObjQuery) throws XmlException {
		List<String> fileUrisToMigrate = new ArrayList<String>();

		// determine which files have to be migrated
		XmlObject[] fileHits = null;
		try {
			fileHits = migObject.selectPath(fileObjQuery
					.getCharacterisationQuery());
		}
		catch (RuntimeException e) {
			throw new XmlException(
					"XPath Syntax of policy template is invalid: "
							+ e.getMessage());
		}

		for (int i = 0; i < fileHits.length; i++) {
			XmlObject fileHit = fileHits[i];
			// if the query result matches the objectcharacter get the
			// corresponding fileur
			if (fileHit.getDomNode().getFirstChild().getNodeValue()
					.equalsIgnoreCase(fileObjQuery.getCharacterisationValue())) {
				fileUrisToMigrate.add(getFileUriOfAncestorNode(migObject,
						fileHit));
			}
		}

		return fileUrisToMigrate;
	}

	/***************************************************************************
	 * Returns the uri for the file which belongs to the node specified.
	 * 
	 * @param migObject
	 * @param fileHit
	 * @return
	 * @throws XmlException
	 **************************************************************************/
	private static String getFileUriOfAncestorNode(XmlObject migObject,
			XmlObject fileHit) throws XmlException {
		XmlObject[] techIdResults = fileHit
				.selectPath(ObjCharacter.ANCESTOR_TECH_ID_QUERY);

		// in mets there is exact one techMD node in any path, so we can safely
		// choose just the
		// the first result
		String techId = ((MdSecType) techIdResults[0]).getID();

		List<XmlObject> fileResults = FileObjCharacterTranslator
				.getQueryResults(migObject, new ObjCharacter(
						ObjCharacter.FILEGRP_FILE_TECHID_QUERY, techId));

		// there should be exactly one filegroup entry with this techId
		if (fileResults.size() != 1) {
			throw new XmlException(
					"More than one filegroup entry with this techId");
		}

		gov.loc.mets.FileGrpType.File fileGrpEntry = (gov.loc.mets.FileGrpType.File) fileResults
				.get(0).selectPath("./..")[0];

		return fileGrpEntry.getFLocatArray(0).getHref();
	}

	/***************************************************************************
	 * Returns the uri for the file which belongs to the node specified.
	 * 
	 * @param migObject
	 * @param fileHit
	 * @return
	 * @throws XmlException
	 **************************************************************************/
	private static String getFileUriForFileId(XmlObject migObject, String fileId)
			throws XmlException {

		List<XmlObject> fileResults = FileObjCharacterTranslator
				.getQueryResults(migObject, new ObjCharacter(
						ObjCharacter.FILEGRP_FILEID_QUERY, fileId));

		// there should be exactly one filegroup entry with this techId
		if (fileResults.size() != 1) {
			throw new XmlException(
					"More than one filegroup entry with this fileId");
		}


		gov.loc.mets.FileGrpType.File fileGrpEntry = (gov.loc.mets.FileGrpType.File) fileResults
				.get(0).selectPath("./..")[0];
		return fileGrpEntry.getFLocatArray(0).getHref();
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * Sets config file and initializes the MigrationPolicyTemplateStore.
	 * 
	 * @param migrationPolicyTemplatesFileName
	 * @throws XmlException
	 * @throws IOException
	 **************************************************************************/
	public static void setMigrationPolicyTemplatesFileName(
			String migrationPolicyTemplatesFileName) throws XmlException,
			IOException {
		migPolTemplateStore = new MigrationPolicyTemplateStore(new File(
				migrationPolicyTemplatesFileName));
	}

	/***************************************************************************
	 * Sets config file and initializes the MigrationServiceRegistry.
	 * 
	 * @param migrationServicesFileName
	 * @throws XmlException
	 * @throws IOException
	 **************************************************************************/
	public static void setMigrationServicesFileName(
			String migrationServicesFileName) throws XmlException, IOException {
		migServRegistry = new MigrationServiceRegistry(new File(
				migrationServicesFileName));
	}

}
