/*******************************************************************************
 * de.langzeitarchivierung.kopal.processstarter.migrationmanager / MigrationPolicyTemplateStore.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 *
 *
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.processstarter.migrationmanager;


import java.io.File;
import java.io.IOException;

import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;

import koLibRIMigrationPolicyTemplates.MigrationPolicyTemplatesDocument;
import koLibRIMigrationPolicyTemplates.CriteriaDocument.Criteria;
import koLibRIMigrationPolicyTemplates.MigrationPolicyTemplateDocument.MigrationPolicyTemplate;
import koLibRIMigrationPolicyTemplates.MigrationPolicyTemplatesDocument.MigrationPolicyTemplates;

/*******************************************************************************
 * Stores the migration policy templates and provides methods for getting the
 * appropriate one.
 * 
 * @author Ludwig
 * @version 0.5
 ******************************************************************************/

public class MigrationPolicyTemplateStore
{

	// CONSTANTS **********************************************************

	public static final String	POLICY_NAME			= "name";
	public static final String	SERVICE_NAME		= "serviceName";
	public static final String	SERVICE_TYPE		= "serviceType";
	public static final String	SERVICE_CRITERIA	= "serviceCriteria";

	// STATE (Instance variables) *****************************************

	MigrationPolicyTemplate[]	migPolTemplates		= null;

	// MANIPULATION (Manipulation - what the object does) *****************

	/***************************************************************************
	 * Constructor which reads the MigrationPolicyTemplate xml file.
	 * 
	 * @param policyTemplateFile
	 * @throws XmlException
	 * @throws IOException
	 **************************************************************************/
	public MigrationPolicyTemplateStore(File policyTemplateFile)
			throws XmlException, IOException {
		super();
		MigrationPolicyTemplatesDocument templatesDoc = MigrationPolicyTemplatesDocument.Factory
				.parse(policyTemplateFile);
		MigrationPolicyTemplates migPols = templatesDoc
				.getMigrationPolicyTemplates();
		this.migPolTemplates = migPols.getMigrationPolicyTemplateArray();
	}

	/***************************************************************************
	 * Get a MigrationPolicyTemplate by name.
	 * 
	 * @param migPolTemplateName
	 * @return
	 **************************************************************************/
	public MigrationPolicyTemplate getMigPolTemplate(String migPolTemplateName) {
		for (int i = 0; i < this.migPolTemplates.length; i++) {
			if (this.migPolTemplates[i].getName().equalsIgnoreCase(
					migPolTemplateName)) {
				return this.migPolTemplates[i];
			}
		}

		return null;
	}

	/***************************************************************************
	 * Returns the first appropriate migration policy template. It is
	 * appropriate if all its criterias are fulfilled if applied to the
	 * migration object.
	 * 
	 * TODO The policy template for each object is chosen independently from the
	 * other objects. But to get the appropriate policy templates all depencies
	 * have to be considerd and analysed.
	 * 
	 * @param migObj xmlbeans representation of the mets.xml TODO use something
	 *            different than an XmlObject in the method signature?
	 * 
	 * @return the appropriate MigrationPolicyTemplate
	 **************************************************************************/
	public MigrationPolicyTemplate getAppropriateMigPolTemplate(XmlObject migObj)
			throws XmlException {
		for (int i = 0; i < this.migPolTemplates.length; i++) {
			if (checkMigPolTemplate(this.migPolTemplates[i], migObj)) {
				return this.migPolTemplates[i];
			}
		}

		return null;
	}

	// INTERNAL (Internal - implementation details, local classes, ...) **

	/***************************************************************************
	 * Checks if a migration policy template matches the migraton candidate.
	 * 
	 * @param migPolTemplate
	 * @param migObj
	 * @return true if policy can be applied, false otherwise
	 **************************************************************************/
	private boolean checkMigPolTemplate(MigrationPolicyTemplate migPolTemplate,
			XmlObject migObj) throws XmlException {
		Criteria[] polCriteria = migPolTemplate.getCriteriaArray();

		// check all criterias of this template
		for (int j = 0; j < polCriteria.length; j++) {

			XmlObject[] critResult = null;
			try {
				critResult = migObj.selectPath(polCriteria[j].getXpath());
			}
			catch (RuntimeException e) {
				throw new XmlException(
						"XPath Syntax of policy template is invalid: "
								+ e.getMessage());
			}

			// one query results has to match the specified criteria result
			boolean critFulfilled = false;
			for (int k = 0; k < critResult.length; k++) {
				if (critResult[k].getDomNode().getFirstChild() != null
						&& critResult[k].getDomNode().getFirstChild()
								.getNodeValue().equals(
										polCriteria[j].getResult())) {
					critFulfilled = true;
					break;
				}
			}

			// this policy template doesn't match if no query result produced
			// the specified criteria result
			if (!critFulfilled) {
				return false;
			}
		}

		return true;
	}

}
