/**********************************************************************
 * Class AccessLayer
 * Copyright 2005-2007 by Project KOPAL
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA
 ******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 14.06.2007	Modifications to return ResponseElemDocument
 * 03.11.2006	Initial version
 ******************************************************************************/

package de.langzeitarchivierung.kopal.retrieval;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;

import org.apache.axis2.AxisFault;

import de.langzeitarchivierung.kopal.util.Configurator;
import de.langzeitarchivierung.kopal.util.KopalXMLParser;
import de.langzeitarchivierung.kopal.util.WSUtils;
import de.langzeitarchivierung.kopal.util.FileUtils;
import de.langzeitarchivierung.kopal.ws.wsdl2java.ResponseElemDocument;
import de.langzeitarchivierung.kopal.ws.wsdl2java.UserType;

/** ********************************************************************
 * Class to provide web service access to DIAS-Core sytem.
 * For more information about how it works, and why, see the
 * DIP Interface Specification provided by IBM-Holland.
 * 
 * @author  Kadir Karaca Koçer, German National Library
 * @version 20070614
 * @see DiasAccess
 * @see DiasAccessResponse
 * @see WSUtils
 * @see ResponseElemDocumentImpl
 * @since 20061103
 **********************************************************************/

public class AccessLayer {
//	CONSTANTS  **********************************************************
//	CLASS VARIABLES (Static variables) **********************************
	//force Administrator to initialise cache and give retrieval free
	private static boolean isRetrievalActive = false;
	private static boolean CACHE_INITIALISED = false;
	//TODO: transfer those hard-coded strings to config file.
	private static String  ABSOLUTE_ROOT_DIR = "c:/temp/cache/";
	private static String  PROTOCOL = "ftp://";
	private static String  BASE_URL = "bernstein.d-nb.de";
	private static String  RELATIVE_ROOT_DIR = "/ftp/download";


//	GET & SET METHODS  **************************************************
	private static void setCacheInitialized(boolean status){
		CACHE_INITIALISED = status;
	}
	/** ********************************************************************
	 * Method to set the used transfer protocol.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param newProtocol Type of the protocol (ftp://, http://, file://  )
	 * @version 20070614
	 * @since 20061106
	 **********************************************************************/
	public static void setProtocol (String newProtocol){
		PROTOCOL = newProtocol;
	}
	
	/** ********************************************************************
	 * Method to set the base URL of this server.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param newURL URL of this server.
	 * @version 20070530
	 * @since 20061106
	 **********************************************************************/
	public static void setBaseURL (String newURL){
		BASE_URL = newURL;
	}
	
	
	/** ********************************************************************
	 * Method to set the relative path to the downloaded files
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param newRelative Relative (to the server root) path to
	 *  root directory the downloaded files
	 * @version 20070530
	 * @since 20061106
	 **********************************************************************/
	public static void setRelativeRootDir (String newRelative){
		RELATIVE_ROOT_DIR = newRelative;
	}
	
	
	/** ********************************************************************
	 * Method to set the absolute path to the root directory of the downloaded files
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param newAbsolute Absolute path to the downloaded files
	 * @version 20070530
	 * @since 20061106
	 **********************************************************************/
	public static void setAbsoluteRootDir (String newAbsolute){
		ABSOLUTE_ROOT_DIR = newAbsolute;
	}
	
	
//	QUERIES (Queries - no change to object's state)  ********************

	/** ********************************************************************
	 * Method to get the METS file of a DIP from DIAS using the class DiasAccess
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param persistentIdentifier Unique ID of DIP
	 * @param dias Connection to DIAS
	 * @return The response object
	 * @see DiasAccess
	 * @see WSUtils
	 * @see ResponseElemDocument
	 * @version 20070615
	 * @since 20061103
	 **********************************************************************/
	public static ResponseElemDocument getMETS(UserType user, String persistentIdentifier) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "getMETS");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		
		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (isRetrievalActive){
			if (!CACHE_INITIALISED) {
				response.getResponseElem().setStatusCode(596);
				response.getResponseElem().setResponseText("Cache not initialised! Please call initCache() method first.");
				return response;
			}
			DiasAccessResponse diasResponse = new DiasAccessResponse();
			java.io.File mets = null;
			try {
				//Read the config file and configure the connection to DIAS 
				Configurator.setConfigDocument(KopalXMLParser.getDocument(WSUtils.CONFIG_FILE));
				//create a connection to DIAS
				DiasAccess dias = new DiasAccess();
				Configurator.configure(dias);
				//contact DIAS and send request
				System.out.println("Asking DIAS for URN: " + persistentIdentifier);
				int trial = 1;
				String type = "";
				String fullAnswer = dias.requestFullMdByExtId(persistentIdentifier, "xml", "zip");
				//System.out.println(fullAnswer);
				diasResponse = DiasAccess.parseDiasResponse(fullAnswer);
				          			//new java.io.StringReader();
				type = diasResponse.getResponseType();
				
				while (!type.equals(DiasAccessResponse.RESPONSE_TYPE_DIP) && trial < 5) {
					//DIAS response is not o.k.
					//check if an error occured. If yes return error code and exit
					if (type.equals(DiasAccessResponse.RESPONSE_TYPE_ERROR)) {
						response.getResponseElem().setStatusCode(595);
						response.getResponseElem().setResponseText("DIAS returned the error: " + diasResponse.getDescription());
						//TODO: check if the document exists
						//if not 404 file not found!
						return response;
					}
					//No error: DIAS was busy. Retry
					System.gc();
					diasResponse = DiasAccess.parseDiasResponse(
				                  new java.io.StringReader(dias.requestFullMdByExtId(
				                                         persistentIdentifier, "xml", "zip")));
					type = diasResponse.getResponseType();
					trial++;
				}
				if (!type.equals(DiasAccessResponse.RESPONSE_TYPE_DIP)) {
					//DIAS is still busy
					response.getResponseElem().setStatusCode(590);
					response.getResponseElem().setResponseText("DIAS is overloaded. Please try again later.");

					return response;
				}
				String internalId = diasResponse.getInternalAssetId();
				System.out.println("DIAS responded. Internal ID is: "+ internalId);
				
				response.getResponseElem().setStatusCode(240);
				response.getResponseElem().setResponseText("Metadata succesfully downloaded. Link follows.");
				response.getResponseElem().setDiasId(internalId);
				
				//Download the mets.zip if it is not in Cache
				String dir = WSUtils.internalIdToDir(internalId);
				if (WSUtils.fileAlreadyThere(ABSOLUTE_ROOT_DIR + dir + "mets.zip")) {
					//This mets file is already there: no need to download it again. Just refresh the statistics
					//TODO: refresh the statistics
					System.out.println("Asset with internal ID : "+ internalId + " is already in cache :-)");
					
				} else {
					// This mets file is not in cache: download it from DIAS
					System.out.println("Asset with internal ID : "+ internalId + " is not in cache.\nDownloading ...");
					mets = downloadDIASFile(dias, diasResponse, ABSOLUTE_ROOT_DIR + dir);
			        mets.renameTo(new java.io.File(mets.getParent(), "mets.zip"));
			        //ask DIAS to "prepare" the DIP, assuming it *will* be asked soon
			        dias.requestAsset(internalId, "xml", "tar");
				}
				//response.setLinkToFile(PROTOCOL + BASE_URL + RELATIVE_ROOT_DIR + dir + "mets.zip");
				response.getResponseElem().setLinkToFile(PROTOCOL + BASE_URL + RELATIVE_ROOT_DIR + dir + "mets.zip");
			} catch (Exception e){
				//Exception. Return an error message.
				response.getResponseElem().setStatusCode(551);
				response.getResponseElem().setResponseText("Exception in AccesLayer.getMETS()." +
				            " Please consult the administrator.\n Exception Message: " + e.getMessage());
				e.printStackTrace();
			} // end try-catch
		} else {
			//Retrieval is not running. Return an error message.
			response.getResponseElem().setStatusCode(550);
			response.getResponseElem().setResponseText("Retrieval is not running. Please consult the administrator.");
		}
		
		return response;
	}	

	
	/** ********************************************************************
	 * Method to get a specific file from a DIP in DIAS
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param dias Connection to DIAS
	 * @param diasInternalID Unique ID of DIP
	 * @param fileName Name of the requested File
	 * @return The response object
	 * @see DiasAccess
	 * @see ResponseElemDocument
	 * @version 20070614
	 * @since 20061103
	 **********************************************************************/
	public static ResponseElemDocument getFile(UserType user, String diasInternalID, String fileName) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "getFile");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		
		ResponseElemDocument response = WSUtils.createDefaultResponse();

		if (isRetrievalActive){
			//Check if the cache is there and initialised
			if (!CACHE_INITIALISED) {
				response.getResponseElem().setStatusCode(596);
				response.getResponseElem().setResponseText("Cache not initialised! Please call initCache() method first.");
				return response;
			}
			
			String[] dirs = WSUtils.internalIdToDirs(diasInternalID);
			//look if the requested file already in cache directory is
			String assetPath = dirs[0] + dirs[1] + dirs[2] + dirs[3];
			java.io.File f = new java.io.File(checkPath(ABSOLUTE_ROOT_DIR + assetPath + "/" +  fileName));
			if (f.exists()) {
				//the requested file is already there unpacked
				f = new java.io.File(checkPath(ABSOLUTE_ROOT_DIR + assetPath));
				//update the Timestamp of the parent directory of this asset
				f.setLastModified(System.currentTimeMillis());
				response.getResponseElem().setStatusCode(250);
				response.getResponseElem().setResponseText("File is already in cache. Link follows.");
				response.getResponseElem().setDiasId(diasInternalID);
				response.getResponseElem().setLinkToFile(PROTOCOL + BASE_URL + RELATIVE_ROOT_DIR +  assetPath + fileName);
				return response;
			}
			System.out.println(f.getPath() + " is not in cache! Looking for .tar.");
			String dipDir = dirs[0] + dirs[1] + "dips/";
			f = new java.io.File(checkPath(ABSOLUTE_ROOT_DIR + dipDir + dirs[3] + ".tar"));
			if (f.exists()) {
				//requested file is not unpacked here but the DIP as TAR archive unpack it
				response.getResponseElem().setStatusCode(251);
				response.getResponseElem().setResponseText("DIP is already in cache. Link follows.");
			} else {
				//its not there :-( Get it:
				try {
					//Read the config file and configure the connection to DIAS 
					Configurator.setConfigDocument(KopalXMLParser.getDocument(WSUtils.CONFIG_FILE));
					//create a connection to DIAS
					DiasAccess dias = new DiasAccess();
					Configurator.configure(dias);
					//request the asset as tar and the response als xml
					String fullAnswer = dias.requestAsset(diasInternalID, "xml", "tar");
					DiasAccessResponse diasResponse = DiasAccess.parseDiasResponse(fullAnswer);
					
					//check if response is o.k.
					String type = diasResponse.getResponseType();
					if (type.equals(DiasAccessResponse.RESPONSE_TYPE_ERROR)) {
						response.getResponseElem().setStatusCode(595);
						response.getResponseElem().setResponseText("DIAS returned the error: " + diasResponse.getDescription());
						return response;
					}
					//No error. Check if DIAS was busy.
					if (type.equals(DiasAccessResponse.RESPONSE_TYPE_BUSY)) {
						response.getResponseElem().setStatusCode(247);
						response.getResponseElem().setResponseText("Your request is already processed by DIAS. See estimated processing time"); 
						int wt = new Integer(diasResponse.getMetadata().get(0).getDipRetrieveWaitTime()).intValue();
						response.getResponseElem().setWaitTime(wt);
						return response;
					}
					 f = downloadDIASFile(dias, diasResponse, checkPath(ABSOLUTE_ROOT_DIR + dipDir));
					//download successfull.
				} catch (Exception e) {
					//Exception. Return an error message.
					response.getResponseElem().setStatusCode(552);
					response.getResponseElem().setResponseText("Exception in AccesLayer.getFile()." +
					            " Please consult the administrator.\n Exception Message: " + e.getMessage());
					e.printStackTrace();
				}
			}
			//untar the dip then delete .tar
			int tarStatus = FileUtils.untar(f, new java.io.File(checkPath(ABSOLUTE_ROOT_DIR + assetPath)));
			if (tarStatus > 0) {
				//error during untar!
				response.getResponseElem().setStatusCode(581);
				response.getResponseElem().setResponseText("Error during untar! Errorcode = " + tarStatus);
				//delete the tar file
				f.delete();
				return response;
			}
			
			//delete the tar file
			f.delete();
			
			//look if the requested file there is
			f = new java.io.File(checkPath(ABSOLUTE_ROOT_DIR + assetPath + "/" + fileName));
			if (!f.exists()) {
				//Error! No such file. Return the error code & string
				response.getResponseElem().setStatusCode(582);
				response.getResponseElem().setResponseText("The requested file " + f.getPath() + " is not in this digital object.");
				return response;
			}
			//the file is there deliver the link
			response.getResponseElem().setStatusCode(245);
			response.getResponseElem().setResponseText("File succesfully downloaded. Link follows.");
			response.getResponseElem().setDiasId(diasInternalID);
			response.getResponseElem().setLinkToFile(PROTOCOL + BASE_URL + RELATIVE_ROOT_DIR + assetPath + "/" + fileName);
		} 
		else {
			//Retrieval is not running. Return an error message.
			response.getResponseElem().setStatusCode(550);
			response.getResponseElem().setResponseText("Retrieval is not running. Please consult the administrator.");
		}
		return response;
	}	

	
	/** ********************************************************************
	 * Method to get a complete DIP from DIAS
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param diasInternalID Unique ID of DIP
	 * @param dias Connection to DIAS
	 * @return The response object
	 * @see DiasAccess
	 * @see ResponseElemDocument
	 * @version 20070614
	 * @since 20061103
	 **********************************************************************/
	public static ResponseElemDocument getDIP(UserType user, String diasInternalID) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "getDIP");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (isRetrievalActive){
			//Check if the cache is there and initialised
			if (!CACHE_INITIALISED) {
				response.getResponseElem().setStatusCode(596);
				response.getResponseElem().setResponseText("Cache not initialised! Please call initCache() method first.");
				return response;
			}
			
			DiasAccessResponse diasResponse = null;
			java.io.File dip = null;
			try {
				response.getResponseElem().setStatusCode(246);
				response.getResponseElem().setResponseText("DIP succesfully downloaded. Link follows.");
				response.getResponseElem().setDiasId(diasInternalID);
				
				//create the path
				String[] dirs = WSUtils.internalIdToDirs(diasInternalID);
				String assetPath = checkPath(ABSOLUTE_ROOT_DIR + dirs[0] + dirs[1] + "dips/");
				System.out.println(assetPath + dirs[3] + ".tar");
				
				//look if the DIP is already in the appropriate DIP directory
				if (WSUtils.fileAlreadyThere(assetPath + dirs[3] + ".tar")) {
					//This asset is already there: no need to download it again. Just refresh the statistics
					//TODO: refresh the statistics
					System.out.println("Asset with internal ID : "+ diasInternalID + " is already in cache :-)");
					
				} else {
					//its not in cache. Contact DIAS and send request
					System.out.println("Asking DIAS for DIP: " + diasInternalID);

					//Read the config file and configure the connection to DIAS 
					Configurator.setConfigDocument(KopalXMLParser.getDocument(WSUtils.CONFIG_FILE));
					//create a connection to DIAS
					DiasAccess dias = new DiasAccess();
					Configurator.configure(dias);
					diasResponse = DiasAccess.parseDiasResponse(
					                  new java.io.StringReader(dias.requestAsset(
					                                         diasInternalID, "xml", "tar")));
					//check if response is o.k.
					String type = diasResponse.getResponseType();
					if (type.equals(DiasAccessResponse.RESPONSE_TYPE_ERROR)) {
						response.getResponseElem().setStatusCode(595);
						response.getResponseElem().setResponseText("DIAS returned the error: " +
						                                           diasResponse.getDescription());
						return response;
					}
					//No error. Check if DIAS was busy.
					if (type.equals(DiasAccessResponse.RESPONSE_TYPE_BUSY)) {
						response.getResponseElem().setStatusCode(247);
						//long waitTime = Long.parseLong(diasResponse.getMetadata().get(0).getDipRetrieveWaitTime());
						//response.setWaitTime(waitTime);
						response.getResponseElem().setResponseText("Your request is being processed by DIAS. Please try later");
						return response;
					}
					//Download the file
					dip = downloadDIASFile(dias, diasResponse, assetPath);
			        dip.renameTo(new java.io.File(dip.getParent(), dirs[3] + ".tar"));
				}
				response.getResponseElem().setLinkToFile(PROTOCOL + BASE_URL + RELATIVE_ROOT_DIR +  dirs[0] + dirs[1] + "dips/" + dirs[3] + ".tar");
			} catch (NumberFormatException nfe){
				//Exception. Return an error message.
				response.getResponseElem().setStatusCode(553);
				response.getResponseElem().setResponseText("NumberFormatException in AccesLayer.getDIP()." +
				                                           "Please consult the administrator.");
				response.getResponseElem().setErrorMessage(nfe.getMessage());
			} // end try-catch
			catch (Exception e){
				//Exception. Return an error message.
				response.getResponseElem().setStatusCode(554);
				response.getResponseElem().setResponseText("Exception in AccesLayer.getDIP(). " +
				                                           "Please consult the administrator.");
				response.getResponseElem().setErrorMessage(e.getMessage());
			} // end try-catch
		} else {
			//Retrieval is not running. Return an error message.
			response.getResponseElem().setStatusCode(550);
			response.getResponseElem().setResponseText("Retrieval is not running. Please consult the administrator.");
		}
		return response;
	}
	
	
	/** ********************************************************************
	 * Method to get a complete DIP from DIAS for the external institutions
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param urn Unique ID of DIP
	 * @param dias Connection to DIAS
	 * @return The response object
	 * @see DiasAccess
	 * @see ResponseElemDocument
	 * @version 20070614
	 * @since 20061103
	 **********************************************************************/
	public static ResponseElemDocument getDIPExternal(UserType user, String urn) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "getDIPExternal");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		//Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (isRetrievalActive){
			try {
				//Read the config file and configure the connection to DIAS 
				Configurator.setConfigDocument(KopalXMLParser.getDocument(WSUtils.CONFIG_FILE));
				//create a connection to DIAS
				DiasAccess dias = new DiasAccess();
				Configurator.configure(dias);
				
				//TODO: implement the algorithm here
				System.out.println(urn);
			}
			catch (InvocationTargetException e) {
				//Return an error message explaining exception.
				response.getResponseElem().setStatusCode(561);
				response.getResponseElem().setResponseText("InvocationTargetException in AccessLayer.getDIPExternal().");
				response.getResponseElem().setErrorMessage(e.getMessage()); 
				e.printStackTrace();
			}
			catch (IllegalAccessException e) {
				//Return an error message explaining exception.
				response.getResponseElem().setStatusCode(562);
				response.getResponseElem().setResponseText("IllegalAccessException in AccessLayer.getDIPExternal().");
				response.getResponseElem().setErrorMessage(e.getMessage()); 
				e.printStackTrace();
			}
			/*
			catch (NoSuchMethodException e) {
				//Return an error message explaining exception.
				response.getResponseElem().setStatusCode(563);
				response.getResponseElem().setResponseText("NoSuchMethodException in AccessLayer.getDIPExternal().");
				response.getResponseElem().setErrorMessage(e.getMessage()); 
				e.printStackTrace();
			}*/
			
		} else {
			//Retrieval is still running. Return an error message.
			response.getResponseElem().setStatusCode(550);
			response.getResponseElem().setResponseText("Retrieval is not running. Please consult the administrator.");
		}
		
		return response;
	}
	
	
	/** ********************************************************************
	 * Gets the version list of an object with a given URN
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param dias Connection to DIAS
	 * @param externalAssetID Unique ID of DIP
	 * @return The response object
	 * @see DiasAccess
	 * @see ResponseElemDocument
	 * @version 20070530
	 * @since 20061103
	 **********************************************************************/
	public static ResponseElemDocument getHistory(UserType user, String externalAssetID) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "getHistory");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		
		ResponseElemDocument response = WSUtils.createDefaultResponse();
	
		if (isRetrievalActive){
			//TODO: implement the logic. Something like that
			try {
				//Read the config file and configure the connection to DIAS 
				Configurator.setConfigDocument(KopalXMLParser.getDocument(WSUtils.CONFIG_FILE));
				//create a connection to DIAS
				DiasAccess dias = new DiasAccess();
				Configurator.configure(dias);
				System.out.println(dias.getKeyStoreFile());
				System.out.println(dias.requestShortMdByExtId(externalAssetID, "xml", false));
				//process the request and write in response
			}
			catch (IOException e) {
				//Return an error message explaining exception.
				response.getResponseElem().setStatusCode(570);
				response.getResponseElem().setResponseText("IOException in AccessLayer.getHistory().");
				response.getResponseElem().setErrorMessage(e.getMessage()); 
				e.printStackTrace();
			}
			catch (InvocationTargetException e) {
				//Return an error message explaining exception.
				response.getResponseElem().setStatusCode(571);
				response.getResponseElem().setResponseText("InvocationTargetException in AccessLayer.getHistory().");
				response.getResponseElem().setErrorMessage(e.getMessage()); 
				e.printStackTrace();
			}
			catch (IllegalAccessException e) {
				//Return an error message explaining exception.
				response.getResponseElem().setStatusCode(572);
				response.getResponseElem().setResponseText("IllegalAccessException in AccessLayer.getHistory().");
				response.getResponseElem().setErrorMessage(e.getMessage()); 
				e.printStackTrace();
			}/*
			catch (NoSuchMethodException e) {
				//Return an error message explaining exception.
				response.getResponseElem().setStatusCode(573);
				response.getResponseElem().setResponseText("NoSuchMethodException in AccessLayer.getHistory().");
				response.getResponseElem().setErrorMessage(e.getMessage()); 
				e.printStackTrace();
			}*/
		} else {
			//Retrieval is not running. Return an error message.
			response.getResponseElem().setStatusCode(550);
			response.getResponseElem().setResponseText("Retrieval is not running. Please consult the administrator.");
		}
		return response;
	}	
	
	
	/** ********************************************************************
	 * Stop the retrieval service. Indeed it does not stop the service.
	 * It just "marks" it stopped to avoid further request.
	 * So can an administrator really shut down the Computer after the last
	 * user got his response.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return The response object
	 * @see ResponseElemDocument
	 * @version 20070530
	 * @since 20061103
	 **********************************************************************/
	public static ResponseElemDocument blockRetrieval(UserType user) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "blockRetrieval");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		//Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (isRetrievalActive){
			//Stop Retrieval
			isRetrievalActive = false;
			response.getResponseElem().setStatusCode(260);
			response.getResponseElem().setResponseText("Retrieval blocked. Please check existing users before shut down.");
		} else {
			//Retrieval is not running. No need to stop
			response.getResponseElem().setStatusCode(261);
			response.getResponseElem().setResponseText("Retrieval is already blocked. No need to block.");
		}
		return response;
	}	
	
	
	/** ********************************************************************
	 * Allow the retrieval service again if it is already marked as stopped.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return The response object
	 * @see DiasAccess
	 * @see ResponseElemDocument
	 * @version 20070530
	 * @since 20061103
	 **********************************************************************/
	public static ResponseElemDocument allowRetrieval(UserType user) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "allowRetrieval");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (!isRetrievalActive){
			//Allow Retrieval
			isRetrievalActive = true;
			response.getResponseElem().setStatusCode(262);
			response.getResponseElem().setResponseText("Retrieval is allowed again.");
		} else {
			//Retrieval is already allowed. Nothing to do. 
			response.getResponseElem().setStatusCode(263);
			response.getResponseElem().setResponseText("Retrieval is not blocked. No action needed.");
		}
		return response;
	}

	
	/** ********************************************************************
	 * Returns some statistical data about Retrieval Web Service 
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return The response object
	 * @see ResponseElemDocument
	 * @version 20070530
	 * @since 20070124
	 **********************************************************************/
	public static ResponseElemDocument getRetrievalStatistics(UserType user) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "getRetrievalStatistics");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		//Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (isRetrievalActive){
			//TODO: implement the algorithm
			
		} else {
			//Retrieval is not running. Return an error message.
			response.getResponseElem().setStatusCode(550);
			response.getResponseElem().setResponseText("Retrieval is not running. Please consult the administrator.");
		}
		return response;
	}
	
	/** ********************************************************************
	 * Resets the statistical data about Retrieval Web Service 
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return The response object
	 * @see ResponseElemDocument 
	 * @version 20070530
	 * @since 20070124
	 **********************************************************************/
	public static ResponseElemDocument resetRetrievalStatistics(UserType user) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "blockRetrieval");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		//Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (!isRetrievalActive){
			//TODO: implement the algorithm
			
		} else {
			//Retrieval is still running. Return an error message.
			response.getResponseElem().setStatusCode(551);
			response.getResponseElem().setResponseText("Retrieval is still running. Please block retrievals first.");
		}
		
		return response;
	}
	
	
	/** ********************************************************************
	 * Gets the list of objects modified since given timestamp (in java.util.Calendar format)
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user
	 * @param timestamp Timestamp
	 * @return The response object
	 * @see ResponseElemDocumentImpl 
	 * @version 20070530
	 * @since 20070124
	 **********************************************************************/
	public static ResponseElemDocument getModifiedSince(UserType user, java.util.Calendar timestamp) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "getModifiedSince");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		//Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (!isRetrievalActive){//Retrieval is not running. Return an error message.
			response.getResponseElem().setStatusCode(550);
			response.getResponseElem().setResponseText("Retrieval is not running. Please consult the administrator.");
		} else {
			//TODO: SUB will implement the algorithm
			System.out.println(timestamp);
		}

		return response;
	}
	

	/** ********************************************************************
	 * Initializes the cache. Please note that this cache is NOT only there
	 * to speed up things. It functions also as a place to temporarily save
	 * the fom DIAS coming objects. It means every koLibRI installation which
	 * uses Web Services to retrieve objects from DIAS MUST initialize some
	 * amount of cache!
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param path Path to cache root directory
	 * @return The response object
	 * @see ResponseElemDocument
	 * @version 20070530
	 * @since 20070124
	 **********************************************************************/
	public static ResponseElemDocument initCache(UserType user, String path){
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "initCache");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		//Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (!isRetrievalActive){
			//create the directory tree, set the response and correct the cache root dir def.
			makeDirectoryTree(path);
			response.getResponseElem().setStatusCode(265);
			response.getResponseElem().setResponseText("Cache is initialised.");
			ABSOLUTE_ROOT_DIR = path;
			setCacheInitialized(true);
		} else {
			//Retrieval is still running. Return an error message.
			response.getResponseElem().setStatusCode(551);
			response.getResponseElem().setResponseText("Retrieval is still running. Please block retrievals first.");
		}
		return response;
	}


	/** ********************************************************************
	 * Initializes the cache. Please note that this cache is NOT only there
	 * to speed up things. It functions also as a place to temporarily save
	 * the fom DIAS coming objects. It means every koLibRI installation which
	 * uses Web Services to retrieve objects from DIAS MUST initialize some
	 * amount of cache!
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @param timelimit Number of days which the older entries must be erased
	 * @return The response object
	 * @see ResponseElemDocument 
	 * @version 20070605
	 * @since 20070605
	 **********************************************************************/
	public static ResponseElemDocument eraseOldCacheEntries(UserType user, int timelimit){
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "eraseOldCacheEntries");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		//Authorised
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		if (!isRetrievalActive){
			//
			eraseOldEntries(timelimit);
			response.getResponseElem().setStatusCode(266);
			response.getResponseElem().setResponseText("Entries that are older than " + timelimit +
			                                           " have been erased.");
		} else {
			//Retrieval is still running. Return an error message.
			response.getResponseElem().setStatusCode(551);
			response.getResponseElem().setResponseText("Retrieval is still running. Please block retrievals first.");
		}
		return response;
	}

	
	/** ********************************************************************
	 * Delete all the temporarily saved files and directories under "PATH"
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param user Login information of the user
	 * @return The response object
	 * @see ResponseElemDocument
	 * @version 20070530
	 * @since 20061122
	 **********************************************************************/
	public static ResponseElemDocument deleteCache(UserType user) {
		//check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "deleteCache");
		if (authorisationResponse > 0){
			//not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}
		
		ResponseElemDocument response = WSUtils.createDefaultResponse();
		
		//look if Retrieval already stopped
		if (isRetrievalActive){
			//Retrieval running. Can not clear.
			response.getResponseElem().setStatusCode(551);
			response.getResponseElem().setResponseText("Retrieval still running. Can not clear the cache.\n" +
				"Please stop the retrieval first and be sure that the system is not being used.");
		} else {
			//Retrieval is not running. Delete the directory three
			try {
				java.io.File dirToDelete = new java.io.File(ABSOLUTE_ROOT_DIR);
				FileUtils.deleteDir(dirToDelete);
				//dirToDelete.mkdir(); //recreate the deleted absolute root directory
				setCacheInitialized(false);
				response.getResponseElem().setStatusCode(290);
				response.getResponseElem().setResponseText("Cache cleared successfully.");
			} catch (Exception e){
				//Exception! Return an error message.
				response.getResponseElem().setStatusCode(552);
				response.getResponseElem().setResponseText("Exception in AccesLayer.clearCache(" + 
				                         ABSOLUTE_ROOT_DIR + ").\nPlease consult the administrator.");
				response.getResponseElem().setExtraInfo(e.getMessage());
			}
		}
		return response;
	}
	
	
	/***************************************************************************
	 * Creates a empty directory tree in the given directory.
	 * 
	 * @param root The root directory
	 **************************************************************************/
	private static void makeDirectoryTree(String root) {
		for (int i = 0; i < 10; i++){
			String dir1 = root + "/" + i;
			java.io.File firstLevel = new java.io.File(dir1);
	    	firstLevel.mkdirs();
	    	for (int j = 0; j < 10; j++) {
	    		String dir2 = dir1 + "/" + j;
				java.io.File secondLevel = new java.io.File(dir2);
		    	secondLevel.mkdir();
		    	for (int k = 0; k < 10; k++) {
		    		String dir3 = dir2 + "/" + k;
					java.io.File thirdLevel = new java.io.File(dir3);
			    	thirdLevel.mkdir();
		    	}
		    	java.io.File dipDir = new java.io.File(dir2 + "/dips");
		    	dipDir.mkdir();
	    	}
		}	
	}

	
	/***************************************************************************
	 * Downloads a file from DIAS
	 * 
	 * @param dias The connection to DIAS
	 * @param diasResponse The response of DIAS
	 * @param destination Download destination
	 * @return Downloaded File
	 * @throws java.io.IOException
	 **************************************************************************/
	private static java.io.File downloadDIASFile(DiasAccess dias, DiasAccessResponse diasResponse,
			                                     String destination) throws java.io.IOException {
		java.io.File f = null;
		//Begin Code from SUB  (Why?)
		String uri = diasResponse.getUri();
		System.out.println("URI = " + uri);
        String uriParts[] = uri.split("/");
        String fileName = uriParts[uriParts.length-1];
        System.out.println("File name = " + fileName);
        String pathName = uri.substring(0, uri.lastIndexOf(fileName));
        diasResponse.setUri(pathName);
        // End SUB code
        f = dias.downloadDiasFile(diasResponse, fileName, destination);
		return f;
	}
	
	private static String checkPath(String path){
		path = path.replaceAll("\\/", "/");
		path = path.replaceAll("//", "/");
		
		return path;
	}


	/***************************************************************************
	 * Deletes the files in cache which are older than given days.
	 * 
	 * @param days Time limit for cache entries in days. Smaller than 1 means 6 hours. 
	 **************************************************************************/
	private static void eraseOldEntries(int days) {
		long now = System.currentTimeMillis();
		long timeLimit = days * 86400000; //milliseconds per day
		if (timeLimit < 21600000) timeLimit = 21600000; //6 hours minimum
		//iterate over the whole directory tree and look which entries are older
		for (int i = 0; i < 10; i++){
			String dir1 = ABSOLUTE_ROOT_DIR + "/" + i;
			//dir1 = dir1.replaceAll("\\", "/"); //Windows to *NIX
			dir1 = dir1.replaceAll("//", "/"); //to be sure that path does not have // at the end
	    	for (int j = 0; j < 10; j++) {
	    		String dir2 = dir1 + "/" + j;
		    	for (int k = 0; k < 10; k++) {
		    		String dir3 = dir2 + "/" + k;
		    		System.out.println(dir3);
					java.io.File dir = new java.io.File(dir3);
					java.io.File[] files = dir.listFiles();
			    	for (int n = 0; n < files.length; n++) {
			    		System.out.println(n + " of " + files.length + "\t" + files[n].getName() + "\t" + files[n].lastModified());
						if ((now - files[n].lastModified()) > timeLimit) {
							//this entry is older than limit. Delete it.
							if (files[n].isDirectory()) FileUtils.deleteDir(files[n]);
							else files[n].delete();
						}
					}
		    	}
		    	//now the directory for complete DIPs
		    	java.io.File dipDir = new java.io.File(dir2 + "/dips");
		    	java.io.File[] files = dipDir.listFiles();
		    	for (int n = 0; n < files.length; n++) {
					if ((now - files[n].lastModified()) > timeLimit) {
						//this entry is older than limit. Delete it.
						if (files[n].isDirectory()) FileUtils.deleteDir(files[n]);
						else files[n].delete();
					}
				}
	    	}
		}	
	}
	
	/**
	 * @param args
	 * @exception AxisFault
	 * @exception RemoteException
	 */
	public static void main(String[] args) {
		ABSOLUTE_ROOT_DIR = "c:/temp/cache/";
		UserType user = UserType.Factory.newInstance();
		user.setUserName("karaca");
		user.setPassword("geheim");
		user.setInstitution(10001);
		
		blockRetrieval(user);
		initCache(user, ABSOLUTE_ROOT_DIR);
		allowRetrieval(user);
		ResponseElemDocument red = getMETS(user, "urn:nbn:de:1111-20070325-3-00548-2");
		System.out.println(red.getResponseElem().getResponseText());
		//System.out.println(getDIP(user, red.getResponseElem().getDiasId()).getResponseElem().getResponseText());
		//System.out.println(getFile(user, red.getResponseElem().getDiasId(), "01380000.tif").getResponseElem().getResponseText());
		//makeDirectoryTree(ABSOLUTE_ROOT_DIR);
		//eraseOldEntries(1);
	}
} //END-Class