/*******************************************************************************
 * de.langzeitarchivierung.kopal.util / FileUtils.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;

/*******************************************************************************
 * Some usefull file load ans store routines.
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20070710
 * @since 20060225
 ******************************************************************************/

public class FileUtils
{

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * Returns the content of a text file as a String.
	 * 
	 * TODO Use no specific encoding as default to avoid the "none" conditional
	 * to improve performance.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param file File to read
	 * @param encoding Encoding of the text. Default is "UTF-8" if called with
	 *            null.
	 * @return Contents of the file as a String
	 * @version 20060225
	 * @since 20060225
	 **************************************************************************/
	public static String readFile(java.io.File file, String encoding) {
		// Default is "UTF-8" if called with null or empty string.
		if (encoding == null || encoding.equals("")) {
			encoding = "UTF-8";
		}

		String text = "";
		java.io.FileInputStream in = null;

		try {
			in = new java.io.FileInputStream(file);
			int fl = (int) file.length();
			byte buffer[] = new byte[fl];
			int len = in.read(buffer, 0, fl);
			// Read all the text file with the given encoding.
			if (encoding.equals("none")) {
				text = new String(buffer, 0, len);
			}
			else {
				text = new String(buffer, 0, len, encoding);
			}
		}
		catch (java.io.IOException e) {
			e.printStackTrace();
		}
		finally {
			try {
				if (in != null)
					in.close();
			}
			catch (java.io.IOException e) {
				e.printStackTrace();
			}
		}

		return text;
	}

	/***************************************************************************
	 * Opens the given text file with the given encoding and saves it as an
	 * other file with another encoding. We use it mainly to convert german
	 * "ISO-8859-1" encoding to "UTF-8" but many other combinations are
	 * possible.
	 * 
	 * @author Kadir Karaca Koçer, German National Library
	 * @param source File to read
	 * @param encodingSource Encoding of the source text.
	 * @param destination Destination file to write into
	 * @param encodingDestination Encoding of the destination text.
	 * @throws IOException
	 * @version 20070414
	 * @since 20070414
	 **************************************************************************/
	public static void convertFile(File source, String encodingSource,
			File destination, String encodingDestination) throws IOException {
		Reader input = null;
		Writer output = null;

		try {
			// Define Buffers to make it faster
			input = new BufferedReader(new InputStreamReader(
					new FileInputStream(source), encodingSource));
			output = new BufferedWriter(new OutputStreamWriter(
					new FileOutputStream(destination), encodingDestination));
			int i = 0;
			while ((i = input.read()) != -1) {
				output.write(i);
			}
		}
		finally {
			if (input != null)
				input.close();
			if (output != null)
				output.close();
		}
	}

	/***************************************************************************
	 * Simply reads a file with UTF-8.
	 * 
	 * @param file
	 * @return The file as text.
	 **************************************************************************/
	public static String readFile(java.io.File file) {
		return readFile(file, "UTF-8");
	}

	/***************************************************************************
	 * Stores a string into a file.
	 * 
	 * TODO Use no specific encoding as default to avoid the "none" conditional
	 * to improve performance.
	 * 
	 * @param file
	 * @param encoding
	 * @param text
	 * @return boolean - true if the store operation was successfull
	 * @author Stefan Funk, Goettingen SUB
	 **************************************************************************/
	public static boolean storeFile(File file, String encoding, String text) {
		// Default is "UTF-8" if called with null or empty string.
		if (encoding == null || encoding.equals("")) {
			encoding = "UTF-8";
		}

		java.io.FileOutputStream os = null;
		try {
			os = new java.io.FileOutputStream(file);
			byte buffer[];
			if (encoding.equals("none")) {
				buffer = text.getBytes();
			}
			else {
				buffer = text.getBytes(encoding);
			}
			os.write(buffer);
		}
		catch (java.io.IOException e) {
			return false;
		}
		finally {
			try {
				if (os != null)
					os.close();
			}
			catch (java.io.IOException e) {
				e.printStackTrace();
			}
		}

		return true;
	}

	/***************************************************************************
	 * Simply stores a file with "UTF-8" encoding.
	 * 
	 * @param file
	 * @param text
	 * @return boolean - true if the store operation was successfull
	 **************************************************************************/
	public static boolean storeFile(File file, String text) {
		return storeFile(file, "UTF-8", text);
	}

	/***************************************************************************
	 * Converts all non letter or digit characters of the asset identifier to
	 * underscores for general filename compliance.
	 * 
	 * @author Jens Ludwig, SUB Goettingen
	 * @param idString
	 * @return the converted identifier
	 **************************************************************************/
	public static String getFilenamePersistentIdentifier(String idString)
			throws Exception {
		if (idString != null) {
			StringBuffer result = new StringBuffer();
			for (int i = 0; i < idString.length(); i++) {
				if (Character.isLetterOrDigit(idString.charAt(i))) {
					result.append(idString.charAt(i));
				}
				else {
					result.append('_');
				}
			}

			return result.toString();
		}

		throw new Exception("The object's persistent identifier is needed to "
				+ "generate a filename persistent identifier");
	}

	/***************************************************************************
	 * Recursive deletion of the given directory.
	 * 
	 * @param dir The directory or file to delete
	 **************************************************************************/
	public static void deleteDir(File dir) {
		if (dir.isDirectory()) {
			File dirEntries[] = dir.listFiles();
			for (int i = 0; i < dirEntries.length; i++) {
				if (dirEntries[i].isDirectory()) {
					deleteDir(dirEntries[i]);
				}
				else {
					dirEntries[i].delete();
				}
			}
			dir.delete();
		}
	}

	/***************************************************************************
	 * Recursive deletion of the Files within the given directory.
	 * 
	 * @param dir The directory or file to delete
	 **************************************************************************/
	public static void deleteFiles(File dir) {
		if (dir.isDirectory()) {
			File dirEntries[] = dir.listFiles();
			for (int i = 0; i < dirEntries.length; i++) {
				if (dirEntries[i].isDirectory()) {
					deleteDir(dirEntries[i]);
				}
				else {
					dirEntries[i].delete();
				}
			}
		}
	}

	/***************************************************************************
	 * Computes the size of a given file or directory.
	 * 
	 * @param currentFile
	 * @return The file size of course :) What else?
	 **************************************************************************/
	public static long getCompleteFileSize(File currentFile) {
		long completeFileLength = 0;

		// If it is a file, directly return it's size.
		if (currentFile.isFile()) {
			completeFileLength = currentFile.length();
		}
		else {
			// Otherwise get the file list and go throught all directories and
			// add their lengths.
			File fileList[] = currentFile.listFiles();

			for (int i = 0; i < fileList.length; i++) {
				if (fileList[i].isDirectory()) {
					completeFileLength += getCompleteFileSize(fileList[i]);
				}
				else {
					completeFileLength += fileList[i].length();
				}
			}
		}

		return completeFileLength;
	}

	/***************************************************************************
	 * Returns the amount of free space on disk.
	 * 
	 * @param dir The directory or file to ask how much free space available
	 * @return The amount of free space in this directory
	 **************************************************************************/
	public static long getFreeSpace(java.io.File dir) {
		long fs = 0;
		// TODO: implement the algorithm

		return fs;
	}

	/***************************************************************************
	 * Untar a tar archive with destination as string.
	 * 
	 * @param tarName Name of the Archive
	 * @param destDir Destination directory
	 * @return Status code -> 0: Success, 1: destination Directory does not
	 *         exist
	 **************************************************************************/
	public static int untar(String tarName, String destDir) {
		// No errors.
		int status = 0;

		java.io.File f = new java.io.File(destDir);
		boolean success = f.mkdirs();
		if (success) {
			status = untar(new java.io.File(tarName), f);
		}
		else {
			status = 1;
		}

		return status;
	}

	/***************************************************************************
	 * Untar a tar archive with destination as file object.
	 * 
	 * TODO Remove all sysouts!
	 * 
	 * @param arcFile The Archive to untar
	 * @param destDir Destination directory
	 * @return Status code -> 0: Success, 1: destination Directory do not exists
	 **************************************************************************/
	public static int untar(java.io.File arcFile, java.io.File destDir) {
		// No errors.
		int status = 0;

		try {
			// Create the destination directory if it does not exist.
			destDir.mkdirs();
			System.out.println("Reading the archive " + "file: "
					+ arcFile.getName());
			java.io.InputStream is = new java.io.FileInputStream(arcFile);

			// Use the tar-Library. Its GPL :-)
			System.out.println("Buffering the contents");
			// com.ice.tar.TarArchive archive = new com.ice.tar.TarArchive(is,
			// BUFFER_SIZE);
			com.ice.tar.TarArchive archive = new com.ice.tar.TarArchive(is);
			System.out.println("Extracting to: " + destDir.getName());
			// Extract the content.
			archive.extractContents(destDir);
			// Close the archive.
			archive.closeArchive();
			System.out.println("TAR archive successfully extracted.");
		}
		catch (FileNotFoundException e) {
			status = 2;
			e.printStackTrace();
		}
		catch (IOException e) {
			status = 3;
			e.printStackTrace();
		}

		return status;
	}

}
