/*******************************************************************************
 * de.langzeitarchivierung.kopal.util / FormatRegistry.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xml.sax.SAXException;

import de.ddb.lmeRfile.LmerFile;
import de.langzeitarchivierung.kopal.ingest.DiasIngest;

/*******************************************************************************
 * Encapsulating class for a table of format specific data
 * 
 * @author Matthias Neubauer, German National Library
 * @version 20070124
 * @see de.langzeitarchivierung.kopal.actionmodule.MetadataGenerator
 * @since 20050809
 ******************************************************************************/

public class FormatRegistry
{

	// CLASS VARIABLES (Static variables) **********************************

	private static Logger			defaultLogger				= Logger
																		.getLogger("de.langzeitarchivierung.kopal");
	private static String			formatRegistryBackupFile	= "";

	/** ID for an unknown filetype */
	// TODO Get the unknown filetype ID from the DIAS?
	public static String			UNKNOWN_FILETYPE			= "urn:diasid:fty:kopal:0000000000000000000000";

	// STATE (Instance variables) *****************************************

	private String					formatMapFileLocation		= "";
	private ArrayList<FormatInfo>	formatRegistry				= new ArrayList<FormatInfo>();
	private FormatMapper			formatMapper				= new FormatMapper();
	private boolean					useBackupFileOnly			= false;

	/***************************************************************************
	 * Default constructor
	 * 
	 * @version 20060703
	 * @since 20060703
	 **************************************************************************/
	public FormatRegistry() {
		super();

		try {
			Configurator.configure(this);
		}
		catch (Exception e) {
			e.printStackTrace();
			defaultLogger.log(Level.SEVERE, "Error configuring class "
					+ this.getClass().getName() + ". System message: "
					+ e.getMessage());
		}

		defaultLogger.log(Level.INFO, "Initializing Format Registry");

		try {
			this.initializeXMLFormatRegistryData();
		}
		catch (SAXException e) {
			defaultLogger.log(Level.SEVERE, "Error parsing the DIAS format "
					+ "registry backup file. No DIAS file types available "
					+ "at all. Exiting!", new File(formatRegistryBackupFile));
			e.printStackTrace();
			System.exit(13);
		}
		catch (IOException e) {
			defaultLogger.log(Level.SEVERE, "Error accessing the format "
					+ "registry backup file. No DIAS file types available "
					+ "at all. Exiting!", new File(formatRegistryBackupFile));
			e.printStackTrace();
			System.exit(16);
		}
	}

	/***************************************************************************
	 * @param diasFormatId
	 * @return
	 **************************************************************************/
	public FormatInfo getFormatInfo(String diasFormatId) {
		for (Iterator formatIter = this.formatRegistry.iterator(); formatIter
				.hasNext();) {
			FormatInfo fInfo = (FormatInfo) formatIter.next();
			if (fInfo.getFileTypeID().equalsIgnoreCase(diasFormatId)) {
				return fInfo;
			}
		}
		return null;
	}

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * Searches the formatregistry list for the specific entry of a formats name
	 * and its version
	 * 
	 * @version 20050831
	 * @since 20050802
	 * @param formatName - The formats name specified by Jhove
	 * @param version - The formats version specified by Jhove
	 * @param lmerfile - LmerFile object which takes the infos
	 **************************************************************************/
	public void getFormat(String formatName, String version, LmerFile lmerfile) {
		// Go through the list...
		for (int i = 0; i < this.formatRegistry.size(); i++) {
			FormatInfo tempform = this.formatRegistry.get(i);

			// If the format has no version specified, do not compare versions.
			if (version == null) {

				// Compare the formatname from JHOVE with the actual entry.
				if (tempform.getFileType().equals(formatName)) {

					// If it equals, set the specific LmerFile fields and
					// return.
					lmerfile.getFormatArray(0).setStringValue(
							tempform.getFileTypeID());

					if (!tempform.getFileTypeCategory().equals("")) {
						// TODO Check if this works.
						lmerfile.setCategory(LmerFile.Category.Enum
								.forString(tempform.getFileTypeCategory()));
					}
					if (!tempform.getFileTypeViewerApplication().equals("")) {
						lmerfile.setViewerApplication(tempform
								.getFileTypeViewerApplication());
					}

					lmerfile.getFormatArray(0).setREGISTRYNAME("DIAS");

					return;
				}
			}
			else {
				// Compare the formats name and version from JHOVE with the
				// actual entry.
				if (tempform.getFileType().equals(formatName)
						&& tempform.getFileTypeVersion().equals(version)) {

					// If they equal, set the specific LmerFile fields and
					// return.
					lmerfile.getFormatArray(0).setStringValue(
							tempform.getFileTypeID());

					if (!tempform.getFileTypeCategory().equals("")) {
						// TODO check if this works
						lmerfile.setCategory(LmerFile.Category.Enum
								.forString(tempform.getFileTypeCategory()));
					}
					if (!tempform.getFileTypeViewerApplication().equals("")) {
						lmerfile.setViewerApplication(tempform
								.getFileTypeViewerApplication());
					}

					lmerfile.getFormatArray(0).setREGISTRYNAME("DIAS");

					return;
				}
			}
		}

		// Fall-through: Set the formatname to the "Unknown File Type".
		lmerfile.getFormatArray(0).setStringValue(UNKNOWN_FILETYPE);

		lmerfile.getFormatArray(0).setREGISTRYNAME("DIAS");
		lmerfile.setCategory(LmerFile.Category.BINARY);
	}

	/***************************************************************************
	 * Method to initialize the format registry by requesting the filetypes list
	 * from DIAS or using a local backupfile
	 * 
	 * @throws IOException
	 * @throws SAXException
	 * 
	 * @version 20060103
	 * @since 20050905
	 **************************************************************************/
	public void initializeXMLFormatRegistryData() throws SAXException,
			IOException {
		String xmlstring = "";

		File xmlfile = new File(formatRegistryBackupFile);

		if (!this.useBackupFileOnly) {
			// Retrieve the FormatRegistry from DIAS.
			DiasIngest diasing = new DiasIngest();

			// Configure the DiasIngest object and get the file type response
			// from DIAS. If not successful, use the backup file.
			try {
				Configurator.configure(diasing);
				defaultLogger.log(Level.FINE, "Requesting format registry "
						+ "file from DIAS");
				xmlstring = diasing.requestFileTypes();

				// Get the ArrayList with format information from the
				// DiasIngest.
				this.formatRegistry = DiasIngest
						.parseDiasFormatResponse(xmlstring);

				// Store the retrieved information locally.
				if (!xmlstring.equals("")) {
					defaultLogger.log(Level.FINE, "Storing format registry "
							+ "file");
					FileUtils.storeFile(xmlfile, xmlstring);
				}
			}
			catch (SAXException e) {
				defaultLogger.log(Level.WARNING, "Parsing the DIAS format "
						+ "registry file failed. Please check the response");
				this.useBackupFileOnly = true;
			}
			catch (Exception e) {
				defaultLogger.log(Level.WARNING, "Request for DIAS format "
						+ "registry file failed");
				this.useBackupFileOnly = true;
			}
		}

		if (this.useBackupFileOnly) {
			if (xmlfile.exists()) {
				// Use the backup file.
				defaultLogger.log(Level.INFO, "Using local format registry "
						+ "backup file");
				xmlstring = FileUtils.readFile(xmlfile);

				// Parse the backup file.
				this.formatRegistry = DiasIngest
						.parseDiasFormatResponse(xmlstring);
			}
			else {
				defaultLogger.log(Level.SEVERE, "No backup file for format "
						+ "registry found! Formats can not be identified",
						xmlfile);
				System.exit(10);
			}
		}

		defaultLogger.log(Level.FINE, "Loaded " + this.formatRegistry.size()
				+ " format descriptions");

		// Map the formats from the JHOVE to the DIAS.
		this.formatMapper.map(this.formatRegistry, this.formatMapFileLocation);
	}

	// GET & SET METHODS **************************************************

	/***************************************************************************
	 * @param _formatRegistryBackupFile The formatRegistryBackupFile to set.
	 **************************************************************************/
	public void setFormatRegistryBackupFile(String _formatRegistryBackupFile) {
		FormatRegistry.formatRegistryBackupFile = _formatRegistryBackupFile;
	}

	/***************************************************************************
	 * @param _formatMapFileLocation The formatMapFileLocation to set.
	 **************************************************************************/
	public void setFormatMapFileLocation(String _formatMapFileLocation) {
		this.formatMapFileLocation = _formatMapFileLocation;
	}

	/***************************************************************************
	 * @param useBackupFile The useBackupFile to set.
	 * @author Stefan Funk, SUB Göttingen 01.06.2007
	 **************************************************************************/
	public void setUseBackupFileOnly(boolean useBackupFileOnly) {
		this.useBackupFileOnly = useBackupFileOnly;
	}

}
