/*******************************************************************************
 * de.langzeitarchivierung.kopal.util / HTTPUtils.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.util;

import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.commons.httpclient.DefaultHttpMethodRetryHandler;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.params.HttpMethodParams;

/*******************************************************************************
 * Some usefull routines to use with HTTP.
 * 
 * @author Stefan Funk, SUB Goettingen
 * @version 20061204
 * @since 20061204
 ******************************************************************************/

public class HTTPUtils
{

	/***************************************************************************
	 * Delivers a list of links contained in an HTML directory index file.
	 * 
	 * @param client
	 * @param url
	 * @return A list of URLs.
	 * @throws Exception
	 **************************************************************************/
	public static ArrayList getAllHrefs(HttpClient client, String url)
			throws Exception {
		ArrayList<String> result = new ArrayList<String>();
		String hrefs[];
		String krams = "";
		String htmlString = "";

		// Get the HTML as a string.
		htmlString = getRootHtmlDataAsString(client, url);
		
		// Split the complete HTML string into pieces beginning with the link's
		// names.
		hrefs = htmlString.split("<a href=\"");

		// Cut all links out of the single strings starting with "http://" and
		// ending with '"'. Add it to the list.
		for (int i = 1; i < hrefs.length; i++) {
			if (!hrefs[i].startsWith("?") && !hrefs[i].startsWith("http://") && !hrefs[i].startsWith("/")) {
				krams = url + hrefs[i].substring(0, hrefs[i].indexOf("\""));
				result.add(krams);
			}
		}

		return result;
	}

	/***************************************************************************
	 * Gets the index dir of an URL and returns it as a string.
	 * 
	 * @param client
	 * @param url
	 * @return String
	 * @throws Exception
	 **************************************************************************/
	public static String getRootHtmlDataAsString(HttpClient client, String url)
			throws Exception {
		StringBuffer result = new StringBuffer();

		// Get an HTTP method.
		GetMethod method = HTTPUtils.httpGetMethod(client, url);

		// Read the response body.
		InputStreamReader isr = new InputStreamReader(method
				.getResponseBodyAsStream());

		// Deal with the response.
		// TODO Use caution: Ensure correct character encoding and that it's not
		// binary data. Check if it really IS HTML.
		char[] buffer = new char[1048576];

		int bytesRead = isr.read(buffer);
		while (bytesRead > 0) {
			result.append(buffer, 0, bytesRead);
			bytesRead = isr.read(buffer);
		}

		// Release the connection.
		method.releaseConnection();

		return result.toString();
	}

	/***************************************************************************
	 * Defines and returns a HTTP get method.
	 * 
	 * @param client
	 * @param url
	 * @return GetMethod
	 * @throws HttpException
	 * @throws IOException
	 **************************************************************************/
	public static GetMethod httpGetMethod(HttpClient client, String url)
			throws HttpException, IOException {
		// Create a method instance.
		GetMethod method = new GetMethod(url);

		// Provide custom retry handler is necessary.
		method.getParams().setParameter(HttpMethodParams.RETRY_HANDLER,
				new DefaultHttpMethodRetryHandler(3, false));

		// Execute the method.
		int statusCode = client.executeMethod(method);

		if (statusCode != HttpStatus.SC_OK) {
			throw new HttpException("HTTP get method failed: "
					+ method.getStatusLine());
		}

		return method;
	}

	/***************************************************************************
	 * Returns the name of a file or folder extracted from an URL, just taking
	 * the values after the last slash, resp. between the last two slashes.
	 * 
	 * @param currentUrl
	 * @return String
	 **************************************************************************/
	public static String getFilenameFromUrl(String currentUrl) {
		// First delete an ending slash.
		if (currentUrl.endsWith("/")) {
			currentUrl = currentUrl.substring(0, currentUrl.length() - 1);
		}
		int filenameStart = currentUrl.lastIndexOf("/");

		return currentUrl.substring(filenameStart + 1);
	}

	/***************************************************************************
	 * Takes recursively all index.html files from a HTTP location and returns a
	 * SHA-1 checksum over all the files or the checksum of the single file.
	 * 
	 * @param currentUrl
	 * @param client
	 * @return String
	 **************************************************************************/
	public static String getCompleteIndexDirChecksum(HttpClient client,
			String currentUrl) throws Exception {
		String result = "";
		
		try {
			if (!currentUrl.endsWith("/")) {
				result = ChecksumGenerator.getSHA1(HTTPUtils
						.getRootHtmlDataAsString(client, currentUrl));
			}
			else {
				result = ChecksumGenerator.getSHA1(getIndexDirChecksum(client,
						currentUrl));
			}
		}
		catch (Exception e) {
			throw new Exception("Error fetching index.html");
		}

		return result;
	}

	// INTERNAL (Internal - implementation details, local classes, ...) ******

	/***************************************************************************
	 * @param currentUrl
	 * @param client
	 * @return String
	 **************************************************************************/
	private static String getIndexDirChecksum(HttpClient client,
			String currentUrl) throws Exception {
		String result = "";
		String currentEntry = "";

		try {
			// Get a list of all contained files in one HTTP location.
			ArrayList source;
			source = HTTPUtils.getAllHrefs(client, currentUrl);

			// Go throught all the entries.
			Iterator iter = source.iterator();
			while (iter.hasNext()) {
				currentEntry = (String) iter.next();

				// If currentEntry is a folder and has more file entries, get
				// the checksum of the index.html and invoke getIndexDirChecksum
				// recursively.
				if (currentEntry.endsWith("/")) {
					result += ChecksumGenerator.getSHA1(HTTPUtils
							.getRootHtmlDataAsString(client, currentEntry))
							+ ChecksumGenerator.getSHA1(getIndexDirChecksum(
									client, currentEntry));
				}
			}

			// If the list is empty, take the index file's content for computing
			// the checksum.
			if (source.isEmpty()) {
				result = ChecksumGenerator.getSHA1(HTTPUtils
						.getRootHtmlDataAsString(client, currentUrl));
			}

		}
		catch (Exception e) {
			throw new Exception("Error fetching index.html");
		}

		return result;
	}

}
