/*******************************************************************************
 * de.langzeitarchivierung.kopal.util / KopalClassLoader.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.util;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;

/*******************************************************************************
 * Helper methods for loading classes, not a real classloader
 * 
 * 
 * @author Jens Ludwig, SUB G�ttingen
 * @version 20060123
 ******************************************************************************/
public class KopalClassLoader
{

	// CONSTANTS **********************************************************

	private static final String	MODUL_PACKAGES				= "de.langzeitarchivierung.kopal.actionmodule";
	private static final String	PROCESS_STARTER_PACKAGES	= "de.langzeitarchivierung.kopal.processstarter";
	private static final String	FORMAT_PACKAGES				= "de.langzeitarchivierung.kopal.formats";

	// MANIPULATION (Manipulation - what the object does) ******************

	/***************************************************************************
	 * Gets the class for the class name of an actionModule.
	 * 
	 * @param actionModuleName The name of the class
	 * @return the class
	 * @throws MalformedURLException
	 * @throws ClassNotFoundException
	 **************************************************************************/
	public static Class getActionModule(String actionModuleName)
			throws MalformedURLException, ClassNotFoundException {
		return getClass(actionModuleName, MODUL_PACKAGES);
	}

	/***************************************************************************
	 * Gets the class for the class name of an ProcessStarter.
	 * 
	 * @param processStarterName The name of the class
	 * @return the class
	 * @throws MalformedURLException
	 * @throws ClassNotFoundException
	 **************************************************************************/
	public static Class getProcessStarter(String processStarterName)
			throws MalformedURLException, ClassNotFoundException {
		return getClass(processStarterName, PROCESS_STARTER_PACKAGES);
	}

	/***************************************************************************
	 * Gets the class for the class name of an MetadataFormat.
	 * 
	 * @param formatName The name of the class
	 * @return the class
	 * @throws MalformedURLException
	 * @throws ClassNotFoundException
	 **************************************************************************/
	public static Class getFormat(String formatName)
			throws MalformedURLException, ClassNotFoundException {
		return getClass(formatName, FORMAT_PACKAGES);
	}

	/***************************************************************************
	 * Gets the class for the class name.
	 * 
	 * @param className The name of the class
	 * @param parentPackage The package where we should for the class
	 * @return The class
	 * @throws ClassNotFoundException
	 * @throws MalformedURLException
	 **************************************************************************/
	private static Class getClass(String className, String parentPackage)
			throws MalformedURLException, ClassNotFoundException {
		// If short form is used, add the missing package add.
		if (!className.startsWith(parentPackage)) {
			className = parentPackage + "." + className;
		}

		// From class name to class.
		URL classURL = new File(MODUL_PACKAGES).toURL();
		URL[] classURLs = new URL[] { classURL };
		URLClassLoader classLoader = new URLClassLoader(classURLs);

		return classLoader.loadClass(className);
	}

	/***************************************************************************
	 * Returns an instance of the specified class with the specified parameters.
	 * It chooses the constructor with the correct number of arguments.
	 * 
	 * @param newClass The class to get an instance of.
	 * @param parameters The parameters of type String to be given to the
	 *            constructor.
	 * @return Instance of the class. An explicit cast is necessary after this
	 *         method.
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 **************************************************************************/
	public static Object getInstanceOfClass(Class newClass, Object[] parameters)
			throws InstantiationException, IllegalAccessException,
			InvocationTargetException {
		// Get available constructors from class.
		Constructor[] constructors = newClass.getConstructors();
		// Check all available constructors for the first constructor that has
		// the correct number of arguments.
		if (parameters.length == 0) {
			return newClass.newInstance();
		}
		for (int i = 0; i < constructors.length; i++) {
			if (constructors[i].getParameterTypes().length == parameters.length) {
				return constructors[i].newInstance(parameters);
			}
		}

		// We haven't found the appropriate constructor for our parameters. That
		// means that the configfile is not valid.
		throw new IllegalArgumentException();
	}

	/***************************************************************************
	 * Returns an instance of the specified class without parameters.
	 * 
	 * @param newClass The class to get an instance of.
	 * @return Instance of the class. An explicit cast is necessary after this
	 *         method.
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 **************************************************************************/
	public static Object getInstanceOfClass(Class newClass)
			throws InstantiationException, IllegalAccessException,
			InvocationTargetException {
		return getInstanceOfClass(newClass, new Object[0]);
	}

}
