/*******************************************************************************
 * de.langzeitarchivierung.kopal.util.db / DBAccess.java
 *  
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.util.db;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.hibernate.exception.LockAcquisitionException;

/*******************************************************************************
 * Class with helper methods for database access and functions.
 * 
 * @author Sabiene Liess, Jens ludwig, SUB G�ttingen
 * 
 * CHANGELOG: 23.01.2007 Ludwig: Removed unnecessary methods and added comments
 ******************************************************************************/

public class DBAccess
{

	// CLASS VARIABLES (Static variables) *********************************

	private static HashMap<Class, HashMap<String, ReentrantLock>>	locks	= new HashMap<Class, HashMap<String, ReentrantLock>>();

	// MANIPULATION (Manipulation - what the object does) **********************

	/***************************************************************************
	 * Tries to lock in the specified database table the specified ID or waits
	 * until a lock is available.
	 * 
	 * @param table
	 * @param id
	 **************************************************************************/
	public static synchronized void lockTable(Class table, String id) {
		// Lazy initialization.
		if (!locks.containsKey(table)) {
			locks.put(table, new HashMap<String, ReentrantLock>());
		}

		// Get all locks for this table.
		HashMap<String, ReentrantLock> tableLocks = new HashMap<String, ReentrantLock>();
		tableLocks = locks.get(table);

		// Lazy initialization.
		if (!tableLocks.containsKey(id)) {
			tableLocks.put(id, new ReentrantLock());
		}

		// If we haven't already the lock, acquire it or wait.
		if (!(tableLocks.get(id)).isHeldByCurrentThread()) {
			(tableLocks.get(id)).lock();
		}
	}

	/***************************************************************************
	 * Tries to unlock an ID in the specified database table. If the lock isn't
	 * held by the current thread nothing happens.
	 * 
	 * @param table
	 * @param id
	 **************************************************************************/
	public static synchronized void unlockTable(Class table, String id) {
		if (locks.containsKey(table)) {
			HashMap<String, ReentrantLock> tableLocks = locks.get(table);
			ReentrantLock lock = tableLocks.get(id);
			if (lock.isHeldByCurrentThread()) {
				lock.unlock();
			}
		}
	}

	/***************************************************************************
	 * Checks if an ID in the specified database table is locked.
	 * 
	 * @param table
	 * @param id
	 * @return true if locked, false otherwise
	 **************************************************************************/
	public static boolean isLocked(Class table, String id) {
		// Lazy initialization.
		if (!locks.containsKey(table)) {
			locks.put(table, new HashMap<String, ReentrantLock>());
			// If a hashmap for this table didn't exist, nothing can be locked.
			return false;
		}

		// Get all locks for this table.
		HashMap<String, ReentrantLock> tableLocks = new HashMap<String, ReentrantLock>();
		tableLocks = locks.get(table);

		// Lazy initialization.
		if (!tableLocks.containsKey(id)) {
			tableLocks.put(id, new ReentrantLock());
			// If a lock didn't exist, nothing can be locked.
			return false;
		}

		if (tableLocks.get(id).isLocked()) {
			return true;
		}
		return false;
	}

	/***************************************************************************
	 * Ensures that an input object has a unique start date in the database and
	 * returns it. If the start date of the input object is already in the
	 * database we increase the start date by one.
	 * 
	 * Note: this method has to be synchronized because it determines the
	 * startdate which maps koLibRI processes to database entries one-to-one
	 * 
	 * TODO check why this method declaration doesn't need to throw an exception
	 * 
	 * @param input
	 * @return the start date which has been added to the database
	 **************************************************************************/
	public static synchronized long addInputUpdateStartdate(Input input) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		long newstartdate = input.getStartdate();
		Input in = (Input) session.createCriteria(Input.class).add(
				Restrictions.eq("startdate", new Long(newstartdate)))
				.uniqueResult();

		// as long as where is already another input object with this start date
		while (in != null) {
			newstartdate += 1;
			in = (Input) session.createCriteria(Input.class).add(
					Restrictions.eq("startdate", new Long(newstartdate)))
					.uniqueResult();
		}

		input.setStartdate(newstartdate);

		tx.commit();
		addObject(input);

		return input.getStartdate();
	}

	/***************************************************************************
	 * Adds a collection of object to the database.
	 * 
	 * TODO check if this method has to be synchronized TODO check why this
	 * method declaration doesn't need to throw an exception
	 * 
	 * @param objects
	 **************************************************************************/
	public static synchronized void addObjects(Collection objects) {
		for (Iterator objIter = objects.iterator(); objIter.hasNext();) {
			addObject(objIter.next());
		}
	}

	/***************************************************************************
	 * Adds an object to the database.
	 * 
	 * TODO Check if this method has to be synchronized.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param hibObject
	 **************************************************************************/
	public static synchronized void addObject(Object hibObject) {
		// Check locks for DIAS uploads with this URN.
		if (hibObject instanceof Dias) {
			Dias newDias = (Dias) hibObject;
			String urn = newDias.getDiasintid();
			// If we commit a SIP we must have a lock, so we need not to check
			// if diasUrnLocks is null.
			HashMap<String, ReentrantLock> diasUrnLocks = locks.get(Dias.class);

			if (diasUrnLocks.containsKey(urn)
					&& !diasUrnLocks.get(urn).isHeldByCurrentThread()) {
				throw new LockAcquisitionException("Dias object with ID " + urn
						+ "is locked", new SQLException());
			}
		}

		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		session.save(hibObject);
		tx.commit();
		HibernateUtil.closeSession();
	}

	/***************************************************************************
	 * Returns a single database object with the specified startdate or null if
	 * it doesn't exists. It should throw an exception if more than one database
	 * entry with the same startdate exist.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param startdate time in milliseconds since the begin of the epoch
	 * @return Input
	 **************************************************************************/
	public static Input getInputPerStartdate(long startdate) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		Input in = (Input) session.createCriteria(Input.class).add(
				Restrictions.eq("startdate", new Long(startdate)))
				.uniqueResult();
		tx.commit();

		return in;
	}

	/***************************************************************************
	 * Returns a list of objects for the specified database table where the
	 * column has the requested value.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param column
	 * @param value
	 * @return List
	 * @throws Exception
	 **************************************************************************/
	public static List getObjects(Class table, String column, Object value)
			throws Exception {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		List list = session.createCriteria(table).add(
				Restrictions.eq(column, value)).list();
		tx.commit();

		return list;
	}

	/***************************************************************************
	 * Returns a list of objects for the specified database table where the
	 * columns have the requested values.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param columnA
	 * @param valueA
	 * @param columnB
	 * @param valueB
	 * @return List
	 **************************************************************************/
	public static List getObjects(Class table, String columnA, Object valueA,
			String columnB, Object valueB) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		List list = session.createCriteria(table).add(
				Restrictions.and(Restrictions.eq(columnA, valueA), Restrictions
						.eq(columnB, valueB))).list();
		tx.commit();

		return list;
	}

	/***************************************************************************
	 * Returns a list of objects for the specified database table where the
	 * columns have the requested values.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param columnA
	 * @param valueA
	 * @param columnB
	 * @param valueB
	 * @return List
	 **************************************************************************/
	public static List getObjects(Class table, String columnA, Object valueA,
			String columnB, Object valueB, String columnC, Object valueC) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();

		List list = session.createCriteria(table).add(
				Restrictions.and(Restrictions.eq(columnA, valueA), Restrictions
						.eq(columnB, valueB))).add(
				Restrictions.eq(columnC, valueC)).list();
		tx.commit();

		return list;
	}

	/***************************************************************************
	 * Returns a sorted list of objects for the specified database table where
	 * the columns have the requested values. The list is sorted by the
	 * specified column.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param columnA
	 * @param valueA
	 * @param columnB
	 * @param valueB
	 * @param sortByColumn
	 * @return List
	 **************************************************************************/
	public static List getObjectsDescending(Class table, String columnA,
			Object valueA, String columnB, Object valueB, String sortByColumn) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		List list = session.createCriteria(table).add(
				Restrictions.and(Restrictions.eq(columnA, valueA), Restrictions
						.eq(columnB, valueB))).addOrder(
				Order.desc(sortByColumn)).list();
		tx.commit();

		return list;
	}

	/***************************************************************************
	 * Returns a sorted list of objects for the specified database table where
	 * the column has the requested value. The list is sorted by the specified
	 * column.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param column
	 * @param value
	 * @param sortByColumn
	 * @return List
	 **************************************************************************/
	public static List getObjectsDescending(Class table, String column,
			Object value, String sortByColumn) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		List list = session.createCriteria(table).add(
				Restrictions.eq(column, value)).addOrder(
				Order.desc(sortByColumn)).list();
		tx.commit();

		return list;
	}

	/***************************************************************************
	 * Returns a database object where the columns have the requested values or
	 * null if it doesn't exists. It should throw an exception if more than one
	 * database entry exists which meets this conditions.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param columnA
	 * @param valueA
	 * @param columnB
	 * @param valueB
	 * @return Object
	 **************************************************************************/
	public static Object getUniqueObject(Class table, String columnA,
			Object valueA, String columnB, Object valueB) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		Object obj = session.createCriteria(table).add(
				Restrictions.and(Restrictions.eq(columnA, valueA), Restrictions
						.eq(columnB, valueB))).uniqueResult();
		tx.commit();

		return obj;
	}

	/***************************************************************************
	 * Returns a database object where the column has the requested value or
	 * null if it doesn't exists. It should throw an exception if more than one
	 * database entry exists which meets this conditions.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param column
	 * @param value
	 * @return Object
	 **************************************************************************/
	public static Object getUniqueObject(Class table, String column,
			Object value) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		Object obj = session.createCriteria(table).add(
				Restrictions.eq(column, value)).uniqueResult();
		tx.commit();

		return obj;
	}

	/***************************************************************************
	 * Returns a list of database objects for the specified database table where
	 * the value of a column is between the specified values.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param column
	 * @param from
	 * @param till
	 * @return List
	 **************************************************************************/
	public static List getObjectPerDate(Class table, String column, Long from,
			Long till) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		List list = session.createCriteria(table).add(
				Restrictions.between(column, from, till)).list();
		tx.commit();

		return list;
	}

	/***************************************************************************
	 * Returns a database object for the specified table with the requested ID
	 * or null if it doesn't exists.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param table
	 * @param id
	 * @return Object
	 **************************************************************************/
	public static Object getObjectPerId(Class table, int id) {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		String col = (table.getSimpleName() + "id").toLowerCase();
		Object obj = session.createCriteria(table).add(
				Restrictions.eq(col.toLowerCase(), new Integer(id)))
				.uniqueResult();
		tx.commit();

		return obj;
	}

	/***************************************************************************
	 * Returns a HashMap table with database constants. The Hashmap uses as key
	 * the value of the constant and as value the database object of the
	 * constant. For example the database entry "creator" in the table dcelement
	 * is in the hashmap ("creator", object representing "creator").
	 * 
	 * A constant is a database entry which does not change during runtime and
	 * which has a column named identical to the table. It is usually only
	 * referenced in other database tables.
	 * 
	 * TODO Check why this method declaration doesn't need to throw hibernate
	 * exceptions.
	 * 
	 * @param table
	 * @return HashMap
	 * @throws NoSuchMethodException
	 * @throws InvocationTargetException
	 * @throws IllegalAccessException
	 **************************************************************************/
	public static HashMap getDbConstantsTable(Class table)
			throws NoSuchMethodException, InvocationTargetException,
			IllegalAccessException {
		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		HashMap<String, Object> constantTable = new HashMap<String, Object>();

		// Get all table entries.
		List list = session.createCriteria(table).list();

		// Put all database objects in this table into the result hashmap with
		// the value of the constant as key.
		for (Iterator iter = list.iterator(); iter.hasNext();) {
			Object o = iter.next();
			Class parameterTypes[] = {};
			// NOTE: All constant tables have a column which has the same name
			// as the table.
			Method getter = table.getMethod("get" + table.getSimpleName(),
					parameterTypes);
			Object parameters[] = {};
			constantTable.put(((String) getter.invoke(o, parameters))
					.toLowerCase(), o);
		}
		tx.commit();

		return constantTable;
	}

	/***************************************************************************
	 * Deletes a Dias database object from the database.
	 * 
	 * TODO Check if this method has to be synchronized.
	 * 
	 * TODO Check why this method declaration doesn't need to throw an
	 * exception.
	 * 
	 * @param hibObject the object to delete
	 **************************************************************************/
	public static synchronized void deleteObject(Object hibObject) {
		if (hibObject instanceof Dias) {
			Dias newDias = (Dias) hibObject;
			String urn = newDias.getDiasintid();
			HashMap diasUrnLocks = locks.get(Dias.class);

			if (diasUrnLocks.containsKey(urn)
					&& !((ReentrantLock) diasUrnLocks.get(urn))
							.isHeldByCurrentThread()) {
				throw new LockAcquisitionException("Dias object with id " + urn
						+ "is locked", new SQLException());
			}
		}

		Session session = HibernateUtil.getSession();
		Transaction tx = session.beginTransaction();
		session.delete(hibObject);
		tx.commit();

		// TODO Must there be a closeSession call??
		// HibernateUtil.closeSession();
	}

}
