/*******************************************************************************
 * de.langzeitarchivierung.kopal.util.db / HibernateUtil.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 ******************************************************************************
 * 
 * 
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.util.db;

import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.hibernate.HibernateException;
import org.hibernate.Interceptor;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.cfg.Configuration;

/*******************************************************************************
 * Basic Hibernate helper class, handles SessionFactory, Session and
 * Transaction.
 * <p>
 * Uses a static initializer for the initial SessionFactory creation and holds
 * Session and Transactions in thread local variables. All exceptions are
 * wrapped in an unchecked InfrastructureException.
 * 
 * @author christian@hibernate.org
 ******************************************************************************/

public class HibernateUtil
{

	// CLASS VARIABLES (Static variables) *********************************

	protected static Logger							defaultLogger				= Logger
																						.getLogger("de.langzeitarchivierung.kopal");

	private static String							hibernateDbHost;
	private static String							hibernateDbName;
	private static String							hibernateConnectionUsername;
	private static String							hibernateConnectionPassword;
	private static String							hibernateConnectionDriver_class;
	private static String							hibernateDialect;
	private static String							hibernateLogLevel;
	private static String							hibernateCharacterEncoding	= "UTF-8";

	private static Level							lLevel;

	private static Configuration					configuration;
	private static SessionFactory					sessionFactory;
	private static final ThreadLocal<Session>		threadSession				= new ThreadLocal<Session>();
	private static final ThreadLocal<Transaction>	threadTransaction			= new ThreadLocal<Transaction>();
	private static final ThreadLocal<Interceptor>	threadInterceptor			= new ThreadLocal<Interceptor>();

	/** Create the initial SessionFactory from the default configuration files. */
	public static void initalize() {
		try {
			// Get the hibernate logger.
			Logger hibernateLogger = defaultLogger.getParent();

			// Check the hibernate log level in config file and set it. Use
			// Level.WARNING if there is no correct level given.
			try {
				lLevel = Level.parse(hibernateLogLevel);
			}
			catch (Exception e) {
				lLevel = Level.parse("WARNING");
			}
			hibernateLogger.setLevel(lLevel);

			// Remove all handlers from the hibernate logger and add all kopal
			// logger handlers to the hibernate logger.
			Handler hibernateHandlers[] = hibernateLogger.getHandlers();
			for (int i = 0; i < hibernateHandlers.length; i++) {
				hibernateLogger.removeHandler(hibernateHandlers[i]);
			}
			Handler defaultHandlers[] = defaultLogger.getHandlers();
			for (int i = 0; i < defaultHandlers.length; i++) {
				hibernateLogger.addHandler(defaultHandlers[i]);
			}

			defaultLogger.log(Level.INFO, "Hibernate log level set to "
					+ lLevel.toString());

			// Configure the Hibernate.
			configuration = new Configuration();
			configuration.configure("hibernate.cfg.xml");

			String hibernateConnectionUrl = "jdbc:mysql://" + hibernateDbHost
					+ "/" + hibernateDbName + "?characterEncoding="
					+ hibernateCharacterEncoding;

			configuration.setProperty("hibernate.connection.url",
					hibernateConnectionUrl).setProperty(
					"hibernate.connection.username",
					hibernateConnectionUsername).setProperty(
					"hibernate.connection.password",
					hibernateConnectionPassword).setProperty(
					"hibernate.connection.driver_class",
					hibernateConnectionDriver_class).setProperty(
					"hibernate.dialect", hibernateDialect);
			// .setProperty("hbm2ddl.auto","update");

			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Identifier.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Fileformat.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/File.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Owner.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Server.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Various.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Techmd.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Dublincore.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Format.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Checksumtype.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Input.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Description.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Event.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Dcelement.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Dias.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Idtype.hbm.xml");
			configuration
					.addResource("de/langzeitarchivierung/kopal/util/db/Modulestatus.hbm.xml");

			Configuration conf = configuration.configure();
			sessionFactory = conf.buildSessionFactory();

			// We could also let Hibernate bind it to JNDI:
			// configuration.configure().buildSessionFactory()
		}
		catch (Throwable ex) {
			// We have to catch Throwable, otherwise we will miss
			// NoClassDefFoundError and other subclasses of Error
			defaultLogger.log(Level.SEVERE, "Building SessionFactory failed.");
			ex.printStackTrace();
			throw new ExceptionInInitializerError(ex);
		}
	}

	/***************************************************************************
	 * Returns the SessionFactory used for this static class.
	 * 
	 * @return SessionFactory
	 **************************************************************************/
	public static SessionFactory getSessionFactory() {
		// Instead of a static variable, use JNDI:
		//		 
		// SessionFactory sessions = null;
		// try {
		// Context ctx = new InitialContext();
		// String jndiName = "java:hibernate/HibernateFactory";
		// sessions = (SessionFactory) ctx.lookup(jndiName);
		// }
		// catch (NamingException ex) {
		// throw new InfrastructureException(ex);
		// }
		// return sessions;
		return sessionFactory;
	}

	/***************************************************************************
	 * Returns the original Hibernate configuration.
	 * 
	 * @return Configuration
	 **************************************************************************/
	public static Configuration getConfiguration() {
		return configuration;
	}

	/***************************************************************************
	 * Rebuild the SessionFactory with the static Configuration.
	 * 
	 * @throws InfrastructureException
	 **************************************************************************/
	public static void rebuildSessionFactory() throws InfrastructureException {
		synchronized (sessionFactory) {
			try {
				sessionFactory = getConfiguration().buildSessionFactory();
			}
			catch (Exception ex) {
				throw new InfrastructureException(ex);
			}
		}
	}

	/***************************************************************************
	 * Rebuild the SessionFactory with the given Hibernate Configuration.
	 * 
	 * @param cfg
	 * @throws InfrastructureException
	 **************************************************************************/
	public static void rebuildSessionFactory(Configuration cfg)
			throws InfrastructureException {
		synchronized (sessionFactory) {
			try {
				sessionFactory = cfg.buildSessionFactory();
				configuration = cfg;
			}
			catch (Exception ex) {
				throw new InfrastructureException(ex);
			}
		}
	}

	/***************************************************************************
	 * Retrieves the current Session local to the thread. <p/> If no Session is
	 * open, opens a new Session for the running thread.
	 * 
	 * @return Session
	 * @throws InfrastructureException
	 **************************************************************************/
	public static Session getSession() throws InfrastructureException {
		Session s = threadSession.get();
		try {
			if (s == null) {
				// log.debug("Opening new Session for this thread.");
				if (getInterceptor() != null) {
					defaultLogger.log(Level.FINER, "Using interceptor: "
							+ getInterceptor().getClass());
					s = getSessionFactory().openSession(getInterceptor());
				}
				else {
					s = getSessionFactory().openSession();
				}
				threadSession.set(s);
			}
		}
		catch (HibernateException ex) {
			throw new InfrastructureException(ex);
		}
		return s;
	}

	/***************************************************************************
	 * Closes the Session local to the thread.
	 * 
	 * @throws InfrastructureException
	 **************************************************************************/
	public static void closeSession() throws InfrastructureException {
		try {
			Session s = threadSession.get();
			threadSession.set(null);
			if (s != null && s.isOpen()) {
				// log.debug("Closing Session of this thread.");
				s.close();
			}
		}
		catch (HibernateException ex) {
			throw new InfrastructureException(ex);
		}
	}

	/***************************************************************************
	 * Start a new database transaction.
	 * 
	 * @throws InfrastructureException
	 **************************************************************************/
	public static void beginTransaction() throws InfrastructureException {
		Transaction tx = threadTransaction.get();
		try {
			if (tx == null) {
				defaultLogger.log(Level.FINER,
						"Starting new database transaction in this thread.");
				tx = getSession().beginTransaction();
				threadTransaction.set(tx);
			}
		}
		catch (HibernateException ex) {
			throw new InfrastructureException(ex);
		}
	}

	/***************************************************************************
	 * Commit the database transaction.
	 * 
	 * @throws InfrastructureException
	 **************************************************************************/
	public static void commitTransaction() throws InfrastructureException {
		Transaction tx = threadTransaction.get();
		try {
			if (tx != null && !tx.wasCommitted() && !tx.wasRolledBack()) {
				defaultLogger.log(Level.FINER,
						"Committing database transaction of this thread.");
				tx.commit();
			}
			threadTransaction.set(null);
		}
		catch (HibernateException ex) {
			rollbackTransaction();
			throw new InfrastructureException(ex);
		}
	}

	/***************************************************************************
	 * Commit the database transaction.
	 * 
	 * @throws InfrastructureException
	 **************************************************************************/
	public static void rollbackTransaction() throws InfrastructureException {
		Transaction tx = threadTransaction.get();
		try {
			threadTransaction.set(null);
			if (tx != null && !tx.wasCommitted() && !tx.wasRolledBack()) {
				defaultLogger
						.log(Level.FINER,
								"Tyring to rollback database transaction of this thread.");
				tx.rollback();
			}
		}
		catch (HibernateException ex) {
			throw new InfrastructureException(ex);
		}
		finally {
			closeSession();
		}
	}

	/***************************************************************************
	 * Reconnects a Hibernate Session to the current Thread.
	 * 
	 * @param session The Hibernate Session to be reconnected.
	 * @throws InfrastructureException
	 **************************************************************************/
	public static void reconnect(Session session)
			throws InfrastructureException {
		try {
			session.reconnect();
			threadSession.set(session);
		}
		catch (HibernateException ex) {
			throw new InfrastructureException(ex);
		}
	}

	/***************************************************************************
	 * Disconnect and return Session from current Thread.
	 * 
	 * @return Session the disconnected Session
	 * @throws InfrastructureException
	 **************************************************************************/
	public static Session disconnectSession() throws InfrastructureException {

		Session session = getSession();
		try {
			threadSession.set(null);
			if (session.isConnected() && session.isOpen())
				session.disconnect();
		}
		catch (HibernateException ex) {
			throw new InfrastructureException(ex);
		}
		return session;
	}

	/***************************************************************************
	 * Register a Hibernate interceptor with the current thread.
	 * <p>
	 * Every Session opened is opened with this interceptor after registration.
	 * Has no effect if the current Session of the thread is already open,
	 * effective on next close()/getSession().
	 * 
	 * @param interceptor
	 **************************************************************************/
	public static void registerInterceptor(Interceptor interceptor) {
		threadInterceptor.set(interceptor);
	}

	// GET & SET METHODS *******************************************************

	/***************************************************************************
	 * @return Interceptor
	 **************************************************************************/
	private static Interceptor getInterceptor() {
		Interceptor interceptor = threadInterceptor.get();
		return interceptor;
	}

	/***************************************************************************
	 * @param _hibernateConnectionUsername The database user name to connect
	 *            with.
	 **************************************************************************/
	public static void setHibernateConnectionUsername(
			String _hibernateConnectionUsername) {
		hibernateConnectionUsername = _hibernateConnectionUsername;
	}

	/***************************************************************************
	 * @return Returns the hibernateConnectionUsername.
	 **************************************************************************/
	public static String getHibernateConnectionUsername() {
		return hibernateConnectionUsername;
	}

	/***************************************************************************
	 * @param _hibernateDbHost The database host name.
	 **************************************************************************/
	public static void setHibernateDbHost(String _hibernateDbHost) {
		hibernateDbHost = _hibernateDbHost;
	}

	/***************************************************************************
	 * @return Returns the hibernateDbHost.
	 **************************************************************************/
	public static String getHibernateDbHost() {
		return hibernateDbHost;
	}

	/***************************************************************************
	 * @param _hibernateDbName The database name.
	 **************************************************************************/
	public static void setHibernateDbName(String _hibernateDbName) {
		hibernateDbName = _hibernateDbName;
	}

	/***************************************************************************
	 * @return Returns the hibernateDbName.
	 **************************************************************************/
	public static String getHibernateDbName() {
		return hibernateDbName;
	}

	/***************************************************************************
	 * @param _hibernateConnectionPassword The database password to connect
	 *            with.
	 **************************************************************************/
	public static void setHibernateConnectionPassword(
			String _hibernateConnectionPassword) {
		hibernateConnectionPassword = _hibernateConnectionPassword;
	}

	/***************************************************************************
	 * @return Returns the hibernateConnectionPassword.
	 **************************************************************************/
	public static String getHibernateConnectionPassword() {
		return hibernateConnectionPassword;
	}

	/***************************************************************************
	 * @param _hibernateConnectionDriver_class The hibernate connection driver
	 *            class to set.
	 **************************************************************************/
	public static void setHibernateConnectionDriver_class(
			String _hibernateConnectionDriver_class) {
		hibernateConnectionDriver_class = _hibernateConnectionDriver_class;
	}

	/***************************************************************************
	 * @return Returns the hibernateConnectionDriver_class.
	 * @author Stefan Funk, SUB Göttingen 25.09.2006
	 **************************************************************************/
	public static String getHibernateConnectionDriver_class() {
		return hibernateConnectionDriver_class;
	}

	/***************************************************************************
	 * @param _hibernateDialect The hibernate dialect to set.
	 **************************************************************************/
	public static void setHibernateDialect(String _hibernateDialect) {
		hibernateDialect = _hibernateDialect;
	}

	/***************************************************************************
	 * @return Returns the hibernateDialect.
	 **************************************************************************/
	public static String getHibernateDialect() {
		return hibernateDialect;
	}

	/***************************************************************************
	 * @param _hibernateLogLevel The hibernate log level to set.
	 **************************************************************************/
	public static void setHibernateLogLevel(String _hibernateLogLevel) {
		hibernateLogLevel = _hibernateLogLevel;
	}

	/***************************************************************************
	 * @return Returns the hibernateLogLevel.
	 **************************************************************************/
	public static String getHibernateLogLevel() {
		return hibernateLogLevel;
	}

	/***************************************************************************
	 * @param hibernateCharacterEncoding The hibernateCharacterEncoding to set.
	 * @author Stefan Funk, SUB Göttingen 05.07.2007
	 **************************************************************************/
	public static void sethibernateCharacterEncoding(
			String hibernateCharacterEncoding) {
		HibernateUtil.hibernateCharacterEncoding = hibernateCharacterEncoding;
	}

	/**
	 * This exception is used to mark (fatal) failures in infrastructure and
	 * system code.
	 * 
	 * @author Christian Bauer <christian@hibernate.org>
	 */
	public static class InfrastructureException extends RuntimeException
	{
		private static final long	serialVersionUID	= -5431126070176417318L;

		/***********************************************************************
		 * Default Constructor
		 **********************************************************************/
		public InfrastructureException() {
			super();
		}

		/***********************************************************************
		 * Constructor
		 * 
		 * @param message
		 **********************************************************************/
		public InfrastructureException(String message) {
			super(message);
		}

		/***********************************************************************
		 * Default Constructor
		 * 
		 * @param message
		 * @param cause
		 **********************************************************************/
		public InfrastructureException(String message, Throwable cause) {
			super(message, cause);
		}

		/***********************************************************************
		 * Default Constructor
		 * 
		 * @param cause
		 **********************************************************************/
		public InfrastructureException(Throwable cause) {
			super(cause);
		}
	}

}
