/*******************************************************************************
 * de.langzeitarchivierung.kopal.util.db / MappToHib.java
 * 
 * Copyright 2005-2007 by Project kopal
 *
 * http://kopal.langzeitarchivierung.de/
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 10.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.util.db;

import gov.loc.mets.FileGrpType;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.hibernate.QueryException;
import org.purl.dc.elements.x11.DcSetDocument.DcSet;

import de.ddb.lmeRfile.LmerFile;
import de.ddb.lmeRfile.LmerFile.Format;
import de.ddb.lmeRfile.LmerFile.XmlData;
import de.langzeitarchivierung.kopal.ProcessData;
import de.langzeitarchivierung.kopal.Step;
import de.langzeitarchivierung.kopal.formats.Uof;
import de.langzeitarchivierung.kopal.retrieval.DiasAccessMetadataResponse;
import de.langzeitarchivierung.kopal.util.ChecksumGenerator;
import de.langzeitarchivierung.kopal.util.KopalTimestamp;

/*******************************************************************************
 * Helper class for mapping logical objects and functions to the database.
 * 
 * @author Jens Ludwig, SUB Goettingen
 * @version 20070612
 ******************************************************************************/

public class MappToHib
{

	// CLASS VARIABLES (Static variables) *********************************

	/** Dublin Core element names */
	public static final String[]			DC_ELEMENT_NAMES	= {
			"contributor", "title", "creator", "subject", "description",
			"publisher", "date", "type", "format", "identifier", "source",
			"language", "relation", "coverage", "rights"		};

	/** Constants in the database for identifier types (idtype) */
	public static final String				ID_GDZPPN			= "GDZ-PPN";
	public static final String				ID_PPN				= "PPN";
	public static final String				ID_LPI				= "LPI";
	public static final String				ID_DIASOLDOBJID		= "DIASOLDOBJID";
	public static final String				ID_DIASINTID		= "DIASINTID";
	public static final String				ID_DIASEXTID		= "DIASEXTID";
	public static final String				ID_GBVPPN			= "GBV-PPN";
	public static final String				ID_GOOBIPROCID		= "GOOBIPROCID";
	public static final String				ID_DIASTICKETID		= "DIASTICKETID";

	/** Constants in the database for checksum types * */
	public static final String				CHECKSUMTYPE_MD5	= "MD5";
	public static final String				CHECKSUMTYPE_SHA1	= "SHA-1";
	public static final String				CHECKSUMTYPE_XOR	= "xor of sha1 file checksums";

	private static HashMap<Class, HashMap>	dbConstantsTables;

	// private static HashMap dbClassIdTables;

	// MANIPULATION (Manipulation - what the object does) **********************

	/***************************************************************************
	 * Initializes MappToHib with constants from the database for caching.
	 * 
	 * A constant is database entry which does not change during runtime and
	 * which has a column named identical to the table. It is usually only
	 * referenced in other database tables.
	 * 
	 * TODO check if this a unnecessary optimization especially as we use
	 * reflection in add- and getDc()
	 * 
	 * @throws Exception
	 **************************************************************************/
	public static void initialize() throws Exception {
		dbConstantsTables = new HashMap<Class, HashMap>();
		dbConstantsTables.put(Owner.class, DBAccess
				.getDbConstantsTable(Owner.class));
		dbConstantsTables.put(Server.class, DBAccess
				.getDbConstantsTable(Server.class));
		dbConstantsTables.put(Idtype.class, DBAccess
				.getDbConstantsTable(Idtype.class));
		dbConstantsTables.put(Dcelement.class, DBAccess
				.getDbConstantsTable(Dcelement.class));
		dbConstantsTables.put(Modulestatus.class, DBAccess
				.getDbConstantsTable(Modulestatus.class));
		dbConstantsTables.put(Checksumtype.class, DBAccess
				.getDbConstantsTable(Checksumtype.class));
	}

	/***************************************************************************
	 * Adds the dublincore metadata of a process data object with the uof
	 * metadata format into the database.
	 * 
	 * @param pData
	 * @throws Exception
	 **************************************************************************/
	public static void addDc(ProcessData pData) throws Exception {
		Input in = getInput(pData);
		DcSet dcXml = ((Uof) pData.getMetadata()).getDcCopy();
		if (dcXml == null) {
			return;
		}
		Set<Dublincore> dcHibs = new HashSet<Dublincore>();

		// Generate database object with the correct Dc values we use reflection
		// to avoid too many boring distinction of cases.
		// TODO Yes, boring distinction of cases would be regarded as better
		// code.
		// TODO So, if you like call 15 times:
		// dcHibs.addAll(createDcelements(in, "XXXXXX",
		// dcXml.getXXXXXXArray()));
		// TODO No, we do not!
		for (int i = 0; i < DC_ELEMENT_NAMES.length; i++) {
			// Construct the method name.
			String methodName = DC_ELEMENT_NAMES[i].substring(0, 1)
					.toUpperCase()
					+ DC_ELEMENT_NAMES[i].substring(1);
			methodName = "get" + methodName + "Array";

			// Get the method.
			Method getArray = DcSet.class.getMethod(methodName);
			// Invoke the method with the correct value.
			Object args[] = {};
			String[] dcValues = (String[]) getArray.invoke(dcXml, args);
			dcHibs.addAll(createDcelements(in, DC_ELEMENT_NAMES[i], dcValues));
		}

		DBAccess.addObjects(dcHibs);
	}

	/***************************************************************************
	 * Returns the dublin core metadata for the specified input object.
	 * 
	 * @param in
	 * @return the dublin core metadata
	 * @throws Exception
	 **************************************************************************/
	public static DcSet getDc(Input in) throws Exception {
		DcSet result = DcSet.Factory.newInstance();
		Set<Dublincore> dcs = in.getDublincores();

		// Add all DC database entries to the DcSet result. We use reflection to
		// avoid too many boring distinctions of cases.
		// TODO Yes, boring distinction of cases would be regarded as better
		// code.
		for (Iterator<Dublincore> dcIter = dcs.iterator(); dcIter.hasNext();) {
			Dublincore dc = dcIter.next();
			// Construct the method's name.
			String dcElement = dc.getDcelement().getDcelement();
			dcElement = dcElement.substring(0, 1).toUpperCase()
					+ dcElement.substring(1);
			// Get the method.
			Class parameterTypes[] = { String.class };
			Method setter = DcSet.class.getMethod("add" + dcElement,
					parameterTypes);
			// Invoke the method with the correct value.
			Object args[] = { dc.getDcvalue() };
			setter.invoke(result, args);
		}

		return result;
	}

	/***************************************************************************
	 * Returns all internal IDs of object modified between two timestamps.
	 * 
	 * @param begin time in milliseconds since the begin of the epoch
	 * @param end time in milliseconds since the begin of the epoch
	 * @return a list internal IDs
	 * @throws Exception
	 **************************************************************************/
	public static List<String> getLastDiasModified(long begin, long end)
			throws Exception {
		List sipList = DBAccess.getObjectPerDate(Dias.class, "diasdate", Long
				.valueOf(begin), Long.valueOf(end));
		List<String> result = new ArrayList<String>(sipList.size());

		for (Iterator sipIter = sipList.iterator(); sipIter.hasNext();) {
			String diasIntId = ((Dias) sipIter.next()).getDiasintid();
			if (diasIntId != null) {
				result.add(diasIntId);
			}
		}

		return result;
	}

	/***************************************************************************
	 * Checks if this user password is correct.
	 * 
	 * @param user
	 * @param password
	 * @return true if the password is correct, false otherwise
	 * @throws Exception
	 **************************************************************************/
	public static boolean isAuthenticated(String user, String password)
			throws Exception {
		Owner own = (Owner) DBAccess.getUniqueObject(Owner.class, "owner",
				user, "pass", password);
		if (own != null) {
			return true;
		}

		return false;
	}

	/***************************************************************************
	 * Checks if the user is the owner of the input object.
	 * 
	 * @param input The input object for which to check the authorization
	 * @param user The user which requests authorization
	 * @return true if user is owner of the object, false otherwise
	 **************************************************************************/
	public static boolean isAuthorized(Input input, String user) {
		if (input.getOwner().getOwner().equals(user)) {
			return true;
		}
		return false;
	}

	/***************************************************************************
	 * Checks if the user is the owner of the object with this identifier type
	 * and value. If there are more than one object with this ID type and value
	 * the latest ingested object is checked.
	 * 
	 * @param idType the identifier type of the object
	 * @param idValue the identifier value of the object
	 * @param user The user which requests authorization
	 * @return true if user is owner of the object, false otherwise or if no
	 *         such object exists.
	 * @throws Exception if something goes wrong with the database access.
	 **************************************************************************/
	public static boolean isAuthorized(String idType, String idValue,
			String user) throws Exception {
		Input input = MappToHib.getInput(idType, idValue);
		if (input == null) {
			return false;
		}
		return MappToHib.isAuthorized(input, user);
	}

	/***************************************************************************
	 * Adds some input data to the input table.
	 * 
	 * @param pData The process data object for which an entry will be added to
	 *            the database.
	 * @param owner The owner which will be added to the object.
	 * @param server The server which received the object.
	 * @throws Exception if something goes wrong with the database access.
	 **************************************************************************/
	public static void addInput(ProcessData pData, String owner, String server)
			throws Exception {
		Input in = new Input();

		// Set the startdate, the server and the owner.
		in.setStartdate(pData.getStartDate());
		in.setServer((Server) getDbConstant(Server.class, server));
		in.setOwner((Owner) getDbConstant(Owner.class, owner));

		// Add the main input entry and set the maybe corrected start date.
		pData.setStartDate(DBAccess.addInputUpdateStartdate(in));
	}

	/***************************************************************************
	 * Only the XOR checksum is added to the database's input table.
	 * 
	 * @param pData The process data object for which the checksum will be added
	 *            to the database.
	 * @param xorChecksum The checksum to add as
	 *            ChecksumGenerator.XOR_OF_FILE_CHECKSUMS
	 * @throws Exception if something goes wrong with the database access.
	 **************************************************************************/
	public static void addInputChecksum(ProcessData pData, String xorChecksum)
			throws Exception {
		Input in = getInput(pData);

		in.setChecksum(xorChecksum);
		in.setChecksumtype((Checksumtype) getDbConstant(Checksumtype.class,
				ChecksumGenerator.XOR_OF_FILE_CHECKSUMS));

		// Add the checksum.
		DBAccess.addObject(in);
	}

	/***************************************************************************
	 * Adds the custom IDs of a process data object to the database if they are
	 * not already present.
	 * 
	 * @param pData The process data object for which the identifiers will be
	 *            added to the database.
	 * @throws Exception if something goes wrong with the database access.
	 **************************************************************************/
	public static void addIdentifier(ProcessData pData) throws Exception {
		if (!pData.getCustomIds().isEmpty()) {
			HashMap ids = pData.getCustomIds();
			Set keys = ids.keySet();

			// Size of hashset is set to two so that no new hashset is
			// necessary.
			HashSet<Identifier> idsToAdd = new HashSet<Identifier>(
					keys.size() * 2);

			for (Iterator keyIter = keys.iterator(); keyIter.hasNext();) {
				String key = (String) keyIter.next();

				Idtype idType = (Idtype) getDbConstant(Idtype.class, key);
				String idValue = (String) ids.get(key);
				Input in = getInput(pData);

				// If this identifier is not already present, add it.
				// TODO Check what to do if there is more than one identifier
				// of this type, because getIdValue() only returns the latest.
				String alreadyInDb = getIdValue(in, idType.getIdtype());
				if (alreadyInDb == null || !alreadyInDb.equals(idValue)) {
					Identifier id = new Identifier();
					id.setInput(in);
					id.setIdtype(idType);
					id.setIdvalue(idValue);
					idsToAdd.add(id);
				}
			}

			DBAccess.addObjects(idsToAdd);
		}
	}

	/***************************************************************************
	 * Returns the database input object for a process data object.
	 * 
	 * @param pData the process data object
	 * @return the database input object
	 * @throws Exception
	 **************************************************************************/
	public static Input getInput(ProcessData pData) throws Exception {
		return DBAccess.getInputPerStartdate(pData.getStartDate());
	}

	/***************************************************************************
	 * Returns the input object with *an* identifier with the specified type and
	 * value. If more than one identifier exist the latest is returned. If no
	 * are available null is returned.
	 * 
	 * @param idType the identifier type for the input object
	 * @param idValue the identifier value for the input object
	 * @return the input object with the specified ID
	 * @throws Exception
	 **************************************************************************/
	public static Input getInput(String idType, String idValue)
			throws Exception {
		Input result = null;

		if (idType.equals(ID_DIASEXTID)) {
			List sips = DBAccess.getObjectsDescending(Dias.class, "urn",
					idValue, "diasdate");
			if (!sips.isEmpty()) {
				result = ((Dias) sips.get(0)).getInput();
			}
		}
		else if (idType.equals(ID_DIASINTID)) {
			List sips = DBAccess.getObjectsDescending(Dias.class, "diasintid",
					idValue, "diasdate");
			if (sips.size() == 1) {
				result = ((Dias) sips.get(0)).getInput();
			}
		}
		else if (idType.equals(ID_DIASOLDOBJID)) {
			List sips = DBAccess.getObjectsDescending(Dias.class,
					"diasoldobjid", idValue, "diasdate");
			if (!sips.isEmpty()) {
				result = ((Dias) sips.get(0)).getInput();
			}
		}
		else {
			int idTypeId = ((Idtype) DBAccess.getUniqueObject(Idtype.class,
					"idtype", idType)).getIdtypeid();
			List ids = DBAccess.getObjectsDescending(Identifier.class,
					"idvalue", idValue, "idtype.idtypeid",
					new Integer(idTypeId), "identifierid");
			if (!ids.isEmpty()) {
				result = ((Identifier) ids.get(0)).getInput();
			}
		}

		return result;
	}

	/***************************************************************************
	 * Adds database entries for all files of the specified processData object.
	 * The information for the files is retrieved from an uof metadata object,
	 * the existence of this is necessary. Optionally technical metadata for
	 * each file can be stored.
	 * 
	 * @param pData the process data object for which the file information will
	 *            be stored
	 * @param storeTechMd flag for storing technical metadata for each file
	 * @throws Exception
	 * @throws Exception if database access fails
	 **************************************************************************/
	public static void addFiles(ProcessData pData, boolean storeTechMd)
			throws Exception {
		Uof metadata = (Uof) pData.getMetadata();
		File mdFiles[] = metadata.getFiles();
		HashSet<de.langzeitarchivierung.kopal.util.db.File> filesToAdd = new HashSet<de.langzeitarchivierung.kopal.util.db.File>(
				mdFiles.length * 2);

		// At first add all the file objects only.
		for (int i = 0; i < mdFiles.length; i++) {
			de.langzeitarchivierung.kopal.util.db.File dbFile = new de.langzeitarchivierung.kopal.util.db.File();
			dbFile.setInput(getInput(pData));

			FileGrpType.File fileMd = metadata.getFileMdMets(mdFiles[i]);

			dbFile.setChecksum(fileMd.getCHECKSUM());
			dbFile.setChecksumtype((Checksumtype) getDbConstant(
					Checksumtype.class, fileMd.getCHECKSUMTYPE().toString()));
			dbFile.setSize(fileMd.getSIZE());
			dbFile.setFilename(fileMd.getFLocatArray(0).getHref().substring(
					Uof.FILE_LOCAT_HREF_PREFIX.length()));

			dbFile.setFiledate(fileMd.getCREATED().getTimeInMillis());
			filesToAdd.add(dbFile);
		}
		DBAccess.addObjects(filesToAdd);

		// Create a hashset for technical metadata only if we need it.
		HashSet<Techmd> techMdToAdd = null;
		if (storeTechMd) {
			techMdToAdd = new HashSet<Techmd>(mdFiles.length * 2);
		}
		HashSet<Fileformat> fileFormatsToAdd = new HashSet<Fileformat>(
				mdFiles.length * 2);

		// Create for the just created database file entries the file format and
		// optionally the techMd database entries.
		Set files = getInput(pData).getFiles();
		for (Iterator fileIter = files.iterator(); fileIter.hasNext();) {
			de.langzeitarchivierung.kopal.util.db.File dbFile = (de.langzeitarchivierung.kopal.util.db.File) fileIter
					.next();
			File contentFile = new File(new URI(metadata.getBasePath()
					.toString()
					+ dbFile.getFilename()));
			LmerFile lmerFileMd = metadata.getLmerFileCopy(contentFile);

			if (storeTechMd) {
				Techmd dbFileTechMd = new Techmd();
				dbFileTechMd.setFile(dbFile);
				XmlData techMd = lmerFileMd.getXmlDataArray(0);

				// TODO Remove fragment of techMd?
				dbFileTechMd.setTechmd(techMd.xmlText());
				techMdToAdd.add(dbFileTechMd);
			}

			Format formats[] = lmerFileMd.getFormatArray();
			for (int i = 0; i < formats.length; i++) {

				String format = lmerFileMd.getFormatArray(i).getStringValue();
				String version = "";

				de.langzeitarchivierung.kopal.util.db.Format formatDb = (de.langzeitarchivierung.kopal.util.db.Format) DBAccess
						.getUniqueObject(
								de.langzeitarchivierung.kopal.util.db.Format.class,
								"format", format, "version", version);

				// Add unknown file formats to database.
				if (formatDb == null) {
					formatDb = new de.langzeitarchivierung.kopal.util.db.Format();
					formatDb.setFormat(format);
					formatDb.setVersion(version);
					DBAccess.addObject(formatDb);

					// After adding the object, we have to ask for the object
					// again, because the key is not correctly updated with
					// addObject().
					formatDb = (de.langzeitarchivierung.kopal.util.db.Format) DBAccess
							.getUniqueObject(
									de.langzeitarchivierung.kopal.util.db.Format.class,
									"format", format, "version", version);
				}

				Fileformat dbFileFormat = new Fileformat();
				dbFileFormat.setFile(dbFile);
				dbFileFormat.setFormat(formatDb);
				fileFormatsToAdd.add(dbFileFormat);
			}
		}

		if (storeTechMd) {
			DBAccess.addObjects(techMdToAdd);
		}

		DBAccess.addObjects(fileFormatsToAdd);
	}


	/***************************************************************************
	 * Adds an event entry to the database.
	 * 
	 * @param pData The process object to which the event belongs
	 * @param pStep The step object where the event occured
	 * @param eventDate The timestamp when the event occured
	 * @throws Exception if something fails
	 **************************************************************************/
	public static void addEvent(ProcessData pData, Step pStep, long eventDate)
			throws Exception {
		addEvent(pData, pStep.getActionName(), pStep.getStatus()
				.getStatusText(), pStep.getStatus().getDescription(), pStep
				.getStatus().getCurrentFile(), eventDate);
	}

	/***************************************************************************
	 * Adds an event entry to the database.
	 * 
	 * @param pData The process object to which the event belongs
	 * @param eventThrowingModul The class which threw the event
	 * @param eventModuleStatus The status of the event
	 * @param eventDescription The description of the event
	 * @param eventFile If the event belongs to a specific it can be specified.
	 * @param eventDate The timestamp when the event occured
	 * @throws Exception if something fails
	 **************************************************************************/
	private static void addEvent(ProcessData pData, String eventThrowingModul,
			String eventModuleStatus, String eventDescription, File eventFile,
			long eventDate) throws Exception {
		Description evDesc = null;

		// Create a new event and set some event data.
		Event ev = new Event();
		ev.setInput(getInput(pData));
		ev.setEventdate(eventDate);
		ev.setModule(eventThrowingModul);
		ev.setModulestatus((Modulestatus) getDbConstant(Modulestatus.class,
				eventModuleStatus));

		String descriptionString = eventDescription;

		// Get the description object from the database, if one with the
		// descriptionString already is existing.
		evDesc = (Description) DBAccess.getUniqueObject(Description.class,
				"description", descriptionString);

		// If this kind of description does not exist yet, create it.
		if (evDesc == null) {

			// DEBUG OUTPUT!
			//
			// System.out.println("NEW DESCRIPTION : " + descriptionString + "
			// ("
			// + eventDate + ")");
			//
			// DEBUG OUTPUT!

			evDesc = new Description();
			evDesc.setDescription(descriptionString);
			DBAccess.addObject(evDesc);

			// We need to retrieve the new insert description again because the
			// retrieved object only gets an ID at retrieval, and we need the
			// ID.
			evDesc = (Description) DBAccess.getUniqueObject(Description.class,
					"description", descriptionString);
		}
		else {
			// DEBUG OUTPUT!
			//
			// System.out.println("OLD DESCRIPTION : " + descriptionString + "
			// ("
			// + eventDate + ")");
			//
			// DEBUG OUTPUT!
		}

		// Get the file reference from the status value, if existing.
		File currentFile = eventFile;

		if (currentFile != null) {
			// Get the relative path of the filename using URIs and add.
			URI ptcf = new File(pData.getPathToContentFiles()).toURI();
			URI ap = new File(currentFile.getAbsolutePath()).toURI();
			String path = ptcf.relativize(ap).toString();
			ev.setFilename(path);
		}

		ev.setDescription(evDesc);
		DBAccess.addObject(ev);
	}

	/***************************************************************************
	 * Removes a Dias table entry from the database.
	 * 
	 * @param pData The process data for which to remove the DIAS entry.
	 * @throws Exception
	 **************************************************************************/
	public static void removeDias(ProcessData pData) throws Exception {
		// The lock is made with the URN because we can only allow simultanous
		// ingests with different URNs if we want to get correct feedback.
		String urn = pData.getMetadata().getObjectId();
		DBAccess.lockTable(Dias.class, urn);

		try {
			// TODO This presupposes that there is only one DIAS entry for each
			// input.
			Dias entryToRemove = getInput(pData).getDiases().iterator().next();
			DBAccess.deleteObject(entryToRemove);
		}
		finally {
			DBAccess.unlockTable(Dias.class, urn);
		}
	}

	/***************************************************************************
	 * Removes all File table entries from the database for the give object
	 * invluding the Techmd entries.
	 * 
	 * @param pData
	 * @throws Exception
	 **************************************************************************/
	public static void removeFile(ProcessData pData) throws Exception {
		MappToHib.removeTableEntry(pData,
				de.langzeitarchivierung.kopal.util.db.File.class);
	}

	/***************************************************************************
	 * Removes all Techmd table entries from the database for the given object.
	 * 
	 * TODO Code redundance here!
	 * 
	 * @param pData
	 * @throws Exception
	 **************************************************************************/
	public static void removeTechmdOnly(ProcessData pData) throws Exception {
		// First get the inputid of the given ProcessData object.
		Input in = DBAccess.getInputPerStartdate(pData.getStartDate());
		int inputid = in.getInputid();

		// Get all File objects for the given inputid.
		List fileList = new ArrayList();
		fileList = DBAccess.getObjects(
				de.langzeitarchivierung.kopal.util.db.File.class,
				"input.inputid", new Integer(inputid));

		de.langzeitarchivierung.kopal.util.db.File file;
		Iterator iter = fileList.iterator();
		while (iter.hasNext()) {
			file = (de.langzeitarchivierung.kopal.util.db.File) iter.next();
			// Only getting the first techmd, if you have more, delete them all!
			Techmd techmd = file.getTechmds().iterator().next();
			DBAccess.deleteObject(techmd);
		}
	}

	/***************************************************************************
	 * Removes all Dublincore table entries from the database for the given
	 * object.
	 * 
	 * @param pData
	 * @throws Exception
	 **************************************************************************/
	public static void removeDublincore(ProcessData pData) throws Exception {
		MappToHib.removeTableEntry(pData, Dublincore.class);
	}

	/***************************************************************************
	 * Removes all Identifier table entries from the database for the given
	 * object.
	 * 
	 * @param pData
	 * @throws Exception
	 **************************************************************************/
	public static void removeIdentifier(ProcessData pData) throws Exception {
		MappToHib.removeTableEntry(pData, Identifier.class);
	}

	/***************************************************************************
	 * Removes all Various table entries from the database for the given object.
	 * 
	 * @param pData
	 * @throws Exception
	 **************************************************************************/
	public static void removeVarious(ProcessData pData) throws Exception {
		MappToHib.removeTableEntry(pData, Various.class);
	}

	/***************************************************************************
	 * Adds the basic DIAS data for a submitted SIP where the successful ingest
	 * is up to now not verified.
	 * 
	 * @param pData The process data object of the SIP
	 * @throws Exception
	 **************************************************************************/
	public static void addDias(ProcessData pData) throws Exception {
		Dias ingestedObject = new Dias();
		ingestedObject.setUrn(pData.getMetadata().getObjectId());
		ingestedObject.setDiasintid(null);
		ingestedObject.setDiasdate(System.currentTimeMillis());

		ingestedObject.setInput(getInput(pData));

		// If the checksum changed between the input entry and now (or no
		// checksum was stored for the input), we store the new checksum (the
		// checksum of the metadata object).
		String newChecksum = ((Uof) pData.getMetadata()).getLmerObjectCopy()
				.getTransferChecksumArray(0).getStringValue();

		if (!newChecksum.equals(ingestedObject.getInput().getChecksum())
				|| !ChecksumGenerator.XOR_OF_FILE_CHECKSUMS
						.equals(ingestedObject.getInput().getChecksumtype()
								.getChecksumtype())) {
			ingestedObject.setChecksumNew(newChecksum);
			ingestedObject
					.setChecksumtype((Checksumtype) getDbConstant(
							Checksumtype.class,
							ChecksumGenerator.XOR_OF_FILE_CHECKSUMS));
		}
		ingestedObject.setDeleted(false);

		DBAccess.lockTable(Dias.class, ingestedObject.getUrn());
		try {
			DBAccess.addObject(ingestedObject);
		}
		finally {
			DBAccess.unlockTable(Dias.class, ingestedObject.getUrn());
		}
	}

	/***************************************************************************
	 * Adds a list of values of custom data elements of the ProcessData to the
	 * various table.
	 * 
	 * @param pData
	 * @throws Exception
	 **************************************************************************/
	public static void addVarious(ProcessData pData, List elementsToAdd)
			throws Exception {
		HashMap customData = pData.getCustomData();

		// Go throught all custom data key elements.
		Iterator iter = customData.keySet().iterator();

		while (iter.hasNext()) {
			String key = (String) iter.next();

			if (elementsToAdd.contains(key)) {
				Various variousObject = new Various();

				// Set the data values.
				variousObject.setInput(getInput(pData));
				variousObject.setType(key);
				variousObject.setValue((String) customData.get(key));

				// Add the various to the database.
				DBAccess.addObject(variousObject);
			}
		}
	}

	/***************************************************************************
	 * This adds the remaining dias database data for a successful ingested sip.
	 * 
	 * @param pData The process data of the sip
	 * @param ingestedAsset The reponse of the dias system for the sip
	 * @throws Exception
	 **************************************************************************/
	public static void updateDias(ProcessData pData,
			DiasAccessMetadataResponse ingestedAsset) throws Exception {
		String urn = ingestedAsset.getExternalAssetId();
		// TODO This presupposes that there is only one DIAS entry for each
		// input.
		Dias entryToUpdate = getInput(pData).getDiases().iterator().next();
		entryToUpdate.setDiasintid(ingestedAsset.getInternalAssetId());
		entryToUpdate.setDiasdate(KopalTimestamp.parseTimestamp(ingestedAsset
				.getAssetIngestTimestamp()));

		if (ingestedAsset.getOldObjectIdentifier() != null) {
			entryToUpdate.setDiasoldobjid(ingestedAsset
					.getOldObjectIdentifier());
		}

		// Update the oldObjectId if available in the custom IDs.
		if (pData.getCustomIds().containsKey(MappToHib.ID_DIASOLDOBJID)) {
			entryToUpdate.setDiasoldobjid((String) pData.getCustomIds().get(
					MappToHib.ID_DIASOLDOBJID));
		}

		DBAccess.lockTable(Dias.class, urn);
		try {
			DBAccess.addObject(entryToUpdate);
		}
		finally {
			DBAccess.unlockTable(Dias.class, urn);
		}
	}

	/***************************************************************************
	 * Checks if an object with the given XORed checksum was already received
	 * for a certain owner. Only checks the XORed checksums in the input table
	 * for the given owner and if a referenced dias table entry is existing.
	 * 
	 * @param checksumToCheck The checksum to control
	 * @param ownerToCheck
	 * @return TRUE if checksum already is in the input table of the database
	 *         with given owner and reference to a dias entry, false otherwise
	 * @throws Exception
	 **************************************************************************/
	public static boolean alreadyReceivedInput(String checksumToCheck,
			String ownerToCheck) throws Exception {
		// Get the proper checksumTypeId and ownerId.
		List inputWithSameXorChecksum = null;

		int checksumTypeId = ((Checksumtype) getDbConstant(Checksumtype.class,
				MappToHib.CHECKSUMTYPE_XOR)).getChecksumtypeid();
		int ownerid = ((Owner) DBAccess.getUniqueObject(Owner.class, "owner",
				ownerToCheck)).getOwnerid();

		// Get a list of input objects with owner equal to given owner.
		inputWithSameXorChecksum = DBAccess.getObjects(Input.class, "checksum",
				checksumToCheck, "checksumtype.checksumtypeid", new Integer(
						checksumTypeId), "owner.ownerid", new Integer(ownerid));

		// Return false if checksum is not existing in the input table.
		if (inputWithSameXorChecksum == null
				|| inputWithSameXorChecksum.isEmpty()) {
			return false;
		}

		// If one of the inputs with the same objectChecksum resulted in an
		// ingest, return true.
		for (Iterator iter = inputWithSameXorChecksum.iterator(); iter
				.hasNext();) {
			Input inputToCheck = (Input) iter.next();
			if (!inputToCheck.getDiases().isEmpty()) {
				return true;
			}
		}

		return false;
	}

	/***************************************************************************
	 * Checks, if an object with the given XORed checksum and owner already
	 * exists in the database. Checks first in the dias table entries and if no
	 * checksum is existing there, it checks the input table, too.
	 * 
	 * @param checksumToCheck
	 * @param ownerToCheck
	 * @return
	 **************************************************************************/
	public static boolean alreadyIngested(String checksumToCheck,
			String ownerToCheck) {
		List diases = null;
		List inputs = null;

		// Get the proper checksumTypeId and ownerId.
		int checksumTypeId = ((Checksumtype) DBAccess.getUniqueObject(
				Checksumtype.class, "checksumtype", MappToHib.CHECKSUMTYPE_XOR))
				.getChecksumtypeid();
		int ownerid = ((Owner) DBAccess.getUniqueObject(Owner.class, "owner",
				ownerToCheck)).getOwnerid();

		// Fetch a list of diases, with checksum and checksumtype matching.
		diases = DBAccess.getObjects(Dias.class, "checksumNew",
				checksumToCheck, "checksumtype.checksumtypeid", new Integer(
						checksumTypeId));

		// Check the owner if diases list is not empty.
		for (Iterator iter = diases.iterator(); iter.hasNext();) {
			Dias diasEntry = (Dias) iter.next();
			String dbOwner = diasEntry.getInput().getOwner().getOwner();

			// Bingo! Owner equals the given owner -> checksum is already
			// existing!
			if (dbOwner.equals(ownerToCheck)) {
				return true;
			}
		}

		// Then fetch a list of inputs with checksum and owner equals given
		// values.
		inputs = DBAccess.getObjects(Input.class, "checksum", checksumToCheck,
				"checksumtype.checksumtypeid", new Integer(checksumTypeId),
				"owner.ownerid", new Integer(ownerid));

		// If one of the inputs with the same objectChecksum resulted in an
		// ingest, return true.
		for (Iterator iter = inputs.iterator(); iter.hasNext();) {
			Input inputToCheck = (Input) iter.next();
			if (!inputToCheck.getDiases().isEmpty()) {
				return true;
			}
		}

		return false;
	}

	/***************************************************************************
	 * Returns the identifier value of a specified input object and identifier
	 * type. Assumes that there is only one identifier of this type resp.
	 * returns the latest one. Return null if not present.
	 * 
	 * @param in
	 * @param idType
	 * @return String - The identifier value of a specified input object and
	 *         identifier type
	 * @throws Exception
	 **************************************************************************/
	public static String getIdValue(Input in, String idType) throws Exception {
		// TODO The DIAS IDs assume that there is only one DIAS entry for one
		// input.
		if (idType.equals(ID_DIASEXTID)) {
			return in.getDiases().iterator().next().getUrn();
		}
		else if (idType.equals(ID_DIASINTID)) {
			return in.getDiases().iterator().next().getDiasintid();
		}
		else if (idType.equals(ID_DIASOLDOBJID)) {
			return in.getDiases().iterator().next().getDiasoldobjid();
		}
		else {
			int idTypeId = ((Idtype) DBAccess.getUniqueObject(Idtype.class,
					"idtype", idType)).getIdtypeid();
			List ids = DBAccess.getObjectsDescending(Identifier.class,
					"input.inputid", new Integer(in.getInputid()),
					"idtype.idtypeid", new Integer(idTypeId), "identifierid");
			if (!ids.isEmpty()) {
				return ((Identifier) ids.get(0)).getIdvalue();
			}
		}

		return null;
	}

	/***************************************************************************
	 * Gets all various entries for one input ID and returns them as a HashMap.
	 * 
	 * @param in
	 * @return
	 * @throws Exception
	 **************************************************************************/
	public static HashMap getVarious(Input in) throws Exception {
		HashMap<String, String> result = new HashMap<String, String>();
		Various entry;

		List various = DBAccess.getObjects(Various.class, "input.inputid",
				new Integer(in.getInputid()));

		Iterator iter = various.iterator();
		while (iter.hasNext()) {
			entry = (Various) iter.next();
			result.put(entry.getType(), entry.getValue());
		}

		return result;
	}

	// INTERNAL (Internal - implementation details, local classes, ...) //

	/***************************************************************************
	 * Removes an item from a database table.
	 * 
	 * @param pData
	 * @param itemClass
	 * @throws Exception
	 **************************************************************************/
	private static void removeTableEntry(ProcessData pData, Class itemClass)
			throws Exception {
		// First get the inputid of the given ProcessData object.
		Input in = DBAccess.getInputPerStartdate(pData.getStartDate());
		int inputid = in.getInputid();

		// Get all objects of the given item class for the given inputid and
		// delete them.
		List fileList = new ArrayList();
		fileList = DBAccess.getObjects(itemClass, "input.inputid", new Integer(
				inputid));

		Iterator iter = fileList.iterator();
		while (iter.hasNext()) {
			DBAccess.deleteObject(iter.next());
		}
	}

	/***************************************************************************
	 * Helper method for retrieving a table of constants for a specific class
	 * 
	 * @param dbClass The class for which constants table is needed
	 * @return the constants table
	 **************************************************************************/
	private static HashMap getDbConstantTable(Class dbClass) {
		return dbConstantsTables.get(dbClass);
	}

	/***************************************************************************
	 * Helper method for retrieving constants from the database.
	 * 
	 * @param dbClass The class of the object to be retrieved resp the database
	 *            table
	 * @param value The value of the constant requested
	 * @return the database object representing the constant
	 * @throws Exception
	 **************************************************************************/
	private static Object getDbConstant(Class dbClass, String value)
			throws Exception {
		HashMap valueToConstantMap = getDbConstantTable(dbClass);
		if (valueToConstantMap == null) {
			throw new QueryException("Unknown requested database class: "
					+ dbClass.getName() + " with value:" + value);
		}

		Object dbConstant = valueToConstantMap.get(value.toLowerCase());
		if (dbConstant == null) {
			throw new QueryException("Unknown value " + value
					+ " in database table: " + dbClass.getName());
		}

		return dbConstant;
	}

	/***************************************************************************
	 * Constructs and returns a set of dublincore database objects for the
	 * specified database input object, dcElement and values.
	 * 
	 * @param input
	 * @param dcElement
	 * @param dcValues
	 * @return Set<Dublincore> - set of dublincore database objects for the
	 *         specified database input object, dcElement and values.
	 * @throws Exception
	 **************************************************************************/
	private static Set<Dublincore> createDcelements(Input input,
			String dcElement, String[] dcValues) throws Exception {
		Set<Dublincore> result = new HashSet<Dublincore>();
		Dcelement dbDcElement = (Dcelement) getDbConstant(Dcelement.class,
				dcElement);
		for (int i = 0; i < dcValues.length; i++) {
			Dublincore dc = new Dublincore();
			dc.setDcelement(dbDcElement);
			dc.setDcvalue(dcValues[i]);
			dc.setInput(input);

			// DEBUG OUTPUT!
			//
			// System.out.println("Element: " +
			// dc.getDcelement().getDcelement());
			// System.out.println("Value: " + dc.getDcvalue());
			// System.out.println("ID: " + dc.getDublincoreid());
			//
			// DEBUG OUTPUT!

			result.add(dc);
		}

		return result;
	}

}
