/*******************************************************************************
 * de.langzeitarchivierung.kopal.ws / WebServiceServer.java
 * 
 * Copyright 2005-2007 by Project kopal
 * 
 * http://kopal.langzeitarchivierung.de/
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307, USA
 * 
 * 
 *******************************************************************************
 *
 *
 * CHANGELOG:
 * 12.07.2007	koLibRI version 1.0
 ******************************************************************************/

package de.langzeitarchivierung.kopal.ws;

import de.langzeitarchivierung.kopal.ws.wsdl2java.*;
// import de.langzeitarchivierung.kopal.util.Configurator;
// import de.langzeitarchivierung.kopal.util.KopalXMLParser;
import de.langzeitarchivierung.kopal.util.WSUtils;
import de.langzeitarchivierung.kopal.ingest.IngestLayer;
import de.langzeitarchivierung.kopal.retrieval.DiasAccess;
import de.langzeitarchivierung.kopal.retrieval.AccessLayer;

/*******************************************************************************
 * Class to provide web service access to DIAS-Core sytem. For more information
 * about how it works, and why, see the SIP and DIP Interface Specifications
 * provided by IBM-Holland.
 * 
 * @author Kadir Karaca Koçer, German National Library
 * @version 20070604
 * @see DiasAccess
 * @see KolibriWebServiceSkeletonInterface
 * @see ResponseElemDocument
 * @since 20070524
 ******************************************************************************/

public class WebServiceServer implements KolibriWebServiceSkeletonInterface
{

	/***************************************************************************
	 * Returns a link to the xml file storing all in DIAS defined file formats
	 * with their DIAS unique identifiers. Please note that you may have to call
	 * reloadFileFormatList() if this file is too old. Actually the list of
	 * known formats does not change often and the administrator should know if
	 * it must be reloaded.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait (not used)</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see IngestLayer
	 * @see GetFileFormatListElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument getFileFormatList(
			GetFileFormatListElemDocument doc) {
		// read the user from message and call IngestLayer
		UserType user = doc.getGetFileFormatListElem().getUser();

		return IngestLayer.getFileFormatList(user);
	}

	/***************************************************************************
	 * Reloads the DIAS Known File Formats List. Should be called each time
	 * after the list modified.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the XML file with the list</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see IngestLayer
	 * @see ReloadFileFormatListElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument reloadFileFormatList(
			ReloadFileFormatListElemDocument doc) {
		// read the user name and password from message and call IngestLayer
		UserType user = doc.getReloadFileFormatListElem().getUser();

		return IngestLayer.reloadFileFormatList(user);
	}

	/***************************************************************************
	 * Ingestst a given digital object into DIAS.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, policy to apply and 1 to 5 Metadata Blocks needed
	 *            to ingest that object.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see IngestLayer
	 * @see IngestDocumentElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument ingestDocument(IngestDocumentElemDocument doc) {
		// read the user name and password from message and call IngestLayer
		UserType user = doc.getIngestDocumentElem().getUser();
		String policy = doc.getIngestDocumentElem().getPolicy();
		CustomXMLType[] metadataArray = doc.getIngestDocumentElem()
				.getMetadataArray();

		return IngestLayer.ingestDocument(user, policy, metadataArray);
	}

	/***************************************************************************
	 * Ingestst a already generated SIP into DIAS. Used for SIP of partner
	 * institutions, which generate their SIPs themself and deliver to DNB or
	 * SUB according to singned partnership contracts.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, policy to apply and 1 to 5 Metadata Blocks needed
	 *            to ingest that object.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see IngestLayer
	 * @see IngestSIPElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument ingestSIP(IngestSIPElemDocument doc) {
		// read the user name and password from message and call IngestLayer
		UserType user = doc.getIngestSIPElem().getUser();
		String urn = doc.getIngestSIPElem().getUrn();
		String linkToSIP = doc.getIngestSIPElem().getLinkToSIP();

		return IngestLayer.ingestSIP(user, urn, linkToSIP);
	}

	/***************************************************************************
	 * Sets the flag that indicates the ingest process is blocked. Good for
	 * maintenence.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready(not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see IngestLayer
	 * @see BlockNewIngestRequestsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument blockNewIngestRequests(
			BlockNewIngestRequestsElemDocument doc) {
		// read the user and call IngestLayer
		UserType user = doc.getBlockNewIngestRequestsElem().getUser();

		return IngestLayer.blockIngest(user);
	}

	/***************************************************************************
	 * Sets the flag that indicates the ingest process is allowed. Must be
	 * called after blockNewIngestRequests() blocked the ingest.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready(not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see IngestLayer
	 * @see AllowIngestRequestsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument allowIngestRequests(
			AllowIngestRequestsElemDocument doc) {// read the user name and
		// password from message and
		// call IngestLayer
		UserType user = doc.getAllowIngestRequestsElem().getUser();

		return IngestLayer.allowIngestRequests(user);
	}

	/***************************************************************************
	 * Returns some statistics about the total ingest activities.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready(not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field containing the statistics as a XML tree</li>
	 * @see IngestLayer
	 * @see GetIngestStatisticsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument getIngestStatistics(
			GetIngestStatisticsElemDocument doc) {
		// read the user and call IngestLayer
		UserType user = doc.getGetIngestStatisticsElem().getUser();

		return IngestLayer.getIngestStatistics(user);
	}

	/***************************************************************************
	 * Resets the statistical data for ingest.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Time to wait if DIAS is not ready(not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see IngestLayer
	 * @see ResetIngestStatisticsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070607
	 **************************************************************************/
	public ResponseElemDocument resetIngestStatistics(
			ResetIngestStatisticsElemDocument doc) {
		// read the user and call IngestLayer
		UserType user = doc.getResetIngestStatisticsElem().getUser();

		return IngestLayer.resetIngestlStatistics(user);
	}

	// R E T R I E V A L

	/***************************************************************************
	 * Method to get a zipped mets.xml file from DIAS using the middle layer
	 * AccessLayer.getMETS() method.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this operation
	 *            and the persistent identifier of the requested object
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object</li>
	 *         <li>Time to wait if DIAS is not ready</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see GetMETSElemDocument
	 * @see ResponseElemDocument
	 * @version 20070524
	 **************************************************************************/
	public ResponseElemDocument getMETS(GetMETSElemDocument doc) {
		// get the persistent identifier from request
		UserType user = doc.getGetMETSElem().getUser();
		String persistentIdentifier = doc.getGetMETSElem().getObjectId();
		// Use the AccesLayer to get the METS file

		return AccessLayer.getMETS(user, persistentIdentifier);
	}

	/***************************************************************************
	 * Method to get a specific file from DIAS using the middle layer
	 * AccessLayer.getFile() method.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, persistent identifier of digital asset and the name
	 *            of the requested file
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see GetFileElemDocument
	 * @see ResponseElemDocument
	 * @version 20070530
	 **************************************************************************/
	public ResponseElemDocument getFile(GetFileElemDocument doc) {
		UserType user = doc.getGetFileElem().getUser();
		// get the persistent identifier from request
		String diasInternalID = doc.getGetFileElem().getObjectId();
		// get the file name from request
		String fileName = doc.getGetFileElem().getFileName();

		// Use the AccesLayer to get the requested file
		return AccessLayer.getFile(user, diasInternalID, fileName);
	}

	/***************************************************************************
	 * Method to get GNU Tarred DIP from DIAS using the middle layer
	 * AccessLayer.getDIP() method.
	 * 
	 * @param doc SOAP Document Message which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, the persistent identifier of the requested object.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (Not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see GetDIPElemDocument
	 * @see ResponseElemDocument
	 * @version 20070530
	 **************************************************************************/
	public ResponseElemDocument getDIP(GetDIPElemDocument doc) {
		UserType user = doc.getGetDIPElem().getUser();
		// get the persistent identifier from request
		String persistentIdentifier = doc.getGetDIPElem().getObjectId();

		// Use the AccesLayer to get the link to the DIP
		return AccessLayer.getDIP(user, persistentIdentifier);
	}

	/***************************************************************************
	 * Method to get GNU Tarred DIP from DIAS using the middle layer
	 * AccessLayer.getDIP() method. The difference to method getDIP is: this
	 * method is for external institutions which like to get their own assets
	 * and that are not in intranet.
	 * 
	 * @param doc Message which contains user, his/her institution & password to
	 *            check if this user is allowed to do this operation and the
	 *            persistent identifier of the requested object.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object</li>
	 *         <li>Link to the requested file</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see GetDIPElemDocument
	 * @see ResponseElemDocument
	 * @version 20070530
	 **************************************************************************/
	public ResponseElemDocument getDIPExternal(GetDIPExternalElemDocument doc) {
		UserType user = doc.getGetDIPExternalElem().getUser();
		// get the persistent identifier from request
		String persistentIdentifier = doc.getGetDIPExternalElem().getObjectId();

		// Use the AccesLayer to get the link to the DIP
		return AccessLayer.getDIPExternal(user, persistentIdentifier);
	}

	/***************************************************************************
	 * Method to get the list of the documents that are modified since given
	 * timestamp.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this operation
	 *            and the requested timestamp.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field containing the list of the objects</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see GetModifiedSinceElemDocument
	 * @see ResponseElemDocument
	 * @version 20070530
	 **************************************************************************/
	public ResponseElemDocument getModifiedSince(
			GetModifiedSinceElemDocument doc) {
		UserType user = doc.getGetModifiedSinceElem().getUser();
		// get the timestamp from request
		java.util.Calendar timestamp = doc.getGetModifiedSinceElem()
				.getModificationdate();

		return AccessLayer.getModifiedSince(user, timestamp);
	}

	/***************************************************************************
	 * Returns the migration history of a specific document.
	 * 
	 * @param doc Message which contains user, his/her institution & password to
	 *            check if this user is allowed to do this operation and the
	 *            persistent identifier of the requested object.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (if necessary)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field containing the list of the objects</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see ResponseElemDocument
	 * @version 20070530
	 **************************************************************************/
	public ResponseElemDocument getHistory(GetHistoryElemDocument doc) {
		UserType user = doc.getGetHistoryElem().getUser();
		// get the persistent identifier from request
		String persistentIdentifier = doc.getGetHistoryElem().getObjectId();

		// Use the AccesLayer to get the history
		return AccessLayer.getHistory(user, persistentIdentifier);
	}

	/***************************************************************************
	 * Sets the "isRetrievalActive" flag in Access to false to block new
	 * retrieval requests. Good for maintenence.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (Not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see BlockNewRetrievalRequestsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070530
	 **************************************************************************/
	public ResponseElemDocument blockNewRetrievalRequests(
			BlockNewRetrievalRequestsElemDocument doc) {
		UserType user = doc.getBlockNewRetrievalRequestsElem().getUser();

		return AccessLayer.blockRetrieval(user);
	}

	/***************************************************************************
	 * Sets the "isRetrievalActive" flag in Access to TRUE to allow new
	 * retrieval requests. Good after maintenence.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see AllowRetrievalRequestsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070531
	 **************************************************************************/
	public ResponseElemDocument allowRetrievalRequests(
			AllowRetrievalRequestsElemDocument doc) {
		UserType user = doc.getAllowRetrievalRequestsElem().getUser();

		return AccessLayer.allowRetrieval(user);
	}

	/***************************************************************************
	 * Gets the statistics for whole retrieval activities.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field containing the statistics data.</li>
	 * @see AccessLayer
	 * @see GetRetrievalStatisticsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070531
	 **************************************************************************/
	public ResponseElemDocument getRetrievalStatistics(
			GetRetrievalStatisticsElemDocument doc) {
		UserType user = doc.getGetRetrievalStatisticsElem().getUser();

		return AccessLayer.getRetrievalStatistics(user);
	}

	/***************************************************************************
	 * Resets the statistics for retrieval.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see ResetRetrievalStatisticsElemDocument
	 * @see ResponseElemDocument
	 * @version 20070531
	 **************************************************************************/
	public ResponseElemDocument resetRetrievalStatistics(
			ResetRetrievalStatisticsElemDocument doc) {
		UserType user = doc.getResetRetrievalStatisticsElem().getUser();

		return AccessLayer.resetRetrievalStatistics(user);
	}

	/***************************************************************************
	 * Initialisiert the cache with creating a new empty directory tree.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, and root directory of the cache.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see DiasAccess
	 * @see AccessLayer
	 * @see InitCacheElemDocument
	 * @see ResponseElemDocument
	 * @version 20070601
	 **************************************************************************/
	public ResponseElemDocument initCache(InitCacheElemDocument doc) {
		UserType user = doc.getInitCacheElem().getUser();
		String path = doc.getInitCacheElem().getRootdir();

		return AccessLayer.initCache(user, path);
	}

	/***************************************************************************
	 * Erases the cache entries which are older than given days.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, and root directory of the cache and number of days
	 *            as time limit.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see AccessLayer
	 * @see EraseOldCacheEntriesElemDocument
	 * @see ResponseElemDocument
	 * @version 20070601
	 **************************************************************************/
	public ResponseElemDocument eraseOldCacheEntries(
			EraseOldCacheEntriesElemDocument doc) {
		UserType user = doc.getEraseOldCacheEntriesElem().getUser();
		int timeLimit = doc.getEraseOldCacheEntriesElem().getTimelimit();

		return AccessLayer.eraseOldCacheEntries(user, timeLimit);
	}

	/***************************************************************************
	 * Deletes the whole cache directory including all subdirectories and files
	 * in it. WARNING: there exists no cache after that! You have to call
	 * initCache() method before you allow the retrieval again.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, and root directory of the cache.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (Not used)</li>
	 *         <li>Link to the requested file (Not used)</li>
	 *         <li>Extra info string (Not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (Not used)</li>
	 * @see AccessLayer
	 * @see EraseOldCacheEntriesElemDocument
	 * @see ResponseElemDocument
	 * @version 20070601
	 **************************************************************************/
	public ResponseElemDocument deleteCache(DeleteCacheElemDocument doc) {
		UserType user = doc.getDeleteCacheElem().getUser();

		return AccessLayer.deleteCache(user);
	}

	/***************************************************************************
	 * Changes the password of the user.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, and the new password to set.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see AccessLayer
	 * @see ChangePasswordElemDocument
	 * @see ResponseElemDocument
	 * @version 20070601
	 **************************************************************************/
	public ResponseElemDocument setParameters(SetParametersElemDocument doc) {
		UserType user = doc.getSetParametersElem().getUser();
		// check if the user is for this action authorised
		int authorisationResponse = WSUtils.checkRights(user, "setParameters");
		if (authorisationResponse > 0) {
			// not authorised!
			return WSUtils.authorisationError(authorisationResponse);
		}

		ResponseElemDocument response = WSUtils.createDefaultResponse();

		String tempStr = doc.getSetParametersElem().getDiasRetrieverURL();
		if ((tempStr != null) && (!tempStr.equals(""))) {
			// setRetrieverURL(tempStr);
		}

		tempStr = doc.getSetParametersElem().getProtocolRetrievalWS();
		if ((tempStr != null) && (!tempStr.equals(""))) {
			AccessLayer.setProtocol(tempStr);
		}

		tempStr = doc.getSetParametersElem().getUrlRetrievalWS();
		if ((tempStr != null) && (!tempStr.equals(""))) {
			AccessLayer.setBaseURL(tempStr);
		}

		tempStr = doc.getSetParametersElem().getCacheRootDir();
		if ((tempStr != null) && (!tempStr.equals(""))) {
			AccessLayer.setAbsoluteRootDir(tempStr);
		}

		response.getResponseElem().setStatusCode(291);
		response.getResponseElem().setResponseText(
				"The parameters are set as wished.");

		return response;
	}

	/***************************************************************************
	 * Changes the password of the user.
	 * 
	 * @param doc SOAP Document which contains user, his/her institution &
	 *            password to check if this user is allowed to do this
	 *            operation, and the new password to set.
	 * @return A ResponseElemDocument containing:
	 *         <li>The status code of the operation</li>
	 *         <li>A human readable text for that code</li>
	 *         <li>DIAS internal ID of this digital object (not used)</li>
	 *         <li>Link to the requested file (not used)</li>
	 *         <li>Extra info string (not used)</li>
	 *         <li>Error Message (if any occured)</li>
	 *         <li>Custom XML field (not used)</li>
	 * @see AccessLayer
	 * @see ChangePasswordElemDocument
	 * @see ResponseElemDocument
	 * @version 20070601
	 **************************************************************************/
	public ResponseElemDocument changePassword(ChangePasswordElemDocument doc) {
		UserType user = doc.getChangePasswordElem().getUser();
		String newPasswd = doc.getChangePasswordElem().getNewPassword();

		return WSUtils.changePassword(user, newPasswd);
	}

}
